/* Copyright(C) 2007 Brazil

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "senna_in.h"
#include <string.h>
#include "lex.h"
#include "token.h"
#include "ql.h"
#include "pat.h"
#include "snip.h"
#include <stdarg.h>
#include <stdio.h>
#include <time.h>

sen_ctx sen_gctx = SEN_CTX_INITIALIZER;
int sen_pagesize;
sen_mutex sen_glock;
uint32_t sen_gtick;

/* fixme by 2038 */

sen_rc
sen_timeval_now(sen_timeval *tv)
{
#ifdef WIN32
  time_t t;
  struct _timeb tb;
  time(&t);
  _ftime(&tb);
  tv->tv_sec = (int32_t) t;
  tv->tv_usec = tb.millitm * 1000;
  return sen_success;
#else /* WIN32 */
  struct timeval t;
  sen_rc rc = gettimeofday(&t, NULL) ? sen_external_error : sen_success;
  if (!rc) {
    tv->tv_sec = (int32_t) t.tv_sec;
    tv->tv_usec = t.tv_usec;
  }
  return rc;
#endif /* WIN32 */
}

sen_rc
sen_timeval2str(sen_timeval *tv, char *buf)
{
  struct tm *ltm;
#ifdef WIN32
  time_t tvsec = (time_t) tv->tv_sec;
  ltm = localtime(&tvsec);
#else /* WIN32 */
  struct tm tm;
  time_t t = tv->tv_sec;
  ltm = localtime_r(&t, &tm);
#endif /* WIN32 */
  if (!ltm) { return sen_external_error; }
  snprintf(buf, SEN_TIMEVAL_STR_SIZE - 1, SEN_TIMEVAL_STR_FORMAT,
           ltm->tm_year + 1900, ltm->tm_mon + 1, ltm->tm_mday,
           ltm->tm_hour, ltm->tm_min, ltm->tm_sec, (int) tv->tv_usec);
  buf[SEN_TIMEVAL_STR_SIZE - 1] = '\0';
  return sen_success;
}

sen_rc
sen_str2timeval(const char *str, uint32_t str_len, sen_timeval *tv)
{
  struct tm tm;
  const char *r1, *r2, *rend = str + str_len;
  uint32_t uv;
  memset(&tm, 0, sizeof(struct tm));

  tm.tm_year = (int)sen_atoui(str, rend, &r1) - 1900;
  if ((r1 + 1) >= rend || (*r1 != '/' && *r1 != '-') ||
      tm.tm_year < 0) { return sen_invalid_argument; }
  r1++;
  tm.tm_mon = (int)sen_atoui(r1, rend, &r1) - 1;
  if ((r1 + 1) >= rend || (*r1 != '/' && *r1 != '-') ||
      tm.tm_mon < 0 || tm.tm_mon >= 12) { return sen_invalid_argument; }
  r1++;
  tm.tm_mday = (int)sen_atoui(r1, rend, &r1);
  if ((r1 + 1) >= rend || *r1 != ' ' ||
      tm.tm_mday < 1 || tm.tm_mday > 31) { return sen_invalid_argument; }

  tm.tm_hour = (int)sen_atoui(++r1, rend, &r2);
  if ((r2 + 1) >= rend || r1 == r2 || *r2 != ':' ||
      tm.tm_hour < 0 || tm.tm_hour >= 24) {
    return sen_invalid_argument;
  }
  r1 = r2 + 1;
  tm.tm_min = (int)sen_atoui(r1, rend, &r2);
  if ((r2 + 1) >= rend || r1 == r2 || *r2 != ':' ||
      tm.tm_min < 0 || tm.tm_min >= 60) {
    return sen_invalid_argument;
  }
  r1 = r2 + 1;
  tm.tm_sec = (int)sen_atoui(r1, rend, &r2);
  if (r1 == r2 ||
      tm.tm_sec < 0 || tm.tm_sec > 61 /* leap 2sec */) {
    return sen_invalid_argument;
  }
  r1 = r2;

  if ((tv->tv_sec = (int32_t) mktime(&tm)) == -1) { return sen_invalid_argument; }
  if ((r1 + 1) < rend && *r1 == '.') { r1++; }
  uv = sen_atoi(r1, rend, &r2);
  while (r2 < r1 + 6) {
    uv *= 10;
    r2++;
  }
  if (uv >= 1000000) { return sen_invalid_argument; }
  tv->tv_usec = uv;
  return sen_success;
}

#ifdef USE_FAIL_MALLOC
int sen_fmalloc_prob = 0;
char *sen_fmalloc_func = NULL;
char *sen_fmalloc_file = NULL;
int sen_fmalloc_line = 0;
#endif /* USE_FAIL_MALLOC */

#define N_SEGMENTS 512

#define SEGMENT_SIZE    (1<<22)
#define SEGMENT_MASK    (SEGMENT_SIZE - 1)

#define SEGMENT_WORD    (1<<31)
#define SEGMENT_VLEN    (1<<30)
#define SEGMENT_LIFO    (1<<29)

#ifdef USE_DYNAMIC_MALLOC_CHANGE
static void
sen_ctx_impl_init_malloc(sen_ctx *ctx)
{
#  ifdef USE_FAIL_MALLOC
  ctx->impl->malloc_func = sen_malloc_fail;
  ctx->impl->calloc_func = sen_calloc_fail;
  ctx->impl->realloc_func = sen_realloc_fail;
  ctx->impl->strdup_func = sen_strdup_fail;
#  else
  ctx->impl->malloc_func = sen_malloc_default;
  ctx->impl->calloc_func = sen_calloc_default;
  ctx->impl->realloc_func = sen_realloc_default;
  ctx->impl->strdup_func = sen_strdup_default;
#  endif
}
#endif

static void
sen_ctx_impl_init(sen_ctx *ctx)
{
  if (!(ctx->impl = SEN_MALLOC(sizeof(struct _sen_ctx_impl)))) { return; }
  if (!(ctx->impl->segs = sen_io_anon_map(ctx, &ctx->impl->mi,
                                          sizeof(sen_io_mapinfo) * N_SEGMENTS))) {
    SEN_FREE(ctx->impl);
    ctx->impl = NULL;
    return;
  }
#ifdef USE_DYNAMIC_MALLOC_CHANGE
  sen_ctx_impl_init_malloc(ctx);
#endif
  if (!(ctx->impl->values = sen_array_create(ctx, NULL, sizeof(sen_tmp_obj_db),
                                             SEN_ARRAY_TINY))) {
    sen_io_anon_unmap(ctx, &ctx->impl->mi, sizeof(sen_io_mapinfo) * N_SEGMENTS);
    SEN_FREE(ctx->impl);
    ctx->impl = NULL;
    return;
  }
  ctx->impl->encoding = ctx->encoding;
  ctx->impl->lifoseg = -1;
  ctx->impl->currseg = -1;
  ctx->impl->db = NULL;
  ctx->impl->phs = NIL;
  ctx->impl->code = NIL;
  ctx->impl->dump = NIL;
  ctx->impl->op = SEN_OP_T0LVL;
  ctx->impl->args = NIL;
  ctx->impl->envir = NIL;
  ctx->impl->value = NIL;
  ctx->impl->ncells = 0;
  ctx->impl->n_entries = 0;
  ctx->impl->seqno = 0;
  ctx->impl->lseqno = 0;
  ctx->impl->nbinds = 0;
  ctx->impl->nunbinds = 0;
  ctx->impl->feed_mode = sen_ql_atonce;
  ctx->impl->cur = NULL;
  ctx->impl->str_end = NULL;
  ctx->impl->batchmode = 0;
  ctx->impl->gc_verbose = 0;
  ctx->impl->inbuf = NULL;
  ctx->impl->co.mode = 0;
  ctx->impl->co.func = NULL;
  ctx->impl->objects = NULL;
  ctx->impl->symbols = NULL;
  ctx->impl->com = NULL;
  SEN_OBJ_INIT(&ctx->impl->outbuf, SEN_BULK, 0);
  SEN_OBJ_INIT(&ctx->impl->subbuf, SEN_BULK, 0);
}

void
sen_ctx_impl_err(sen_ctx *ctx)
{
  if (ctx->impl) {
    ctx->impl->cur = ctx->impl->str_end;
    ctx->impl->op = SEN_OP_ERR0;
  }
}

static void
sen_ctx_ql_init(sen_ctx *ctx, int flags)
{
  if (!ctx->impl) {
    sen_ctx_impl_init(ctx);
    if (ERRP(ctx, SEN_ERROR)) { return; }
  }
  if (flags & SEN_CTX_BATCH_MODE) { ctx->impl->batchmode = 1; }
  if ((ctx->impl->objects = sen_array_create(ctx, NULL, sizeof(sen_cell),
                                             SEN_ARRAY_TINY))) {
    if ((ctx->impl->symbols = sen_hash_create(ctx, NULL, SEN_TABLE_MAX_KEY_SIZE,
                                              sizeof(sen_cell),
                                              SEN_OBJ_KEY_VAR_SIZE|SEN_HASH_TINY,
                                              sen_enc_none))) {
      if (!ERRP(ctx, SEN_ERROR)) {
        sen_ql_init_globals(ctx);
        if (!ERRP(ctx, SEN_ERROR)) {
          return;
        }
      }
      sen_hash_close(ctx, ctx->impl->symbols);
      ctx->impl->symbols = NULL;
    } else {
      MERR("ctx->impl->symbols init failed");
    }
    sen_array_close(ctx, ctx->impl->objects);
    ctx->impl->objects = NULL;
  } else {
    MERR("ctx->impl->objects init failed");
  }
}

sen_rc
sen_ctx_init(sen_ctx *ctx, int flags, sen_encoding encoding)
{
  if (!ctx) { return sen_invalid_argument; }
  // if (ctx->stat != SEN_CTX_FIN) { return sen_invalid_argument; }
  ERRCLR(ctx);
  ctx->flags = flags;
  ctx->stat = SEN_QL_WAIT_EXPR;
  ctx->encoding = encoding;
  ctx->seqno = 0;
  ctx->subno = 0;
  ctx->impl = NULL;
  if (flags & SEN_CTX_USE_DB) {
    sen_ctx_impl_init(ctx);
    if (ERRP(ctx, SEN_ERROR)) { return ctx->rc; }
  }
  if (flags & SEN_CTX_USE_QL) {
    sen_ctx_ql_init(ctx, flags);
    if (ERRP(ctx, SEN_ERROR)) { return ctx->rc; }
  }
  MUTEX_LOCK(sen_glock);
  ctx->next = sen_gctx.next;
  ctx->prev = &sen_gctx;
  sen_gctx.next->prev = ctx;
  sen_gctx.next = ctx;
  MUTEX_UNLOCK(sen_glock);
  return ctx->rc;
}

sen_rc
sen_ctx_fin(sen_ctx *ctx)
{
  sen_rc rc = sen_success;
  if (!ctx) { return sen_invalid_argument; }
  if (ctx->stat == SEN_CTX_FIN) { return sen_invalid_argument; }
  MUTEX_LOCK(sen_glock);
  ctx->next->prev = ctx->prev;
  ctx->prev->next = ctx->next;
  MUTEX_UNLOCK(sen_glock);
  if (ctx->impl) {
    if (ctx->impl->objects) {
      sen_cell *o;
      SEN_ARRAY_EACH(ctx->impl->objects, 0, 0, id, &o, {
        sen_cell_clear(ctx, o);
      });
      sen_array_close(ctx, ctx->impl->objects);
    }
    if (ctx->impl->values) {
      sen_tmp_obj_db *o;
      SEN_ARRAY_EACH(ctx->impl->values, 0, 0, id, &o, {
        sen_obj_close(ctx, (sen_obj *)o->obj);
      });
      sen_array_close(ctx, ctx->impl->values);
    }
    if (ctx->impl->symbols) {
      sen_hash_close(ctx, ctx->impl->symbols);
    }
    if (ctx->impl->com) {
      if (ctx->stat != SEN_CTX_QUIT) {
        int flags;
        char *str;
        unsigned int str_len;
        sen_ctx_send(ctx, "(quit)", 6, SEN_CTX_HEAD);
        sen_ctx_recv(ctx, &str, &str_len, &flags);
      }
      sen_ctx_send(ctx, "ACK", 3, SEN_CTX_HEAD);
      rc = sen_com_sqtp_close(NULL, ctx->impl->com);
    }
    rc = sen_bulk_fin(ctx, &ctx->impl->outbuf);
    rc = sen_bulk_fin(ctx, &ctx->impl->subbuf);
    {
      int i;
      sen_io_mapinfo *mi;
      for (i = 0, mi = ctx->impl->segs; i < N_SEGMENTS; i++, mi++) {
        if (mi->map) {
          if (mi->count & SEGMENT_VLEN) {
            sen_io_anon_unmap(ctx, mi, mi->nref * sen_pagesize);
          } else {
            sen_io_anon_unmap(ctx, mi, SEGMENT_SIZE);
          }
        }
      }
    }
    sen_io_anon_unmap(ctx, &ctx->impl->mi, sizeof(sen_io_mapinfo) * N_SEGMENTS);
    SEN_FREE(ctx->impl);
  }
  ctx->stat = SEN_CTX_FIN;
  return rc;
}

static void
expand_stack(void)
{
#ifndef WIN32
  struct rlimit rlim;
  if (!getrlimit(RLIMIT_STACK, &rlim)) {
    SEN_LOG(sen_log_notice, "RLIMIT_STACK is %d (%d)", rlim.rlim_cur, getpid());
    if (rlim.rlim_cur != RLIM_INFINITY && rlim.rlim_cur < SEN_STACK_SIZE) {
      rlim.rlim_cur = SEN_STACK_SIZE;
      if (!setrlimit(RLIMIT_STACK, &rlim)) {
        SEN_LOG(sen_log_notice, "expanded RLIMIT_STACK to %d", rlim.rlim_cur);
      }
    }
  }
#endif /* WIN32 */
}

//sen_rc sen_index_init(void);
//void sen_index_fin(void);

sen_rc
sen_init(void)
{
  sen_rc rc;
  MUTEX_INIT(sen_glock);
  sen_gtick = 0;
  sen_ql_init_const();
  sen_gctx.next = &sen_gctx;
  sen_gctx.prev = &sen_gctx;
  sen_ctx_init(&sen_gctx, 0, sen_strtoenc(SENNA_DEFAULT_ENCODING));
#ifdef WIN32
  {
    SYSTEM_INFO si;
    GetSystemInfo(&si);
    sen_pagesize = si.dwAllocationGranularity;
  }
#else /* WIN32 */
  if ((sen_pagesize = sysconf(_SC_PAGESIZE)) == -1) {
    GSERR("_SC_PAGESIZE");
    return sen_other_error;
  }
#endif /* WIN32 */
  if (sen_pagesize & (sen_pagesize - 1)) {
    SEN_LOG(sen_log_crit, "pagesize=%x", sen_pagesize);
  }
  // expand_stack();
#ifdef USE_AIO
  if (getenv("SEN_DEBUG_PRINT")) {
    sen_debug_print = atoi(getenv("SEN_DEBUG_PRINT"));
  } else {
    sen_debug_print = 0;
  }
  if (getenv("SEN_AIO_ENABLED")) {
    sen_aio_enabled = atoi(getenv("SEN_AIO_ENABLED"));
  } else {
    sen_aio_enabled = 0;
  }
  if (sen_aio_enabled) {
    SEN_LOG(sen_log_notice, "AIO and DIO enabled");
    sen_cache_open();
  }
#endif /* USE_AIO */
#ifdef USE_FAIL_MALLOC
  if (getenv("SEN_FMALLOC_PROB")) {
    sen_fmalloc_prob = strtod(getenv("SEN_FMALLOC_PROB"), 0) * RAND_MAX;
    if (getenv("SEN_FMALLOC_SEED")) {
      srand((unsigned int)atoi(getenv("SEN_FMALLOC_SEED")));
    } else {
      srand((unsigned int)time(NULL));
    }
  }
  if (getenv("SEN_FMALLOC_FUNC")) {
    sen_fmalloc_func = getenv("SEN_FMALLOC_FUNC");
  }
  if (getenv("SEN_FMALLOC_FILE")) {
    sen_fmalloc_file = getenv("SEN_FMALLOC_FILE");
  }
  if (getenv("SEN_FMALLOC_LINE")) {
    sen_fmalloc_line = atoi(getenv("SEN_FMALLOC_LINE"));
  }
#endif /* USE_FAIL_MALLOC */
  if ((rc = sen_token_init())) {
    SEN_LOG(sen_log_alert, "sen_token_init failed (%d)", rc);
    return rc;
  }
  if ((rc = sen_lex_init())) {
    SEN_LOG(sen_log_alert, "sen_lex_init failed (%d)", rc);
    return rc;
  }
  if ((rc = sen_com_init())) {
    SEN_LOG(sen_log_alert, "sen_com_init failed (%d)", rc);
    return rc;
  }
  sen_ctx_ql_init(&sen_gctx, 0);
  if ((rc = sen_gctx.rc)) {
    SEN_LOG(sen_log_alert, "gctx initialize failed (%d)", rc);
    return rc;
  }
  if ((rc = sen_io_init())) {
    SEN_LOG(sen_log_alert, "io initialize failed (%d)", rc);
    return rc;
  }
  /*
  if ((rc = sen_index_init())) {
    SEN_LOG(sen_log_alert, "index initialize failed (%d)", rc);
    return rc;
  }
  */
  SEN_LOG(sen_log_notice, "sen_init");
  return rc;
}

static int alloc_count = 0;

sen_rc
sen_fin(void)
{
  sen_rc rc = sen_success;
  /*
  sen_index_fin();
  */
  sen_io_fin();
  sen_ctx_fin(&sen_gctx);
  sen_lex_fin();
  sen_token_fin();
  sen_com_fin();
  SEN_LOG(sen_log_notice, "sen_fin (%d)", alloc_count);
  sen_logger_fin();
  MUTEX_DESTROY(sen_glock);
  return rc;
}

sen_ctx *
sen_ctx_open(sen_db *db, int flags)
{
  sen_ctx *ctx = SEN_GMALLOCN(sen_ctx, 1);
  if (ctx) {
    sen_ctx_init(ctx, flags, db ?
                 ((sen_pat *)sen_space_keys(db))->encoding :
                 sen_enc_default);
    if (ERRP(ctx, SEN_ERROR)) {
      sen_ctx_close(ctx);
      return NULL;
    }
    if (ctx->impl && (ctx->impl->db = db) && ctx->impl->symbols) {
      if (sen_ql_def_db_funcs(ctx)) {
        sen_ctx_close(ctx);
        return NULL;
      }
    }
  }
  return ctx;
}

sen_ctx *
sen_ctx_connect(const char *host, int port, int flags)
{
  sen_com_sqtp *com = sen_com_sqtp_copen(NULL, host, port);
  if (com) {
    sen_ctx *ctx;
    if ((ctx = SEN_GMALLOCN(sen_ctx, 1))) {
      sen_ctx_init(ctx, flags, sen_enc_default);
      if (!ctx->impl) { sen_ctx_impl_init(ctx); }
      if (ctx->impl) {
        ctx->impl->com = com;
        return ctx;
      }
      SEN_FREE(ctx);
    }
    sen_com_sqtp_close(NULL, com);
  }
  return NULL;
}

sen_rc
sen_ctx_close(sen_ctx *ctx)
{
  sen_rc rc = sen_ctx_fin(ctx);
  SEN_GFREE(ctx);
  return rc;
}

sen_rc
sen_ctx_send(sen_ctx *ctx, char *str, unsigned int str_len, int flags)
{
  if (!ctx) { return sen_invalid_argument; }
  ERRCLR(ctx);
  if (ctx->impl) {
    if (ctx->impl->com) {
      sen_rc rc;
      static uint32_t info = 0;
      sen_com_sqtp_header sheader;
      if ((flags & SEN_CTX_MORE)) { flags |= SEN_CTX_QUIET; }
      sheader.flags = flags;
      sheader.size = str_len;
      sheader.qtype = 0;
      sheader.level = 0;
      sheader.status = 0;
      sheader.info = info++; /* for debug */
      if ((rc = sen_com_sqtp_send(ctx->impl->com, &sheader, (char *)str))) {
        ERR(rc, "sen_com_sqtp_send failed");
      }
      goto exit;
    } else {
      if (ctx->impl->symbols) {
        sen_ql_feed(ctx, str, str_len, flags);
        if (ctx->stat == SEN_QL_QUITTING) { ctx->stat = SEN_CTX_QUIT; }
        if (!ERRP(ctx, SEN_CRIT)) {
          if (!(flags & SEN_CTX_QUIET) && ctx->impl->output) {
            ctx->impl->output(ctx, 0, ctx->impl->data.ptr);
          }
        }
        goto exit;
      }
    }
  }
  ERR(sen_invalid_argument, "invalid ctx assigned");
exit :
  return ctx->rc;
}

sen_rc
sen_ctx_recv(sen_ctx *ctx, char **str, unsigned int *str_len, int *flags)
{
  if (!ctx) { return sen_invalid_argument; }
  ERRCLR(ctx);
  if (ctx->stat == SEN_CTX_QUIT) {
    *flags = SEN_CTX_QUIT;
    return ctx->rc;
  }
  if (ctx->impl) {
    if (ctx->impl->com) {
      if (sen_com_sqtp_recv(ctx->impl->com, &ctx->impl->com->msg, &ctx->impl->com_status, &ctx->impl->com_info)) {
        *str = NULL;
        *str_len = 0;
        *flags = 0;
      } else {
        sen_com_sqtp_header *rheader = SEN_COM_SQTP_MSG_HEADER(&ctx->impl->com->msg);
        *str = SEN_COM_SQTP_MSG_BODY(&ctx->impl->com->msg);
        *str_len = rheader->size;
        if (rheader->flags & SEN_CTX_QUIT) {
          ctx->stat = SEN_CTX_QUIT;
          *flags = SEN_CTX_QUIT;
        } else {
          *flags = (rheader->flags & SEN_CTX_TAIL) ? 0 : SEN_CTX_MORE;
        }
      }
      if (ctx->impl->com->rc) {
        ERR(ctx->impl->com->rc, "sen_com_sqtp_recv failed!");
      }
      goto exit;
    } else {
      if (ctx->impl->symbols) {
        sen_obj *buf = &ctx->impl->outbuf;
        unsigned int head, tail;
        unsigned int *offsets = (unsigned int *) ctx->impl->subbuf.u.b.head;
        int npackets = SEN_BULK_VSIZE(&ctx->impl->subbuf) / sizeof(unsigned int);
        if (npackets < ctx->impl->bufcur) { return sen_invalid_argument; }
        head = ctx->impl->bufcur ? offsets[ctx->impl->bufcur - 1] : 0;
        tail = ctx->impl->bufcur < npackets ? offsets[ctx->impl->bufcur] : SEN_BULK_VSIZE(buf);
        *str = buf->u.b.head + head;
        *str_len = tail - head;
        *flags = ctx->impl->bufcur++ < npackets ? SEN_CTX_MORE : 0;
        goto exit;
      }
    }
  }
  ERR(sen_invalid_argument, "invalid ctx assigned");
exit :
  return ctx->rc;
}

void
sen_ctx_concat_func(sen_ctx *ctx, int flags, void *dummy)
{
  if (ctx && ctx->impl && (flags & SEN_CTX_MORE)) {
    unsigned int size = SEN_BULK_VSIZE(&ctx->impl->outbuf);
    sen_bulk_write(ctx, &ctx->impl->subbuf, (char *) &size, sizeof(unsigned int));
  }
}

void
sen_ctx_stream_out_func(sen_ctx *ctx, int flags, void *stream)
{
  if (ctx && ctx->impl) {
    sen_obj *buf = &ctx->impl->outbuf;
    uint32_t size = SEN_BULK_VSIZE(buf);
    if (size) {
      fwrite(buf->u.b.head, 1, size, (FILE *)stream);
      fputc('\n', (FILE *)stream);
      fflush((FILE *)stream);
      SEN_BULK_REWIND(buf);
    }
  }
}

void
sen_ctx_recv_handler_set(sen_ctx *ctx, void (*func)(sen_ctx *, int, void *), void *func_arg)
{
  if (ctx && ctx->impl) {
    ctx->impl->output = func;
    ctx->impl->data.ptr = func_arg;
  }
}

sen_rc
sen_ctx_info_get(sen_ctx *ctx, sen_ctx_info *info)
{
  if (!ctx || !ctx->impl) { return sen_invalid_argument; }
  if (ctx->impl->com) {
    info->fd = ctx->impl->com->com.fd;
    info->com_status = ctx->impl->com_status;
    info->com_info = ctx->impl->com_info;
    info->outbuf = &ctx->impl->com->msg;
    info->stat = ctx->stat;
  } else {
    info->fd = -1;
    info->com_status = 0;
    info->com_info = 0;
    info->outbuf = &ctx->impl->outbuf;
    info->stat = ctx->stat;
  }
  return sen_success;
}


sen_cell *
sen_get(const char *key)
{
  sen_cell *obj;
  sen_search_flags f = SEN_TABLE_ADD;
  if (!sen_gctx.impl || !sen_gctx.impl->symbols ||
      !sen_hash_get(&sen_gctx, sen_gctx.impl->symbols, key, strlen(key),
                    (void **) &obj, &f)) {
    SEN_LOG(sen_log_warning, "sen_get(%s) failed", key);
    return F;
  }
  if (!obj->header.impl_flags) {
    obj->header.impl_flags |= SEN_CELL_SYMBOL;
    obj->header.type = SEN_VOID;
  }
  return obj;
}

sen_cell *
sen_at(const char *key)
{
  sen_cell *obj;
  if (!sen_gctx.impl || sen_gctx.impl->symbols ||
      !sen_hash_at(&sen_gctx, sen_gctx.impl->symbols,
                   key, strlen(key), (void **) &obj)) {
    return F;
  }
  return obj;
}

sen_rc
sen_del(const char *key)
{
  if (!sen_gctx.impl || !sen_gctx.impl->symbols) {
    SEN_LOG(sen_log_warning, "sen_del(%s) failed", key);
    return sen_invalid_argument;
  }
  return sen_hash_delete(&sen_gctx, sen_gctx.impl->symbols, key, strlen(key), NULL);
}

/**** memory allocation ****/

#define ALIGN_SIZE (1<<3)
#define ALIGN_MASK (ALIGN_SIZE-1)

void *
sen_ctx_alloc(sen_ctx *ctx, size_t size, int flags,
              const char* file, int line, const char *func)
{
  if (!ctx) { return NULL; }
  if (!ctx->impl) {
    sen_ctx_impl_init(ctx);
    if (ERRP(ctx, SEN_ERROR)) { return NULL; }
  }
  {
    int32_t i;
    uint64_t *header;
    sen_io_mapinfo *mi;
    size = ((size + ALIGN_MASK) & ~ALIGN_MASK) + ALIGN_SIZE;
    if (size > SEGMENT_SIZE) {
      uint64_t npages = (size + (sen_pagesize - 1)) / sen_pagesize;
      if (npages >= (1LL<<32)) {
        MERR("too long request size=%zu", size);
        return NULL;
      }
      for (i = 0, mi = ctx->impl->segs;; i++, mi++) {
        if (i >= N_SEGMENTS) {
          MERR("all segments are full");
          return NULL;
        }
        if (!mi->map) { break; }
      }
      if (!sen_io_anon_map(ctx, mi, npages * sen_pagesize)) { return NULL; }
      //SEN_LOG(sen_log_notice, "map i=%d (%d)", i, npages * sen_pagesize);
      mi->nref = (uint32_t) npages;
      mi->count = SEGMENT_VLEN;
      ctx->impl->currseg = -1;
      header = mi->map;
      *header = (uint64_t) i;
      return &header[1];
    } else {
      i = ctx->impl->currseg;
      mi = &ctx->impl->segs[i];
      if (i < 0 || size + mi->nref > SEGMENT_SIZE) {
        for (i = 0, mi = ctx->impl->segs;; i++, mi++) {
          if (i >= N_SEGMENTS) {
            MERR("all segments are full");
            return NULL;
          }
          if (!mi->map) { break; }
        }
        if (!sen_io_anon_map(ctx, mi, SEGMENT_SIZE)) { return NULL; }
        //SEN_LOG(sen_log_notice, "map i=%d", i);
        mi->nref = 0;
        mi->count = SEGMENT_WORD;
        ctx->impl->currseg = i;
      }
      header = (uint64_t *)((byte *)mi->map + mi->nref);
      mi->nref += size;
      mi->count++;
      *header = (uint64_t) i;
      return &header[1];
    }
  }
}

void
sen_ctx_free(sen_ctx *ctx, void *ptr,
             const char* file, int line, const char *func)
{
  if (!ctx) { return; }
  if (!ctx->impl) {
    ERR(sen_invalid_argument,"ctx without impl passed.");
    return;
  }
  {
    uint64_t *header = &((uint64_t *)ptr)[-1];
    if (*header >= N_SEGMENTS) {
      ERR(sen_invalid_argument,"invalid ptr passed. ptr=%p seg=%zu", ptr, *header);
      return;
    }
    {
      int32_t i = (int32_t)*header;
      sen_io_mapinfo *mi = &ctx->impl->segs[i];
      if (mi->count & SEGMENT_VLEN) {
        if (mi->map != header) {
          ERR(sen_invalid_argument,"invalid ptr passed. ptr=%p seg=%d", ptr, i);
          return;
        }
        //SEN_LOG(sen_log_notice, "umap i=%d (%d)", i, mi->nref * sen_pagesize);
        sen_io_anon_unmap(ctx, mi, mi->nref * sen_pagesize);
        mi->map = NULL;
      } else {
        if (!mi->map) {
          ERR(sen_invalid_argument,"invalid ptr passed. ptr=%p seg=%d", ptr, i);
          return;
        }
        mi->count--;
        if (!(mi->count & SEGMENT_MASK)) {
          //SEN_LOG(sen_log_notice, "umap i=%d", i);
          sen_io_anon_unmap(ctx, mi, SEGMENT_SIZE);
          mi->map = NULL;
          if (i == ctx->impl->currseg) { ctx->impl->currseg = -1; }
        }
      }
    }
  }
}

void *
sen_ctx_alloc_lifo(sen_ctx *ctx, size_t size, int flags,
                   const char* file, int line, const char *func)
{
  if (!ctx) { return NULL; }
  if (!ctx->impl) {
    sen_ctx_impl_init(ctx);
    if (ERRP(ctx, SEN_ERROR)) { return NULL; }
  }
  {
    int32_t i = ctx->impl->lifoseg;
    sen_io_mapinfo *mi = &ctx->impl->segs[i];
    if (size > SEGMENT_SIZE) {
      uint64_t npages = (size + (sen_pagesize - 1)) / sen_pagesize;
      if (npages >= (1LL<<32)) {
        MERR("too long request size=%zu", size);
        return NULL;
      }
      for (;;) {
        if (++i >= N_SEGMENTS) {
          MERR("all segments are full");
          return NULL;
        }
        mi++;
        if (!mi->map) { break; }
      }
      if (!sen_io_anon_map(ctx, mi, npages * sen_pagesize)) { return NULL; }
      mi->nref = (uint32_t) npages;
      mi->count = SEGMENT_VLEN|SEGMENT_LIFO;
      ctx->impl->lifoseg = i;
      return mi->map;
    } else {
      size = (size + ALIGN_MASK) & ~ALIGN_MASK;
      if (i < 0 || (mi->count & SEGMENT_VLEN) || size + mi->nref > SEGMENT_SIZE) {
        for (;;) {
          if (++i >= N_SEGMENTS) {
            MERR("all segments are full");
            return NULL;
          }
          if (!(++mi)->map) { break; }
        }
        if (!sen_io_anon_map(ctx, mi, SEGMENT_SIZE)) { return NULL; }
        mi->nref = 0;
        mi->count = SEGMENT_WORD|SEGMENT_LIFO;
        ctx->impl->lifoseg = i;
      }
      {
        uint32_t u = mi->nref;
        mi->nref += size;
        return (byte *)mi->map + u;
      }
    }
  }
}

void
sen_ctx_free_lifo(sen_ctx *ctx, void *ptr,
                  const char* file, int line, const char *func)
{
  if (!ctx) { return; }
  if (!ctx->impl) {
    ERR(sen_invalid_argument,"ctx without impl passed.");
    return;
  }
  {
    int32_t i = ctx->impl->lifoseg, done = 0;
    sen_io_mapinfo *mi = &ctx->impl->segs[i];
    if (i < 0) {
      ERR(sen_invalid_argument, "lifo buffer is void");
      return;
    }
    for (; i >= 0; i--, mi--) {
      if (!(mi->count & SEGMENT_LIFO)) { continue; }
      if (done) { break; }
      if (mi->count & SEGMENT_VLEN) {
        if (mi->map == ptr) { done = 1; }
        sen_io_anon_unmap(ctx, mi, mi->nref * sen_pagesize);
        mi->map = NULL;
      } else {
        if (mi->map == ptr) {
          done = 1;
        } else {
          if (mi->map < ptr && ptr < (void *)((byte*)mi->map + mi->nref)) {
            mi->nref = (uint32_t) ((uintptr_t)ptr - (uintptr_t)mi->map);
            break;
          }
        }
        sen_io_anon_unmap(ctx, mi, SEGMENT_SIZE);
        mi->map = NULL;
      }
    }
    ctx->impl->lifoseg = i;
  }
}

#if USE_DYNAMIC_MALLOC_CHANGE
sen_malloc_func
sen_ctx_get_malloc(sen_ctx *ctx)
{
  if (!ctx || !ctx->impl) { return NULL; }
  return ctx->impl->malloc_func;
}

void
sen_ctx_set_malloc(sen_ctx *ctx, sen_malloc_func malloc_func)
{
  if (!ctx || !ctx->impl) { return; }
  ctx->impl->malloc_func = malloc_func;
}

sen_calloc_func
sen_ctx_get_calloc(sen_ctx *ctx)
{
  if (!ctx || !ctx->impl) { return NULL; }
  return ctx->impl->calloc_func;
}

void
sen_ctx_set_calloc(sen_ctx *ctx, sen_calloc_func calloc_func)
{
  if (!ctx || !ctx->impl) { return; }
  ctx->impl->calloc_func = calloc_func;
}

sen_realloc_func
sen_ctx_get_realloc(sen_ctx *ctx)
{
  if (!ctx || !ctx->impl) { return NULL; }
  return ctx->impl->realloc_func;
}

void
sen_ctx_set_realloc(sen_ctx *ctx, sen_realloc_func realloc_func)
{
  if (!ctx || !ctx->impl) { return; }
  ctx->impl->realloc_func = realloc_func;
}

sen_strdup_func
sen_ctx_get_strdup(sen_ctx *ctx)
{
  if (!ctx || !ctx->impl) { return NULL; }
  return ctx->impl->strdup_func;
}

void
sen_ctx_set_strdup(sen_ctx *ctx, sen_strdup_func strdup_func)
{
  if (!ctx || !ctx->impl) { return; }
  ctx->impl->strdup_func = strdup_func;
}

void *
sen_malloc(sen_ctx *ctx, size_t size, const char* file, int line, const char *func)
{
  if (ctx && ctx->impl && ctx->impl->malloc_func) {
    return ctx->impl->malloc_func(ctx, size, file, line, func);
  } else {
    return sen_malloc_default(ctx, size, file, line, func);
  }
}

void *
sen_calloc(sen_ctx *ctx, size_t size, const char* file, int line, const char *func)
{
  if (ctx && ctx->impl && ctx->impl->calloc_func) {
    return ctx->impl->calloc_func(ctx, size, file, line, func);
  } else {
    return sen_calloc_default(ctx, size, file, line, func);
  }
}

void *
sen_realloc(sen_ctx *ctx, void *ptr, size_t size, const char* file, int line, const char *func)
{
  if (ctx && ctx->impl && ctx->impl->realloc_func) {
    return ctx->impl->realloc_func(ctx, ptr, size, file, line, func);
  } else {
    return sen_realloc_default(ctx, ptr, size, file, line, func);
  }
}

char *
sen_strdup(sen_ctx *ctx, const char *string, const char* file, int line, const char *func)
{
  if (ctx && ctx->impl && ctx->impl->strdup_func) {
    return ctx->impl->strdup_func(ctx, string, file, line, func);
  } else {
    return sen_strdup_default(ctx, string, file, line, func);
  }
}
#endif

void *
sen_malloc_default(sen_ctx *ctx, size_t size, const char* file, int line, const char *func)
{
  if (!ctx) { return NULL; }
  {
    void *res = malloc(size);
    if (res) {
      alloc_count++;
    } else {
      sen_index_expire();
      if (!(res = malloc(size))) {
        MERR("malloc fail (%d)=%p (%s:%d) <%d>", size, res, file, line, alloc_count);
      }
    }
    return res;
  }
}

void *
sen_calloc_default(sen_ctx *ctx, size_t size, const char* file, int line, const char *func)
{
  if (!ctx) { return NULL; }
  {
    void *res = calloc(size, 1);
    if (res) {
      alloc_count++;
    } else {
      sen_index_expire();
      if (!(res = calloc(size, 1))) {
        MERR("calloc fail (%d)=%p (%s:%d) <%d>", size, res, file, line, alloc_count);
      }
    }
    return res;
  }
}

void
sen_free(sen_ctx *ctx, void *ptr, const char* file, int line)
{
  if (!ctx) { return; }
  {
    free(ptr);
    if (ptr) {
      alloc_count--;
    } else {
      SEN_LOG(sen_log_alert, "free fail (%p) (%s:%d) <%d>", ptr, file, line, alloc_count);
    }
  }
}

void *
sen_realloc_default(sen_ctx *ctx, void *ptr, size_t size, const char* file, int line, const char *func)
{
  if (!ctx) { return NULL; }
  {
    void *res;
    if (!size) {
      alloc_count--;
#if defined __FreeBSD__
      free(ptr);
      return NULL;
#endif /* __FreeBSD__ */
    }
    res = realloc(ptr, size);
    if (!ptr && res) { alloc_count++; }
    if (size && !res) {
      sen_index_expire();
      if (!(res = realloc(ptr, size))) {
        MERR("realloc fail (%p,%zu)=%p (%s:%d) <%d>", ptr, size, res, file, line, alloc_count);
      }
    }
    if (!size && res) {
      SEN_LOG(sen_log_alert, "realloc(%p,%zu)=%p (%s:%d) <%d>", ptr, size, res, file, line, alloc_count);
      // sen_free(ctx, res, file, line);
    }
    return res;
  }
}

int
sen_alloc_count(void)
{
  return alloc_count;
}

char *
sen_strdup_default(sen_ctx *ctx, const char *s, const char* file, int line, const char *func)
{
  if (!ctx) { return NULL; }
  {
    char *res = strdup(s);
    if (res) {
      alloc_count++;
    } else  {
      sen_index_expire();
      if (!(res = strdup(s))) {
        MERR("strdup(%p)=%p (%s:%d) <%d>", s, res, file, line, alloc_count);
      }
    }
    return res;
  }
}

#ifdef USE_FAIL_MALLOC
int
fail_malloc_check(size_t size, const char *file, int line, const char *func)
{
  if ((sen_fmalloc_file && strcmp(file, sen_fmalloc_file)) ||
      (sen_fmalloc_line && line != sen_fmalloc_line) ||
      (sen_fmalloc_func && strcmp(func, sen_fmalloc_func))) {
    return 1;
  }
  if (sen_fmalloc_prob && sen_fmalloc_prob >= rand()) {
    return 0;
  }
  return 1;
}

void *
sen_malloc_fail(sen_ctx *ctx, size_t size, const char* file, int line, const char *func)
{
  if (fail_malloc_check(size, file, line, func)) {
    return sen_malloc_default(ctx, size, file, line, func);
  } else {
    sen_index_expire();
    MERR("fail_malloc (%d) (%s:%d@%s) <%d>", size, file, line, func, alloc_count);
    return NULL;
  }
}

void *
sen_calloc_fail(sen_ctx *ctx, size_t size, const char* file, int line, const char *func)
{
  if (fail_malloc_check(size, file, line, func)) {
    return sen_calloc_default(ctx, size, file, line, func);
  } else {
    sen_index_expire();
    MERR("fail_calloc (%d) (%s:%d@%s) <%d>", size, file, line, func, alloc_count);
    return NULL;
  }
}

void *
sen_realloc_fail(sen_ctx *ctx, void *ptr, size_t size, const char* file, int line,
                 const char *func)
{
  if (fail_malloc_check(size, file, line, func)) {
    return sen_realloc_default(ctx, ptr, size, file, line, func);
  } else {
    sen_index_expire();
    MERR("fail_realloc (%p,%zu) (%s:%d@%s) <%d>", ptr, size, file, line, func, alloc_count);
    return NULL;
  }
}

char *
sen_strdup_fail(sen_ctx *ctx, const char *s, const char* file, int line, const char *func)
{
  if (fail_malloc_check(strlen(s), file, line, func)) {
    return sen_strdup_default(ctx, s, file, line, func);
  } else {
    sen_index_expire();
    MERR("fail_strdup(%p) (%s:%d@%s) <%d>", s, file, line, func, alloc_count);
    return NULL;
  }
}
#endif /* USE_FAIL_MALLOC */

sen_cell *
sen_cell_new(sen_ctx *ctx)
{
  sen_cell *o = NULL;
  if (ctx && ctx->impl) {
    sen_array_add(ctx, ctx->impl->objects, (void **)&o);
    if (o) {
      o->header.impl_flags = 0;
      ctx->impl->n_entries++;
    }
  }
  return o;
}

sen_cell *
sen_cell_cons(sen_ctx *ctx, sen_cell *a, sen_cell *b)
{
  if (!ctx) { return NULL; }
  {
    sen_cell *o;
    SEN_CELL_NEW(ctx, o);
    if (o) {
      o->header.type = SEN_CELL_LIST;
      o->header.impl_flags = 0;
      o->u.l.car = a;
      o->u.l.cdr = b;
    }
    return o;
  }
}

sen_cell *
sen_cell_alloc(sen_ctx *ctx, uint32_t size)
{
  if (!ctx) { return NULL; }
  {
    void *value = SEN_MALLOC(size + 1);
    if (value) {
      sen_cell *o = sen_cell_new(ctx);
      if (!ERRP(ctx, SEN_ERROR)) {
        o->header.impl_flags = SEN_OBJ_ALLOCATED;
        o->header.type = SEN_CELL_STR;
        o->u.b.size = size;
        o->u.b.value = value;
        return o;
      }
      SEN_FREE(value);
    } else {
      MERR("malloc(%d) failed", size + 1);
    }
    return NULL;
  }
}

void
sen_cell_clear(sen_ctx *ctx, sen_cell *o)
{
  if (!ctx || !ctx->impl) { return; }
  if (o->header.impl_flags & SEN_OBJ_ALLOCATED) {
    switch (o->header.type) {
    case SEN_SNIP :
      if (o->u.p.value) { sen_snip_close((sen_snip *)o->u.p.value); }
      break;
    case SEN_TABLE_HASH_KEY :
    case SEN_TABLE_PAT_KEY :
    case SEN_TABLE_NO_KEY :
      sen_obj_close(ctx, sen_ctx_get(ctx, o->u.o.id));
      break;
    case SEN_CELL_STR :
      if (o->u.b.value) {
        SEN_FREE(o->u.b.value);
      }
      break;
    case SEN_QUERY :
      if (o->u.p.value) { sen_query_close((sen_query *)o->u.p.value); }
      break;
    case SEN_VERSES :
      if (o->u.p.value) { sen_obj_close(ctx, o->u.p.value); }
      break;
    case SEN_PATSNIP :
      sen_obj_patsnip_spec_close(ctx, (patsnip_spec *)o->u.p.value);
      break;
    default :
      SEN_LOG(sen_log_warning, "obj_clear: invalid type(%x)", o->header.type);
      break;
    }
    o->header.impl_flags &= ~SEN_OBJ_ALLOCATED;
  }
}

/* don't handle error inside logger functions */

void
sen_ctx_log(sen_ctx *ctx, char *fmt, ...)
{
  va_list argp;
  va_start(argp, fmt);
  vsnprintf(ctx->errbuf, SEN_CTX_MSGSIZE - 1, fmt, argp);
  va_end(argp);
  ctx->errbuf[SEN_CTX_MSGSIZE - 1] = '\0';
}

static FILE *default_logger_fp = NULL;

static void
default_logger_func(int level, const char *time, const char *title,
                    const char *msg, const char *location, void *func_arg)
{
  const char slev[] = " EACewnid-";
  if (!default_logger_fp) {
    MUTEX_LOCK(sen_glock);
    if (!default_logger_fp) {
      default_logger_fp = fopen(SENNA_LOG_PATH, "a");
    }
    MUTEX_UNLOCK(sen_glock);
  }
  if (default_logger_fp) {
    fprintf(default_logger_fp, "%s|%c|%s %s %s\n",
            time, *(slev + level), title, msg, location);
    fflush(default_logger_fp);
  }
}

void
sen_logger_fin(void)
{
  if (default_logger_fp) {
    fclose(default_logger_fp);
    default_logger_fp = NULL;
  }
}

static sen_logger_info default_logger = {
  SEN_LOG_DEFAULT_LEVEL,
  SEN_LOG_TIME|SEN_LOG_MESSAGE,
  default_logger_func
};

static const sen_logger_info *sen_logger = &default_logger;

sen_rc
sen_logger_info_set(const sen_logger_info *info)
{
  if (info) {
    sen_logger = info;
  } else {
    sen_logger = &default_logger;
  }
  return sen_success;
}

int
sen_logger_pass(sen_log_level level)
{
  return level <= sen_logger->max_level;
}

#define TBUFSIZE SEN_TIMEVAL_STR_SIZE
#define MBUFSIZE 0x1000
#define LBUFSIZE 0x400

void
sen_logger_put(sen_log_level level,
               const char *file, int line, const char *func, char *fmt, ...)
{
  if (level <= sen_logger->max_level) {
    char tbuf[TBUFSIZE];
    char mbuf[MBUFSIZE];
    char lbuf[LBUFSIZE];
    if (sen_logger->flags & SEN_LOG_TIME) {
      sen_timeval tv;
      if (sen_timeval_now(&tv) || sen_timeval2str(&tv, tbuf)) { tbuf[0] = '\0'; }
    } else {
      tbuf[0] = '\0';
    }
    if (sen_logger->flags & SEN_LOG_MESSAGE) {
      va_list argp;
      va_start(argp, fmt);
      vsnprintf(mbuf, MBUFSIZE - 1, fmt, argp);
      va_end(argp);
      mbuf[MBUFSIZE - 1] = '\0';
    } else {
      mbuf[0] = '\0';
    }
    if (sen_logger->flags & SEN_LOG_LOCATION) {
      snprintf(lbuf, LBUFSIZE - 1, "%04x %s:%d %s()",
               getpid(), file, line, func);
      lbuf[LBUFSIZE - 1] = '\0';
    } else {
      lbuf[0] = '\0';
    }
    if (sen_logger->func) {
      sen_logger->func(level, tbuf, "", mbuf, lbuf, sen_logger->func_arg);
    } else {
      default_logger_func(level, tbuf, "", mbuf, lbuf, sen_logger->func_arg);
    }
  }
}

void
sen_assert(int cond, const char* file, int line, const char* func)
{
  if (!cond) {
    SEN_LOG(sen_log_warning, "ASSERT fail on %s %s:%d", func, file, line);
  }
}
