/* Copyright(C) 2004-2007 Brazil

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#ifndef SEN_CTX_H
#define SEN_CTX_H

#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif /* HAVE_ERRNO_H */

#ifdef HAVE_EXECINFO_H
#include <execinfo.h>
#endif /* HAVE_EXECINFO_H */

#ifndef SEN_COM_H
#include "com.h"
#endif /* SEN_COM_H */

#ifndef SEN_IO_H
#include "io.h"
#endif /* SEN_IO_H */

#ifdef __cplusplus
extern "C" {
#endif

/**** api in/out ****/

#define SEN_API_ENTER \
{\
  (ctx)->errlvl = SEN_OK;\
  (ctx)->rc = sen_success;\
  if ((ctx)->seqno & 1) {\
    (ctx)->subno++;\
  } else {\
    (ctx)->seqno++;\
  }\
}

/* CAUTION!! : pass only variables or constants as r */
#define SEN_API_RETURN(r) \
{\
  if (ctx->subno) {\
    ctx->subno--;\
  } else {\
    ctx->seqno++;\
  }\
  return r;\
}

/**** error handling ****/

#define SEN_OP_T0LVL 0
#define SEN_OP_ERR0  1

#define  SEN_EMERG  sen_log_emerg
#define  SEN_ALERT  sen_log_alert
#define  SEN_CRIT   sen_log_crit
#define  SEN_ERROR  sen_log_error
#define  SEN_WARN   sen_log_warning
#define  SEN_OK     sen_log_notice

#define ERRCLR(ctx) do {\
  if (ctx) {\
    ((sen_ctx *)ctx)->errlvl = SEN_OK;\
    ((sen_ctx *)ctx)->rc = sen_success;\
  }\
  sen_gctx.errlvl = SEN_OK;\
  sen_gctx.rc = sen_success;\
} while (0)

#ifdef HAVE_EXECINFO_H
#define BACKTRACE(ctx) ((ctx)->ntrace = (unsigned char)backtrace((ctx)->trace, 16))
#else /* HAVE_EXECINFO_H */
#define BACKTRACE(ctx)
#endif /* HAVE_EXECINFO_H */

void sen_ctx_impl_err(sen_ctx *ctx);

#define ERRSET(ctx,lvl,r,...) do {\
  sen_ctx *ctx_ = (sen_ctx *)ctx;\
  ctx_->errlvl = (lvl);\
  ctx_->rc = (r);\
  ctx_->errfile = __FILE__;\
  ctx_->errline = __LINE__;\
  ctx_->errfunc = __FUNCTION__;\
  sen_ctx_impl_err(ctx);\
  SEN_LOG(lvl, __VA_ARGS__);\
  sen_ctx_log((sen_ctx *)ctx, __VA_ARGS__);\
  BACKTRACE(ctx);\
} while (0)

#define ERRP(ctx,lvl) (((ctx) && ((sen_ctx *)(ctx))->errlvl <= (lvl)) || (sen_gctx.errlvl <= (lvl)))

#define QLERR(...) do {\
  ERRSET(ctx, SEN_WARN, sen_invalid_argument, __VA_ARGS__);\
  return F;\
} while (0)

#define QLASSERT(expr) do {\
  if (!(expr)) { QLERR("syntax error"); }\
} while (0)

#define ERR(rc,...) ERRSET(ctx, SEN_ERROR, (rc),  __VA_ARGS__)
#define MERR(...) ERRSET(ctx, SEN_ALERT, sen_memory_exhausted,  __VA_ARGS__)
#define SERR(str) ERR(sen_other_error, "syscall error '%s' (%s)", str, strerror(errno))

#define GERR(rc,...) ERRSET(&sen_gctx, SEN_ERROR, (rc),  __VA_ARGS__)
#define GMERR(...) ERRSET(&sen_gctx, SEN_ALERT, sen_memory_exhausted,  __VA_ARGS__)
#define GSERR(str) GERR(sen_other_error, "syscall error '%s' (%s)", str, strerror(errno))

#define SEN_MALLOC(s) sen_malloc(ctx,s,__FILE__,__LINE__,__FUNCTION__)
#define SEN_CALLOC(s) sen_calloc(ctx,s,__FILE__,__LINE__,__FUNCTION__)
#define SEN_REALLOC(p,s) sen_realloc(ctx,p,s,__FILE__,__LINE__,__FUNCTION__)
#define SEN_STRDUP(s) sen_strdup(ctx,s,__FILE__,__LINE__,__FUNCTION__)
#define SEN_GMALLOC(s) sen_malloc(&sen_gctx,s,__FILE__,__LINE__,__FUNCTION__)
#define SEN_GCALLOC(s) sen_calloc(&sen_gctx,s,__FILE__,__LINE__,__FUNCTION__)
#define SEN_GREALLOC(p,s) sen_realloc(&sen_gctx,p,s,__FILE__,__LINE__,__FUNCTION__)
#define SEN_GSTRDUP(s) sen_strdup(&sen_gctx,s,__FILE__,__LINE__,__FUNCTION__)
#define SEN_FREE(p) sen_free(ctx,p,__FILE__,__LINE__)
#define SEN_MALLOCN(t,n) ((t *)(SEN_MALLOC(sizeof(t) * (n))))
#define SEN_GFREE(p) sen_free(&sen_gctx,p,__FILE__,__LINE__)
#define SEN_GMALLOCN(t,n) ((t *)(SEN_GMALLOC(sizeof(t) * (n))))

#ifdef DEBUG
#define SEN_ASSERT(s) sen_assert((s),__FILE__,__LINE__,__FUNCTION__)
#else
#define SEN_ASSERT(s)
#endif

#define SEN_CTX_ALLOC(ctx,s,f) sen_ctx_alloc(ctx,s,f,__FILE__,__LINE__,__FUNCTION__)
#define SEN_CTX_FREE(ctx,p) sen_ctx_free(ctx,p,__FILE__,__LINE__,__FUNCTION__)
#define SEN_CTX_ALLOC_L(ctx,s,f) sen_ctx_alloc_lifo(ctx,s,f,__FILE__,__LINE__,__FUNCTION__)
#define SEN_CTX_FREE_L(ctx,p) sen_ctx_free_lifo(ctx,p,__FILE__,__LINE__,__FUNCTION__)

void *sen_ctx_alloc(sen_ctx *ctx, size_t size, int flags,
                     const char* file, int line, const char *func);
void sen_ctx_free(sen_ctx *ctx, void *ptr,
                  const char* file, int line, const char *func);
void *sen_ctx_alloc_lifo(sen_ctx *ctx, size_t size, int flags,
                         const char* file, int line, const char *func);
void sen_ctx_free_lifo(sen_ctx *ctx, void *ptr,
                       const char* file, int line, const char *func);

#ifdef USE_DYNAMIC_MALLOC_CHANGE
typedef void *(*sen_malloc_func) (sen_ctx *ctx, size_t size,
                                  const char *file, int line, const char *func);
typedef void *(*sen_calloc_func) (sen_ctx *ctx, size_t size,
                                  const char *file, int line, const char *func);
typedef void *(*sen_realloc_func) (sen_ctx *ctx, void *ptr, size_t size,
                                   const char *file, int line, const char *func);
typedef char *(*sen_strdup_func) (sen_ctx *ctx, const char *string,
                                  const char *file, int line, const char *func);
sen_malloc_func sen_ctx_get_malloc(sen_ctx *ctx);
void sen_ctx_set_malloc(sen_ctx *ctx, sen_malloc_func malloc_func);
sen_calloc_func sen_ctx_get_calloc(sen_ctx *ctx);
void sen_ctx_set_calloc(sen_ctx *ctx, sen_calloc_func calloc_func);
sen_realloc_func sen_ctx_get_realloc(sen_ctx *ctx);
void sen_ctx_set_realloc(sen_ctx *ctx, sen_realloc_func realloc_func);
sen_strdup_func sen_ctx_get_strdup(sen_ctx *ctx);
void sen_ctx_set_strdup(sen_ctx *ctx, sen_strdup_func strdup_func);

void *sen_malloc(sen_ctx *ctx, size_t size, const char* file, int line, const char *func);
void *sen_calloc(sen_ctx *ctx, size_t size, const char* file, int line, const char *func);
void *sen_realloc(sen_ctx *ctx, void *ptr, size_t size, const char* file, int line, const char *func);
char *sen_strdup(sen_ctx *ctx, const char *s, const char* file, int line, const char *func);
#else
#  define sen_malloc sen_malloc_default
#  define sen_calloc sen_calloc_default
#  define sen_realloc sen_realloc_default
#  define sen_strdup sen_strdup_default
#endif

void *sen_malloc_default(sen_ctx *ctx, size_t size, const char* file, int line, const char *func);
void *sen_calloc_default(sen_ctx *ctx, size_t size, const char* file, int line, const char *func);
void *sen_realloc_default(sen_ctx *ctx, void *ptr, size_t size, const char* file, int line, const char *func);
char *sen_strdup_default(sen_ctx *ctx, const char *s, const char* file, int line, const char *func);

#ifdef USE_FAIL_MALLOC
void *sen_malloc_fail(sen_ctx *ctx, size_t size, const char* file, int line, const char *func);
void *sen_calloc_fail(sen_ctx *ctx, size_t size, const char* file, int line, const char *func);
void *sen_realloc_fail(sen_ctx *ctx, void *ptr, size_t size, const char* file, int line, const char *func);
char *sen_strdup_fail(sen_ctx *ctx, const char *s, const char* file, int line, const char *func);
#endif

void sen_free(sen_ctx *ctx, void *ptr, const char* file, int line);

void sen_assert(int cond, const char* file, int line, const char* func);

void sen_index_expire(void);

/**** sen_ctx ****/

extern sen_ctx sen_gctx;
extern int sen_pagesize;
extern sen_mutex sen_glock;
extern uint32_t sen_gtick;

typedef struct {
  int32_t tv_sec;
  int32_t tv_usec;
} sen_timeval;

#ifndef SEN_TIMEVAL_STR_SIZE
#define SEN_TIMEVAL_STR_SIZE 0x100
#endif /* SEN_TIMEVAL_STR_SIZE */
#ifndef SEN_TIMEVAL_STR_FORMAT
#define SEN_TIMEVAL_STR_FORMAT "%04d-%02d-%02d %02d:%02d:%02d.%06d"
#endif /* SEN_TIMEVAL_STR_FORMAT */

sen_rc sen_timeval_now(sen_timeval *tv);
sen_rc sen_timeval2str(sen_timeval *tv, char *buf);
sen_rc sen_str2timeval(const char *str, uint32_t str_len, sen_timeval *tv);

void sen_ctx_log(sen_ctx *ctx, char *fmt, ...);

/**** receive handler ****/

void sen_ctx_recv_handler_set(sen_ctx *c, void (*func)(sen_ctx *, int, void *),
                              void *func_arg);

void sen_ctx_concat_func(sen_ctx *ctx, int flags, void *dummy);
void sen_ctx_stream_out_func(sen_ctx *c, int flags, void *stream);

#ifdef __cplusplus
}
#endif

#endif /* SEN_CTX_H */
