/* -*- c-basic-offset: 2; coding: utf-8 -*- */
/* Copyright(C) 2008 Brazil

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "senna_in.h"
#include "db.h"
#include "hash.h"
#include "pat.h"
#include "ii.h"
#include "ql.h"
#include <string.h>

#define NEXT_ADDR(p) (((byte *)(p)) + sizeof *(p))

#define OBJ_DB(obj) ((sen_obj_db *)obj)

#define WITH_NORMALIZE(table,key,key_size,block) {\
  if ((table)->obj.flags & SEN_OBJ_KEY_NORMALIZE) {\
    sen_nstr *nstr;\
    if ((nstr = sen_nstr_open(key, key_size, (table)->encoding, 0))) {\
      char *key = nstr->norm;\
      unsigned key_size = nstr->norm_blen;\
      block\
      sen_nstr_close(nstr);\
    }\
  } else {\
    block\
  }\
}

struct _sen_space {
  sen_obj_db obj;
  sen_pat *keys;
  sen_ja *specs;
  sen_tiny_array values;
  sen_mutex lock;
};

static sen_rc sen_obj_db_init(sen_ctx *ctx, sen_obj *space, sen_id id, sen_obj_db *obj);

inline static void
gen_pathname(const char *path, char *buffer, int fno)
{
  size_t len = strlen(path);
  memcpy(buffer, path, len);
  if (fno >= 0) {
    buffer[len] = '.';
    sen_itoh(fno, buffer + len + 1, 7);
  } else {
    buffer[len] = '\0';
  }
}

#define SPACE_P(s) ((s) && ((sen_space *)s)->obj.header.type == SEN_SPACE)
#define PERSISTENT_SPACE_P(s) (((sen_space *)s)->specs)

sen_obj *
sen_space_create(sen_ctx *ctx, const char *path, sen_encoding encoding)
{
  sen_space *s;
  SEN_API_ENTER;
  if (!path || strlen(path) <= PATH_MAX - 14) {
    if ((s = SEN_MALLOC(sizeof(sen_space)))) {
      sen_tiny_array_init(&s->values, ctx, sizeof(sen_obj *),
                          SEN_TINY_ARRAY_CLEAR|
                          SEN_TINY_ARRAY_THREADSAFE|
                          SEN_TINY_ARRAY_USE_MALLOC);
      if ((s->keys = sen_pat_create(ctx, path, SEN_PAT_MAX_KEY_SIZE, 0,
                                    SEN_OBJ_KEY_VAR_SIZE, encoding))) {
        MUTEX_INIT(s->lock);
        SEN_OBJ_DB_SET_TYPE(s, SEN_SPACE);
        sen_obj_db_init(ctx, NULL, SEN_ID_NIL, &s->obj);
        // prepare builtin classes and load builtin plugins.
        if (path) {
          char buffer[PATH_MAX];
          gen_pathname(path, buffer, 0);
          if ((s->specs = sen_ja_create(ctx, buffer, 65536, 0))) {
            ctx->impl->db = (sen_obj *)s;
            SEN_API_RETURN((sen_obj *)s);
          } else {
            ERR(sen_memory_exhausted, "ja create failed");
          }
        } else {
          s->specs = NULL;
          ctx->impl->db = (sen_obj *)s;
          SEN_API_RETURN((sen_obj *)s);
        }
        sen_pat_close(ctx, s->keys);
        sen_pat_remove(ctx, path);
      } else {
        ERR(sen_memory_exhausted, "s->keys create failed");
      }
      sen_tiny_array_fin(&s->values);
      SEN_FREE(s);
    } else {
      ERR(sen_memory_exhausted, "sen_space alloc failed");
    }
  } else {
    ERR(sen_invalid_argument, "too long path");
  }
  SEN_API_RETURN(NULL);
}

sen_obj *
sen_space_open(sen_ctx *ctx, const char *path)
{
  sen_space *s;
  SEN_API_ENTER;
  if (path && strlen(path) <= PATH_MAX - 14) {
    if ((s = SEN_MALLOC(sizeof(sen_space)))) {
      sen_tiny_array_init(&s->values, ctx, sizeof(sen_obj *),
                          SEN_TINY_ARRAY_CLEAR|
                          SEN_TINY_ARRAY_THREADSAFE|
                          SEN_TINY_ARRAY_USE_MALLOC);
      if ((s->keys = sen_pat_open(ctx, path))) {
        char buffer[PATH_MAX];
        gen_pathname(path, buffer, 0);
        if ((s->specs = sen_ja_open(ctx, buffer))) {
          MUTEX_INIT(s->lock);
          SEN_OBJ_DB_SET_TYPE(s, SEN_SPACE);
          sen_obj_db_init(ctx, NULL, SEN_ID_NIL, &s->obj);
          // prepare builtin classes and load builtin plugins.
          ctx->impl->db = (sen_obj *)s;
          SEN_API_RETURN((sen_obj *)s);
        } else {
          ERR(sen_memory_exhausted, "ja open failed");
        }
        sen_pat_close(ctx, s->keys);
      } else {
        ERR(sen_memory_exhausted, "s->keys open failed");
      }
      sen_tiny_array_fin(&s->values);
      SEN_FREE(s);
    } else {
      ERR(sen_memory_exhausted, "sen_space alloc failed");
    }
  } else {
    ERR(sen_invalid_argument, "inappropriate path");
  }
  SEN_API_RETURN(NULL);
}

/* s must be validated by caller */
sen_rc
sen_space_close(sen_ctx *ctx, sen_space *s)
{
  sen_id id;
  sen_obj **vp;
  if (!s) { return sen_invalid_argument; }
  SEN_API_ENTER;
  SEN_TINY_ARRAY_EACH(&s->values, 1, sen_pat_curr_id(ctx, s->keys), id, vp, {
    if (*vp) { sen_obj_close(ctx, *vp); }
  });
  sen_tiny_array_fin(&s->values);
  sen_pat_close(ctx, s->keys);
  if (s->specs) { sen_ja_close(ctx, s->specs); }
  SEN_FREE(s);
  SEN_API_RETURN(sen_success);
}

static sen_id sen_obj_register(sen_ctx *ctx, sen_obj *space,
                               const char *name, unsigned name_size);
static sen_rc sen_obj_delete_by_id(sen_ctx *ctx, sen_obj *space, sen_id id, int removep);

sen_obj *
sen_ctx_lookup(sen_ctx *ctx, const char *name, unsigned name_size)
{
  sen_id id;
  sen_search_flags flags = 0;
  sen_obj *obj = NULL;
  sen_obj *space = ctx->impl->db;
  SEN_API_ENTER;
  if (SPACE_P(space)) {
    sen_space *s = (sen_space *)space;
    if ((id = sen_pat_lookup(ctx, s->keys, name, name_size, NULL, &flags))) {
      obj = sen_ctx_get(ctx, id);
    }
  }
  SEN_API_RETURN(obj);
}

sen_rc
sen_ctx_use(sen_ctx *ctx, sen_obj *space)
{
  sen_rc rc = sen_success;
  SEN_API_ENTER;
  if (ctx->impl && (!space || SPACE_P(space))) {
    ctx->impl->db = space;
  } else {
    rc = sen_invalid_argument;
  }
  SEN_API_RETURN(rc);
}

sen_obj *
sen_ctx_space(sen_ctx *ctx)
{
  return (ctx->impl) ? ctx->impl->db : NULL;
}

#define SEN_PROC_INIT_PREFIX "sen_init_"

sen_rc
sen_space_load(sen_ctx *ctx, sen_obj *space, const char *path)
{
  sen_id id;
  SEN_API_ENTER;
  if (SPACE_P(space)) {
    if ((id = sen_dl_open(ctx, path))) {
      sen_proc_init_func *func;
      const char *p;
      char buffer[PATH_MAX];
      for (p = path + strlen(path); path < p; p--) {
        if (*p == PATH_SEPARATOR[0]) { p++; break; }
      }
      strcpy(buffer, SEN_PROC_INIT_PREFIX);
      strcat(buffer, p);
      if ((func = sen_dl_sym(ctx, id, buffer))) {
        ctx->rc = func(ctx, path);
      } else {
        ERR(sen_invalid_format, "init_func not found(%s)", buffer);
      }
    }
  } else {
    ERR(sen_invalid_argument, "invalid space assigned");
  }
  SEN_API_RETURN(ctx->rc);
}

sen_obj *
sen_space_keys(sen_obj *s)
{
  return (sen_obj *)(((sen_space *)s)->keys);
}

#define SEN_DB_DELIMITER '.'

static sen_rc
check_name(sen_ctx *ctx, const char *name, unsigned int name_size)
{
  const char *name_end = name + name_size;
  while (name < name_end) {
    if (*name++ == SEN_DB_DELIMITER) {
      return sen_invalid_argument;
    }
  }
  return sen_success;
}

#define SEN_TYPE_SIZE(type) ((type)->range)

struct _sen_type {
  sen_obj_db obj;
};

sen_obj *
sen_type_create(sen_ctx *ctx, const char *name, unsigned name_size,
                sen_obj_flags flags, unsigned int size)
{
  sen_id id;
  struct _sen_type *res = NULL;
  sen_obj *space = ctx->impl->db;
  SEN_API_ENTER;
  if (check_name(ctx, name, name_size)) {
    ERR(sen_invalid_argument, "name contains '%c'", SEN_DB_DELIMITER);
    SEN_API_RETURN(NULL);
  }
  if (!SPACE_P(space)) {
    ERR(sen_invalid_argument, "invalid space assigned");
    SEN_API_RETURN(NULL);
  }
  id = sen_obj_register(ctx, space, name, name_size);
  if (id && (res = SEN_MALLOC(sizeof(sen_obj_db)))) {
    SEN_OBJ_DB_SET_TYPE(res, SEN_TYPE);
    res->obj.header.flags = flags;
    SEN_TYPE_SIZE(&res->obj) = size;
    if (sen_obj_db_init(ctx, space, id, OBJ_DB(res))) {
      // sen_obj_delete(ctx, space, id);
      SEN_FREE(res);
      SEN_API_RETURN(NULL);
    }
  }
  SEN_API_RETURN((sen_obj *)res);
}

typedef struct {
  sen_obj_header header;
  sen_id range;
} sen_obj_spec;

static sen_obj *
sen_type_open(sen_ctx *ctx, sen_obj_spec *spec)
{
  struct _sen_type *res;
  res = SEN_MALLOC(sizeof(struct _sen_type));
  if (res) {
    SEN_OBJ_DB_SET_TYPE(res, SEN_TYPE);
    res->obj.header = spec->header;
    SEN_TYPE_SIZE(&res->obj) = SEN_TYPE_SIZE(spec);
  }
  return (sen_obj *)res;
}

struct _sen_proc {
  sen_obj_db obj;
  sen_proc_func *funcs[3];
};

sen_obj *
sen_proc_create(sen_ctx *ctx,
                const char *name, unsigned name_size,
                const char *path, sen_proc_type type,
                sen_proc_func *init, sen_proc_func *next, sen_proc_func *fin)
{
  sen_proc *res;
  sen_id id = SEN_ID_NIL;
  sen_id domain;
  sen_search_flags f = SEN_TABLE_ADD;
  sen_obj *space = ctx->impl->db;
  SEN_API_ENTER;
  domain = path ? sen_dl_lookup(ctx, path) : SEN_ID_NIL;
  if (check_name(ctx, name, name_size)) {
    ERR(sen_invalid_argument, "name contains '%c'", SEN_DB_DELIMITER);
    SEN_API_RETURN(NULL);
  }
  if (!SPACE_P(space)) {
    ERR(sen_invalid_argument, "invalid space assigned");
    SEN_API_RETURN(NULL);
  }
  if (name && name_size) {
    sen_space *s = (sen_space *)space;
    if (!(id = sen_pat_lookup(ctx, s->keys, name, name_size, NULL, &f))) {
      ERR(sen_memory_exhausted, "sen_pat_lookup failed");
      SEN_API_RETURN(NULL);
    }
    if (!(f & SEN_TABLE_ADDED)) {
      ERR(sen_invalid_argument, "already used name");
      SEN_API_RETURN(NULL);
    }
  }
  if ((res = SEN_MALLOCN(sen_proc, 1))) {
    SEN_OBJ_DB_SET_TYPE(res, SEN_PROC);
    res->obj.space = space;
    res->obj.id = id;
    res->obj.header.domain = domain;
    res->obj.range = SEN_ID_NIL;
    res->funcs[PROC_INIT] = init;
    res->funcs[PROC_NEXT] = next;
    res->funcs[PROC_FIN] = fin;
    if (f & SEN_TABLE_ADDED) {
      if (sen_obj_db_init(ctx, space, id, OBJ_DB(res))) {
        // sen_obj_delete(ctx, space, id);
        SEN_FREE(res);
        SEN_API_RETURN(NULL);
      }
    }
  }
  SEN_API_RETURN((sen_obj *)res);
}

/* sen_table */

static void
calc_rec_size(sen_obj_flags flags, uint32_t *max_n_subrecs,
              uint8_t *subrec_size, uint8_t *subrec_offset,
              uint32_t *key_size, uint32_t *value_size)
{
  *max_n_subrecs = 0;
  *subrec_size = 0;
  *subrec_offset = 0;
  if (flags & SEN_OBJ_WITH_SUBREC) {
    switch (flags & SEN_OBJ_UNIT_MASK) {
    case SEN_OBJ_UNIT_DOCUMENT_NONE :
      break;
    case SEN_OBJ_UNIT_DOCUMENT_SECTION :
      *max_n_subrecs = *value_size;
      *subrec_offset = sizeof(sen_id);
      *subrec_size = sizeof(uint32_t);
      break;
    case SEN_OBJ_UNIT_DOCUMENT_POSITION :
      *max_n_subrecs = *value_size;
      *subrec_offset = sizeof(sen_id);
      *subrec_size = sizeof(uint32_t) + sizeof(uint32_t);
      break;
    case SEN_OBJ_UNIT_SECTION_NONE :
      *key_size += sizeof(uint32_t);
      break;
    case SEN_OBJ_UNIT_SECTION_POSITION :
      *key_size += sizeof(uint32_t);
      *max_n_subrecs = *value_size;
      *subrec_offset = sizeof(sen_id) + sizeof(uint32_t);
      *subrec_size = sizeof(uint32_t);
      break;
    case SEN_OBJ_UNIT_POSITION_NONE :
      *key_size += sizeof(uint32_t) + sizeof(uint32_t);
      break;
    case SEN_OBJ_UNIT_USERDEF_DOCUMENT :
      *max_n_subrecs = *value_size;
      *subrec_size = sizeof(sen_id);
      break;
    case SEN_OBJ_UNIT_USERDEF_SECTION :
      *max_n_subrecs = *value_size;
      *subrec_size = sizeof(sen_id) + sizeof(uint32_t);
      break;
    case SEN_OBJ_UNIT_USERDEF_POSITION :
      *max_n_subrecs = *value_size;
      *subrec_size = sizeof(sen_id) + sizeof(uint32_t) + sizeof(uint32_t);
      break;
    }
    *value_size = (uintptr_t)SEN_RSET_SUBRECS_NTH((((sen_rset_recinfo *)0)->subrecs),
                                                  *subrec_size, *max_n_subrecs);
  }
}

sen_obj *
sen_table_create(sen_ctx *ctx, const char *name, unsigned name_size,
                 const char *path, sen_obj_flags flags,
                 sen_obj *key_type, unsigned value_size, sen_encoding encoding)
{
  sen_id id;
  unsigned char impl_flags = 0;
  sen_id domain = SEN_ID_NIL;
  uint32_t key_size, max_n_subrecs;
  uint8_t subrec_size, subrec_offset;
  sen_obj *res = NULL;
  sen_obj *space = ctx->impl->db;
  char buffer[PATH_MAX];
  SEN_API_ENTER;
  if (check_name(ctx, name, name_size)) {
    ERR(sen_invalid_argument, "name contains '%c'", SEN_DB_DELIMITER);
    SEN_API_RETURN(NULL);
  }
  if (!SPACE_P(space)) {
    ERR(sen_invalid_argument, "invalid space assigned");
    SEN_API_RETURN(NULL);
  }
  if (key_type) {
    domain = OBJ_DB(key_type)->id;
    switch (key_type->header.type) {
    case SEN_TYPE :
      {
        sen_obj_db *t = (sen_obj_db *)key_type;
        flags |= t->header.flags;
        key_size = SEN_TYPE_SIZE(t);
      }
      break;
    case SEN_TABLE_HASH_KEY :
    case SEN_TABLE_NO_KEY :
    case SEN_TABLE_PAT_KEY :
      key_size = sizeof(sen_id);
      break;
    default :
      /*
      if (key_type == sen_type_any) {
        key_size = sizeof(sen_id) + sizeof(sen_id);
      }
      */
      key_size = sizeof(sen_id);
    }
  } else {
    key_size = (flags & SEN_OBJ_KEY_VAR_SIZE) ? SEN_TABLE_MAX_KEY_SIZE : sizeof(sen_id);
  }
  id = sen_obj_register(ctx, space, name, name_size);
  if (ERRP(ctx, SEN_ERROR)) { SEN_API_RETURN(NULL);  }
  if (SEN_OBJ_PERSISTENT & flags) {
    if (!path) {
      if (PERSISTENT_SPACE_P(space)) {
        gen_pathname(((sen_space *)space)->keys->io->path, buffer, id);
        path = buffer;
      } else {
        ERR(sen_invalid_argument, "path not assigend for persistent table");
        SEN_API_RETURN(NULL);
      }
    } else {
      impl_flags |= SEN_OBJ_CUSTOM_NAME;
    }
  } else {
    if (path) {
      ERR(sen_invalid_argument, "path assigend for temporary table");
      SEN_API_RETURN(NULL);
    }
  }
  calc_rec_size(flags, &max_n_subrecs, &subrec_size,
                &subrec_offset, &key_size, &value_size);
  switch (flags & SEN_OBJ_TABLE_TYPE_MASK) {
  case SEN_OBJ_TABLE_HASH_KEY :
    res = (sen_obj *)sen_hash_create(ctx, path, key_size, value_size, flags, encoding);
    break;
  case SEN_OBJ_TABLE_PAT_KEY :
    res = (sen_obj *)sen_pat_create(ctx, path, key_size, value_size, flags, encoding);
    break;
  case SEN_OBJ_TABLE_NO_KEY :
    res = (sen_obj *)sen_array_create(ctx, path, value_size, flags);
    break;
  }
  if (res) {
    OBJ_DB(res)->header.impl_flags = impl_flags;
    OBJ_DB(res)->header.domain = domain;
    OBJ_DB(res)->range = SEN_ID_NIL;
    OBJ_DB(res)->max_n_subrecs = max_n_subrecs;
    OBJ_DB(res)->subrec_size = subrec_size;
    OBJ_DB(res)->subrec_offset = subrec_offset;
    if (sen_obj_db_init(ctx, space, id, OBJ_DB(res))) {
      sen_obj_close(ctx, res);
      sen_obj_remove(ctx, path);
      res = NULL;
    }
  } else {
    sen_obj_delete_by_id(ctx, space, id, 1);
  }
  SEN_API_RETURN(res);
}

sen_obj *
sen_table_open(sen_ctx *ctx, const char *name, unsigned name_size, const char *path)
{
  sen_obj *space = ctx->impl->db;
  SEN_API_ENTER;
  if (!SPACE_P(space)) {
    ERR(sen_invalid_argument, "invalid space assigned");
    SEN_API_RETURN(NULL);
  } else {
    sen_obj *res = sen_ctx_lookup(ctx, name, name_size);
    if (res) {
      const char *path2 = sen_obj_path(ctx, res);
      if (path && (!path2 || strcmp(path, path2))) { SEN_API_RETURN(NULL); }
    } else if (path) {
      uint32_t type = sen_io_detect_type(ctx, path);
      if (!type) { SEN_API_RETURN(NULL); }
      switch (type) {
      case SEN_TABLE_HASH_KEY :
        res = (sen_obj *)sen_hash_open(ctx, path);
        break;
      case SEN_TABLE_PAT_KEY :
        res = (sen_obj *)sen_pat_open(ctx, path);
        break;
      case SEN_TABLE_NO_KEY :
        res = (sen_obj *)sen_array_open(ctx, path);
        break;
      }
      if (res) {
        sen_id id = sen_obj_register(ctx, space, name, name_size);
        res->header.impl_flags |= SEN_OBJ_CUSTOM_NAME;
        res->header.domain = SEN_ID_NIL; /* unknown */
        sen_obj_db_init(ctx, space, id, OBJ_DB(res));
      }
    }
    SEN_API_RETURN(res);
  }
}

sen_id
sen_table_lookup(sen_ctx *ctx, sen_obj *table, const void *key, unsigned key_size,
                 sen_search_flags *flags)
{
  sen_id id = SEN_ID_NIL;
  SEN_API_ENTER;
  if (table) {
    switch (table->header.type) {
    case SEN_TABLE_PAT_KEY :
      WITH_NORMALIZE((sen_pat *)table, key, key_size, {
        if (*flags & SEN_SEARCH_LCP) {
          id = sen_pat_lcp_search(ctx, (sen_pat *)table, key, key_size);
        } else {
          id = sen_pat_lookup(ctx, (sen_pat *)table, key, key_size, NULL, flags);
        }
      });
      break;
    case SEN_TABLE_HASH_KEY :
      WITH_NORMALIZE((sen_hash *)table, key, key_size, {
        id = sen_hash_lookup(ctx, (sen_hash *)table, key, key_size, NULL, flags);
      });
      break;
    }
  }
  SEN_API_RETURN(id);
}

sen_id
sen_table_add(sen_ctx *ctx, sen_obj *table)
{
  sen_id id = SEN_ID_NIL;
  SEN_API_ENTER;
  if (table) {
    switch (table->header.type) {
    case SEN_TABLE_NO_KEY :
      id = sen_array_add(ctx, (sen_array *)table, NULL);
      break;
    default :
      break;
    }
  }
  SEN_API_RETURN(id);
}

int
sen_table_get_key(sen_ctx *ctx, sen_obj *table, sen_id id, void *keybuf, int buf_size)
{
  int r = 0;
  SEN_API_ENTER;
  if (table) {
    switch (table->header.type) {
    case SEN_TABLE_HASH_KEY :
      r = sen_hash_get_key(ctx, (sen_hash *)table, id, keybuf, buf_size);
      break;
    case SEN_TABLE_PAT_KEY :
      r = sen_pat_get_key(ctx, (sen_pat *)table, id, keybuf, buf_size);
      break;
    case SEN_TABLE_NO_KEY :
      {
        sen_array *a = (sen_array *)table;
        if (a->obj.header.domain) {
          if (buf_size >= a->value_size) {
            r = sen_array_get_value(ctx, a, id, keybuf);
          } else {
            r = a->value_size;
          }
        }
      }
      break;
    }
  }
  SEN_API_RETURN(r);
}

int
sen_table_get_key2(sen_ctx *ctx, sen_obj *table, sen_id id, sen_obj *bulk)
{
  int r = 0;
  SEN_API_ENTER;
  if (table) {
    switch (table->header.type) {
    case SEN_TABLE_HASH_KEY :
      r = sen_hash_get_key2(ctx, (sen_hash *)table, id, bulk);
      break;
    case SEN_TABLE_PAT_KEY :
      r = sen_pat_get_key2(ctx, (sen_pat *)table, id, bulk);
      break;
    case SEN_TABLE_NO_KEY :
      {
        sen_array *a = (sen_array *)table;
        if (a->obj.header.domain) {
          if (!sen_bulk_space(ctx, bulk, a->value_size)) {
            r = sen_array_get_value(ctx, a, id, bulk->u.b.curr - a->value_size);
          }
        }
      }
      break;
    }
  }
  SEN_API_RETURN(r);
}

sen_rc
sen_table_delete(sen_ctx *ctx, sen_obj *table, const void *key, unsigned key_size)
{
  sen_rc rc = sen_invalid_argument;
  SEN_API_ENTER;
  if (table) {
    switch (table->header.type) {
    case SEN_SPACE :
      /* todo : delete tables and columns from space */
      break;
    case SEN_TABLE_PAT_KEY :
      WITH_NORMALIZE((sen_pat *)table, key, key_size, {
        rc = sen_pat_delete(ctx, (sen_pat *)table, key, key_size, NULL);
      });
      break;
    case SEN_TABLE_HASH_KEY :
      WITH_NORMALIZE((sen_hash *)table, key, key_size, {
        rc = sen_hash_delete(ctx, (sen_hash *)table, key, key_size, NULL);
      });
      break;
    }
    /* todo : clear_all_column_values */
  }
  SEN_API_RETURN(rc);
}

sen_rc
_sen_table_delete_by_id(sen_ctx *ctx, sen_obj *table, sen_id id,
                       sen_table_delete_optarg *optarg)
{
  if (table) {
    // todo : support optarg
    switch (table->header.type) {
    case SEN_TABLE_PAT_KEY :
      return sen_pat_delete_by_id(ctx, (sen_pat *)table, id, optarg);
    case SEN_TABLE_HASH_KEY :
      return sen_hash_delete_by_id(ctx, (sen_hash *)table, id, optarg);
    case SEN_TABLE_NO_KEY :
      return sen_array_delete_by_id(ctx, (sen_array *)table, id, optarg);
    }
  }
  return sen_invalid_argument;
}

sen_rc
sen_table_delete_by_id(sen_ctx *ctx, sen_obj *table, sen_id id)
{
  sen_rc rc;
  SEN_API_ENTER;
  rc = _sen_table_delete_by_id(ctx, table, id, NULL);
  SEN_API_RETURN(rc);
}

sen_rc
sen_table_truncate(sen_ctx *ctx, sen_obj *table)
{
  sen_rc rc = sen_invalid_argument;
  SEN_API_ENTER;
  if (table) {
    rc = sen_success;
  }
  SEN_API_RETURN(rc);
}

sen_id
sen_table_at(sen_ctx *ctx, sen_obj *table, const void *key, int key_size,
             void **value)
{
  sen_id id = SEN_ID_NIL;
  SEN_API_ENTER;
  if (table) {
    switch (table->header.type) {
    case SEN_TABLE_PAT_KEY :
      WITH_NORMALIZE((sen_pat *)table, key, key_size, {
        id = sen_pat_at(ctx, (sen_pat *)table, key, key_size, value);
      });
      break;
    case SEN_TABLE_HASH_KEY :
      WITH_NORMALIZE((sen_hash *)table, key, key_size, {
        id = sen_hash_at(ctx, (sen_hash *)table, key, key_size, value);
      });
      break;
    }
  }
  SEN_API_RETURN(id);
}

sen_id
sen_table_get(sen_ctx *ctx, sen_obj *table, const void *key, int key_size,
              void **value, sen_search_flags *flags)
{
  sen_id id = SEN_ID_NIL;
  if (!key || !key_size) { return SEN_ID_NIL; }
  SEN_API_ENTER;
  if (table) {
    switch (table->header.type) {
    case SEN_TABLE_PAT_KEY :
      WITH_NORMALIZE((sen_pat *)table, key, key_size, {
        id = sen_pat_get(ctx, (sen_pat *)table, key, key_size, value, flags);
      });
      break;
    case SEN_TABLE_HASH_KEY :
      WITH_NORMALIZE((sen_hash *)table, key, key_size, {
        id = sen_hash_get(ctx, (sen_hash *)table, key, key_size, value, flags);
      });
      break;
    }
  }
  SEN_API_RETURN(id);
}

sen_rc
sen_table_get_info(sen_ctx *ctx, sen_obj *table, sen_obj_flags *flags,
                   sen_encoding *encoding, uint8_t *ngram_unit)
{
  sen_rc rc = sen_invalid_argument;
  SEN_API_ENTER;
  if (table) {
    switch (table->header.type) {
    case SEN_TABLE_PAT_KEY :
      if (flags ) { *flags = ((sen_pat *)table)->obj.flags; }
      if (encoding ) { *encoding = ((sen_pat *)table)->encoding; }
      if (ngram_unit ) { *ngram_unit = ((sen_pat *)table)->ngram_unit; }
      rc = sen_success;
      break;
    case SEN_TABLE_HASH_KEY :
      if (flags ) { *flags = ((sen_hash *)table)->obj.flags; }
      if (encoding ) { *encoding = ((sen_hash *)table)->encoding; }
      if (ngram_unit ) { *ngram_unit = ((sen_hash *)table)->ngram_unit; }
      rc = sen_success;
      break;
    }
  }
  SEN_API_RETURN(rc);
}

unsigned int
sen_table_size(sen_ctx *ctx, sen_obj *table)
{
  unsigned int n = 0;
  SEN_API_ENTER;
  if (table) {
    switch (table->header.type) {
    case SEN_TABLE_PAT_KEY :
      n = sen_pat_size(ctx, (sen_pat *)table);
      break;
    case SEN_TABLE_HASH_KEY :
      n = SEN_HASH_SIZE((sen_hash *)table);
      break;
    case SEN_TABLE_NO_KEY :
      n = SEN_ARRAY_SIZE((sen_array *)table);
      break;
    default :
      ERR(sen_invalid_argument, "not supported");
      break;
    }
  } else {
    ERR(sen_invalid_argument, "invalid table assigned");
  }
  SEN_API_RETURN(n);
}

inline static void
subrecs_push(byte *subrecs, int size, int n_subrecs, int score, void *body, int dir)
{
  byte *v;
  int *c2;
  int n = n_subrecs - 1, n2;
  while (n) {
    n2 = (n - 1) >> 1;
    c2 = SEN_RSET_SUBRECS_NTH(subrecs,size,n2);
    if (SEN_RSET_SUBRECS_CMP(score, *c2, dir)) { break; }
    SEN_RSET_SUBRECS_COPY(subrecs,size,n,c2);
    n = n2;
  }
  v = subrecs + n * (size + SEN_RSET_SCORE_SIZE);
  *((int *)v) = score;
  memcpy(v + SEN_RSET_SCORE_SIZE, body, size);
}

inline static void
subrecs_replace_min(byte *subrecs, int size, int n_subrecs, int score, void *body, int dir)
{
  byte *v;
  int n = 0, n1, n2, *c1, *c2;
  for (;;) {
    n1 = n * 2 + 1;
    n2 = n1 + 1;
    c1 = n1 < n_subrecs ? SEN_RSET_SUBRECS_NTH(subrecs,size,n1) : NULL;
    c2 = n2 < n_subrecs ? SEN_RSET_SUBRECS_NTH(subrecs,size,n2) : NULL;
    if (c1 && SEN_RSET_SUBRECS_CMP(score, *c1, dir)) {
      if (c2 &&
          SEN_RSET_SUBRECS_CMP(score, *c2, dir) &&
          SEN_RSET_SUBRECS_CMP(*c1, *c2, dir)) {
        SEN_RSET_SUBRECS_COPY(subrecs,size,n,c2);
        n = n2;
      } else {
        SEN_RSET_SUBRECS_COPY(subrecs,size,n,c1);
        n = n1;
      }
    } else {
      if (c2 && SEN_RSET_SUBRECS_CMP(score, *c2, dir)) {
        SEN_RSET_SUBRECS_COPY(subrecs,size,n,c2);
        n = n2;
      } else {
        break;
      }
    }
  }
  v = subrecs + n * (size + SEN_RSET_SCORE_SIZE);
  memcpy(v, &score, SEN_RSET_SCORE_SIZE);
  memcpy(v + SEN_RSET_SCORE_SIZE, body, size);
}

void
sen_table_add_subrec(sen_obj *table, sen_rset_recinfo *ri, int score,
                     sen_rset_posinfo *pi, int dir)
{
  if (OBJ_DB(table)->flags & SEN_OBJ_WITH_SUBREC) {
    int limit = OBJ_DB(table)->max_n_subrecs;
    ri->score += score;
    ri->n_subrecs += 1;
    if (limit) {
      int subrec_size = OBJ_DB(table)->subrec_size;
      int n_subrecs = SEN_RSET_N_SUBRECS(ri);
      if (pi) {
        byte *body = (byte *)pi + OBJ_DB(table)->subrec_offset;
        if (limit < n_subrecs) {
          if (SEN_RSET_SUBRECS_CMP(score, *ri->subrecs, dir)) {
            subrecs_replace_min((byte *)ri->subrecs, subrec_size, limit, score, body, dir);
          }
        } else {
          subrecs_push((byte *)ri->subrecs, subrec_size, n_subrecs, score, body, dir);
        }
      }
    }
  }
}

sen_table_cursor *
sen_table_cursor_open(sen_ctx *ctx, sen_obj *table,
                      const void *min, unsigned min_size,
                      const void *max, unsigned max_size,
                      int flags)
{
  sen_table_cursor *tc = NULL;
  SEN_API_ENTER;
  if (table) {
    switch (table->header.type) {
    case SEN_TABLE_PAT_KEY :
      tc = (sen_table_cursor *)sen_pat_cursor_open(ctx, (sen_pat *)table,
                                                   min, min_size,
                                                   max, max_size, flags);
      break;
    case SEN_TABLE_HASH_KEY :
      tc = (sen_table_cursor *)sen_hash_cursor_open(ctx, (sen_hash *)table,
                                                    min, min_size,
                                                    max, max_size, flags);
      break;
    case SEN_TABLE_NO_KEY :
      tc = (sen_table_cursor *)sen_array_cursor_open(ctx, (sen_array *)table,
                                                     SEN_ID_NIL, SEN_ID_NIL, flags);
      break;
    }
  }
  SEN_API_RETURN(tc);
}

sen_table_cursor *
sen_table_cursor_open_by_id(sen_ctx *ctx, sen_obj *table,
                            sen_id min, sen_id max, int flags)
{
  sen_table_cursor *tc = NULL;
  SEN_API_ENTER;
  if (table) {
    switch (table->header.type) {
    case SEN_TABLE_PAT_KEY :
      tc = (sen_table_cursor *)sen_pat_cursor_open(ctx, (sen_pat *)table,
                                                   NULL, 0, NULL, 0, flags);
      break;
    case SEN_TABLE_HASH_KEY :
      tc = (sen_table_cursor *)sen_hash_cursor_open(ctx, (sen_hash *)table,
                                                    NULL, 0, NULL, 0, flags);
      break;
    case SEN_TABLE_NO_KEY :
      tc = (sen_table_cursor *)sen_array_cursor_open(ctx, (sen_array *)table,
                                                     min, max, flags);
      break;
    }
  }
  SEN_API_RETURN(tc);
}

sen_rc
sen_table_cursor_close(sen_ctx *ctx, sen_table_cursor *tc)
{
  sen_rc rc = sen_success;
  SEN_API_ENTER;
  if (!tc) {
    ERR(sen_invalid_argument, "tc is null");
    rc = sen_invalid_argument;
  } else {
    switch (tc->header.type) {
    case SEN_CURSOR_TABLE_PAT_KEY :
      sen_pat_cursor_close(ctx, (sen_pat_cursor *)tc);
      break;
    case SEN_CURSOR_TABLE_HASH_KEY :
      sen_hash_cursor_close(ctx, (sen_hash_cursor *)tc);
      break;
    case SEN_CURSOR_TABLE_NO_KEY :
      sen_array_cursor_close(ctx, (sen_array_cursor *)tc);
      break;
    default :
      rc = sen_invalid_argument;
      break;
    }
  }
  SEN_API_RETURN(rc);
}

sen_id
sen_table_cursor_next(sen_ctx *ctx, sen_table_cursor *tc)
{
  sen_id id = SEN_ID_NIL;
  SEN_API_ENTER;
  if (!tc) {
    ERR(sen_invalid_argument, "tc is null");
  } else {
    switch (tc->header.type) {
    case SEN_CURSOR_TABLE_PAT_KEY :
      id = sen_pat_cursor_next(ctx, (sen_pat_cursor *)tc);
      break;
    case SEN_CURSOR_TABLE_HASH_KEY :
      id = sen_hash_cursor_next(ctx, (sen_hash_cursor *)tc);
      break;
    case SEN_CURSOR_TABLE_NO_KEY :
      id = sen_array_cursor_next(ctx, (sen_array_cursor *)tc);
      break;
    }
  }
  SEN_API_RETURN(id);
}

int
sen_table_cursor_get_key(sen_ctx *ctx, sen_table_cursor *tc, void **key)
{
  int len = 0;
  SEN_API_ENTER;
  if (!tc) {
    ERR(sen_invalid_argument, "tc is null");
  } else {
    switch (tc->header.type) {
    case SEN_CURSOR_TABLE_PAT_KEY :
      len = sen_pat_cursor_get_key(ctx, (sen_pat_cursor *)tc, key);
      break;
    case SEN_CURSOR_TABLE_HASH_KEY :
      len = sen_hash_cursor_get_key(ctx, (sen_hash_cursor *)tc, key);
      break;
    default :
      ERR(sen_invalid_argument, "invalid type %d", tc->header.type);
      break;
    }
  }
  SEN_API_RETURN(len);
}

int
sen_table_cursor_get_value(sen_ctx *ctx, sen_table_cursor *tc, void **value)
{
  int len = 0;
  SEN_API_ENTER;
  if (!tc) {
    ERR(sen_invalid_argument, "tc is null");
  } else {
    switch (tc->header.type) {
    case SEN_CURSOR_TABLE_PAT_KEY :
      len = sen_pat_cursor_get_value(ctx, (sen_pat_cursor *)tc, value);
      break;
    case SEN_CURSOR_TABLE_HASH_KEY :
      len = sen_hash_cursor_get_value(ctx, (sen_hash_cursor *)tc, value);
      break;
    case SEN_CURSOR_TABLE_NO_KEY :
      len = sen_array_cursor_get_value(ctx, (sen_array_cursor *)tc, value);
      break;
    default :
      ERR(sen_invalid_argument, "invalid type %d", tc->header.type);
      break;
    }
  }
  SEN_API_RETURN(len);
}

sen_rc
sen_table_cursor_set_value(sen_ctx *ctx, sen_table_cursor *tc,
                           void *value, int flags)
{
  sen_rc rc = sen_invalid_argument;
  SEN_API_ENTER;
  if (!tc) {
    ERR(sen_invalid_argument, "tc is null");
  } else {
    switch (tc->header.type) {
    case SEN_CURSOR_TABLE_PAT_KEY :
      rc = sen_pat_cursor_set_value(ctx, (sen_pat_cursor *)tc, value, flags);
      break;
    case SEN_CURSOR_TABLE_HASH_KEY :
      rc = sen_hash_cursor_set_value(ctx, (sen_hash_cursor *)tc, value, flags);
      break;
    case SEN_CURSOR_TABLE_NO_KEY :
      rc = sen_array_cursor_set_value(ctx, (sen_array_cursor *)tc, value, flags);
      break;
    default :
      ERR(sen_invalid_argument, "invalid type %d", tc->header.type);
      break;
    }
  }
  SEN_API_RETURN(rc);
}

sen_rc
sen_table_cursor_delete(sen_ctx *ctx, sen_table_cursor *tc)
{
  sen_rc rc = sen_invalid_argument;
  SEN_API_ENTER;
  if (!tc) {
    ERR(sen_invalid_argument, "tc is null");
  } else {
    switch (tc->header.type) {
    case SEN_CURSOR_TABLE_PAT_KEY :
      rc = sen_pat_cursor_delete(ctx, (sen_pat_cursor *)tc, NULL);
      break;
    case SEN_CURSOR_TABLE_HASH_KEY :
      rc = sen_hash_cursor_delete(ctx, (sen_hash_cursor *)tc, NULL);
      break;
    case SEN_CURSOR_TABLE_NO_KEY :
      rc = sen_array_cursor_delete(ctx, (sen_array_cursor *)tc, NULL);
      break;
    default :
      ERR(sen_invalid_argument, "invalid type %d", tc->header.type);
      break;
    }
  }
  SEN_API_RETURN(rc);
}

sen_rc
sen_table_search(sen_ctx *ctx, sen_obj *table, const void *key, uint32_t key_size,
                 sen_search_flags flags, sen_obj *res, sen_sel_operator op)
{
  sen_rc rc = sen_success;
  SEN_API_ENTER;
  switch (table->header.type) {
  case SEN_TABLE_PAT_KEY :
    {
      sen_pat *pat = (sen_pat *)table;
      WITH_NORMALIZE(pat, key, key_size, {
        switch (flags) {
        case SEN_SEARCH_EXACT :
          {
            sen_search_flags f = 0;
            sen_id id = sen_pat_lookup(ctx, pat, key, key_size, NULL, &f);
            if (id) {
              sen_search_flags fl = SEN_TABLE_ADD;
              sen_table_lookup(ctx, res, &id, sizeof(sen_id), &fl);
            }
          }
          // todo : support op;
          break;
        case SEN_SEARCH_LCP :
          {
            sen_id id = sen_pat_lcp_search(ctx, pat, key, key_size);
            if (id) {
              sen_search_flags fl = SEN_TABLE_ADD;
              sen_table_lookup(ctx, res, &id, sizeof(sen_id), &fl);
            }
          }
          // todo : support op;
          break;
        case SEN_SEARCH_SUFFIX :
          rc = sen_pat_suffix_search(ctx, pat, key, key_size, (sen_hash *)res);
          // todo : support op;
          break;
        case SEN_SEARCH_PREFIX :
          rc = sen_pat_prefix_search(ctx, pat, key, key_size, (sen_hash *)res);
          // todo : support op;
          break;
        case SEN_SEARCH_TERM_EXTRACT :
          {
            int len;
            sen_id tid;
            const char *sp = key;
            const char *se = sp + key_size;
            for (; sp < se; sp += len) {
              if ((tid = sen_pat_lcp_search(ctx, pat, sp, se - sp))) {
                sen_search_flags fl = SEN_TABLE_ADD;
                sen_table_lookup(ctx, res, &tid, sizeof(sen_id), &fl);
                /* todo : nsubrec++ if SEN_OBJ_TABLE_SUBSET assigned */
              }
              if (!(len = sen_str_charlen_nonnull(sp, se, pat->encoding))) { break; }
            }
          }
          // todo : support op;
          break;
        default :
          rc = sen_invalid_argument;
          ERR(rc, "invalid flag %d", flags);
        }
      });
    }
    break;
  case SEN_TABLE_HASH_KEY :
    {
      sen_hash *hash = (sen_hash *)table;
      sen_search_flags f = 0;
      sen_id id;
      WITH_NORMALIZE(hash, key, key_size, {
        id = sen_hash_lookup(ctx, hash, key, key_size, NULL, &f);
      });
      if (id) {
        sen_search_flags fl = SEN_TABLE_ADD;
        sen_table_lookup(ctx, res, &id, sizeof(sen_id), &fl);
      }
    }
    break;
  }
  SEN_API_RETURN(rc);
}

sen_id
sen_table_next(sen_ctx *ctx, sen_obj *table, sen_id id)
{
  sen_id r = SEN_ID_NIL;
  SEN_API_ENTER;
  if (table) {
    switch (table->header.type) {
    case SEN_TABLE_PAT_KEY :
      r = sen_pat_next(ctx, (sen_pat *)table, id);
      break;
    case SEN_TABLE_HASH_KEY :
      r = sen_hash_next(ctx, (sen_hash *)table, id);
      break;
    case SEN_TABLE_NO_KEY :
      r = sen_array_next(ctx, (sen_array *)table, id);
      break;
    }
  }
  SEN_API_RETURN(r);
}

sen_rc
sen_obj_search(sen_ctx *ctx, sen_obj *obj, sen_obj *query,
               sen_obj *res, sen_sel_operator op, sen_search_optarg *optarg)
{
  sen_rc rc = sen_invalid_argument;
  SEN_API_ENTER;
  if (SEN_OBJ_DBP(obj)) {
    switch (obj->header.type) {
    case SEN_TABLE_PAT_KEY :
    case SEN_TABLE_HASH_KEY :
      {
        const void *key = SEN_BULK_VALUE(query);
        uint32_t key_size = SEN_BULK_LEN(query);
        sen_search_flags flags = optarg ? optarg->flags : SEN_SEARCH_EXACT;
        if (!key || !key_size) {
          return sen_invalid_argument;
        }
        rc = sen_table_search(ctx, obj, key, key_size, flags, res, op);
      }
      break;
    case SEN_COLUMN_INDEX :
      switch (query->header.type) {
      case SEN_BULK :
        {
          const char *str = SEN_BULK_VALUE(query);
          unsigned int str_len = SEN_BULK_LEN(query);
          rc = sen_ii_sel(ctx, (sen_ii *)obj, str, str_len, (sen_hash *)res);
        }
        break;
      case SEN_QUERY :
        rc = sen_query_search(ctx, (sen_ii *)obj, (sen_query *)query, (sen_hash *)res, op);
        break;
      }
      break;
    }
  }
  SEN_API_RETURN(rc);
}

#define SEN_TABLE_GROUP_BY_KEY           0
#define SEN_TABLE_GROUP_BY_VALUE         1
#define SEN_TABLE_GROUP_BY_COLUMN_VALUE  2

#define SEN_TABLE_GROUP_FILTER_PREFIX    0
#define SEN_TABLE_GROUP_FILTER_SUFFIX    (1L<<2)

sen_rc
sen_table_group(sen_ctx *ctx, sen_obj *table,
                sen_table_sort_key *keys, int n_keys,
                sen_table_group_result *results, int n_results)
{
  if (!table || !n_keys || !n_results) {
    ERR(sen_invalid_argument, "table or n_keys or n_results is void");
    return sen_invalid_argument;
  }
  SEN_API_ENTER;
  {
    int k, r;
    void *key;
    sen_obj bulk;
    sen_table_sort_key *kp;
    sen_table_group_result *rp;
    for (k = 0, kp = keys; k < n_keys; k++, kp++) {
      if ((kp->flags & SEN_TABLE_GROUP_BY_COLUMN_VALUE) && !kp->key) {
        ERR(sen_invalid_argument, "column missing in (%d)", k);
        goto exit;
      }
    }
    for (r = 0, rp = results; r < n_results; r++, rp++) {
      if (!rp->table) {
        ERR(sen_invalid_argument, "table missing in (%d)", r);
        goto exit;
      }
    }
    SEN_OBJ_INIT(&bulk, SEN_BULK, 0);
    {
      sen_table_cursor *tc = sen_table_cursor_open(ctx, table, NULL, 0, NULL, 0, 0);
      if (tc) {
        sen_id id;
        while ((id = sen_table_cursor_next(ctx, tc))) {
          sen_rset_recinfo *ri = NULL;
          SEN_BULK_REWIND(&bulk);
          if (OBJ_DB(table)->flags & SEN_OBJ_WITH_SUBREC) {
            sen_table_cursor_get_value(ctx, tc, (void **)&ri);
          }
          for (k = 0, kp = keys; k < n_keys; k++, kp++) {
            kp->offset = SEN_BULK_VSIZE(&bulk);
            sen_obj_get_value(ctx, kp->key, id, &bulk);
          }
          for (r = 0, rp = results; r < n_results; r++, rp++) {
            void *value;
            sen_search_flags f = SEN_TABLE_ADD;
            int begin = keys[rp->key_begin].offset;
            int end = rp->key_end >= n_keys
              ? SEN_BULK_VSIZE(&bulk)
              : keys[rp->key_end].offset;
            key = bulk.u.b.head + begin;
            if (sen_table_get(ctx, rp->table, key, end - begin, &value, &f)) {
              sen_table_add_subrec(rp->table, value, ri ? ri->score : 0, NULL, 0);
            }
          }
        }
        sen_table_cursor_close(ctx, tc);
      }
    }
    sen_obj_close(ctx, &bulk);
  }
exit :
  SEN_API_RETURN(sen_success);
}

sen_rc
sen_table_setoperation(sen_ctx *ctx, sen_obj *table1, sen_obj *table2, sen_obj *res,
                       sen_sel_operator op)
{
  sen_rc rc = sen_success;
  void *key, *value1, *value2;
  uint32_t value_size;
  uint32_t key_size;
  if (table1 != res) {
    if (table2 == res) {
      sen_obj *t = table1;
      table1 = table2;
      table2 = t;
    } else {
      return sen_invalid_argument;
    }
  }
  switch (table1->header.type) {
  case SEN_TABLE_HASH_KEY :
    value_size = ((sen_hash *)table1)->value_size;
    break;
  case SEN_TABLE_PAT_KEY :
    value_size = ((sen_pat *)table1)->value_size;
    break;
  case SEN_TABLE_NO_KEY :
    value_size = ((sen_array *)table1)->value_size;
    break;
  }
  switch (table2->header.type) {
  case SEN_TABLE_HASH_KEY :
    if (value_size < ((sen_hash *)table2)->value_size) {
      value_size = ((sen_hash *)table2)->value_size;
    }
    break;
  case SEN_TABLE_PAT_KEY :
    if (value_size < ((sen_pat *)table2)->value_size) {
      value_size = ((sen_pat *)table2)->value_size;
    }
    break;
  case SEN_TABLE_NO_KEY :
    if (value_size < ((sen_array *)table2)->value_size) {
      value_size = ((sen_array *)table2)->value_size;
    }
    break;
  }
  switch (op) {
  case sen_sel_or :
    SEN_TABLE_EACH(ctx, table2, 0, 0, id, &key, &key_size, &value2, {
      sen_search_flags f = SEN_TABLE_ADD;
      sen_table_get(ctx, table1, key, key_size, &value1, &f);
      memcpy(value1, value2, value_size);
    });
    break;
  case sen_sel_and :
    SEN_TABLE_EACH(ctx, table1, 0, 0, id, &key, &key_size, &value1, {
      if (!sen_table_at(ctx, table2, key, key_size, &value2)) {
        _sen_table_delete_by_id(ctx, table1, id, NULL);
      }
    });
    break;
  case sen_sel_but :
    SEN_TABLE_EACH(ctx, table2, 0, 0, id, &key, &key_size, &value2, {
      sen_table_delete(ctx, table1, key, key_size);
    });
    break;
  case sen_sel_adjust :
    SEN_TABLE_EACH(ctx, table2, 0, 0, id, &key, &key_size, &value2, {
      if (sen_table_at(ctx, table1, key, key_size, &value1)) {
        memcpy(value1, value2, value_size);
      }
    });
    break;
  }
  return rc;
}

sen_rc
sen_table_difference(sen_ctx *ctx, sen_obj *table1, sen_obj *table2,
                     sen_obj *res1, sen_obj *res2)
{
  void *key;
  uint32_t key_size;
  if (table1 != res1 || table2 != res2) { return sen_invalid_argument; }
  if (sen_table_size(ctx, table1) > sen_table_size(ctx, table2)) {
    SEN_TABLE_EACH(ctx, table2, 0, 0, id, &key, &key_size, NULL, {
      sen_id id1;
      if ((id1 = sen_table_at(ctx, table1, key, key_size, NULL))) {
        _sen_table_delete_by_id(ctx, table1, id1, NULL);
        _sen_table_delete_by_id(ctx, table2, id, NULL);
      }
    });
  } else {
    SEN_TABLE_EACH(ctx, table1, 0, 0, id, &key, &key_size, NULL, {
      sen_id id2;
      if ((id2 = sen_table_at(ctx, table2, key, key_size, NULL))) {
        _sen_table_delete_by_id(ctx, table1, id, NULL);
        _sen_table_delete_by_id(ctx, table2, id2, NULL);
      }
    });
  }
  return sen_success;
}

sen_obj *
sen_table_column(sen_ctx *ctx, sen_obj *table, const char *name, unsigned name_size)
{
  sen_obj *column = NULL;
  SEN_API_ENTER;
  if (SEN_OBJ_TABLEP(table)) {
    if (check_name(ctx, name, name_size)) {
      column = sen_obj_get_accessor(ctx, table, name, name_size);
    } else {
      char buf[SEN_PAT_MAX_KEY_SIZE];
      int len = sen_obj_name(ctx, table, buf, SEN_PAT_MAX_KEY_SIZE);
      if (len) {
        buf[len++] = SEN_DB_DELIMITER;
        if (len + name_size <= SEN_PAT_MAX_KEY_SIZE) {
          memcpy(buf + len, name, name_size);
          column = sen_ctx_lookup(ctx, buf, len + name_size);
        } else {
          ERR(sen_invalid_argument, "name is too long");
        }
      } else {
        /* todo : support temporary table */
      }
    }
  } else {
    if (table->header.type == SEN_ACCESSOR) {
      column = sen_obj_get_accessor(ctx, table, name, name_size);
    } else {
      ERR(sen_invalid_argument, "invalid table assigned");
    }
  }
  SEN_API_RETURN(column);
}

int
sen_table_columns(sen_ctx *ctx, sen_obj *table, const char *name, unsigned name_size,
                  sen_obj *res)
{
  int n = 0;
  SEN_API_ENTER;
  if (SEN_OBJ_TABLEP(table)) {
    sen_obj bulk;
    sen_space *s = (sen_space *)OBJ_DB(table)->space;
    SEN_OBJ_INIT(&bulk, SEN_BULK, 0);
    sen_pat_get_key2(ctx, s->keys, OBJ_DB(table)->id, &bulk);
    SEN_BULK_PUTC(ctx, &bulk, SEN_DB_DELIMITER);
    sen_bulk_write(ctx, &bulk, name, name_size);
    sen_pat_prefix_search(ctx, s->keys, SEN_BULK_HEAD(&bulk), SEN_BULK_VSIZE(&bulk),
                          (sen_hash *)res);
    sen_obj_close(ctx, &bulk);
    n = sen_table_size(ctx, res);
  }
  SEN_API_RETURN(n);
}

const char *
_sen_table_key(sen_ctx *ctx, sen_obj *table, sen_id id, uint32_t *key_size)
{
  SEN_ASSERT(table);
  switch (table->header.type) {
  case SEN_TABLE_HASH_KEY :
    return _sen_hash_key((sen_hash *)table, id, key_size);
  case SEN_TABLE_PAT_KEY :
    return _sen_pat_key((sen_pat *)table, id, key_size);
  case SEN_TABLE_NO_KEY :
    {
      sen_array *a = (sen_array *)table;
      if (a->obj.header.domain && a->value_size) {
        *key_size = a->value_size;
        return _sen_array_get_value(ctx, a, id);
      }
    }
    break;
  }
  return NULL;
}

/* column */

sen_obj *
sen_column_create(sen_ctx *ctx, sen_obj *table,
                  const char *name, unsigned name_size,
                  const char *path, sen_obj_flags flags, sen_obj *type)
{
  sen_space *s;
  uint32_t value_size;
  sen_obj *space, *res = NULL;
  unsigned char impl_flags = 0;
  sen_id id = SEN_ID_NIL;
  sen_id range = SEN_ID_NIL;
  sen_id domain = SEN_ID_NIL;
  char fullname[SEN_PAT_MAX_KEY_SIZE];
  char buffer[PATH_MAX];
  SEN_API_ENTER;
  if (!table || !type || !name || !name_size) {
    ERR(sen_invalid_argument, "missing type or name");
    goto exit;
  }
  space = OBJ_DB(table)->space;
  s = (sen_space *)space;
  if (!SPACE_P(s)) {
    ERR(sen_invalid_argument, "invalid space assigned");
    goto exit;
  }
  if (check_name(ctx, name, name_size)) {
    ERR(sen_invalid_argument, "name contains '%c'", SEN_DB_DELIMITER);
    goto exit;
  }
  if ((domain = OBJ_DB(table)->id)) {
    int len = sen_pat_get_key(ctx, s->keys, domain,
                              fullname, SEN_PAT_MAX_KEY_SIZE);
    if (name_size + 1 + len > SEN_PAT_MAX_KEY_SIZE) {
      ERR(sen_invalid_argument, "too long column name");
      goto exit;
    }
    fullname[len] = SEN_DB_DELIMITER;
    memcpy(fullname + len + 1, name, name_size);
    name_size += len + 1;
  } else {
    ERR(sen_invalid_argument, "todo : not supported yet");
    goto exit;
  }
  range = OBJ_DB(type)->id;
  switch (type->header.type) {
  case SEN_TYPE :
    {
      sen_obj_db *t = (sen_obj_db *)type;
      flags |= t->header.flags;
      value_size = SEN_TYPE_SIZE(t);
    }
    break;
  case SEN_TABLE_HASH_KEY :
  case SEN_TABLE_NO_KEY :
  case SEN_TABLE_PAT_KEY :
    value_size = sizeof(sen_id);
    break;
  default :
    /*
    if (type == sen_type_any) {
      value_size = sizeof(sen_id) + sizeof(sen_id);
    }
    */
    value_size = sizeof(sen_id);
  }
  id = sen_obj_register(ctx, space, fullname, name_size);
  if (ERRP(ctx, SEN_ERROR)) { goto exit;  }
  if (SEN_OBJ_PERSISTENT & flags) {
    if (!path) {
      if (PERSISTENT_SPACE_P(space)) {
        gen_pathname(s->keys->io->path, buffer, id);
        path = buffer;
      } else {
        ERR(sen_invalid_argument, "path not assigend for persistent table");
        goto exit;
      }
    } else {
      impl_flags |= SEN_OBJ_CUSTOM_NAME;
    }
  } else {
    if (path) {
      ERR(sen_invalid_argument, "path assigend for temporary table");
      goto exit;
    }
  }
  switch (flags & SEN_OBJ_COLUMN_TYPE_MASK) {
  case SEN_OBJ_COLUMN_SCALAR :
    if ((flags & SEN_OBJ_KEY_VAR_SIZE) || value_size > sizeof(int64_t)) {
      res = (sen_obj *)sen_ja_create(ctx, path, value_size, 0); //todo : zlib support
    } else {
      res = (sen_obj *)sen_ra_create(ctx, path, value_size);
    }
    break;
  case SEN_OBJ_COLUMN_ARRAY :
  case SEN_OBJ_COLUMN_VERSES :
  case SEN_OBJ_COLUMN_POSTINGS :
    res = (sen_obj *)sen_ja_create(ctx, path, value_size * 16/*todo*/, 0);
    //todo : zlib support
    break;
  case SEN_OBJ_COLUMN_INDEX :
    res = (sen_obj *)sen_ii_create(ctx, path, table, 0); //todo : ii layout support
    break;
  }
  if (res) {
    OBJ_DB(res)->header.domain = domain;
    OBJ_DB(res)->range = range;
    OBJ_DB(res)->flags = flags;
    res->header.flags = flags;
    if (sen_obj_db_init(ctx, space, id, OBJ_DB(res))) {
      sen_obj_close(ctx, res);
      sen_obj_remove(ctx, path);
      res = NULL;
    }
  }
exit :
  if (!res && id) { sen_obj_delete_by_id(ctx, space, id, 1); }
  SEN_API_RETURN(res);
}

sen_obj *
sen_column_open(sen_ctx *ctx, sen_obj *table,
                const char *name, unsigned name_size,
                const char *path, sen_obj *type)
{
  sen_id domain;
  sen_obj *res = NULL;
  sen_space *s;
  char fullname[SEN_PAT_MAX_KEY_SIZE];
  SEN_API_ENTER;
  if (!table || !type || !name || !name_size) {
    ERR(sen_invalid_argument, "missing type or name");
    goto exit;
  }
  s = (sen_space *)OBJ_DB(table)->space;
  if (!SPACE_P(s)) {
    ERR(sen_invalid_argument, "invalid space assigned");
    goto exit;
  }
  if (check_name(ctx, name, name_size)) {
    ERR(sen_invalid_argument, "name contains '%c'", SEN_DB_DELIMITER);
    goto exit;
  }
  if ((domain = OBJ_DB(table)->id)) {
    int len = sen_pat_get_key(ctx, s->keys, domain, fullname, SEN_PAT_MAX_KEY_SIZE);
    if (name_size + 1 + len > SEN_PAT_MAX_KEY_SIZE) {
      ERR(sen_invalid_argument, "too long column name");
      goto exit;
    }
    fullname[len] = SEN_DB_DELIMITER;
    memcpy(fullname + len + 1, name, name_size);
    name_size += len + 1;
  } else {
    ERR(sen_invalid_argument, "todo : not supported yet");
    goto exit;
  }
  res = sen_ctx_lookup(ctx, fullname, name_size);
  if (res) {
    const char *path2 = sen_obj_path(ctx, res);
    if (path && (!path2 || strcmp(path, path2))) { goto exit; }
  } else if (path) {
    uint32_t dbtype = sen_io_detect_type(ctx, path);
    if (!dbtype) { goto exit; }
    switch (dbtype) {
    case SEN_COLUMN_VAR_SIZE :
      res = (sen_obj *)sen_ja_open(ctx, path);
      break;
    case SEN_COLUMN_FIX_SIZE :
      res = (sen_obj *)sen_ra_open(ctx, path);
      break;
    case SEN_COLUMN_INDEX :
      res = (sen_obj *)sen_ii_open(ctx, path, table);
      break;
    }
    if (res) {
      sen_id id = sen_obj_register(ctx, (sen_obj *)s, fullname, name_size);
      OBJ_DB(res)->header.domain = domain;
      OBJ_DB(res)->range = OBJ_DB(type)->id;
      res->header.impl_flags |= SEN_OBJ_CUSTOM_NAME;
      sen_obj_db_init(ctx, (sen_obj *)s, id, OBJ_DB(res));
    }
  }
exit :
  SEN_API_RETURN(res);
}

/*
typedef struct {
  sen_id id;
  int flags;
} sen_column_set_value_arg;

static sen_rc
default_column_set_value(sen_ctx *ctx, sen_proc_ctx *pctx, sen_obj *in, sen_obj *out)
{
  sen_proc_data *data = sen_proc_ctx_get_local_data(pctx);
  if (data) {
    sen_column_set_value_arg *arg = data->ptr;
    unsigned int value_size = in->u.p.size; //todo
    if (!pctx->obj) { return SEN_ID_NIL; }
    switch (pctx->obj->header.type) {
    case SEN_COLUMN_VAR_SIZE :
      return sen_ja_put(ctx, (sen_ja *)pctx->obj, arg->id,
                        in->u.p.ptr, value_size, 0); // todo type->flag
    case SEN_COLUMN_FIX_SIZE :
      if (((sen_ra *)pctx->obj)->header->element_size < value_size) {
        ERR(sen_invalid_argument, "too long value (%d)", value_size);
        return sen_invalid_argument;
      } else {
        void *v = sen_ra_get(ctx, (sen_ra *)pctx->obj, arg->id);
        if (!v) {
          ERR(sen_memory_exhausted, "ra get failed");
          return sen_memory_exhausted;
        }
        memcpy(v, in->u.p.ptr, value_size);
      }
      break;
    case SEN_COLUMN_INDEX :
      // todo : how??
      break;
    }
    return sen_success;
  } else {
    return sen_other_error;
  }
}
*/

typedef struct {
  sen_id target;
  unsigned int section;
} default_set_value_hook_data;

static sen_rc
default_set_value_hook(sen_ctx *ctx, sen_proc_ctx *pctx,
                       int argc, sen_proc_data *argv)
{
  if (!pctx || argc != 4) {
    ERR(sen_invalid_argument, "default_set_value_hook failed");
    return sen_invalid_argument;
  } else {
    //sen_obj *obj = pctx->obj;
    sen_id id = argv[0].id;
    sen_obj *oldvalue = argv[1].ptr;
    sen_obj *newvalue = argv[2].ptr;
    //    int flags = argv[3].int_value;
    sen_hook *h = pctx->currh;
    default_set_value_hook_data *data = (void *)NEXT_ADDR(h);
    sen_obj *target = sen_ctx_get(ctx, data->target);
    int section = data->section;
    switch (target->header.type) {
    case SEN_COLUMN_INDEX :
      return sen_ii_column_update(ctx, (sen_ii *)target, id, section, oldvalue, newvalue);
    default :
      return sen_success;
    }
  }
}

/**** sen_vector ****/

#define VECTOR(obj) ((sen_vector *)obj)

#define INITIAL_VECTOR_SIZE 256

sen_obj *
sen_vector_open(sen_ctx *ctx, sen_obj_flags flags)
{
  sen_obj *obj = (sen_obj *)SEN_MALLOCN(sen_vector, 1);
  if (obj) {
    SEN_OBJ_INIT(obj, SEN_VECTOR, flags);
    obj->header.impl_flags |= SEN_OBJ_ALLOCATED;
    VECTOR(obj)->offsets = NULL;
    VECTOR(obj)->n_entries = 0;
  }
  return obj;
}

int
sen_vector_delimit(sen_ctx *ctx, sen_obj *vector)
{
  sen_vector *v = VECTOR(vector);
  uint32_t *offsets;
  if (!(v->n_entries & (INITIAL_VECTOR_SIZE - 1))) {
    offsets = SEN_REALLOC(v->offsets, sizeof(uint32_t) *
                          (v->n_entries + INITIAL_VECTOR_SIZE));
    if (!offsets) { return -1; }
    v->offsets = offsets;
  }
  v->offsets[v->n_entries] = SEN_BULK_VSIZE(vector);
  return ++(v->n_entries);
}

int
sen_vector_size(sen_ctx *ctx, sen_obj *vector)
{
  return VECTOR(vector)->n_entries;
}

const char *
sen_vector_fetch(sen_ctx *ctx, sen_obj *vector, int i, unsigned int *size)
{
  sen_vector *v = VECTOR(vector);
  if (i < 0) {
    i += v->n_entries;
    if (i < 0) { return NULL; }
  }
  if (v->n_entries <= i) { return NULL; }
  {
    size_t off = i ? v->offsets[i - 1] : 0;
    *size = v->offsets[i] - off;
    return vector->u.b.head + off;
  }
}

sen_rc
sen_vector_decode(sen_ctx *ctx, sen_obj *vector, const char *data, uint32_t data_size)
{
  sen_vector *v = VECTOR(vector);
  uint8_t *p = (uint8_t *)data;
  uint8_t *pe = p + data_size;
  SEN_B_DEC(v->n_entries, p);
  {
    int n = (v->n_entries + (INITIAL_VECTOR_SIZE - 1)) & ~(INITIAL_VECTOR_SIZE - 1);
    uint32_t *offsets = SEN_REALLOC(v->offsets, sizeof(uint32_t) * n);
    if (!offsets) { return sen_memory_exhausted; }
    v->offsets = offsets;
  }
  {
    int i;
    uint32_t o = 0, l;
    for (i = 0; i < v->n_entries; i++) {
      if (pe <= p) { return sen_invalid_argument; }
      SEN_B_DEC(l, p);
      o += l;
      v->offsets[i] = o;
    }
  }
  return sen_bulk_write(ctx, vector, p, pe - p);
}

/**** sen_verses ****/

#define INITIAL_VERSES_SIZE 256

sen_rc
sen_verses_add(sen_ctx *ctx, sen_obj *v, const char *str, unsigned int str_len,
               unsigned int weight, sen_id domain)
{
  sen_verse *vp;
  if (!str_len) { return sen_success; }
  if (!v || !str) {
    SEN_LOG(sen_log_warning, "sen_verses_add: invalid argument");
    return sen_invalid_argument;
  }
  if (!(v->u.v.n_verses & (INITIAL_VERSES_SIZE - 1))) {
    vp = SEN_REALLOC(v->u.v.verses, sizeof(sen_verse) *
                     (v->u.v.n_verses + INITIAL_VERSES_SIZE));
    if (!vp) { return sen_memory_exhausted; }
    v->u.v.verses = vp;
  }
  vp = &v->u.v.verses[v->u.v.n_verses];
  if (v->header.flags & SEN_OBJ_DO_SHALLOW_COPY) {
    vp->str = (char *)str;
  } else {
    if (!(vp->str = SEN_MALLOC(str_len))) { return sen_memory_exhausted; }
    memcpy(vp->str, str, str_len);
  }
  vp->str_len = str_len;
  vp->weight = weight;
  vp->domain = domain;
  v->u.v.n_verses++;
  return sen_success;
}

sen_obj *
sen_verses_to_vector(sen_ctx *ctx, sen_obj *verses)
{
  sen_obj *vector = sen_vector_open(ctx, 0);
  if (vector) {
    sen_verse *vp;
    int i;
    for (i = verses->u.v.n_verses, vp = verses->u.v.verses; i; i--, vp++) {
      sen_bulk_benc(ctx, vector, vp->weight);
      sen_bulk_benc(ctx, vector, vp->domain);
      sen_bulk_write(ctx, vector, vp->str, vp->str_len);
      sen_vector_delimit(ctx, vector);
    }
  }
  return vector;
}

sen_obj *
sen_vector_to_verses(sen_ctx *ctx, sen_obj *vector, sen_obj *verses)
{
  if (!verses) {
    verses = sen_obj_open(ctx, SEN_VERSES, SEN_OBJ_DO_SHALLOW_COPY);
  }
  if (verses) {
    int i, n = sen_vector_size(ctx, vector);
    verses->u.v.src = vector;
    for (i = 0; i < n; i++) {
      unsigned int size;
      const char *pe, *p = sen_vector_fetch(ctx, vector, i, &size);
      if (p) {
        sen_id domain;
        unsigned int weight;
        pe = p + size;
        if (p < pe) {
          SEN_B_DEC(weight, p);
          if (p < pe) {
            SEN_B_DEC(domain, p);
            if (p <= pe) {
              sen_verses_add(ctx, verses, p, pe - p, weight, domain);
            }
          }
        }
      }
    }
  }
  return verses;
}

/**** accessor ****/

typedef struct _sen_accessor sen_accessor;

struct _sen_accessor {
  sen_obj_header header;
  uint8_t action;
  int offset;
  sen_obj *obj;
  sen_accessor *next;
};

enum {
  SEN_ACCESSOR_VOID = 0,
  SEN_ACCESSOR_GET_KEY,
  SEN_ACCESSOR_GET_VALUE,
  SEN_ACCESSOR_GET_SCORE,
  SEN_ACCESSOR_GET_NSUBRECS,
  SEN_ACCESSOR_GET_COLUMN_VALUE,
  SEN_ACCESSOR_GET_DB_OBJ,
  SEN_ACCESSOR_LOOKUP,
  SEN_ACCESSOR_FUNCALL
};

sen_accessor *
accessor_new(sen_ctx *ctx)
{
  sen_accessor *res = SEN_MALLOCN(sen_accessor, 1);
  if (res) {
    res->header.type = SEN_ACCESSOR;
    res->header.impl_flags = SEN_OBJ_ALLOCATED;
    res->header.flags = 0;
    res->action = SEN_ACCESSOR_VOID;
    res->offset = 0;
    res->next = NULL;
  }
  return res;
}

sen_obj *
sen_obj_get_accessor(sen_ctx *ctx, sen_obj *obj, const char *name, unsigned name_size)
{
  sen_accessor *res = NULL, **rp = NULL, **rp0 = NULL;
  if (!obj) { return NULL; }
  SEN_API_ENTER;
  if (obj->header.type == SEN_ACCESSOR) {
    for (rp0 = (sen_accessor **)&obj; *rp0; rp0 = &(*rp0)->next) {
      res = *rp0;
    }
    switch (res->action) {
    case SEN_ACCESSOR_GET_KEY :
      obj = sen_ctx_get(ctx, res->obj->header.domain);
      break;
    case SEN_ACCESSOR_GET_VALUE :
    case SEN_ACCESSOR_GET_SCORE :
    case SEN_ACCESSOR_GET_NSUBRECS :
      obj = sen_ctx_get(ctx, OBJ_DB(res->obj)->range);
      break;
    case SEN_ACCESSOR_GET_COLUMN_VALUE :
      obj = sen_ctx_get(ctx, OBJ_DB(res->obj)->range);
      break;
    case SEN_ACCESSOR_LOOKUP :
      /* todo */
      break;
    case SEN_ACCESSOR_FUNCALL :
      /* todo */
      break;
    }
  }
  {
    size_t len;
    const char *sp, *se = name + name_size;
    if (*name == SEN_DB_DELIMITER) { name++; }
    for (sp = name; (len = sen_str_charlen_nonnull(sp, se, ctx->encoding)); sp += len) {
      if (*sp == SEN_DB_DELIMITER) { break; }
    }
    if (!(len = sp - name)) { goto exit; }
    if (*name == ':') { /* pseudo column */
      int done = 0;
      switch (name[1]) {
      case 'k' : /* key */
      case 'K' :
        for (rp = &res; !done; rp = &(*rp)->next) {
          *rp = accessor_new(ctx);
          (*rp)->obj = obj;
          obj = sen_ctx_get(ctx, obj->header.domain);
          switch (obj->header.type) {
          case SEN_SPACE :
            (*rp)->action = SEN_ACCESSOR_GET_KEY;
            rp = &(*rp)->next;
            *rp = accessor_new(ctx);
            (*rp)->obj = obj;
            (*rp)->action = SEN_ACCESSOR_GET_DB_OBJ;
            done++;
            break;
          case SEN_TYPE :
            (*rp)->action = SEN_ACCESSOR_GET_KEY;
            done++;
            break;
          case SEN_TABLE_PAT_KEY :
          case SEN_TABLE_HASH_KEY :
            (*rp)->action = SEN_ACCESSOR_GET_KEY;
            break;
          case SEN_TABLE_NO_KEY :
            if (obj->header.domain) {
              (*rp)->action = SEN_ACCESSOR_GET_VALUE;
              break;
            }
            /* fallthru */
          default :
            /* lookup failed */
            sen_obj_close(ctx, (sen_obj *)res);
            res = NULL;
            goto exit;
          }
        }
        break;
      case 's' : /* score */
      case 'S' :
        for (rp = &res; !done; rp = &(*rp)->next) {
          *rp = accessor_new(ctx);
          (*rp)->obj = obj;
          if (OBJ_DB(obj)->flags & SEN_OBJ_WITH_SUBREC) {
            (*rp)->action = SEN_ACCESSOR_GET_SCORE;
            done++;
          } else {
            switch (obj->header.type) {
            case SEN_TABLE_PAT_KEY :
            case SEN_TABLE_HASH_KEY :
              (*rp)->action = SEN_ACCESSOR_GET_KEY;
              break;
            case SEN_TABLE_NO_KEY :
              if (obj->header.domain) {
                (*rp)->action = SEN_ACCESSOR_GET_VALUE;
                break;
              }
              /* fallthru */
            default :
              /* lookup failed */
              sen_obj_close(ctx, (sen_obj *)res);
              res = NULL;
              goto exit;
            }
            obj = sen_ctx_get(ctx, obj->header.domain);
          }
        }
        break;
      case 'n' : /* nsubrecs */
      case 'N' :
        for (rp = &res; !done; rp = &(*rp)->next) {
          *rp = accessor_new(ctx);
          (*rp)->obj = obj;
          if (OBJ_DB(obj)->flags & SEN_OBJ_WITH_SUBREC) {
            (*rp)->action = SEN_ACCESSOR_GET_NSUBRECS;
            done++;
          } else {
            switch (obj->header.type) {
            case SEN_TABLE_PAT_KEY :
            case SEN_TABLE_HASH_KEY :
              (*rp)->action = SEN_ACCESSOR_GET_KEY;
              break;
            case SEN_TABLE_NO_KEY :
              if (obj->header.domain) {
                (*rp)->action = SEN_ACCESSOR_GET_VALUE;
                break;
              }
              /* fallthru */
            default :
              /* lookup failed */
              sen_obj_close(ctx, (sen_obj *)res);
              res = NULL;
              goto exit;
            }
            obj = sen_ctx_get(ctx, obj->header.domain);
          }
        }
        break;
      }
    } else {
      /* if obj->header.type == SEN_TYPE ... lookup table */
      for (rp = &res; ; rp = &(*rp)->next) {
        sen_obj *column = sen_table_column(ctx, obj, name, len);
        *rp = accessor_new(ctx);
        if (column) {
          (*rp)->obj = column;
          /*
          switch (column->header.type) {
          case SEN_COLUMN_VAR_SIZE :
            break;
          case SEN_COLUMN_FIX_SIZE :
            break;
          case SEN_COLUMN_INDEX :
            break;
          }
          */
          (*rp)->action = SEN_ACCESSOR_GET_COLUMN_VALUE;
          break;
        } else {
          (*rp)->obj = obj;
          obj = sen_ctx_get(ctx, obj->header.domain);
          switch (obj->header.type) {
          case SEN_TABLE_PAT_KEY :
          case SEN_TABLE_HASH_KEY :
            (*rp)->action = SEN_ACCESSOR_GET_KEY;
            break;
          case SEN_TABLE_NO_KEY :
            /* todo */
            break;
          default :
            /* lookup failed */
            sen_obj_close(ctx, (sen_obj *)res);
            res = NULL;
            goto exit;
          }
        }
      }
    }
    if (sp != se) { sen_obj_get_accessor(ctx, (sen_obj *)res, sp, se - sp); }
  }
  if (rp0) { *rp0 = res; }
 exit :
  SEN_API_RETURN((sen_obj *)res);
}

sen_id
sen_obj_get_range(sen_ctx *ctx, sen_obj *obj)
{
  sen_id range = SEN_ID_NIL;
  if (SEN_OBJ_DBP(obj)) {
    range = OBJ_DB(obj)->range;
  } else if (obj->header.type == SEN_ACCESSOR) {
    sen_accessor *a;
    for (a = (sen_accessor *)obj; a; a = a->next) {
      switch (a->action) {
      case SEN_ACCESSOR_GET_VALUE :
      case SEN_ACCESSOR_GET_SCORE :
      case SEN_ACCESSOR_GET_NSUBRECS :
        range = 1; /* todo : fix it (SEN_QL_INT) */
        break;
      case SEN_ACCESSOR_GET_COLUMN_VALUE :
        if (SEN_OBJ_DBP(a->obj)) { range = OBJ_DB(a->obj)->range; }
        break;
      case SEN_ACCESSOR_GET_KEY :
        if (SEN_OBJ_DBP(a->obj)) { range = OBJ_DB(a->obj)->header.domain; }
        break;
      default :
        if (SEN_OBJ_DBP(a->obj)) { range = OBJ_DB(a->obj)->range; }
        break;
      }
    }
  }
  return range;
}

const char *sen_obj_get_value_(sen_ctx *ctx, sen_obj *obj, sen_id id, uint32_t *size);

const char *
sen_accessor_get_value_(sen_ctx *ctx, sen_accessor *a, sen_id id, uint32_t *size)
{
  const char *value;
  for (;;) {
    switch (a->action) {
    case SEN_ACCESSOR_GET_KEY :
      value = _sen_table_key(ctx, a->obj, id, size);
      break;
    case SEN_ACCESSOR_GET_VALUE :
      value = sen_obj_get_value_(ctx, a->obj, id, size);
      break;
    case SEN_ACCESSOR_GET_SCORE :
      if ((value = sen_obj_get_value_(ctx, a->obj, id, size))) {
        value = (const char *)&((sen_rset_recinfo *)value)->score;
        *size = sizeof(int);
      }
      break;
    case SEN_ACCESSOR_GET_NSUBRECS :
      if ((value = sen_obj_get_value_(ctx, a->obj, id, size))) {
        value = (const char *)&((sen_rset_recinfo *)value)->n_subrecs;
        *size = sizeof(int);
      }
      break;
    case SEN_ACCESSOR_GET_COLUMN_VALUE :
      /* todo : support vector */
      value = sen_obj_get_value_(ctx, a->obj, id, size);
      break;
    case SEN_ACCESSOR_GET_DB_OBJ :
      value = _sen_pat_key(((sen_space *)ctx->impl->db)->keys, id, size);
      break;
    case SEN_ACCESSOR_LOOKUP :
      /* todo */
      break;
    case SEN_ACCESSOR_FUNCALL :
      /* todo */
      break;
    }
    if (value && (a = a->next)) {
      id = *((sen_id *)value);
    } else {
      break;
    }
  }
  return value;
}

static sen_obj *
sen_accessor_get_value(sen_ctx *ctx, sen_accessor *a, sen_id id, sen_obj *value)
{
  sen_obj buf;
  void *vp;
  size_t vs;
  SEN_OBJ_INIT(&buf, SEN_BULK, 0);
  for (;;) {
    SEN_BULK_REWIND(&buf);
    switch (a->action) {
    case SEN_ACCESSOR_GET_KEY :
      sen_table_get_key2(ctx, a->obj, id, &buf);
      vp = SEN_BULK_VALUE(&buf);
      vs = SEN_BULK_VSIZE(&buf);
      break;
    case SEN_ACCESSOR_GET_VALUE :
      sen_obj_get_value(ctx, a->obj, id, &buf);
      vp = SEN_BULK_VALUE(&buf);
      vs = SEN_BULK_VSIZE(&buf);
      break;
    case SEN_ACCESSOR_GET_SCORE :
      sen_obj_get_value(ctx, a->obj, id, &buf);
      {
        sen_rset_recinfo *ri = (sen_rset_recinfo *)buf.u.b.head;
        vp = &ri->score;
        vs = sizeof(int);
      }
      break;
    case SEN_ACCESSOR_GET_NSUBRECS :
      sen_obj_get_value(ctx, a->obj, id, &buf);
      {
        sen_rset_recinfo *ri = (sen_rset_recinfo *)buf.u.b.head;
        vp = &ri->n_subrecs;
        vs = sizeof(int);
      }
      break;
    case SEN_ACCESSOR_GET_COLUMN_VALUE :
      /* todo : support vector */
      sen_obj_get_value(ctx, a->obj, id, &buf);
      vp = SEN_BULK_VALUE(&buf);
      vs = SEN_BULK_VSIZE(&buf);
      break;
    case SEN_ACCESSOR_GET_DB_OBJ :
      value = sen_ctx_get(ctx, id);
      sen_obj_close(ctx, &buf);
      return value;
      break;
    case SEN_ACCESSOR_LOOKUP :
      /* todo */
      break;
    case SEN_ACCESSOR_FUNCALL :
      /* todo */
      break;
    }
    if ((a = a->next)) {
      id = *((sen_id *)vp);
    } else {
      break;
    }
  }
  if (!value) { value = sen_obj_open(ctx, SEN_BULK, 0); }
  if (value) { sen_bulk_write(ctx, value, vp, vs); }
  sen_obj_close(ctx, &buf);
  return value;
}

static sen_rc
sen_accessor_set_value(sen_ctx *ctx, sen_accessor *a, sen_id id,
                       sen_obj *value, int flags)
{
  sen_rc rc = sen_success;
  if (!value) { value = sen_obj_open(ctx, SEN_BULK, 0); }
  if (value) {
    sen_obj buf;
    void *vp;
    size_t vs;
    SEN_OBJ_INIT(&buf, SEN_BULK, 0);
    for (;;) {
      SEN_BULK_REWIND(&buf);
      switch (a->action) {
      case SEN_ACCESSOR_GET_KEY :
        sen_table_get_key2(ctx, a->obj, id, &buf);
        vp = SEN_BULK_VALUE(&buf);
        vs = SEN_BULK_VSIZE(&buf);
        break;
      case SEN_ACCESSOR_GET_VALUE :
        if (a->next) {
          sen_obj_get_value(ctx, a->obj, id, &buf);
          vp = SEN_BULK_VALUE(&buf);
          vs = SEN_BULK_VSIZE(&buf);
        } else {
          rc = sen_obj_set_value(ctx, a->obj, id, value, flags);
        }
        break;
      case SEN_ACCESSOR_GET_SCORE :
        sen_obj_get_value(ctx, a->obj, id, &buf);
        {
          sen_rset_recinfo *ri = (sen_rset_recinfo *)buf.u.b.head;
          vp = &ri->score;
          vs = sizeof(int);
        }
        break;
      case SEN_ACCESSOR_GET_NSUBRECS :
        sen_obj_get_value(ctx, a->obj, id, &buf);
        {
          sen_rset_recinfo *ri = (sen_rset_recinfo *)buf.u.b.head;
          vp = &ri->n_subrecs;
          vs = sizeof(int);
        }
        break;
      case SEN_ACCESSOR_GET_COLUMN_VALUE :
        /* todo : support vector */
        if (a->next) {
          sen_obj_get_value(ctx, a->obj, id, &buf);
          vp = SEN_BULK_VALUE(&buf);
          vs = SEN_BULK_VSIZE(&buf);
        } else {
          rc = sen_obj_set_value(ctx, a->obj, id, value, flags);
        }
        break;
      case SEN_ACCESSOR_LOOKUP :
        /* todo */
        break;
      case SEN_ACCESSOR_FUNCALL :
        /* todo */
        break;
      }
      if ((a = a->next)) {
        id = *((sen_id *)vp);
      } else {
        break;
      }
    }
    sen_obj_close(ctx, &buf);
  }
  return rc;
}

sen_rc
sen_obj_set_value(sen_ctx *ctx, sen_obj *obj, sen_id id,
                  sen_obj *value, int flags)
{
  sen_rc rc = sen_invalid_argument;
  SEN_API_ENTER;
  if (!SEN_OBJ_DBP(obj)) {
    if (obj->header.type == SEN_ACCESSOR) {
      rc = sen_accessor_set_value(ctx, (sen_accessor *)obj, id, value, flags);
    } else {
      ERR(sen_invalid_argument, "not obj_db");
    }
  } else {
    void *v;
    unsigned int s;
    sen_hook *hooks = OBJ_DB(obj)->hooks[SEN_HOOK_SET];
    sen_obj oldbuf, *oldvalue;
    SEN_OBJ_INIT(&oldbuf, SEN_BULK, 0);
    oldvalue = sen_obj_get_value(ctx, obj, id, &oldbuf);
    if (flags & SEN_OBJ_SET) {
      void *ov;
      unsigned int os;
      v = SEN_BULK_VALUE(value);
      s = SEN_BULK_LEN(value);
      ov = SEN_BULK_VALUE(oldvalue);
      os = SEN_BULK_LEN(oldvalue);
      if (ov && v && os == s && !memcmp(ov, v, s)) {
        sen_bulk_fin(ctx, oldvalue);
        rc = sen_success;
        goto exit;
      }
    }
    if (hooks) {
      // todo : sen_proc_ctx_open()
      sen_proc_ctx pctx = { ctx, obj, hooks, hooks, PROC_INIT, 4, 4};
      pctx.data[0].id = id;
      pctx.data[1].ptr = oldvalue;
      pctx.data[2].ptr = value;
      pctx.data[3].int_value = flags;
      while (hooks) {
        pctx.currh = hooks;
        if (hooks->proc) {
          rc = hooks->proc->funcs[PROC_INIT](ctx, &pctx, 4, pctx.data);
        } else {
          rc = default_set_value_hook(ctx, &pctx, 4, pctx.data);
        }
        if (rc) { goto exit; }
        hooks = hooks->next;
        pctx.offset++;
      }
    }
    sen_obj_close(ctx, oldvalue);
    v = SEN_BULK_VALUE(value);
    s = SEN_BULK_LEN(value);
    switch (obj->header.type) {
    case SEN_TABLE_PAT_KEY :
      rc = sen_pat_set_value(ctx, (sen_pat *)obj, id, v, flags);
      break;
    case SEN_TABLE_HASH_KEY :
      rc = sen_hash_set_value(ctx, (sen_hash *)obj, id, v, flags);
      break;
    case SEN_TABLE_NO_KEY :
      rc = sen_array_set_value(ctx, (sen_array *)obj, id, v, flags);
      break;
    case SEN_COLUMN_VAR_SIZE :
      if (obj->header.flags & SEN_OBJ_COLUMN_TYPE_MASK) {
        sen_obj *vector;
        /* (obj->header.flags & SEN_OBJ_COLUMN_TYPE_MASK) == SEN_OBJ_COLUMN_VERSES */
        /* if OBJ_DB(obj)->range is a table, then verses2updspecs() */
        if (value->header.type != SEN_VERSES) {
          /* todo : convert */
          ERR(sen_invalid_argument, "verses required");
          rc = sen_invalid_argument;
          goto exit;
        }
        vector = sen_verses_to_vector(ctx, value);
        rc = sen_ja_putv(ctx, (sen_ja *)obj, id, vector, 0);
        sen_obj_close(ctx, vector);
      } else {
        rc = sen_ja_put(ctx, (sen_ja *)obj, id, v, s, 0);
      }
      break;
      // todo type->flag
    case SEN_COLUMN_FIX_SIZE :
      if (((sen_ra *)obj)->header->element_size < s) {
        ERR(sen_invalid_argument, "too long value (%d)", s);
      } else {
        void *p = sen_ra_get(ctx, (sen_ra *)obj, id);
        if (!p) {
          ERR(sen_memory_exhausted, "ra get failed");
          rc = sen_memory_exhausted;
          goto exit;
        }
        memcpy(p, v, s);
        rc = sen_success;
      }
      break;
    case SEN_COLUMN_INDEX :
      // todo : how??
      break;
    }
  }
exit :
  SEN_API_RETURN(rc);
}

const char *
sen_obj_get_value_(sen_ctx *ctx, sen_obj *obj, sen_id id, uint32_t *size)
{
  const char *value;
  switch (obj->header.type) {
  case SEN_ACCESSOR :
    value = sen_accessor_get_value_(ctx, (sen_accessor *)obj, id, size);
    break;
  case SEN_TABLE_PAT_KEY :
    value = sen_pat_get_value_(ctx, (sen_pat *)obj, id, size);
    break;
  case SEN_TABLE_HASH_KEY :
    value = sen_hash_get_value_(ctx, (sen_hash *)obj, id, size);
    break;
  case SEN_TABLE_NO_KEY :
    if ((value = _sen_array_get_value(ctx, (sen_array *)obj, id))) {
      *size = ((sen_array *)obj)->value_size;
    }
    break;
  case SEN_COLUMN_VAR_SIZE :
    if ((value = sen_ja_ref(ctx, (sen_ja *)obj, id, size))) {
      sen_ja_unref(ctx, (sen_ja *)obj, id, (void *)value, *size);
    }
    break;
  case SEN_COLUMN_FIX_SIZE :
    if ((value = sen_ra_at(ctx, (sen_ra *)obj, id))) {
      *size = ((sen_ra *)obj)->header->element_size;
    }
    break;
  case SEN_COLUMN_INDEX :
    // todo : how??
    break;
  }
  return value;
}

sen_obj *
sen_obj_get_value(sen_ctx *ctx, sen_obj *obj, sen_id id, sen_obj *value)
{
  unsigned int len = 0;
  SEN_API_ENTER;
  if (!obj) {
    ERR(sen_invalid_argument, "sen_obj_get_value failed");
    goto exit;
  }
  if (!value) {
    if (!(value = sen_obj_open(ctx, SEN_BULK, 0))) {
      ERR(sen_invalid_argument, "sen_obj_get_value failed");
      goto exit;
    }
  }
  switch (value->header.type) {
  case SEN_VOID :
    value->header.type = SEN_BULK;
    sen_bulk_init(ctx, value, 0);
    break;
  case SEN_BULK :
    break;
  default :
    ERR(sen_invalid_argument, "sen_obj_get_value failed");
    goto exit;
  }
  switch (obj->header.type) {
  case SEN_ACCESSOR :
    value = sen_accessor_get_value(ctx, (sen_accessor *)obj, id, value);
    break;
  case SEN_TABLE_PAT_KEY :
    {
      sen_pat *pat = (sen_pat *)obj;
      uint32_t size = pat->value_size;
      if (sen_bulk_space(ctx, value, size)) {
        MERR("sen_bulk_space failed");
        goto exit;
      }
      len = sen_pat_get_value(ctx, pat, id, value->u.b.curr - size);
    }
    break;
  case SEN_TABLE_HASH_KEY :
    {
      sen_hash *hash = (sen_hash *)obj;
      uint32_t size = hash->value_size;
      if (sen_bulk_space(ctx, value, size)) {
        MERR("sen_bulk_space failed");
        goto exit;
      }
      len = sen_hash_get_value(ctx, hash, id, value->u.b.curr - size);
    }
    break;
  case SEN_TABLE_NO_KEY :
    {
      sen_array *array = (sen_array *)obj;
      uint32_t size = array->value_size;
      if (sen_bulk_space(ctx, value, size)) {
        MERR("sen_bulk_space failed");
        goto exit;
      }
      len = sen_array_get_value(ctx, array, id, value->u.b.curr - size);
    }
    break;
  case SEN_COLUMN_VAR_SIZE :
    if (obj->header.flags & SEN_OBJ_COLUMN_TYPE_MASK) {
      /* (obj->header.flags & SEN_OBJ_COLUMN_TYPE_MASK) == SEN_OBJ_COLUMN_VERSES */
      sen_obj *vector;
      void *v = sen_ja_ref(ctx, (sen_ja *)obj, id, &len);
      if (!v) { len = 0; goto exit; }
      vector = sen_vector_open(ctx, 0);
      if (!sen_vector_decode(ctx, vector, v, len)) {
        value->header.type = SEN_VERSES;
        value->header.flags |= SEN_OBJ_DO_SHALLOW_COPY;
        value = sen_vector_to_verses(ctx, vector, value);
      } else {
        sen_obj_close(ctx, vector);
      }
      sen_ja_unref(ctx, (sen_ja *)obj, id, v, len);
    } else {
      void *v = sen_ja_ref(ctx, (sen_ja *)obj, id, &len);
      if (!v) { len = 0; goto exit; }
      // todo : reduce copy
      sen_bulk_write(ctx, value, v, len);
      sen_ja_unref(ctx, (sen_ja *)obj, id, v, len);
    }
    break;
  case SEN_COLUMN_FIX_SIZE :
    {
      unsigned element_size;
      void *v = sen_ra_at(ctx, (sen_ra *)obj, id);
      if (!v) { goto exit; }
      element_size = ((sen_ra *)obj)->header->element_size;
      sen_bulk_write(ctx, value, v, element_size);
      len = element_size;
    }
    break;
  case SEN_COLUMN_INDEX :
    // todo : how??
    break;
  }
  if (value->header.type == SEN_BULK) {
    value->header.domain = sen_obj_get_range(ctx, obj);
  }
exit :
  SEN_API_RETURN(value);
}

sen_rc
sen_column_index_update(sen_ctx *ctx, sen_obj *column,
                        sen_id id, unsigned int section,
                        sen_obj *oldvalue, sen_obj *newvalue)
{
  sen_rc rc = sen_invalid_argument;
  SEN_API_ENTER;
  if (column->header.type != SEN_COLUMN_INDEX) {
    ERR(sen_invalid_argument, "invalid column assigned");
  } else {
    rc = sen_ii_column_update(ctx, (sen_ii *)column, id, section, oldvalue, newvalue);
  }
  SEN_API_RETURN(rc);
}

sen_obj *
sen_column_table(sen_ctx *ctx, sen_obj *column)
{
  sen_obj *obj = NULL;
  sen_obj_db *col = OBJ_DB(column);
  SEN_API_ENTER;
  if (col) {
    obj = sen_ctx_get(ctx, col->header.domain);
  }
  SEN_API_RETURN(obj);
}

sen_obj *
sen_obj_get_info(sen_ctx *ctx, sen_obj *obj, sen_info_type type, sen_obj *valuebuf)
{
  SEN_API_ENTER;
  if (!obj) {
    ERR(sen_invalid_argument, "sen_obj_get_info failed");
    goto exit;
  }
  if (!valuebuf) {
    if (!(valuebuf = sen_obj_open(ctx, SEN_BULK, 0))) {
      ERR(sen_invalid_argument, "sen_obj_get_info failed");
      goto exit;
    }
  }
  switch (type) {
  case SEN_INFO_SOURCE :
    if (!SEN_OBJ_DBP(obj)) {
      ERR(sen_invalid_argument, "only db_obj can accept SEN_INFO_SOURCE");
      goto exit;
    }
    SEN_OBJ_INIT(valuebuf, SEN_BULK, 0);
    SEN_BULK_SET(ctx, valuebuf, OBJ_DB(obj)->source, OBJ_DB(obj)->source_size);
    break;
  default :
    /* todo */
    break;
  }
exit :
  SEN_API_RETURN(valuebuf);
}

static void
update_source_hook(sen_ctx *ctx, sen_obj *obj)
{
  sen_id *s = OBJ_DB(obj)->source;
  int i, n = OBJ_DB(obj)->source_size / sizeof(sen_id);
  default_set_value_hook_data hook_data = { OBJ_DB(obj)->id, 0 };
  sen_obj *source, data;
  SEN_OBJ_INIT(&data, SEN_BULK, SEN_OBJ_DO_SHALLOW_COPY);
  SEN_BULK_SET(ctx, &data, &hook_data, sizeof hook_data);
  for (i = 1; i <= n; i++, s++) {
    hook_data.section = i;
    if ((source = sen_ctx_get(ctx, *s))) {
      sen_obj_add_hook(ctx, source, SEN_HOOK_SET, 0, NULL, &data);
    }
  }
  sen_obj_close(ctx, &data);
}

#define N_HOOK_ENTRIES 5

sen_rc
sen_hook_pack(sen_ctx *ctx, sen_obj_db *obj, sen_obj *buf)
{
  sen_rc rc;
  sen_hook_entry e;
  for (e = 0; e < N_HOOK_ENTRIES; e++) {
    sen_hook *hooks;
    for (hooks = obj->hooks[e]; hooks; hooks = hooks->next) {
      sen_id id = hooks->proc ? hooks->proc->obj.id : 0;
      if ((rc = sen_bulk_benc(ctx, buf, id + 1))) { goto exit; }
      if ((rc = sen_bulk_benc(ctx, buf, hooks->hld_size))) { goto exit; }
      if ((rc = sen_bulk_write(ctx, buf, NEXT_ADDR(hooks), hooks->hld_size))) { goto exit; }
    }
    if ((rc = sen_bulk_benc(ctx, buf, 0))) { goto exit; }
  }
exit :
  return rc;
}

static sen_rc
sen_hook_unpack(sen_ctx *ctx, sen_obj_db *obj, const char *buf, uint32_t buf_size)
{
  sen_hook_entry e;
  const char *p = buf, *pe = p + buf_size;
  for (e = 0; e < N_HOOK_ENTRIES; e++) {
    sen_hook *new, **last = &obj->hooks[e];
    for (;;) {
      sen_id id;
      uint32_t hld_size;
      SEN_B_DEC(id, p);
      if (!id--) { break; }
      if (p >= pe) { return sen_invalid_format; }
      SEN_B_DEC(hld_size, p);
      if (p >= pe) { return sen_invalid_format; }
      if (!(new = SEN_MALLOC(sizeof(sen_hook) + hld_size))) {
        return sen_memory_exhausted;
      }
      if (id) {
        new->proc = (sen_proc *)sen_ctx_get(ctx, id);
        if (!new->proc) {
          SEN_FREE(new);
          return ctx->rc;
        }
      } else {
        new->proc = NULL;
      }
      if ((new->hld_size = hld_size)) {
        memcpy(NEXT_ADDR(new), p, hld_size);
        p += hld_size;
      }
      *last = new;
      last = &new->next;
      if (p >= pe) { return sen_invalid_format; }
    }
    *last = NULL;
  }
  return sen_success;
}

static void
sen_obj_spec_save(sen_ctx *ctx, sen_obj_db *obj)
{
  sen_obj *v;
  sen_space *s;
  sen_obj_spec spec;
  if (obj->id & SEN_OBJ_TMP_OBJECT) { return; }
  if (!ctx->impl || !SEN_OBJ_DBP(obj)) { return; }
  if (!(s = (sen_space *)ctx->impl->db) || !s->specs) { return; }
  v = sen_vector_open(ctx, 0);
  spec.header = obj->header;
  spec.range = obj->range;
  sen_bulk_write(ctx, v, (void *)&spec, sizeof(sen_obj_spec));
  sen_vector_delimit(ctx, v);
  if (obj->header.impl_flags & SEN_OBJ_CUSTOM_NAME) {
    SEN_BULK_PUTS(ctx, v, sen_obj_path(ctx, (sen_obj *)obj));
  }
  sen_vector_delimit(ctx, v);
  sen_bulk_write(ctx, v, obj->source, obj->source_size);
  sen_vector_delimit(ctx, v);
  sen_hook_pack(ctx, obj, v);
  sen_vector_delimit(ctx, v);
  sen_ja_putv(ctx, s->specs, obj->id, v, 0);
  sen_obj_close(ctx, v);
}

sen_rc
sen_obj_set_info(sen_ctx *ctx, sen_obj *obj, sen_info_type type, sen_obj *value)
{
  sen_rc rc = sen_invalid_argument;
  SEN_API_ENTER;
  if (!obj) {
    ERR(sen_invalid_argument, "sen_obj_get_value failed");
    goto exit;
  }
  switch (type) {
  case SEN_INFO_SOURCE :
    if (!SEN_OBJ_DBP(obj)) {
      ERR(sen_invalid_argument, "only db_obj can accept SEN_INFO_SOURCE");
      goto exit;
    }
    {
      void *v = SEN_BULK_VALUE(value);
      uint32_t s = SEN_BULK_LEN(value);
      if (s) {
        void *v2 = SEN_MALLOC(s);
        if (!v2) {
          rc = ctx->rc;
          goto exit;
        }
        memcpy(v2, v, s);
        OBJ_DB(obj)->source = v2;
        OBJ_DB(obj)->source_size = s;

        if (obj->header.type == SEN_COLUMN_INDEX) {
          update_source_hook(ctx, obj);
        }

      } else {
        OBJ_DB(obj)->source = NULL;
        OBJ_DB(obj)->source_size = 0;
      }
    }
    sen_obj_spec_save(ctx, OBJ_DB(obj));
    rc = sen_success;
    break;
  default :
    /* todo */
    break;
  }
exit :
  SEN_API_RETURN(rc);
}

sen_obj *
sen_obj_get_element_info(sen_ctx *ctx, sen_obj *obj, sen_id id,
                         sen_info_type type, sen_obj *valuebuf)
{
  SEN_API_ENTER;
  SEN_API_RETURN(valuebuf);
}

sen_rc
sen_obj_set_element_info(sen_ctx *ctx, sen_obj *obj, sen_id id,
                         sen_info_type type, sen_obj *value)
{
  SEN_API_ENTER;
  SEN_API_RETURN(sen_success);
}

sen_proc_data *
sen_proc_ctx_get_local_data(sen_proc_ctx *pctx)
{
  return &pctx->data[pctx->offset];
}

static void
sen_hook_free(sen_ctx *ctx, sen_hook *h)
{
  sen_hook *curr, *next;
  for (curr = h; curr; curr = next) {
    next = curr->next;
    SEN_FREE(curr);
  }
}

sen_rc
sen_obj_add_hook(sen_ctx *ctx, sen_obj *obj, sen_hook_entry entry,
                 int offset, sen_obj *proc, sen_obj *hld)
{
  sen_rc rc = sen_success;
  SEN_API_ENTER;
  if (!SEN_OBJ_DBP(obj)) {
    rc = sen_invalid_argument;
  } else {
    int i;
    void *hld_value;
    uint32_t hld_size = 0;
    sen_hook *new, **last = &OBJ_DB(obj)->hooks[entry];
    if (hld) {
      hld_value = SEN_BULK_VALUE(hld);
      hld_size = SEN_BULK_LEN(hld);
    }
    if (!(new = SEN_MALLOC(sizeof(sen_hook) + hld_size))) {
      rc = sen_memory_exhausted;
      goto exit;
    }
    new->proc = (sen_proc *)proc;
    new->hld_size = hld_size;
    if (hld_size) {
      memcpy(NEXT_ADDR(new), hld_value, hld_size);
    }
    for (i = 0; i != offset && *last; i++) { last = &(*last)->next; }
    new->next = *last;
    *last = new;
  }
  sen_obj_spec_save(ctx, OBJ_DB(obj));
exit :
  SEN_API_RETURN(rc);
}

sen_obj *
sen_obj_get_hook(sen_ctx *ctx, sen_obj *obj, sen_hook_entry entry,
                      int offset, sen_obj *hldbuf)
{
  sen_obj *res = NULL;
  SEN_API_ENTER;
  {
    int i;
    sen_hook *hook = OBJ_DB(obj)->hooks[entry];
    for (i = 0; i < offset; i++) {
      hook = hook->next;
      if (!hook) { return NULL; }
    }
    res = (sen_obj *)hook->proc;
    sen_bulk_write(ctx, hldbuf, NEXT_ADDR(hook), hook->hld_size);
  }
  SEN_API_RETURN(res);
}

sen_rc
sen_obj_delete_hook(sen_ctx *ctx, sen_obj *obj, sen_hook_entry entry, int offset)
{
  SEN_API_ENTER;
  {
    int i;
    sen_hook *h, **last = &OBJ_DB(obj)->hooks[entry];
    for (i = 0; i < offset; i++) {
      if (!(h = *last)) { return sen_invalid_argument; }
      last = &(*last)->next;
    }
    *last = h->next;
    SEN_FREE(h);
  }
  sen_obj_spec_save(ctx, OBJ_DB(obj));
  SEN_API_RETURN(sen_success);
}

sen_rc
sen_obj_remove(sen_ctx *ctx, const char *path)
{
  SEN_API_ENTER;
  SEN_API_RETURN(sen_success);
}

sen_rc
sen_obj_rename(sen_ctx *ctx, const char *old_path, const char *new_path)
{
  SEN_API_ENTER;
  SEN_API_RETURN(sen_success);
}

/* space must be validate by caller */
static sen_id
sen_obj_register(sen_ctx *ctx, sen_obj *space, const char *name, unsigned name_size)
{
  sen_id id = SEN_ID_NIL;
  if (name && name_size) {
    sen_space *s = (sen_space *)space;
    sen_search_flags f = SEN_TABLE_ADD;
    if (!(id = sen_pat_lookup(ctx, s->keys, name, name_size, NULL, &f))) {
      ERR(sen_memory_exhausted, "sen_pat_lookup failed");
    } else if (!(f & SEN_TABLE_ADDED)) {
      ERR(sen_invalid_argument, "already used name was assigend");
      id = SEN_ID_NIL;
    }
  } else if (ctx->impl && ctx->impl->values) {
    id = sen_array_add(ctx, ctx->impl->values, NULL) | SEN_OBJ_TMP_OBJECT;
  }
  return id;
}

static sen_rc
sen_obj_delete_by_id(sen_ctx *ctx, sen_obj *space, sen_id id, int removep)
{
  if (id) {
    if (id & SEN_OBJ_TMP_OBJECT) {
      if (ctx->impl && ctx->impl->values) {
        return sen_array_delete_by_id(ctx, ctx->impl->values,
                                      id & ~SEN_OBJ_TMP_OBJECT, NULL);
      }
    } else {
      sen_obj **vp;
      sen_space *s = (sen_space *)space;
      if ((vp = sen_tiny_array_at(&s->values, id))) {
        *vp = NULL;
      }
      return removep ? sen_pat_delete_by_id(ctx, s->keys, id, NULL) : sen_success;
    }
  }
  return sen_invalid_argument;
}

/* space must be validate by caller */
static sen_rc
sen_obj_db_init(sen_ctx *ctx, sen_obj *space, sen_id id, sen_obj_db *obj)
{
  sen_rc rc = sen_success;
  if (id) {
    if (id & SEN_OBJ_TMP_OBJECT) {
      if (ctx->impl && ctx->impl->values) {
        sen_tmp_obj_db tmp_obj;
        tmp_obj.obj = obj;
        rc = sen_array_set_value(ctx, ctx->impl->values,
                                 id & ~SEN_OBJ_TMP_OBJECT, &tmp_obj, SEN_OBJ_SET);
      }
    } else {
      void **vp;
      vp = sen_tiny_array_at(&((sen_space *)space)->values, id);
      if (!vp) {
        rc = sen_memory_exhausted;
        ERR(rc, "sen_tiny_array_at failed (%d)", id);
        return rc;
      }
      *vp = (sen_obj *)obj;
    }
  }
  obj->id = id;
  obj->space = space;
  obj->source = NULL;
  obj->source_size = 0;
  {
    sen_hook_entry entry;
    for (entry = 0; entry < N_HOOK_ENTRIES; entry++) {
      obj->hooks[entry] = NULL;
    }
  }
  sen_obj_spec_save(ctx, obj);
  return rc;
}

#define GET_PATH(spec,buffer,s,id) {\
  if (spec->header.impl_flags & SEN_OBJ_CUSTOM_NAME) {\
    const char *path = sen_vector_fetch(ctx, v, 1, &size);\
    if (size > PATH_MAX) { ERR(sen_invalid_format, "too long path"); }\
    memcpy(buffer, path, size);\
    buffer[size] = '\0';\
  } else {\
    gen_pathname(s->keys->io->path, buffer, id);\
  }\
}

sen_obj *
sen_ctx_get(sen_ctx *ctx, sen_id id)
{
  sen_obj *res = NULL;
  if (!ctx || !ctx->impl) { return res; }
  SEN_API_ENTER;
  if (id & SEN_OBJ_TMP_OBJECT) {
    if (ctx->impl->values) {
      sen_array_get_value(ctx, ctx->impl->values, id & ~SEN_OBJ_TMP_OBJECT, &res);
    }
  } else {
    sen_obj *space = ctx->impl->db;
    if (space) {
      sen_space *s = (sen_space *)space;
      sen_obj **vp;
      if (id == SEN_ID_NIL) { res = ctx->impl->db; goto exit; }
      if (!(vp = sen_tiny_array_at(&s->values, id))) { goto exit; }
      if (s->specs && !*vp) {
        uint32_t value_len;
        char *value = sen_ja_ref(ctx, s->specs, id, &value_len);
        if (value) {
          sen_obj *v = sen_vector_open(ctx, 0);
          if (v) {
            uint32_t size;
            sen_obj_spec *spec;
            char buffer[PATH_MAX];
            sen_vector_decode(ctx, v, value, value_len);
            spec = (sen_obj_spec *)sen_vector_fetch(ctx, v, 0, &size);
            if (spec) {
              switch (spec->header.type) {
              case SEN_TYPE :
                MUTEX_LOCK(s->lock);
                if (!*vp) { *vp = (sen_obj *)sen_type_open(ctx, spec); }
                MUTEX_UNLOCK(s->lock);
                break;
              case SEN_TABLE_HASH_KEY :
                GET_PATH(spec, buffer, s, id);
                MUTEX_LOCK(s->lock);
                if (!*vp) { *vp = (sen_obj *)sen_hash_open(ctx, buffer); }
                MUTEX_UNLOCK(s->lock);
                break;
              case SEN_TABLE_PAT_KEY :
                GET_PATH(spec, buffer, s, id);
                MUTEX_LOCK(s->lock);
                if (!*vp) { *vp = (sen_obj *)sen_pat_open(ctx, buffer); }
                MUTEX_UNLOCK(s->lock);
                break;
              case SEN_TABLE_NO_KEY :
                GET_PATH(spec, buffer, s, id);
                MUTEX_LOCK(s->lock);
                if (!*vp) { *vp = (sen_obj *)sen_array_open(ctx, buffer); }
                MUTEX_UNLOCK(s->lock);
                break;
              case SEN_COLUMN_VAR_SIZE :
                GET_PATH(spec, buffer, s, id);
                MUTEX_LOCK(s->lock);
                if (!*vp) { *vp = (sen_obj *)sen_ja_open(ctx, buffer); }
                MUTEX_UNLOCK(s->lock);
                break;
              case SEN_COLUMN_FIX_SIZE :
                GET_PATH(spec, buffer, s, id);
                MUTEX_LOCK(s->lock);
                if (!*vp) { *vp = (sen_obj *)sen_ra_open(ctx, buffer); }
                MUTEX_UNLOCK(s->lock);
                break;
              case SEN_COLUMN_INDEX :
                GET_PATH(spec, buffer, s, id);
                {
                  sen_obj *table = sen_ctx_get(ctx, spec->header.domain);
                  MUTEX_LOCK(s->lock);
                  if (!*vp) { *vp = (sen_obj *)sen_ii_open(ctx, buffer, table); }
                  MUTEX_UNLOCK(s->lock);
                }
                break;
              }
              if (*vp) {
                const char *p;
                sen_obj_db *r = OBJ_DB(*vp);
                r->header = spec->header;
                r->id = id;
                r->range = spec->range;
                r->space = space;
                p  = sen_vector_fetch(ctx, v, 2, &size);
                if (size) {
                  if ((r->source = SEN_MALLOC(size))) {
                    memcpy(r->source, p, size);
                    r->source_size = size;
                  }
                }
                p  = sen_vector_fetch(ctx, v, 3, &size);
                sen_hook_unpack(ctx, r, p, size);
              }
            }
            sen_obj_close(ctx, v);
          }
          sen_ja_unref(ctx, s->specs, id, value, value_len);
        }
      }
      res = *vp;
    }
  }
exit :
  SEN_API_RETURN(res);
}

sen_obj *
sen_obj_open(sen_ctx *ctx, unsigned char type, sen_obj_flags flags)
{
  switch (type) {
  case SEN_VECTOR :
    return sen_vector_open(ctx, flags);
  default :
    {
      sen_obj *obj = SEN_MALLOCN(sen_obj, 1);
      if (obj) {
        SEN_OBJ_INIT(obj, type, flags);
        obj->header.impl_flags |= SEN_OBJ_ALLOCATED;
      }
      return obj;
    }
  }
}

sen_obj *
sen_obj_graft(sen_ctx *ctx, sen_obj *obj)
{
  sen_obj *new = sen_obj_open(ctx, obj->header.type, obj->header.flags);
  if (new) {
    /* todo : deep copy if (obj->header.flags & SEN_OBJ_DO_SHALLOW_COPY) */
    new->u.b.head = obj->u.b.head;
    new->u.b.curr = obj->u.b.curr;
    new->u.b.tail = obj->u.b.tail;
    obj->u.b.head = NULL;
    obj->u.b.curr = NULL;
    obj->u.b.tail = NULL;
  }
  return new;
}

sen_rc
sen_obj_close(sen_ctx *ctx, sen_obj *obj)
{
  sen_rc rc = sen_invalid_argument;
  SEN_API_ENTER;
  if (obj) {
    if (SEN_OBJ_DBP(obj)) {
      sen_hook_entry entry;
      if (OBJ_DB(obj)->source) {
        SEN_FREE(OBJ_DB(obj)->source);
      }
      for (entry = 0; entry < N_HOOK_ENTRIES; entry++) {
        sen_hook_free(ctx, OBJ_DB(obj)->hooks[entry]);
      }
      sen_obj_delete_by_id(ctx, OBJ_DB(obj)->space, OBJ_DB(obj)->id, 0);
    }
    switch (obj->header.type) {
    case SEN_BULK :
      obj->header.type = SEN_VOID;
      if (obj->header.flags & SEN_OBJ_DO_SHALLOW_COPY) {
        obj->u.b.head = NULL;
        obj->u.b.tail = NULL;
        rc = sen_success;
      } else {
        rc = sen_bulk_fin(ctx, obj);
      }
      if (obj->header.impl_flags & SEN_OBJ_ALLOCATED) {
        SEN_FREE(obj);
      }
      break;
    case SEN_ACCESSOR :
      {
        sen_accessor *p, *n;
        for (p = (sen_accessor *)obj; p; p = n) {
          n = p->next;
          SEN_FREE(p);
        }
      }
      break;
    case SEN_TYPE :
      SEN_FREE(obj);
      rc = sen_success;
      break;
    case SEN_SPACE :
      rc = sen_space_close(ctx, (sen_space *)obj);
      break;
    case SEN_TABLE_PAT_KEY :
      rc = sen_pat_close(ctx, (sen_pat *)obj);
      break;
    case SEN_TABLE_HASH_KEY :
      rc = sen_hash_close(ctx, (sen_hash *)obj);
      break;
    case SEN_TABLE_NO_KEY :
      rc = sen_array_close(ctx, (sen_array *)obj);
      break;
    case SEN_VERSES :
      if (obj->u.v.verses) {
        if (!(obj->header.flags & SEN_OBJ_DO_SHALLOW_COPY)) {
          int n = obj->u.v.n_verses;
          sen_verse *vp = obj->u.v.verses;
          for (n = obj->u.v.n_verses, vp = obj->u.v.verses; n; n--, vp++) {
            SEN_FREE(vp->str);
          }
        }
        SEN_FREE(obj->u.v.verses);
      }
      if (obj->u.v.src) { sen_obj_close(ctx, obj->u.v.src); }
      if (obj->header.impl_flags & SEN_OBJ_ALLOCATED) {
        SEN_FREE(obj);
      }
      rc = sen_success;
      break;
    case SEN_VECTOR :
      sen_bulk_fin(ctx, obj);
      if (VECTOR(obj)->offsets) {
        SEN_FREE(VECTOR(obj)->offsets);
      }
      if (obj->header.impl_flags & SEN_OBJ_ALLOCATED) {
        SEN_FREE(obj);
      }
      rc = sen_success;
      break;
    case SEN_QUERY :
      rc = sen_query_close((sen_query *)obj);
      break;
    case SEN_COLUMN_VAR_SIZE :
      rc = sen_ja_close(ctx, (sen_ja *)obj);
      break;
    case SEN_COLUMN_FIX_SIZE :
      rc = sen_ra_close(ctx, (sen_ra *)obj);
      break;
    case SEN_COLUMN_INDEX :
      rc = sen_ii_close(ctx, (sen_ii *)obj);
      break;
    }
  }
  SEN_API_RETURN(rc);
}

static sen_io*
sen_obj_io(sen_obj *obj)
{
  sen_io *io = NULL;
  if (obj) {
    switch (obj->header.type) {
    case SEN_SPACE :
      io = ((sen_space *)obj)->keys->io;
      break;
    case SEN_TABLE_PAT_KEY :
      io = ((sen_pat *)obj)->io;
      break;
    case SEN_TABLE_HASH_KEY :
      io = ((sen_hash *)obj)->io;
      break;
    case SEN_TABLE_NO_KEY :
      io = ((sen_array *)obj)->io;
      break;
    case SEN_COLUMN_VAR_SIZE :
      io = ((sen_ja *)obj)->io;
      break;
    case SEN_COLUMN_FIX_SIZE :
      io = ((sen_ra *)obj)->io;
      break;
    case SEN_COLUMN_INDEX :
      io = ((sen_ii *)obj)->seg;
      break;
    }
  }
  return io;
}

const char *
sen_obj_path(sen_ctx *ctx, sen_obj *obj)
{
  sen_io *io;
  char *path = NULL;
  SEN_API_ENTER;
  io = sen_obj_io(obj);
  path = io ? io->path : NULL;
  SEN_API_RETURN(path);
}

int
sen_obj_name(sen_ctx *ctx, sen_obj *obj, char *namebuf, int buf_size)
{
  int len = 0;
  SEN_API_ENTER;
  if (SEN_OBJ_DBP(obj)) {
    if (OBJ_DB(obj)->id && OBJ_DB(obj)->id < SEN_ID_MAX) {
      sen_space *s = (sen_space *)OBJ_DB(obj)->space;
      len = sen_pat_get_key(ctx, s->keys, OBJ_DB(obj)->id, namebuf, buf_size);
    }
  }
  SEN_API_RETURN(len);
}

int
sen_obj_expire(sen_ctx *ctx, sen_obj *obj, int threshold)
{
  SEN_API_ENTER;
  SEN_API_RETURN(0);
}

int
sen_obj_check(sen_ctx *ctx, sen_obj *obj)
{
  SEN_API_ENTER;
  SEN_API_RETURN(0);
}

sen_rc
sen_obj_lock(sen_ctx *ctx, sen_obj *obj, sen_id id, int timeout)
{
  sen_rc rc = sen_success;
  SEN_API_ENTER;
  rc = sen_io_lock(ctx, sen_obj_io(obj), timeout);
  SEN_API_RETURN(rc);
}

sen_rc
sen_obj_unlock(sen_ctx *ctx, sen_obj *obj, sen_id id)
{
  SEN_API_ENTER;
  sen_io_unlock(sen_obj_io(obj));
  SEN_API_RETURN(sen_success);
}

sen_rc
sen_obj_clear_lock(sen_ctx *ctx, sen_obj *obj)
{
  SEN_API_ENTER;
  sen_io_clear_lock(sen_obj_io(obj));
  SEN_API_RETURN(sen_success);
}

unsigned int
sen_obj_is_locked(sen_ctx *ctx, sen_obj *obj)
{
  unsigned int res = 0;
  SEN_API_ENTER;
  res = sen_io_is_locked(sen_obj_io(obj));
  SEN_API_RETURN(res);
}

sen_obj *
sen_obj_space(sen_ctx *ctx, sen_obj *obj)
{
  SEN_API_ENTER;
  SEN_API_RETURN(obj);
}

/**** db ****/

sen_db *
sen_db_create(const char *path, int flags, sen_encoding encoding)
{
  return sen_space_create(&sen_gctx, path, encoding);
}

sen_db *
sen_db_open(const char *path)
{
  return sen_space_open(&sen_gctx, path);
}

sen_rc
sen_db_close(sen_db *s)
{
  return sen_space_close(&sen_gctx, (sen_space *)s);
}

/**** sort ****/

typedef struct {
  sen_id id;
  int32_t size;
  const void *value;
} sort_entry;

enum {
  KEY_BULK = 0,
  KEY_INT8,
  KEY_INT16,
  KEY_INT32,
  KEY_INT64,
  KEY_UINT8,
  KEY_UINT16,
  KEY_UINT32,
  KEY_UINT64,
  KEY_FLOAT32,
  KEY_FLOAT64,
};

#define CMPNUM(type) {\
  type va = *((type *)(ap));\
  type vb = *((type *)(bp));\
  if (va != vb) { return va > vb; }\
}

inline static int
compare_value(sen_ctx *ctx, sort_entry *a, sort_entry *b,
              sen_table_sort_key *keys, int n_keys)
{
  int i;
  uint8_t type;
  uint32_t as, bs;
  const char *ap, *bp;
  for (i = 0; i < n_keys; i++, keys++) {
    if (i) {
      if (keys->flags & SEN_TABLE_SORT_ASC) {
        ap = sen_obj_get_value_(ctx, keys->key, a->id, &as);
        bp = sen_obj_get_value_(ctx, keys->key, b->id, &bs);
      } else {
        ap = sen_obj_get_value_(ctx, keys->key, b->id, &as);
        bp = sen_obj_get_value_(ctx, keys->key, a->id, &bs);
      }
    } else {
      if (keys->flags & SEN_TABLE_SORT_ASC) {
        ap = a->value; as = a->size;
        bp = b->value; bs = b->size;
      } else {
        ap = b->value; as = b->size;
        bp = a->value; bs = a->size;
      }
    }
    type = keys->offset;
    switch (type) {
    case KEY_BULK :
      for (;; ap++, bp++, as--, bs--) {
        if (!as) { if (bs) { return 0; } else { break; } }
        if (!bs) { return 1; }
        if (*ap < *bp) { return 0; }
        if (*ap > *bp) { return 1; }
      }
      break;
    case KEY_INT8 :
      CMPNUM(int8_t);
      break;
    case KEY_INT16 :
      CMPNUM(int16_t);
      break;
    case KEY_INT32 :
      CMPNUM(int32_t);
      break;
    case KEY_INT64 :
      CMPNUM(int64_t);
      break;
    case KEY_UINT8 :
      CMPNUM(uint8_t);
      break;
    case KEY_UINT16 :
      CMPNUM(uint16_t);
      break;
    case KEY_UINT32 :
      CMPNUM(uint32_t);
      break;
    case KEY_UINT64 :
      CMPNUM(uint64_t);
      break;
    case KEY_FLOAT32 :
      {
        float va = *((float *)(ap));
        float vb = *((float *)(bp));
        if (va < vb || va > vb) { return va > vb; }
      }
      break;
    case KEY_FLOAT64 :
      {
        double va = *((double *)(ap));
        double vb = *((double *)(bp));
        if (va < vb || va > vb) { return va > vb; }
      }
      break;
    }
  }
  return 0;
}

inline static void
swap(sort_entry *a, sort_entry *b)
{
  sort_entry c_ = *a;
  *a = *b;
  *b = c_;
}

inline static sort_entry *
part(sen_ctx *ctx, sort_entry *b, sort_entry *e, sen_table_sort_key *keys, int n_keys)
{
  sort_entry *c;
  intptr_t d = e - b;
  if (compare_value(ctx, b, e, keys, n_keys)) {
    swap(b, e);
  }
  if (d < 2) { return NULL; }
  c = b + (d >> 1);
  if (compare_value(ctx, b, c, keys, n_keys)) {
    swap(b, c);
  } else {
    if (compare_value(ctx, c, e, keys, n_keys)) {
      swap(c, e);
    }
  }
  if (d < 3) { return NULL; }
  b++;
  swap(b, c);
  c = b;
  for (;;) {
    do {
      b++;
    } while (compare_value(ctx, c, b, keys, n_keys));
    do {
      e--;
    } while (compare_value(ctx, e, c, keys, n_keys));
    if (b >= e) { break; }
    swap(b, e);
  }
  swap(c, e);
  return e;
}

static void
_sort(sen_ctx *ctx, sort_entry *head, sort_entry *tail, int limit,
      sen_table_sort_key *keys, int n_keys)
{
  sort_entry *c;
  if (head < tail && (c = part(ctx, head, tail, keys, n_keys))) {
    intptr_t rest = limit - 1 - (c - head);
    _sort(ctx, head, c - 1, limit, keys, n_keys);
    if (rest > 0) { _sort(ctx, c + 1, tail, (int)rest, keys, n_keys); }
  }
}

static sort_entry *
pack(sen_ctx *ctx, sen_obj *table, sort_entry *head, sort_entry *tail,
     sen_table_sort_key *keys, int n_keys)
{
  int i = 0;
  sort_entry e, c;
  sen_table_cursor *tc = sen_table_cursor_open(ctx, table, NULL, 0, NULL, 0, 0);
  if (!tc) { return NULL; }
  if ((c.id = sen_table_cursor_next(ctx, tc))) {
    c.value = sen_obj_get_value_(ctx, keys->key, c.id, &c.size);
    while ((e.id = sen_table_cursor_next(ctx, tc))) {
      e.value = sen_obj_get_value_(ctx, keys->key, e.id, &e.size);
      if (compare_value(ctx, &c, &e, keys, n_keys)) {
        *head++ = e;
      } else {
        *tail-- = e;
      }
      i++;
    }
    *head = c;
    i++;
  }
  sen_table_cursor_close(ctx, tc);
  return i > 2 ? head : NULL;
}

int
sen_table_sort(sen_ctx *ctx, sen_obj *table, int limit,
               sen_obj *result, sen_table_sort_key *keys, int n_keys)
{
  int n, i = 0;
  sort_entry *array, *ep;
  SEN_API_ENTER;
  if (!n_keys || !keys) {
    ERR(sen_invalid_argument, "keys is null");
    goto exit;
  }
  if (!table) {
    ERR(sen_invalid_argument, "table is null");
    goto exit;
  }
  if (!(result && result->header.type == SEN_TABLE_NO_KEY)) {
    ERR(sen_invalid_argument, "result is not a array");
    goto exit;
  }
  n = sen_table_size(ctx, table);
  if (limit <= 0) {
    limit += n;
    if (limit <= 0) {
      SEN_LOG(sen_log_alert, "limit is too small in sen_table_sort !");
      goto exit;
    }
  }
  {
    int j;
    sen_table_sort_key *kp;
    for (kp = keys, j = n_keys; j; kp++, j--) {
      sen_obj *range = sen_ctx_get(ctx, sen_obj_get_range(ctx, kp->key));
      if (range->header.type == SEN_TYPE) {
        if (range->header.flags & SEN_OBJ_KEY_VAR_SIZE) {
          kp->offset = KEY_BULK;
        } else {
          uint8_t key_type = range->header.flags & SEN_OBJ_KEY_MASK;
          switch (key_type) {
          case SEN_OBJ_KEY_UINT :
            switch (SEN_TYPE_SIZE(OBJ_DB(range))) {
            case 1 :
              kp->offset = KEY_UINT8;
              break;
            case 2 :
              kp->offset = KEY_UINT16;
              break;
            case 4 :
              kp->offset = KEY_UINT32;
              break;
            case 8 :
              kp->offset = KEY_UINT64;
              break;
            default :
              ERR(sen_invalid_argument, "unsupported uint value");
              goto exit;
            }
          case SEN_OBJ_KEY_INT :
            switch (SEN_TYPE_SIZE(OBJ_DB(range))) {
            case 1 :
              kp->offset = KEY_INT8;
              break;
            case 2 :
              kp->offset = KEY_INT16;
              break;
            case 4 :
              kp->offset = KEY_INT32;
              break;
            case 8 :
              kp->offset = KEY_INT64;
              break;
            default :
              ERR(sen_invalid_argument, "unsupported int value");
              goto exit;
            }
          case SEN_OBJ_KEY_FLOAT :
            switch (SEN_TYPE_SIZE(OBJ_DB(range))) {
            case 4 :
              kp->offset = KEY_FLOAT32;
              break;
            case 8 :
              kp->offset = KEY_FLOAT64;
              break;
            default :
              ERR(sen_invalid_argument, "unsupported float value");
              goto exit;
            }
          }
        }
      } else {
        kp->offset = KEY_UINT32;
      }
    }
  }
  if (!(array = SEN_MALLOC(sizeof(sort_entry) * n))) {
    goto exit;
  }
  if (limit > n) { limit = n; }
  if ((ep = pack(ctx, table, array, array + n - 1, keys, n_keys))) {
    intptr_t rest = limit - 1 - (ep - array);
    _sort(ctx, array, ep - 1, limit, keys, n_keys);
    if (rest > 0 ) {
      _sort(ctx, ep + 1, array + n - 1, (int)rest, keys, n_keys);
    }
  }
  {
    sen_id *v;
    for (i = 0, ep = array; i < limit; i++, ep++) {
      if (!sen_array_add(ctx, (sen_array *)result, (void **)&v)) { break; }
      if (!(*v = ep->id)) { break; }
    }
    SEN_FREE(array);
  }
exit :
  SEN_API_RETURN(i);
}
