/* Copyright(C) 2008 Brazil

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "hash.h"
#include "pat.h"
#include <string.h>
#include <limits.h>

#include "store.h"

/* sen_tiny_array */

void
sen_tiny_array_init(sen_tiny_array *a, sen_ctx *ctx, uint16_t element_size, uint16_t flags)
{
  a->ctx = ctx;
  a->element_size = element_size;
  a->flags = flags;
  a->max = 0;
  if (flags & SEN_TINY_ARRAY_THREADSAFE) { MUTEX_INIT(a->lock); }
  memset(a->elements, 0, sizeof(a->elements));
}

void
sen_tiny_array_fin(sen_tiny_array *a)
{
  int i;
  sen_ctx *ctx = a->ctx;
  for (i = 0; i < SEN_TINY_ARRAY_N; i++) {
    void **e = &a->elements[i];
    if (*e) {
      if (a->flags & SEN_TINY_ARRAY_USE_MALLOC) {
        SEN_FREE(*e);
      } else {
        SEN_CTX_FREE(ctx, *e);
      }
      *e = NULL;
    }
  }
}

void *
sen_tiny_array_at(sen_tiny_array *a, sen_id id)
{
  void *e;
  SEN_TINY_ARRAY_AT(a, id, e);
  return e;
}

sen_id
sen_tiny_array_id(sen_tiny_array *a, void *p)
{
  uint32_t i, s, n;
  uintptr_t o, p_ = (uintptr_t) p;
  for (i = 0, s = 0; i < SEN_TINY_ARRAY_N; i++, s += n) {
    n = SEN_TINY_ARRAY_S * SEN_TINY_ARRAY_R(i);
    o = (uintptr_t) a->elements[i];
    if (o <= p_ && p_ < o + n * a->element_size) {
      return s + ((p_ - o) / a->element_size);
    }
  }
  return SEN_ID_NIL;
}

/* sen_array */

#define SEN_ARRAY_HEADER_SIZE 0x9000
#define SEN_ARRAY_IDSTR "SENNA:AR:01.000"
#define SEN_ARRAY_SEGMENT_SIZE 0x400000

struct sen_array_header {
  char idstr[16];
  uint32_t flags;
  uint32_t curr_rec;
  uint32_t value_size;
  uint32_t n_entries;
  uint32_t n_garbages;
  sen_id garbages;
  uint32_t lock;
  uint32_t reserved[5];
};

enum {
  array_seg_value = 0,
  array_seg_bitmap = 1
};

#define IO_ARRAYP(array) ((array)->io)

#define ARRAY_ENTRY_AT_(array,id,value,addp) {\
  int flags = addp;\
  SEN_IO_ARRAY_AT(array->io, array_seg_value, id, &flags, value);\
}

#define ARRAY_ENTRY_AT(array,id,value,addp) {\
  if (IO_ARRAYP(array)) {\
    ARRAY_ENTRY_AT_(array, id, value, addp);\
  } else {\
    SEN_TINY_ARRAY_AT(&array->a, id, value);\
  }\
}

#define ARRAY_BITMAP_AT(array,id,value) {\
  if (IO_ARRAYP(array)) {\
    SEN_IO_ARRAY_BIT_AT((array)->io, array_seg_bitmap, (id), (value));\
  } else {\
    SEN_TINY_ARRAY_BIT_AT(&(array)->bitmap, (id), (value));\
  }\
}

inline static sen_rc
tiny_array_init(sen_array *array, sen_ctx *ctx, const char *path,
               uint32_t value_size, uint32_t flags)
{
  if (path) { return sen_invalid_argument; }
  SEN_OBJ_DB_SET_TYPE(array, SEN_TABLE_NO_KEY);
  array->obj.flags = flags;
  array->ctx = ctx;
  array->value_size = value_size;
  array->n_garbages = &array->n_garbages_;
  array->n_entries = &array->n_entries_;
  array->n_garbages_ = 0;
  array->n_entries_ = 0;
  array->io = NULL;
  array->garbages = SEN_ID_NIL;
  sen_tiny_array_init(&array->a, ctx, value_size, SEN_TINY_ARRAY_CLEAR);
  sen_tiny_array_init(&array->bitmap, ctx, 1, SEN_TINY_ARRAY_CLEAR);
  return sen_success;
}

inline static sen_rc
io_array_init(sen_array *array, sen_ctx *ctx, const char *path,
               uint32_t value_size, uint32_t flags)
{
  sen_io *io;
  struct sen_array_header *header;
  uint32_t w_of_element;
  for (w_of_element = 0; (1 << w_of_element) < value_size; w_of_element++);
  {
    sen_io_array_spec array_spec[2];
    array_spec[array_seg_value].w_of_element = w_of_element;
    array_spec[array_seg_value].max_n_segments = 1 << (30 - (22 - w_of_element));
    array_spec[array_seg_bitmap].w_of_element = 0;
    array_spec[array_seg_bitmap].max_n_segments = 1 << (30 - (22 + 3));
    io = sen_io_create_with_array(path, sizeof(struct sen_array_header),
                                  SEN_ARRAY_SEGMENT_SIZE, sen_io_auto, 2, array_spec);
  }
  if (!io) { return sen_memory_exhausted; }
  header = sen_io_header(io);
  memcpy(header->idstr, SEN_ARRAY_IDSTR, 16);
  sen_io_set_type(io, SEN_TABLE_NO_KEY);
  header->flags = flags;
  header->curr_rec = 0;
  header->lock = 0;
  header->value_size = value_size;
  header->n_entries = 0;
  header->n_garbages = 0;
  header->garbages = SEN_ID_NIL;
  SEN_OBJ_DB_SET_TYPE(array, SEN_TABLE_NO_KEY);
  array->obj.flags = flags;
  array->ctx = ctx;
  array->value_size = value_size;
  array->n_garbages = &header->n_garbages;
  array->n_entries = &header->n_entries;
  array->io = io;
  array->header = header;
  array->lock = &header->lock;
  return sen_success;
}

sen_array *
sen_array_create(sen_ctx *ctx, const char *path, uint32_t value_size, uint32_t flags)
{
  sen_array *array;
  if (ctx && (array = SEN_MALLOC(sizeof(sen_array)))) {
    if (!((flags & SEN_ARRAY_TINY) ?
          tiny_array_init(array, ctx, path, value_size, flags) :
          io_array_init(array, ctx, path, value_size, flags))) {
      return array;
    }
    SEN_FREE(array);
  }
  return NULL;
}

sen_array *
sen_array_open(sen_ctx *ctx, const char *path)
{
  if (ctx) {
    sen_io *io = sen_io_open(path, sen_io_auto);
    if (io) {
      struct sen_array_header *header = sen_io_header(io);
      if (!memcmp(header->idstr, SEN_ARRAY_IDSTR, 16)) {
        sen_array *array = SEN_MALLOC(sizeof(sen_array));
        if (array) {
          if (!(header->flags & SEN_ARRAY_TINY)) {
            SEN_OBJ_DB_SET_TYPE(array, SEN_TABLE_NO_KEY);
            array->obj.flags = header->flags;
            array->ctx = ctx;
            array->value_size = header->value_size;
            array->n_garbages = &header->n_garbages;
            array->n_entries = &header->n_entries;
            array->io = io;
            array->header = header;
            array->lock = &header->lock;
            return (sen_array *)array;
          } else {
            SEN_LOG(sen_log_notice, "invalid array flag. (%x)", header->flags);
          }
          SEN_FREE(array);
        }
      } else {
        SEN_LOG(sen_log_notice, "invalid array file. array_idstr (%s)", header->idstr);
      }
      sen_io_close(io);
    }
  }
  return NULL;
}

sen_rc
sen_array_close(sen_ctx *ctx, sen_array *array)
{
  sen_rc rc = sen_success;
  if (!ctx || !array) { return sen_invalid_argument; }
  if (IO_ARRAYP(array)) {
    rc = sen_io_close(array->io);
  } else {
    SEN_ASSERT(ctx == array->ctx);
    sen_tiny_array_fin(&array->a);
    sen_tiny_array_fin(&array->bitmap);
  }
  SEN_FREE(array);
  return rc;
}

sen_rc
sen_array_remove(sen_ctx *ctx, const char *path)
{
  if (!ctx || !path) { return sen_invalid_argument; }
  return sen_io_remove(path);
}

int
sen_array_get_value(sen_ctx *ctx, sen_array *array, sen_id id, void *valuebuf)
{
  if (ctx && array) {
    void *ee;
    uint8_t res;
    ARRAY_BITMAP_AT(array, id, res);
    if (!res) { return 0; }
    ARRAY_ENTRY_AT(array, id, ee, 0);
    if (ee) {
      if (valuebuf) { memcpy(valuebuf, ee, array->value_size); }
      return array->value_size;
    }
  }
  return 0;
}

void *
_sen_array_get_value(sen_ctx *ctx, sen_array *array, sen_id id)
{
  if (ctx && array) {
    void *ee;
    uint8_t res;
    ARRAY_BITMAP_AT(array, id, res);
    if (!res) { return NULL; }
    ARRAY_ENTRY_AT(array, id, ee, 0);
    return ee;
  }
  return NULL;
}

sen_rc
sen_array_set_value(sen_ctx *ctx, sen_array *array, sen_id id, void *value, int flags)
{
  if (ctx && array && value) {
    void *ee;
    uint8_t res;
    ARRAY_BITMAP_AT(array, id, res);
    if (!res) { return sen_invalid_argument; }
    ARRAY_ENTRY_AT(array, id, ee, 0);
    if (ee) {
      switch ((flags & SEN_OBJ_SET_MASK)) {
      case SEN_OBJ_SET :
        memcpy(ee, value, array->value_size);
        return sen_success;
      default :
        // todo : support other types.
        return sen_invalid_argument;
      }
    } else {
      return sen_memory_exhausted;
    }
  }
  return sen_invalid_argument;
}

sen_rc
sen_array_delete_by_id(sen_ctx *ctx, sen_array *array, sen_id id,
                       sen_table_delete_optarg *optarg)
{
  sen_rc rc = sen_success;
  if (!ctx || !array) { return sen_invalid_argument; }
  /* lock */
  if (IO_ARRAYP(array)) {
    if (array->value_size >= sizeof(sen_id)) {
      void *ee;
      struct sen_array_header *hh = array->header;
      ARRAY_ENTRY_AT_(array, id, ee, 0);
      if (!ee) { rc = sen_invalid_argument; goto exit; }
      *((sen_id *)ee) = hh->garbages;
      hh->garbages = id;
    }
    SEN_IO_ARRAY_BIT_OFF(array->io, array_seg_bitmap, id);
  } else {
    if (array->value_size >= sizeof(sen_id)) {
      void *ee;
      SEN_TINY_ARRAY_AT(&array->a, id, ee);
      if (!ee) { rc = sen_invalid_argument; goto exit; }
      *((sen_id *)ee) = array->garbages;
      array->garbages = id;
    }
    SEN_TINY_ARRAY_BIT_OFF(&array->bitmap, id);
  }
  (*array->n_entries)--;
  (*array->n_garbages)++;
exit :
  /* unlock */
  return rc;
}

void
sen_array_cursor_close(sen_ctx *ctx, sen_array_cursor *c)
{
  SEN_ASSERT(c->ctx == ctx);
  SEN_FREE(c);
}

#define ARRAY_CURR_MAX(array) \
  ((IO_ARRAYP(array)) ? (array)->header->curr_rec : (array)->a.max)

sen_array_cursor *
sen_array_cursor_open(sen_ctx *ctx, sen_array *array, sen_id min, sen_id max, int flags)
{
  sen_array_cursor *c;
  if (!array || !ctx) { return NULL; }
  if (!(c = SEN_MALLOCN(sen_array_cursor, 1))) { return NULL; }
  SEN_OBJ_DB_SET_TYPE(c, SEN_CURSOR_TABLE_NO_KEY);
  c->array = array;
  c->ctx = ctx;
  c->obj.flags = flags;
  if (flags & SEN_CURSOR_ASCENDING) {
    c->dir = 1;
    if (min) {
      c->curr_rec = min;
      if (!(flags & SEN_CURSOR_GT)) { c->curr_rec--; }
    } else {
      c->curr_rec = SEN_ID_NIL;
    }
    if (max) {
      c->limit = max;
      if ((flags & SEN_CURSOR_LT)) { c->limit--; }
    } else {
      c->limit = ARRAY_CURR_MAX(array);
    }
    if (c->limit < c->curr_rec) { c->limit = c->curr_rec; }
  } else {
    c->dir = -1;
    if (max) {
      c->curr_rec = max;
      if (!(flags & SEN_CURSOR_LT)) { c->curr_rec++; }
    } else {
      c->curr_rec = ARRAY_CURR_MAX(array) + 1;
    }
    if (min) {
      c->limit = min;
      if ((flags & SEN_CURSOR_GT)) { c->limit++; }
    } else {
      c->limit = SEN_ID_NIL + 1;
    }
    if (c->curr_rec < c->limit) { c->limit = c->curr_rec; }
  }
  return c;
}

sen_id
sen_array_cursor_next(sen_ctx *ctx, sen_array_cursor *c)
{
  if (c) {
    while (c->curr_rec != c->limit) {
      uint8_t res;
      c->curr_rec += c->dir;
      ARRAY_BITMAP_AT(c->array, c->curr_rec, res);
      if (res) { return c->curr_rec; }
    }
  }
  return SEN_ID_NIL;
}

sen_id
sen_array_next(sen_ctx *ctx, sen_array *array, sen_id id)
{
  sen_id max = ARRAY_CURR_MAX(array);
  while (++id <= max) {
    uint8_t res;
    ARRAY_BITMAP_AT(array, id, res);
    if (res) { return id; }
  }
  return SEN_ID_NIL;
}

int
sen_array_cursor_get_value(sen_ctx *ctx, sen_array_cursor *c, void **value)
{
  void *ee;
  if (c) {
    ARRAY_ENTRY_AT(c->array, c->curr_rec, ee, 0);
    if (ee) {
      *value = ee;
      return c->array->value_size;
    }
  }
  return 0;
}

sen_rc
sen_array_cursor_set_value(sen_ctx *ctx, sen_array_cursor *c, void *value, int flags)
{
  return sen_array_set_value(ctx, c->array, c->curr_rec, value, flags);
}

sen_rc
sen_array_cursor_delete(sen_ctx *ctx, sen_array_cursor *c,
                        sen_table_delete_optarg *optarg)
{
  return sen_array_delete_by_id(ctx, c->array, c->curr_rec, optarg);
}

inline static sen_id
array_entry_new(sen_array *array)
{
  sen_id e;
  if (IO_ARRAYP(array)) {
    struct sen_array_header *hh = array->header;
    if ((e = hh->garbages)) {
      void *ee;
      ARRAY_ENTRY_AT_(array, e, ee, SEN_TABLE_ADD);
      if (!ee) { return SEN_ID_NIL; }
      hh->garbages = *((sen_id *)ee);
      memset(ee, 0, hh->value_size);
    } else {
      e = ++hh->curr_rec;
    }
    SEN_IO_ARRAY_BIT_ON(array->io, array_seg_bitmap, e);
  } else {
    if ((e = array->garbages)) {
      void *ee;
      SEN_TINY_ARRAY_AT(&array->a, e, ee);
      array->garbages = *((sen_id *)ee);
      memset(ee, 0, array->value_size);
    } else {
      e = ++array->a.max;
    }
    SEN_TINY_ARRAY_BIT_ON(&array->bitmap, e);
  }
  return e;
}

sen_id
sen_array_add(sen_ctx *ctx, sen_array *array, void **value)
{
  sen_id e;
  if (!ctx || !array || !(e = array_entry_new(array))) { return SEN_ID_NIL; }
  (*array->n_entries)++;
  if (value) {
    void *ee;
    ARRAY_ENTRY_AT(array, e, ee, SEN_TABLE_ADD);
    *value = ee;
  }
  return e;
}

/* sen_hash : hash table */

#define SEN_HASH_MAX_SEGMENT 0x400
#define SEN_HASH_HEADER_SIZE 0x9000
#define SEN_IDSTR "SENNA:HSH:01.00"
#define SEN_HASH_SEGMENT_SIZE 0x400000
#define W_OF_KEY_IN_A_SEGMENT 22
#define IDX_MASK_IN_A_SEGMENT 0xfffff

typedef struct {
  uint32_t key;
  uint8_t dummy[1];
} entry;

typedef struct {
  uint32_t key;
  uint16_t flag;
  uint16_t size;
  uint32_t str;
  uint8_t dummy[1];
} entry_str;

typedef struct {
  uint32_t key;
  uint16_t flag;
  uint16_t size;
  char *str;
  uint8_t dummy[1];
} entry_astr;

#define LOGICAL_MAX_SEGMENT ((SEN_HASH_MAX_SEGMENT) * 4)

#define IO_HASHP(hash) ((hash)->io)

enum {
  segment_key = 0,
  segment_entry = 1,
  segment_index = 2,
  segment_bitmap = 3
};

#define ENTRY_AT_(hash,id,ee,addp) {\
  int flags = addp;\
  SEN_IO_ARRAY_AT(hash->io, segment_entry, id, &flags, ee);\
}

// todo : error handling
#define ENTRY_AT(hash,id,ee,addp) {\
  if (IO_HASHP(hash)) {\
    ENTRY_AT_(hash, id, ee, addp);\
  } else {\
    SEN_TINY_ARRAY_AT(&hash->a, id, ee);\
  }\
}

#define BITMAP_AT(hash,id,value) {\
  if (IO_HASHP(hash)) {\
    SEN_IO_ARRAY_BIT_AT((hash)->io, segment_bitmap, (id), (value));\
  } else {\
    SEN_TINY_ARRAY_BIT_AT(&(hash)->bitmap, (id), (value));\
  }\
}

inline static sen_id *
idx_at_(sen_hash *hash, sen_id id)
{
  int flags = SEN_TABLE_ADD;
  void *pp;
  SEN_IO_ARRAY_AT(hash->io, segment_index, id, &flags, pp);
  return pp;
}

#define IDX_AT(h,i) (\
  IO_HASHP(h) ?\
  idx_at_(h, ((i) & *(h)->max_offset) + h->header->idx_offset) :\
  h->index + ((i) & *(h)->max_offset)\
)

#define KEY_AT(hash,pos,ptr) \
{\
  int flags = SEN_TABLE_ADD;\
  SEN_IO_ARRAY_AT(hash->io, segment_key, pos, &flags, ptr);\
}

#define HASH_IMMEDIATE 1

#define MAX_INDEX_SIZE ((SEN_HASH_MAX_SEGMENT * (IDX_MASK_IN_A_SEGMENT + 1)) >> 1)

inline static char *
get_key(sen_hash *hash, entry_str *n)
{
  if (hash->obj.flags & SEN_OBJ_KEY_VAR_SIZE) {
    if (n->flag & HASH_IMMEDIATE) {
      return (char *)&n->str;
    } else {
      if (IO_HASHP(hash)) {
        char *res;
        KEY_AT(hash, n->str, res);
        return res;
      } else {
        return ((entry_astr *)n)->str;
      }
    }
  } else {
    if (hash->key_size == sizeof(uint32_t)) {
      return (char *)(&((entry *)n)->key);
    } else {
      return (char *)(((entry *)n)->dummy);
    }
  }
}

inline static void *
get_value(sen_hash *hash, entry_str *n)
{
  if (hash->obj.flags & SEN_OBJ_KEY_VAR_SIZE) {
    if (IO_HASHP(hash)) {
      return ((entry_str *)n)->dummy;
    } else {
      return ((entry_astr *)n)->dummy;
    }
  } else {
    if (hash->key_size == sizeof(uint32_t)) {
      return ((entry *)n)->dummy;
    } else {
      return &((entry *)n)->dummy[hash->key_size];
    }
  }
}

inline static uint32_t
put_key_(sen_hash *hash, const char *key, int len)
{
  uint32_t res, ts;
  if (len >= SEN_HASH_SEGMENT_SIZE) { return 0; /* error */ }
  res = hash->header->curr_key;
  ts = (res + len) >> W_OF_KEY_IN_A_SEGMENT;
  if (res >> W_OF_KEY_IN_A_SEGMENT != ts) {
    res = hash->header->curr_key = ts << W_OF_KEY_IN_A_SEGMENT;
  }
  {
    uint8_t *dest;
    KEY_AT(hash, res, dest);
    if (!dest) { return 0; }
    memcpy(dest, key, len);
  }
  hash->header->curr_key += len;
  return res;
}

inline static sen_rc
put_key(sen_hash *hash, entry_str *n, uint32_t h, const char *key, unsigned int len)
{
  n->key = h;
  if (hash->obj.flags & SEN_OBJ_KEY_VAR_SIZE) {
    if (IO_HASHP(hash)) {
      if (len <= sizeof(uint32_t)) {
        n->flag = HASH_IMMEDIATE;
        memcpy(&n->str, key, len);
      } else {
        n->flag = 0;
        n->str = put_key_(hash, key, len);
      }
    } else {
      if (len <= sizeof(char *)) {
        n->flag = HASH_IMMEDIATE;
        memcpy(&((entry_astr *)n)->str, key, len);
      } else {
        sen_ctx *ctx = hash->ctx;
        if (!(((entry_astr *)n)->str = SEN_CTX_ALLOC(ctx, len, 0))) {
          return sen_memory_exhausted;
        }
        memcpy(((entry_astr *)n)->str, key, len);
        n->flag = 0;
      }
    }
    n->size = len;
  } else {
    if (hash->key_size != sizeof(uint32_t)) {
      memcpy(((entry *)n)->dummy, key, len);
    }
  }
  return sen_success;
}

inline static int
match_key(sen_hash *hash, entry_str *ee, uint32_t h, const char *key, unsigned int len)
{
  if (hash->obj.flags & SEN_OBJ_KEY_VAR_SIZE) {
    return (ee->key == h &&
            ee->size == len &&
            !memcmp(key, get_key(hash, ee), len));
  } else {
    return (ee->key == h &&
            ((len == sizeof(uint32_t)) ||
             !memcmp(key, get_key(hash, ee), len)));
  }
}

#define GARBAGE (0xffffffff)

#define STEP(x) (((x) >> 2) | 0x1010101)

inline static sen_rc
io_hash_init(sen_hash *ih, sen_ctx *ctx, const char *path, uint32_t key_size,
               uint32_t value_size, uint32_t flags, sen_encoding encoding,
               uint32_t init_size)
{
  sen_io *io;
  struct sen_hash_header *header;
  uint32_t entry_size, w_of_element, m;
  for (m = IDX_MASK_IN_A_SEGMENT + 1; m < init_size * 2; m *= 2);
  if (flags & SEN_OBJ_KEY_VAR_SIZE) {
    entry_size = (intptr_t)(&((entry_str *)0)->dummy) + value_size;
  } else {
    if (key_size == sizeof(uint32_t)) {
      entry_size = (intptr_t)(&((entry *)0)->dummy) + value_size;
    } else {
      entry_size = (intptr_t)(&((entry *)0)->dummy) + key_size + value_size;
    }
  }
  for (w_of_element = 0; (1 << w_of_element) < entry_size; w_of_element++);
  {
    sen_io_array_spec array_spec[4];
    array_spec[segment_key].w_of_element = 0;
    array_spec[segment_key].max_n_segments = 0x400;
    array_spec[segment_entry].w_of_element = w_of_element;
    array_spec[segment_entry].max_n_segments = 1 << (30 - (22 - w_of_element));
    array_spec[segment_index].w_of_element = 2;
    array_spec[segment_index].max_n_segments = 1 << (30 - (22 - 2));
    array_spec[segment_bitmap].w_of_element = 0;
    array_spec[segment_bitmap].max_n_segments = 1 << (30 - (22 + 3));
    io = sen_io_create_with_array(path, SEN_HASH_HEADER_SIZE, SEN_HASH_SEGMENT_SIZE,
                                  sen_io_auto, 4, array_spec);
  }
  if (!io) { return sen_memory_exhausted; }
  if (encoding == sen_enc_default) { encoding = ctx->encoding; }
  header = sen_io_header(io);
  memcpy(header->idstr, SEN_IDSTR, 16);
  sen_io_set_type(io, SEN_TABLE_HASH_KEY);
  header->flags = flags;
  header->encoding = encoding;
  header->key_size = key_size;
  header->curr_rec = 0;
  header->curr_key = 0;
  header->lock = 0;
  header->idx_offset = 0;
  header->value_size = value_size;
  header->entry_size = entry_size;
  header->max_offset = m - 1;
  header->n_entries = 0;
  header->n_garbages = 0;
  header->ngram_unit = SEN_TABLE_DEFAULT_NGRAM_UNIT_SIZE;
  SEN_OBJ_DB_SET_TYPE(ih, SEN_TABLE_HASH_KEY);
  ih->obj.flags = flags;
  ih->ctx = ctx;
  ih->key_size = key_size;
  ih->encoding = encoding;
  ih->value_size = value_size;
  ih->entry_size = entry_size;
  ih->n_garbages = &header->n_garbages;
  ih->n_entries = &header->n_entries;
  ih->max_offset = &header->max_offset;
  ih->io = io;
  ih->header = header;
  ih->lock = &header->lock;
  ih->ngram_unit = header->ngram_unit;
  return sen_success;
}

#define INITIAL_INDEX_SIZE 256U

inline static sen_rc
tiny_hash_init(sen_hash *ah, sen_ctx *ctx, const char *path, uint32_t key_size,
               uint32_t value_size, uint32_t flags, sen_encoding encoding)
{
  uint32_t entry_size;
  if (path) { return sen_invalid_argument; }
  if (!(ah->index = SEN_CTX_ALLOC(ctx, INITIAL_INDEX_SIZE * sizeof(sen_id), 0))) {
    return sen_memory_exhausted;
  }
  if (flags & SEN_OBJ_KEY_VAR_SIZE) {
    entry_size = (intptr_t)(&((entry_astr *)0)->dummy) + value_size;
  } else {
    if (key_size == sizeof(uint32_t)) {
      entry_size = (intptr_t)(&((entry *)0)->dummy) + value_size;
    } else {
      entry_size = (intptr_t)(&((entry *)0)->dummy) + key_size + value_size;
    }
  }
  if (entry_size != sizeof(uint32_t)) {
    entry_size = ((entry_size + (sizeof(intptr_t)) - 1) & ~((sizeof(intptr_t)) - 1));
  }
  SEN_OBJ_DB_SET_TYPE(ah, SEN_TABLE_HASH_KEY);
  ah->obj.flags = flags;
  ah->ctx = ctx;
  ah->key_size = key_size;
  ah->encoding = encoding;
  ah->value_size = value_size;
  ah->entry_size = entry_size;
  ah->n_garbages = &ah->n_garbages_;
  ah->n_entries = &ah->n_entries_;
  ah->max_offset = &ah->max_offset_;
  ah->max_offset_ = INITIAL_INDEX_SIZE - 1;
  ah->io = NULL;
  ah->n_garbages_ = 0;
  ah->n_entries_ = 0;
  ah->garbages = SEN_ID_NIL;
  ah->ngram_unit = SEN_TABLE_DEFAULT_NGRAM_UNIT_SIZE;
  sen_tiny_array_init(&ah->a, ctx, entry_size, SEN_TINY_ARRAY_CLEAR);
  sen_tiny_array_init(&ah->bitmap, ctx, 1, SEN_TINY_ARRAY_CLEAR);
  return sen_success;
}

sen_hash *
sen_hash_create(sen_ctx *ctx, const char *path, uint32_t key_size, uint32_t value_size,
                uint32_t flags, sen_encoding encoding)
{
  sen_hash *hash;
  if (key_size > SEN_HASH_MAX_KEY_SIZE) { return NULL; }
  if ((hash = SEN_MALLOC(sizeof(sen_hash)))) {
    if (!((flags & SEN_HASH_TINY) ?
          tiny_hash_init(hash, ctx, path, key_size, value_size, flags, encoding) :
          io_hash_init(hash, ctx, path, key_size, value_size, flags, encoding, 0))) {
      return hash;
    }
    SEN_FREE(hash);
  }
  return NULL;
}

sen_hash *
sen_hash_open(sen_ctx *ctx, const char *path)
{
  sen_io *io = sen_io_open(path, sen_io_auto);
  if (io) {
    struct sen_hash_header *header = sen_io_header(io);
    if (!memcmp(header->idstr, SEN_IDSTR, 16)) {
      sen_hash *hash = SEN_MALLOC(sizeof(sen_hash));
      if (hash) {
        if (!(header->flags & SEN_HASH_TINY)) {
          SEN_OBJ_DB_SET_TYPE(hash, SEN_TABLE_HASH_KEY);
          hash->obj.flags = header->flags;
          hash->ctx = ctx;
          hash->key_size = header->key_size;
          hash->encoding = header->encoding;
          hash->value_size = header->value_size;
          hash->entry_size = header->entry_size;
          hash->n_garbages = &header->n_garbages;
          hash->n_entries = &header->n_entries;
          hash->max_offset = &header->max_offset;
          hash->io = io;
          hash->header = header;
          hash->lock = &header->lock;
          hash->ngram_unit = header->ngram_unit;
          return (sen_hash *)hash;
        } else {
          SEN_LOG(sen_log_notice, "invalid hash flag. (%x)", header->flags);
        }
        SEN_FREE(hash);
      }
    } else {
      SEN_LOG(sen_log_notice, "invalid hash file. hash_idstr (%s)", header->idstr);
    }
    sen_io_close(io);
  }
  return NULL;
}

inline static sen_rc
tiny_hash_fin(sen_ctx *ctx, sen_hash *hash)
{
  if (hash->index) {
    if (hash->obj.flags & SEN_OBJ_KEY_VAR_SIZE) {
      uint32_t i;
      sen_id e, *sp;
      for (i = *hash->n_entries, sp = hash->index; i; sp++) {
        entry_astr *n;
        e = *sp;
        if (!e || (e == GARBAGE)) { continue; }
        SEN_TINY_ARRAY_AT(&hash->a, e, n);
        SEN_ASSERT(n);
        i--;
        if (!n || (n->flag & HASH_IMMEDIATE)) { continue; }
        SEN_CTX_FREE(ctx, n->str);
      }
    }
    sen_tiny_array_fin(&hash->a);
    sen_tiny_array_fin(&hash->bitmap);
    SEN_CTX_FREE(ctx, hash->index);
    return sen_success;
  } else {
    return sen_invalid_argument;
  }
}

sen_rc
sen_hash_close(sen_ctx *ctx, sen_hash *hash)
{
  sen_rc rc;
  if (!hash) { return sen_invalid_argument; }
  if (IO_HASHP(hash)) {
    rc = sen_io_close(hash->io);
  } else {
    SEN_ASSERT(ctx == hash->ctx);
    rc = tiny_hash_fin(ctx, hash);
  }
  SEN_FREE(hash);
  return rc;
}

sen_rc
sen_hash_remove(sen_ctx *ctx, const char *path)
{
  if (!path) { return sen_invalid_argument; }
  return sen_io_remove(path);
}

sen_rc
sen_hash_reset(sen_ctx *ctx, sen_hash *hash, uint32_t ne)
{
  entry *ee;
  sen_id e, *index, *sp, *dp;
  uint32_t n, n0 = *hash->n_entries, offs, offd;
  if (!ne) { ne = n0 * 2; }
  if (ne > INT_MAX) { return sen_memory_exhausted; }
  for (n = INITIAL_INDEX_SIZE; n <= ne; n *= 2);
  if (IO_HASHP(hash)) {
    uint32_t i;
    offs = hash->header->idx_offset;
    offd = MAX_INDEX_SIZE - offs;
    for (i = 0; i < n; i += (IDX_MASK_IN_A_SEGMENT + 1)) {
      dp = idx_at_(hash, i + offd); // todo : use idx_at
      if (!dp) { return sen_memory_exhausted; }
      memset(dp, 0, SEN_HASH_SEGMENT_SIZE);
    }
  } else {
    SEN_ASSERT(ctx == hash->ctx);
    if (!(index = SEN_CTX_ALLOC(ctx, n * sizeof(sen_id), 0))) {
      return sen_memory_exhausted;
    }
    sp = hash->index;
  }
  {
    uint32_t i, j, k, m0 = *hash->max_offset, m = n - 1, s;
    for (k = 0, j = 0; k < n0 && j <= m0; j++, sp++) {
      if (IO_HASHP(hash) && !(j & IDX_MASK_IN_A_SEGMENT)) {
        sp = idx_at_(hash, j + offs);
        if (!sp) { return sen_memory_exhausted; }
      }
      e = *sp;
      if (!e || (e == GARBAGE)) { continue; }
      ENTRY_AT(hash, e, ee, SEN_TABLE_ADD);
      if (!ee) { return sen_memory_exhausted; }
      for (i = ee->key, s = STEP(i); ; i += s) {
        if (IO_HASHP(hash)) {
          dp = idx_at_(hash, (i & m) + offd);
          if (!dp) { return sen_memory_exhausted; }
        } else {
          dp = index + (i & m);
        }
        if (!*dp) { break; }
      }
      *dp = e;
      k++;
    }
    *hash->max_offset = m;
    *hash->n_garbages = 0;
  }
  if (IO_HASHP(hash)) {
    hash->header->idx_offset = offd;
  } else {
    sen_id *i0 = hash->index;
    hash->index = index;
    SEN_CTX_FREE(ctx, i0);
  }
  return sen_success;
}

inline static uint32_t
bin_hash(const uint8_t *p, uint32_t length)
{
  uint32_t r;
  for (r = 0; length--; p++) { r = (r * 1021) + *p; }
  return r;
}

inline static sen_id
entry_new(sen_hash *hash, uint32_t size)
{
  sen_id e;
  if (IO_HASHP(hash)) {
    struct sen_hash_header *hh = hash->header;
    size -= 1;
    if ((e = hh->garbages[size])) {
      entry *ee;
      ENTRY_AT_(hash, e, ee, SEN_TABLE_ADD);
      if (!ee) { return SEN_ID_NIL; }
      hh->garbages[size] = ee->key;
      memset(ee, 0, hh->entry_size);
    } else {
      e = ++hh->curr_rec;
    }
    SEN_IO_ARRAY_BIT_ON(hash->io, segment_bitmap, e);
  } else {
    if (hash->garbages) {
      entry *ee;
      e = hash->garbages;
      SEN_TINY_ARRAY_AT(&hash->a, e, ee);
      hash->garbages = ee->key;
      memset(ee, 0, hash->entry_size);
    } else {
      e = hash->a.max + 1;
    }
    SEN_TINY_ARRAY_BIT_ON(&hash->bitmap, e);
  }
  return e;
}

sen_rc
sen_hash_lock(sen_ctx *ctx, sen_hash *hash, int timeout)
{
  static int _ncalls = 0, _ncolls = 0;
  uint32_t count;
  _ncalls++;
  for (count = 0;; count++) {
    uint32_t lock;
    SEN_ATOMIC_ADD_EX(hash->lock, 1, lock);
    if (lock) {
      SEN_ATOMIC_ADD_EX(hash->lock, -1, lock);
      if (!timeout || (timeout > 0 && timeout == count)) { break; }
      if (!(++_ncolls % 1000000) && (_ncolls > _ncalls)) {
        if (_ncolls < 0 || _ncalls < 0) {
          _ncolls = 0; _ncalls = 0;
        } else {
          SEN_LOG(sen_log_notice, "hash(%p) collisions(%d/%d)", hash, _ncolls, _ncalls);
        }
      }
      usleep(1000);
      continue;
    }
    return sen_success;
  }
  return sen_other_error;
}

sen_rc
sen_hash_unlock(sen_ctx *ctx, sen_hash *hash)
{
  uint32_t lock;
  SEN_ATOMIC_ADD_EX(hash->lock, -1, lock);
  return sen_success;
}

sen_rc
sen_hash_clear_lock(sen_ctx *ctx, sen_hash *hash)
{
  *hash->lock = 0;
  return sen_success;
}

sen_id
sen_hash_get(sen_ctx *ctx, sen_hash *hash, const void *key, int key_size, void **value,
             sen_search_flags *flags)
{
  entry_str *ee;
  uint32_t h, i, m, s;
  sen_id e, *ep, *np = NULL;
  if (hash->obj.flags & SEN_OBJ_KEY_VAR_SIZE) {
    if (key_size > hash->key_size) { return SEN_ID_NIL; }
    h = bin_hash((unsigned char *)key, key_size);
  } else {
    if (key_size != hash->key_size) { return SEN_ID_NIL; }
    if (key_size == sizeof(uint32_t)) {
      h = *((uint32_t *)key);
    } else {
      h = bin_hash((unsigned char *)key, key_size);
    }
  }
  s = STEP(h);
  /* lock */
  if ((*hash->n_entries + *hash->n_garbages) * 2 > *hash->max_offset) {
    sen_hash_reset(ctx, hash, 0);
  }
  m = *hash->max_offset;
  for (i = h; ; i += s) {
    if (!(ep = IDX_AT(hash, i))) { return SEN_ID_NIL; }
    if (!(e = *ep)) { break; }
    if (e == GARBAGE) {
      if (!np) { np = ep; }
      continue;
    }
    ENTRY_AT(hash, e, ee, SEN_TABLE_ADD);
    if (!ee) { return SEN_ID_NIL; }
    if (match_key(hash, ee, h, key, key_size)) { goto exit; }
  }
  if (!(e = entry_new(hash, key_size))) { /* unlock */ return SEN_ID_NIL; }
  ENTRY_AT(hash, e, ee, SEN_TABLE_ADD);
  if (!ee) { return SEN_ID_NIL; }
  put_key(hash, ee, h, key, key_size);
  if (np) {
    (*hash->n_garbages)--;
    ep = np;
  }
  *ep = e;
  (*hash->n_entries)++;
  if (flags) { *flags |= SEN_TABLE_ADDED; }
exit :
  /* unlock */
  if (value) { *value = get_value(hash,ee); }
  return e;
}

sen_id
sen_hash_at(sen_ctx *ctx, sen_hash *hash, const void *key, int key_size, void **value)
{
  sen_id e, *ep;
  uint32_t h, i, m, s;
  if (hash->obj.flags & SEN_OBJ_KEY_VAR_SIZE) {
    if (key_size > hash->key_size) { return SEN_ID_NIL; }
    h = bin_hash((unsigned char *)key, key_size);
  } else {
    if (key_size != hash->key_size) { return SEN_ID_NIL; }
    if (key_size == sizeof(uint32_t)) {
      h = *((uint32_t *)key);
    } else {
      h = bin_hash((unsigned char *)key, key_size);
    }
  }
  s = STEP(h);
  m = *hash->max_offset;
  for (i = h; ; i += s) {
    if (!(ep = IDX_AT(hash, i))) { return SEN_ID_NIL; }
    if (!(e = *ep)) { break; }
    if (e == GARBAGE) { continue; }
    {
      entry_str *ee;
      ENTRY_AT(hash,e,ee, 0);
      if (ee && match_key(hash, ee, h, key, key_size)) {
        if (value) { *value = get_value(hash,ee); }
        return e;
      }
    }
  }
  return SEN_ID_NIL;
}

sen_id
sen_hash_lookup(sen_ctx *ctx, sen_hash *hash, const void *key, int key_size, void **value,
                sen_search_flags *flags)
{
  if (*flags & SEN_TABLE_ADD) {
    return sen_hash_get(ctx, hash, key, key_size, value, flags);
  } else {
    return sen_hash_at(ctx, hash, key, key_size, value);
  }
}

const char *
_sen_hash_key(sen_hash *hash, sen_id id, uint32_t *key_size)
{
  uint8_t res;
  entry_str *ee;
  BITMAP_AT(hash, id, res);
  if (!res) { return NULL; }
  ENTRY_AT(hash, id, ee, 0);
  if (!ee) { return NULL; }
  *key_size = (hash->obj.flags & SEN_OBJ_KEY_VAR_SIZE) ? ee->size : hash->key_size;
  return get_key(hash, ee);
}

int
sen_hash_get_key(sen_ctx *ctx, sen_hash *hash, sen_id id, void *keybuf, int bufsize)
{
  uint8_t res;
  int key_size;
  entry_str *ee;
  BITMAP_AT(hash, id, res);
  if (!res) { return 0; }
  ENTRY_AT(hash, id, ee, 0);
  if (!ee) { return 0; }
  key_size = (hash->obj.flags & SEN_OBJ_KEY_VAR_SIZE) ? ee->size : hash->key_size;
  if (bufsize >= key_size) { memcpy(keybuf, get_key(hash, ee), key_size); }
  return key_size;
}

int
sen_hash_get_key2(sen_ctx *ctx, sen_hash *hash, sen_id id, sen_obj *bulk)
{
  uint8_t res;
  int key_size;
  char *key;
  entry_str *ee;
  BITMAP_AT(hash, id, res);
  if (!res) { return 0; }
  ENTRY_AT(hash, id, ee, 0);
  if (!ee) { return 0; }
  key_size = (hash->obj.flags & SEN_OBJ_KEY_VAR_SIZE) ? ee->size : hash->key_size;
  key = get_key(hash, ee);
  if (bulk->header.flags & SEN_OBJ_DO_SHALLOW_COPY) {
    bulk->u.b.head = key;
    bulk->u.b.curr = key + key_size;
  } else {
    sen_bulk_write(ctx, bulk, key, key_size);
  }
  return key_size;
}

int
sen_hash_get_value(sen_ctx *ctx, sen_hash *hash, sen_id id, void *valuebuf)
{
  void *v;
  uint8_t res;
  entry_str *ee;
  BITMAP_AT(hash, id, res);
  if (!res) { return 0; }
  ENTRY_AT(hash, id, ee, 0);
  if (ee && (v = get_value(hash, ee))) {
    if (valuebuf) { memcpy(valuebuf, v, hash->value_size); }
    return hash->value_size;
  }
  return 0;
}

const char *
sen_hash_get_value_(sen_ctx *ctx, sen_hash *hash, sen_id id, uint32_t *size)
{
  uint8_t res;
  entry_str *ee;
  const char *value = NULL;
  BITMAP_AT(hash, id, res);
  if (!res) { return NULL; }
  ENTRY_AT(hash, id, ee, 0);
  if (ee && (value = get_value(hash, ee))) {
    *size = hash->value_size;
  }
  return value;
}

int
sen_hash_get_key_value(sen_ctx *ctx, sen_hash *hash, sen_id id,
                       void *keybuf, int bufsize, void *valuebuf)
{
  void *v;
  uint8_t res;
  int key_size;
  entry_str *ee;
  BITMAP_AT(hash, id, res);
  if (!res) { return 0; }
  ENTRY_AT(hash, id, ee, 0);
  if (!ee) { return 0; }
  key_size = (hash->obj.flags & SEN_OBJ_KEY_VAR_SIZE) ? ee->size : hash->key_size;
  if (bufsize >= key_size) { memcpy(keybuf, get_key(hash, ee), key_size); }
  if (valuebuf && (v = get_value(hash, ee))) {
    memcpy(valuebuf, v, hash->value_size);
  }
  return key_size;
}

int
_sen_hash_get_key_value(sen_ctx *ctx, sen_hash *hash, sen_id id,
                        void **key, void **value)
{
  uint8_t res;
  int key_size;
  entry_str *ee;
  BITMAP_AT(hash, id, res);
  if (!res) { return 0; }
  ENTRY_AT(hash, id, ee, 0);
  if (!ee) { return 0; }
  key_size = (hash->obj.flags & SEN_OBJ_KEY_VAR_SIZE) ? ee->size : hash->key_size;
  *key = get_key(hash, ee);
  return (*value = get_value(hash, ee)) ? key_size : 0;
}

sen_rc
sen_hash_set_value(sen_ctx *ctx, sen_hash *hash, sen_id id, void *value, int flags)
{
  if (value) {
    void *v;
    uint8_t res;
    entry_str *ee;
    BITMAP_AT(hash, id, res);
    if (!res) { return sen_invalid_argument; }
    ENTRY_AT(hash, id, ee, 0);
    if (ee && (v = get_value(hash, ee))) {
      switch ((flags & SEN_OBJ_SET_MASK)) {
      case SEN_OBJ_SET :
        memcpy(v, value, hash->value_size);
        return sen_success;
      default :
        // todo : support other types
        return sen_invalid_argument;
      }
    } else {
      return sen_memory_exhausted;
    }
  }
  return sen_invalid_argument;
}

#define DELETE_IT {\
  *ep = GARBAGE;\
  if (IO_HASHP(hash)) {\
    uint32_t size = key_size - 1;\
    struct sen_hash_header *hh = hash->header;\
    ee->key = hh->garbages[size];\
    hh->garbages[size] = e;\
    SEN_IO_ARRAY_BIT_OFF(hash->io, segment_bitmap, e);\
  } else {\
    ee->key = hash->garbages;\
    hash->garbages = e;\
    if ((hash->obj.flags & SEN_OBJ_KEY_VAR_SIZE) && !(ee->flag & HASH_IMMEDIATE)) {\
      sen_ctx *ctx = hash->ctx;\
      SEN_CTX_FREE(ctx, ((entry_astr *)ee)->str);\
    }\
    SEN_TINY_ARRAY_BIT_OFF(&hash->bitmap, e);\
  }\
  (*hash->n_entries)--;\
  (*hash->n_garbages)++;\
  rc = sen_success;\
}

sen_rc
sen_hash_delete_by_id(sen_ctx *ctx, sen_hash *hash, sen_id id,
                      sen_table_delete_optarg *optarg)
{
  entry_str *ee;
  sen_rc rc = sen_invalid_argument;
  if (!hash || !id) { return rc; }
  /* lock */
  ENTRY_AT(hash, id, ee, 0);
  if (ee) {
    sen_id e, *ep;
    uint32_t i, key_size, h = ee->key, s = STEP(h);
    key_size = (hash->obj.flags & SEN_OBJ_KEY_VAR_SIZE) ? ee->size : hash->key_size;
    for (i = h; ; i += s) {
      if (!(ep = IDX_AT(hash, i))) { return sen_memory_exhausted; }
      if (!(e = *ep)) { break; }
      if (e == id) {
        DELETE_IT;
        break;
      }
    }
  }
  /* unlock */
  return rc;
}

sen_rc
sen_hash_delete(sen_ctx *ctx, sen_hash *hash, const void *key, uint32_t key_size,
                sen_table_delete_optarg *optarg)
{
  uint32_t h, i, m, s;
  sen_rc rc = sen_invalid_argument;
  if (hash->obj.flags & SEN_OBJ_KEY_VAR_SIZE) {
    if (key_size > hash->key_size) { return sen_invalid_argument; }
    h = bin_hash((unsigned char *)key, key_size);
  } else {
    if (key_size != hash->key_size) { return sen_invalid_argument; }
    if (key_size == sizeof(uint32_t)) {
      h = *((uint32_t *)key);
    } else {
      h = bin_hash((unsigned char *)key, key_size);
    }
  }
  s = STEP(h);
  {
    sen_id e, *ep;
    /* lock */
    m = *hash->max_offset;
    for (i = h; ; i += s) {
      if (!(ep = IDX_AT(hash, i))) { return sen_memory_exhausted; }
      if (!(e = *ep)) { break; }
      if (e == GARBAGE) { continue; }
      {
        entry_str *ee;
        ENTRY_AT(hash, e, ee, 0);
        if (ee && match_key(hash, ee, h, key, key_size)) {
          DELETE_IT;
          break;
        }
      }
    }
    /* unlock */
    return rc;
  }
}

/* only valid for hash tables, SEN_OBJ_KEY_VAR_SIZE && SEN_HASH_TINY */
const char *
_sen_hash_strkey_by_val(void *v, uint16_t *size)
{
  entry_astr *n = (entry_astr *)((uintptr_t)v -
                                 (uintptr_t)&((entry_astr *)0)->dummy);
  *size = n->size;
  return (n->flag & HASH_IMMEDIATE) ? (char *)&n->str : n->str;
}

void
sen_hash_cursor_close(sen_ctx *ctx, sen_hash_cursor *c)
{
  SEN_ASSERT(c->ctx == ctx);
  SEN_FREE(c);
}

#define HASH_CURR_MAX(hash) ((IO_HASHP(hash)) ? (hash)->header->curr_rec : (hash)->a.max)

sen_hash_cursor *
sen_hash_cursor_open(sen_ctx *ctx, sen_hash *hash,
                    const void *min, uint32_t min_size,
                    const void *max, uint32_t max_size, int flags)
{
  sen_hash_cursor *c;
  sen_search_flags f = 0;
  if (!hash || !ctx) { return NULL; }
  if (!(c = SEN_MALLOCN(sen_hash_cursor, 1))) { return NULL; }
  SEN_OBJ_DB_SET_TYPE(c, SEN_CURSOR_TABLE_HASH_KEY);
  c->hash = hash;
  c->ctx = ctx;
  c->obj.flags = flags;
  if (flags & SEN_CURSOR_ASCENDING) {
    c->dir = 1;
    if (min) {
      if (!(c->curr_rec = sen_hash_lookup(ctx, hash, min, min_size, NULL, &f))) {
        c->limit = SEN_ID_NIL;
        goto exit;
      }
      if (!(flags & SEN_CURSOR_GT)) { c->curr_rec--; }
    } else {
      c->curr_rec = SEN_ID_NIL;
    }
    if (max) {
      if (!(c->limit = sen_hash_lookup(ctx, hash, max, max_size, NULL, &f))) {
        c->curr_rec = SEN_ID_NIL;
        goto exit;
      }
      if ((flags & SEN_CURSOR_LT)) { c->limit--; }
    } else {
      c->limit = HASH_CURR_MAX(hash);
    }
    if (c->limit < c->curr_rec) { c->limit = c->curr_rec; }
  } else {
    c->dir = -1;
    if (max) {
      if (!(c->curr_rec = sen_hash_lookup(ctx, hash, max, max_size, NULL, &f))) {
        c->limit = SEN_ID_NIL;
        goto exit;
      }
      if (!(flags & SEN_CURSOR_LT)) { c->curr_rec++; }
    } else {
      c->curr_rec = HASH_CURR_MAX(hash) + 1;
    }
    if (min) {
      if (!(c->limit = sen_hash_lookup(ctx, hash, min, min_size, NULL, &f))) {
        c->curr_rec = SEN_ID_NIL;
        goto exit;
      }
      if ((flags & SEN_CURSOR_GT)) { c->limit++; }
    } else {
      c->limit = SEN_ID_NIL + 1;
    }
    if (c->curr_rec < c->limit) { c->limit = c->curr_rec; }
  }
exit :
  return c;
}

sen_id
sen_hash_cursor_next(sen_ctx *ctx, sen_hash_cursor *c)
{
  if (c) {
    while (c->curr_rec != c->limit) {
      uint8_t res;
      c->curr_rec += c->dir;
      BITMAP_AT(c->hash, c->curr_rec, res);
      if (res) { return c->curr_rec; }
    }
  }
  return SEN_ID_NIL;
}

sen_id
sen_hash_next(sen_ctx *ctx, sen_hash *hash, sen_id id)
{
  sen_id max = HASH_CURR_MAX(hash);
  while (++id <= max) {
    uint8_t res;
    BITMAP_AT(hash, id, res);
    if (res) { return id; }
  }
  return SEN_ID_NIL;
}

int
sen_hash_cursor_get_key(sen_ctx *ctx, sen_hash_cursor *c, void **key)
{
  int key_size;
  entry_str *ee;
  if (!c) { return 0; }
  ENTRY_AT(c->hash, c->curr_rec, ee, 0);
  if (!ee) { return 0; }
  key_size = (c->hash->obj.flags & SEN_OBJ_KEY_VAR_SIZE) ? ee->size : c->hash->key_size;
  *key = get_key(c->hash, ee);
  return key_size;
}

int
sen_hash_cursor_get_value(sen_ctx *ctx, sen_hash_cursor *c, void **value)
{
  void *v;
  entry_str *ee;
  if (!c) { return 0; }
  ENTRY_AT(c->hash, c->curr_rec, ee, 0);
  if (ee && (v = get_value(c->hash, ee))) {
    *value = v;
    return c->hash->value_size;
  }
  return 0;
}

int
sen_hash_cursor_get_key_value(sen_ctx *ctx, sen_hash_cursor *c,
                              void **key, uint32_t *key_size, void **value)
{
  entry_str *ee;
  if (!c) { return sen_invalid_argument; }
  ENTRY_AT(c->hash, c->curr_rec, ee, 0);
  if (!ee) { return sen_invalid_argument; }
  if (key_size) {
    *key_size = (c->hash->obj.flags & SEN_OBJ_KEY_VAR_SIZE) ? ee->size : c->hash->key_size;
  }
  if (key) { *key = get_key(c->hash, ee); }
  if (value) { *value = get_value(c->hash, ee); }
  return c->hash->value_size;
}

sen_rc
sen_hash_cursor_set_value(sen_ctx *ctx, sen_hash_cursor *c, void *value, int flags)
{
  if (!c) { return sen_invalid_argument; }
  return sen_hash_set_value(ctx, c->hash, c->curr_rec, value, flags);
}

sen_rc
sen_hash_cursor_delete(sen_ctx *ctx, sen_hash_cursor *c,
                       sen_table_delete_optarg *optarg)
{
  if (!c) { return sen_invalid_argument; }
  return sen_hash_delete_by_id(ctx, c->hash, c->curr_rec, optarg);
}

/* sort */

#define PREPARE_VAL(e,ep,es) {\
  if ((arg->flags & SEN_TABLE_SORT_BY_VALUE)) {\
    ep = ((const uint8_t *)(get_value(hash, (entry_str *)(e))));\
    es = hash->value_size;\
  } else {\
    ep = ((const uint8_t *)(get_key(hash, (entry_str *)(e))));\
    es = ((hash->obj.flags & SEN_OBJ_KEY_VAR_SIZE)\
          ? ((entry_str *)(e))->size : hash->key_size); \
  }\
  ep += arg->offset;\
  es -= arg->offset;\
}

#define COMPARE_VAL_(ap,as,bp,bs)\
  (arg->compar\
   ? arg->compar(ctx,\
                 (sen_obj *)hash, (void *)(ap), as,\
                 (sen_obj *)hash, (void *)(bp), bs, arg->compar_arg)\
   : ((arg->flags & SEN_TABLE_SORT_AS_NUMBER)\
      ? ((arg->flags & SEN_TABLE_SORT_AS_UNSIGNED)\
         ? ((arg->flags & SEN_TABLE_SORT_AS_INT64)\
            ? *((uint64_t *)(ap)) > *((uint64_t *)(bp))\
            : *((uint32_t *)(ap)) > *((uint32_t *)(bp)))\
         : ((arg->flags & SEN_TABLE_SORT_AS_INT64)\
            ? *((int64_t *)(ap)) > *((int64_t *)(bp))\
            : *((int32_t *)(ap)) > *((int32_t *)(bp))))\
      : sen_str_greater(ap, as, bp, bs)))

#define COMPARE_VAL(ap,as,bp,bs)\
  ((dir) ? COMPARE_VAL_((ap),(as),(bp),(bs)) : COMPARE_VAL_((bp),(bs),(ap),(as)))

inline static entry **
pack(sen_ctx *ctx, sen_hash *hash, entry **res, sen_table_sort_optarg *arg, int dir)
{
  uint32_t n;
  uint8_t exist;
  uint32_t cs, es;
  const uint8_t *cp, *ep;
  entry **head, **tail, *e, *c;
  sen_id id, m = HASH_CURR_MAX(hash);
  for (id = m >> 1;;id = (id == m) ? 1 : id + 1) {
    BITMAP_AT(hash, id, exist);
    if (exist) { break; }
  }
  ENTRY_AT(hash, id, c, 0);
  if (!c) { return NULL; }
  PREPARE_VAL(c, cp, cs);
  head = res;
  n = *hash->n_entries - 1;
  tail = res + n;
  while (n--) {
    do {
      id = (id == m) ? 1 : id + 1;
      BITMAP_AT(hash, id, exist);
    } while (!exist);
    ENTRY_AT(hash, id, e, 0);
    if (!e) { return NULL; }
    PREPARE_VAL(e, ep, es);
    if (COMPARE_VAL(cp, cs, ep, es)) {
      *head++ = e;
    } else {
      *tail-- = e;
    }
  }
  *head = c;
  return *hash->n_entries > 2 ? head : NULL;
}

inline static void
swap(entry **a, entry **b)
{
  entry *c_ = *a;
  *a = *b;
  *b = c_;
}

#define SWAP(a,ap,as,b,bp,bs) {\
  const uint8_t *cp_ = ap;\
  uint32_t cs_ = as;\
  ap = bp; bp = cp_;\
  as = bs; bs = cs_;\
  swap(a,b);\
}

inline static entry **
part(sen_ctx *ctx, entry **b, entry **e, sen_table_sort_optarg *arg, sen_hash *hash, int dir)
{
  entry **c;
  const uint8_t *bp, *cp, *ep;
  uint32_t bs, cs, es;
  intptr_t d = e - b;
  PREPARE_VAL(*b, bp, bs);
  PREPARE_VAL(*e, ep, es);
  if (COMPARE_VAL(bp, bs, ep, es)) {
    SWAP(b, bp, bs, e, ep, es);
  }
  if (d < 2) { return NULL; }
  c = b + (d >> 1);
  PREPARE_VAL(*c, cp, cs);
  if (COMPARE_VAL(bp, bs, cp, cs)) {
    SWAP(b, bp, bs, c, cp, cs);
  } else {
    if (COMPARE_VAL(cp, cs, ep, es)) {
      SWAP(c, cp, cs, e, ep, es);
    }
  }
  if (d < 3) { return NULL; }
  b++;
  swap(b, c);
  c = b;
  PREPARE_VAL(*c, cp, cs);
  for (;;) {
    do {
      b++;
      PREPARE_VAL(*b, bp, bs);
    } while (COMPARE_VAL(cp, cs, bp, bs));
    do {
      e--;
      PREPARE_VAL(*e, ep, es);
    } while (COMPARE_VAL(ep, es, cp, cs));
    if (b >= e) { break; }
    SWAP(b, bp, bs, e, ep, es);
  }
  SWAP(c, cp, cs, e, ep, es);
  return e;
}

static void
_sort(sen_ctx *ctx, entry **head, entry **tail, int limit,
      sen_table_sort_optarg *arg, sen_hash *hash, int dir)
{
  entry **c;
  if (head < tail && (c = part(ctx, head, tail, arg, hash, dir))) {
    intptr_t rest = limit - 1 - (c - head);
    _sort(ctx, head, c - 1, limit, arg, hash, dir);
    if (rest > 0) { _sort(ctx, c + 1, tail, (int)rest, arg, hash, dir); }
  }
}

static void
sort(sen_ctx *ctx,
     sen_hash *hash, entry **res, int limit, sen_table_sort_optarg *arg, int dir)
{
  entry **c = pack(ctx, hash, res, arg, dir);
  if (c) {
    intptr_t rest = limit - 1 - (c - res);
    _sort(ctx, res, c - 1, limit, arg, hash, dir);
    if (rest > 0 ) {
      _sort(ctx, c + 1, res + *hash->n_entries - 1, (int)rest, arg, hash, dir);
    }
  }
}

typedef struct {
  sen_id id;
  int32_t v;
} val32;

#define PREPARE_VAL32(id,e,ep) {\
  (ep)->id = id;\
  (ep)->v = (arg->flags & SEN_TABLE_SORT_BY_ID)\
    ? (int32_t) id\
    : (*((int32_t *)((byte *)((arg->flags & SEN_TABLE_SORT_BY_VALUE)\
                              ? get_value(hash, (e))\
                              : get_key(hash, (e))) + arg->offset)));\
}

#define COMPARE_VAL32_(ap,bp) \
  (arg->compar\
   ? arg->compar(ctx,\
                 (sen_obj *)hash, (void *)&(ap)->v, sizeof(uint32_t),\
                 (sen_obj *)hash, (void *)&(bp)->v, sizeof(uint32_t),\
                 arg->compar_arg)\
   : ((arg->flags & SEN_TABLE_SORT_AS_NUMBER)\
      ? ((arg->flags & SEN_TABLE_SORT_AS_UNSIGNED)\
         ? *((uint32_t *)&(ap)->v) > *((uint32_t *)&(bp)->v)\
         : *((int32_t *)&(ap)->v) > *((int32_t *)&(bp)->v))\
      : memcmp(&(ap)->v, &(bp)->v, sizeof(uint32_t)) > 0))

#define COMPARE_VAL32(ap,bp)\
  ((dir) ? COMPARE_VAL32_((ap),(bp)) : COMPARE_VAL32_((bp),(ap)))

inline static val32 *
pack_val32(sen_ctx *ctx, sen_hash *hash, val32 *res, sen_table_sort_optarg *arg, int dir)
{
  uint32_t n;
  uint8_t exist;
  entry_str *e, *c;
  val32 *head, *tail, cr, er;
  sen_id id, m = HASH_CURR_MAX(hash);
  for (id = m >> 1;;id = (id == m) ? 1 : id + 1) {
    BITMAP_AT(hash, id, exist);
    if (exist) { break; }
  }
  ENTRY_AT(hash, id, c, 0);
  if (!c) { return NULL; }
  PREPARE_VAL32(id, c, &cr);
  head = res;
  n = *hash->n_entries - 1;
  tail = res + n;
  while (n--) {
    do {
      id = (id == m) ? 1 : id + 1;
      BITMAP_AT(hash, id, exist);
    } while (!exist);
    ENTRY_AT(hash, id, e, 0);
    if (!e) { return NULL; }
    PREPARE_VAL32(id, e, &er);
    if (COMPARE_VAL32(&cr, &er)) {
      *head++ = er;
    } else {
      *tail-- = er;
    }
  }
  *head = cr;
  return *hash->n_entries > 2 ? head : NULL;
}

#define SWAP_VAL32(ap,bp) {\
  val32 cr_ = *ap;\
  *ap = *bp;\
  *bp = cr_;\
}

inline static val32 *
part_val32(sen_ctx *ctx,
           val32 *b, val32 *e, sen_table_sort_optarg *arg, sen_hash *hash, int dir)
{
  val32 *c;
  intptr_t d = e - b;
  if (COMPARE_VAL32(b, e)) { SWAP_VAL32(b, e); }
  if (d < 2) { return NULL; }
  c = b + (d >> 1);
  if (COMPARE_VAL32(b, c)) {
    SWAP_VAL32(b, c);
  } else {
    if (COMPARE_VAL32(c, e)) { SWAP_VAL32(c, e); }
  }
  if (d < 3) { return NULL; }
  b++;
  SWAP_VAL32(b, c);
  c = b;
  for (;;) {
    do { b++; } while (COMPARE_VAL32(c, b));
    do { e--; } while (COMPARE_VAL32(e, c));
    if (b >= e) { break; }
    SWAP_VAL32(b, e);
  }
  SWAP_VAL32(c, e);
  return e;
}

static void
_sort_val32(sen_ctx *ctx, val32 *head, val32 *tail, int limit,
      sen_table_sort_optarg *arg, sen_hash *hash, int dir)
{
  val32 *c;
  if (head < tail && (c = part_val32(ctx, head, tail, arg, hash, dir))) {
    intptr_t rest = limit - 1 - (c - head);
    _sort_val32(ctx, head, c - 1, limit, arg, hash, dir);
    if (rest > 0) { _sort_val32(ctx, c + 1, tail, (int)rest, arg, hash, dir); }
  }
}

static void
sort_val32(sen_ctx *ctx,
           sen_hash *hash, val32 *res, int limit, sen_table_sort_optarg *arg, int dir)
{
  val32 *c = pack_val32(ctx, hash, res, arg, dir);
  if (c) {
    intptr_t rest = limit - 1 - (c - res);
    _sort_val32(ctx, res, c - 1, limit, arg, hash, dir);
    if (rest > 0 ) {
      _sort_val32(ctx, c + 1, res + *hash->n_entries - 1, (int)rest, arg, hash, dir);
    }
  }
}

inline static sen_id
entry2id(sen_hash *hash, entry *e)
{
  entry *e2;
  sen_id id, *ep;
  uint32_t i, h = e->key, s = STEP(h);
  for (i = h; ; i += s) {
    if (!(ep = IDX_AT(hash, i))) { return SEN_ID_NIL; }
    if (!(id = *ep)) { break; }
    if (id != GARBAGE) {
      ENTRY_AT(hash, id, e2, 0);
      if (!e2) { return SEN_ID_NIL; }
      if (e2 == e) { break; }
    }
  }
  return id;
}

int
sen_hash_sort(sen_ctx *ctx, sen_hash *hash,
              int limit, sen_array *result, sen_table_sort_optarg *optarg)
{
  entry **res;
  if (!result || !*hash->n_entries) { return 0; }
  if (!(res = SEN_MALLOC(sizeof(entry *) * *hash->n_entries))) {
    SEN_LOG(sen_log_alert, "allocation of entries failed on sen_hash_sort !");
    return 0;
  }
  if (limit <= 0) {
    limit += *hash->n_entries;
    if (limit <= 0) {
      SEN_LOG(sen_log_alert, "limit is too small in sen_hash_sort !");
      return 0;
    }
  }
  if (limit > *hash->n_entries) { limit = *hash->n_entries; }
  //  hash->limit = limit;
  if (optarg) {
    int dir = (optarg->flags & SEN_TABLE_SORT_ASC);
    if ((optarg->flags & SEN_TABLE_SORT_BY_ID) ||
        (optarg->flags & SEN_TABLE_SORT_BY_VALUE)
        ? ((hash->value_size - optarg->offset) == sizeof(uint32_t))
        : (!(hash->obj.flags & SEN_OBJ_KEY_VAR_SIZE)
           && hash->key_size == sizeof(uint32_t))) {
      if (sizeof(entry *) != sizeof(val32)) {
        SEN_FREE(res);
        if (!(res = SEN_MALLOC(sizeof(val32) * *hash->n_entries))) {
          SEN_LOG(sen_log_alert, "allocation of entries failed on sen_hash_sort !");
          return 0;
        }
      }
      sort_val32(ctx, hash, (val32 *)res, limit, optarg, dir);
      {
        int i;
        sen_id *v;
        val32 *rp = (val32 *)res;
        for (i = 0; i < limit; i++, rp++) {
          if (!sen_array_add(ctx, result, (void **)&v)) { break; }
          if (!(*v = rp->id)) { break; }
        }
        SEN_FREE(res);
        return i;
      }
    } else {
      sort(ctx, hash, res, limit, optarg, dir);
    }
  } else {
    sen_table_sort_optarg opt = {0, NULL, NULL, 0};
    sort(ctx, hash, res, limit, &opt, 0);
  }
  {
    int i;
    sen_id *v;
    entry **rp = res;
    for (i = 0; i < limit; i++, rp++) {
      if (!sen_array_add(ctx, result, (void **)&v)) { break; }
      if (!(*v = entry2id(hash, *rp))) { break; }
    }
    SEN_FREE(res);
    return i;
  }
}

/* todo : not completed yet
sen_rc
sen_hash_group(sen_ctx *ctx, sen_hash *hash,
               int n_args, sen_table_group_arg *args)
{
  sen_id id;
  uint32_t n = *hash->n_entries;
  SEN_ASSERT(n_args && args);
  {
    int i;
    for (i = 0; i < n_args; i++) {
      ((sen_obj_db *)args[i].result)->range = ((sen_obj_db *)hash)->id;
    }
  }
  for (id = 1; n; id++) {
    uint32_t es;
    entry_str *e;
    uint8_t exist;
    const uint8_t *ep;
    sen_table_sort_optarg *arg = &args->sortarg;
    BITMAP_AT(hash, id, exist);
    if (!exist) { continue; }
    n--;
    ENTRY_AT(hash, id, e, 0);
    if (!e) { return sen_internal_error; }
    PREPARE_VAL(e, ep, es);
    {
      uint32_t *val;
      sen_search_flags flags = SEN_TABLE_ADD;
      sen_id gid = (args->result->header.type == SEN_TABLE_PAT_KEY)
        ? sen_pat_lookup(ctx, (sen_pat *)args->result, ep, es, (void **)&val, &flags)
        : sen_hash_lookup(ctx, (sen_hash *)args->result, ep, es, (void **)&val, &flags);
      if (!gid) { return sen_internal_error; }
      if (args->flags & SEN_TABLE_GROUP_CALC_COUNT) {
        *val += 1;
        val++;
      }
    }
  }
  return sen_success;
}
*/

/* rhash : sen_hash with subrecs */

#ifdef USE_SEN_INDEX2

static uint32_t default_flags = SEN_HASH_TINY;

sen_rc
sen_rhash_init(sen_ctx *ctx, sen_hash *hash, sen_rec_unit record_unit, int record_size,
              sen_rec_unit subrec_unit, int subrec_size, unsigned int max_n_subrecs)
{
  sen_rc rc;
  record_size = sen_rec_unit_size(record_unit, record_size);
  subrec_size = sen_rec_unit_size(subrec_unit, subrec_size);
  if (record_unit != sen_rec_userdef && subrec_unit != sen_rec_userdef) {
    subrec_size -= record_size;
  }
  if (!hash) { return sen_invalid_argument; }
  if (record_size < 0) { return sen_invalid_argument; }
  if ((default_flags & SEN_HASH_TINY)) {
    rc = tiny_hash_init(hash, ctx, NULL, record_size,
                        max_n_subrecs * (SEN_RSET_SCORE_SIZE + subrec_size),
                        default_flags, sen_enc_none);
  } else {
    rc = io_hash_init(hash, ctx, NULL, record_size,
                      max_n_subrecs * (SEN_RSET_SCORE_SIZE + subrec_size),
                      default_flags, sen_enc_none, 0);
  }
  if (rc) { return rc; }
  hash->record_unit = record_unit;
  hash->subrec_unit = subrec_unit;
  hash->subrec_size = subrec_size;
  hash->max_n_subrecs = max_n_subrecs;
  return rc;
}

sen_rc
sen_rhash_fin(sen_ctx *ctx, sen_hash *hash)
{
  sen_rc rc;
  if (IO_HASHP(hash)) {
    rc = sen_io_close(hash->io);
  } else {
    SEN_ASSERT(ctx == hash->ctx);
    rc = tiny_hash_fin(ctx, hash);
  }
  return rc;
}

inline static void
subrecs_push(byte *subrecs, int size, int n_subrecs, int score, void *body, int dir)
{
  byte *v;
  int *c2;
  int n = n_subrecs - 1, n2;
  while (n) {
    n2 = (n - 1) >> 1;
    c2 = SEN_RSET_SUBRECS_NTH(subrecs,size,n2);
    if (SEN_RSET_SUBRECS_CMP(score, *c2, dir)) { break; }
    SEN_RSET_SUBRECS_COPY(subrecs,size,n,c2);
    n = n2;
  }
  v = subrecs + n * (size + SEN_RSET_SCORE_SIZE);
  *((int *)v) = score;
  memcpy(v + SEN_RSET_SCORE_SIZE, body, size);
}

inline static void
subrecs_replace_min(byte *subrecs, int size, int n_subrecs, int score, void *body, int dir)
{
  byte *v;
  int n = 0, n1, n2, *c1, *c2;
  for (;;) {
    n1 = n * 2 + 1;
    n2 = n1 + 1;
    c1 = n1 < n_subrecs ? SEN_RSET_SUBRECS_NTH(subrecs,size,n1) : NULL;
    c2 = n2 < n_subrecs ? SEN_RSET_SUBRECS_NTH(subrecs,size,n2) : NULL;
    if (c1 && SEN_RSET_SUBRECS_CMP(score, *c1, dir)) {
      if (c2 &&
          SEN_RSET_SUBRECS_CMP(score, *c2, dir) &&
          SEN_RSET_SUBRECS_CMP(*c1, *c2, dir)) {
        SEN_RSET_SUBRECS_COPY(subrecs,size,n,c2);
        n = n2;
      } else {
        SEN_RSET_SUBRECS_COPY(subrecs,size,n,c1);
        n = n1;
      }
    } else {
      if (c2 && SEN_RSET_SUBRECS_CMP(score, *c2, dir)) {
        SEN_RSET_SUBRECS_COPY(subrecs,size,n,c2);
        n = n2;
      } else {
        break;
      }
    }
  }
  v = subrecs + n * (size + SEN_RSET_SCORE_SIZE);
  memcpy(v, &score, SEN_RSET_SCORE_SIZE);
  memcpy(v + SEN_RSET_SCORE_SIZE, body, size);
}

void
sen_rhash_add_subrec(sen_hash *s, sen_rset_recinfo *ri, int score, void *body, int dir)
{
  int limit = s->max_n_subrecs;
  ri->score += score;
  ri->n_subrecs += 1;
  if (limit) {
    int ssize = s->subrec_size;
    int n_subrecs = SEN_RSET_N_SUBRECS(ri);
    if (limit < n_subrecs) {
      if (SEN_RSET_SUBRECS_CMP(score, *ri->subrecs, dir)) {
        subrecs_replace_min(ri->subrecs, ssize, limit, score, body, dir);
      }
    } else {
      subrecs_push(ri->subrecs, ssize, n_subrecs, score, body, dir);
    }
  }
}

sen_hash *
sen_rhash_group(sen_hash *s, int limit, sen_group_optarg *optarg)
{
  sen_ctx *ctx = s->ctx;
  sen_hash *g, h;
  sen_rset_recinfo *ri;
  sen_rec_unit unit;
  sen_hash_cursor *c;
  sen_id rh;
  byte *key, *ekey, *gkey = NULL;
  int funcp, dir;
  unsigned int rsize;
  if (!s || !s->index) { return NULL; }
  if (optarg) {
    unit = sen_rec_userdef;
    rsize = optarg->key_size;
    funcp = optarg->func ? 1 : 0;
    dir = (optarg->mode == sen_sort_ascending) ? -1 : 1;
  } else {
    unit = sen_rec_document;
    rsize = sen_rec_unit_size(unit, sizeof(sen_id));
    funcp = 0;
    dir = 1;
  }
  if (funcp) {
    gkey = SEN_MALLOC(rsize ? rsize : 8192);
    if (!gkey) {
      SEN_LOG(sen_log_alert, "allocation for gkey failed !");
      return NULL;
    }
  } else {
    if (s->key_size <= rsize) { return NULL; }
  }
  if (!(c = sen_hash_cursor_open(s->ctx, s, NULL, 0, NULL, 0, 0))) {
    SEN_LOG(sen_log_alert, "sen_hash_cursor_open on sen_hash_group failed !");
    if (gkey) { SEN_FREE(gkey); }
    return NULL;
  }
  memcpy(&h, s, sizeof(sen_hash));
  g = s;
  s = &h;
  if (sen_rhash_init(ctx, g, unit, rsize, s->record_unit, s->key_size, limit)) {
    SEN_LOG(sen_log_alert, "sen_rhash_init in sen_hash_group failed !");
    sen_hash_cursor_close(s->ctx, c);
    if (gkey) { SEN_FREE(gkey); }
    return NULL;
  }
  while ((rh = sen_hash_cursor_next(ctx, c))) {
    sen_hash_cursor_get_key_value(ctx, c, (void **)&key, NULL, (void **)&ri);
    if (funcp) {
      if (optarg->func((sen_records *)s,
                       (sen_recordh *)(intptr_t)rh, gkey, optarg->func_arg)) { continue; }
      ekey = key;
    } else {
      gkey = key;
      ekey = key + rsize;
    }
    {
      sen_rset_recinfo *gri;
      sen_search_flags f = SEN_TABLE_ADD;
      if (sen_hash_get(ctx, g, gkey, rsize, (void **)&gri, &f)) {
        sen_rhash_add_subrec(g, gri, ri->score, ekey, dir);
      }
    }
  }
  sen_hash_cursor_close(s->ctx, c);
  sen_rhash_fin(s->ctx, s);
  if (funcp) { SEN_FREE(gkey); }
  return g;
}

sen_rc
sen_rhash_subrec_info(sen_hash *s, sen_id rh, int index,
                     sen_id *rid, int *section, int *pos, int *score, void **subrec)
{
  sen_rset_posinfo *pi;
  sen_rset_recinfo *ri;
  int *p, unit_size = SEN_RSET_SCORE_SIZE + s->subrec_size;
  if (!s || !rh || index < 0) { return sen_invalid_argument; }
  if ((unsigned int)index >= s->max_n_subrecs) { return sen_invalid_argument; }
  {
    uint8_t res;
    entry_str *ee;
    BITMAP_AT(s, rh, res);
    if (!res) { return sen_invalid_argument; }
    ENTRY_AT(s, rh, ee, 0);
    if (!ee) { return sen_invalid_argument; }
    pi = (sen_rset_posinfo *)get_key(s, ee);
    ri = get_value(s, ee);
    if (!pi || !ri) { return sen_invalid_argument; }
  }
  if (index >= ri->n_subrecs) { return sen_invalid_argument; }
  p = (int *)(ri->subrecs + index * unit_size);
  if (score) { *score = p[0]; }
  if (subrec) { *subrec = &p[1]; }
  switch (s->record_unit) {
  case sen_rec_document :
    if (rid) { *rid = pi->rid; }
    if (section) { *section = (s->subrec_unit != sen_rec_userdef) ? p[1] : 0; }
    if (pos) { *pos = (s->subrec_unit == sen_rec_position) ? p[2] : 0; }
    break;
  case sen_rec_section :
    if (rid) { *rid = pi->rid; }
    if (section) { *section = pi->sid; }
    if (pos) { *pos = (s->subrec_unit == sen_rec_position) ? p[1] : 0; }
    break;
  default :
    pi = (sen_rset_posinfo *)&p[1];
    switch (s->subrec_unit) {
    case sen_rec_document :
      if (rid) { *rid = pi->rid; }
      if (section) { *section = 0; }
      if (pos) { *pos = 0; }
      break;
    case sen_rec_section :
      if (rid) { *rid = pi->rid; }
      if (section) { *section = pi->sid; }
      if (pos) { *pos = 0; }
      break;
    case sen_rec_position :
      if (rid) { *rid = pi->rid; }
      if (section) { *section = pi->sid; }
      if (pos) { *pos = pi->pos; }
      break;
    default :
      if (rid) { *rid = 0; }
      if (section) { *section = 0; }
      if (pos) { *pos = 0; }
      break;
    }
    break;
  }
  return sen_success;
}
#endif /* USE_SEN_INDEX2 */
