/* Copyright(C) 2004-2007 Brazil

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "senna_in.h"
#include <stdio.h>
#include <string.h>
#include "sym.h"
#include "inv.h"
#include "str.h"
#include "set.h"
#include "lex.h"
#include "cache.h"
#include "store.h"
#include "ctx.h"
#include "com.h"

/* todo : config_path will be disappeared */
sen_rc
sen_info(char **version,
         char **configure_options,
         char **config_path,
         sen_encoding *default_encoding,
         unsigned int *initial_n_segments,
         unsigned int *partial_match_threshold)
{
  if (version) {
    *version = PACKAGE_VERSION;
  }
  if (configure_options) {
    *configure_options = CONFIGURE_OPTIONS;
  }
  if (default_encoding) {
    *default_encoding = sen_gctx.encoding;
  }
  if (initial_n_segments) {
    *initial_n_segments = SENNA_DEFAULT_INITIAL_N_SEGMENTS;
  }
  if (partial_match_threshold) {
    *partial_match_threshold = SENNA_DEFAULT_QUERY_ESCALATION_THRESHOLD;
  }
  return sen_success;
}

/* sen_index */

void
sen_index_expire(void)
{
  sen_inv_expire();
}

sen_rc
sen_index_remove(const char *path)
{
  sen_rc rc;
  char buffer[PATH_MAX];
  if (!path || strlen(path) > PATH_MAX - 8) { return sen_invalid_argument; }
  snprintf(buffer, PATH_MAX, "%s.SEN", path);
  if ((rc = sen_sym_remove(buffer))) { goto exit; }
  snprintf(buffer, PATH_MAX, "%s.SEN.i", path);
  if ((rc = sen_inv_remove(buffer))) { goto exit; }
  snprintf(buffer, PATH_MAX, "%s.SEN.l", path);
  if ((rc = sen_sym_remove(buffer))) { goto exit; }
  snprintf(buffer, PATH_MAX, "%s.SEN.v", path);
  sen_io_remove(buffer); // sen_vgram_remove
exit :
  return rc;
}

sen_rc
sen_index_rename(const char *old_name, const char *new_name)
{
  sen_rc rc;
  char old_buffer[PATH_MAX];
  char new_buffer[PATH_MAX];
  if (!old_name || strlen(old_name) > PATH_MAX - 8) { return sen_invalid_argument; }
  if (!new_name || strlen(new_name) > PATH_MAX - 8) { return sen_invalid_argument; }
  snprintf(old_buffer, PATH_MAX, "%s.SEN", old_name);
  snprintf(new_buffer, PATH_MAX, "%s.SEN", new_name);
  if ((rc = sen_io_rename(old_buffer, new_buffer))) { return rc; }
  snprintf(old_buffer, PATH_MAX, "%s.SEN.i", old_name);
  snprintf(new_buffer, PATH_MAX, "%s.SEN.i", new_name);
  if ((rc = sen_io_rename(old_buffer, new_buffer))) { return rc; }
  snprintf(old_buffer, PATH_MAX, "%s.SEN.i.c", old_name);
  snprintf(new_buffer, PATH_MAX, "%s.SEN.i.c", new_name);
  if ((rc = sen_io_rename(old_buffer, new_buffer))) { return rc; }
  snprintf(old_buffer, PATH_MAX, "%s.SEN.l", old_name);
  snprintf(new_buffer, PATH_MAX, "%s.SEN.l", new_name);
  if ((rc = sen_io_rename(old_buffer, new_buffer))) { return rc; }
  snprintf(old_buffer, PATH_MAX, "%s.SEN.v", old_name);
  snprintf(new_buffer, PATH_MAX, "%s.SEN.v", new_name);
  sen_io_rename(old_buffer, new_buffer);
  return sen_success;
}

#define INITIAL_VALUE_SIZE 1024

sen_values *
sen_values_open(void)
{
  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
  sen_values *v = SEN_MALLOC(sizeof(sen_values));
  if (v) {
    v->n_values = 0;
    v->values = NULL;
  }
  return v;
}

sen_rc
sen_values_close(sen_values *v)
{
  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
  if (!v) { return sen_invalid_argument; }
  if (v->values) { SEN_FREE(v->values); }
  SEN_FREE(v);
  return sen_success;
}

sen_rc
sen_values_add(sen_values *v, const char *str, unsigned int str_len, unsigned int weight)
{
  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
  sen_value *vp;
  if (!str_len) { return sen_success; }
  if (!v || !str) {
    SEN_LOG(sen_log_warning, "sen_values_add: invalid argument");
    return sen_invalid_argument;
  }
  if (!(v->n_values & (INITIAL_VALUE_SIZE - 1))) {
    vp = SEN_REALLOC(v->values, sizeof(sen_value) * (v->n_values + INITIAL_VALUE_SIZE));
    SEN_LOG(sen_log_debug, "expanded values to %d,%p", v->n_values + INITIAL_VALUE_SIZE, vp);
    if (!vp) { return sen_memory_exhausted; }
    v->values = vp;
  }
  vp = &v->values[v->n_values];
  vp->str = str;
  vp->str_len = str_len;
  vp->weight = weight;
  v->n_values++;
  return sen_success;
}

struct _sen_index {
  int foreign_flags;
  sen_sym *keys;
  sen_sym *lexicon;
  sen_inv *inv;
  sen_vgram *vgram;
};

inline static int
build_flags(int flags)
{
  if (flags & SEN_INDEX_ENABLE_SUFFIX_SEARCH) {
    return flags | SEN_SYM_WITH_SIS;
  } else if (flags & SEN_INDEX_DISABLE_SUFFIX_SEARCH) {
    return flags & ~SEN_SYM_WITH_SIS;
  } else {   /* default */
    switch (flags & SEN_INDEX_TOKENIZER_MASK) {
    case SEN_INDEX_MORPH_ANALYSE :
      return flags | SEN_SYM_WITH_SIS;
    case SEN_INDEX_NGRAM :
      return flags & ~SEN_SYM_WITH_SIS;
    case SEN_INDEX_DELIMITED :
      return flags & ~SEN_SYM_WITH_SIS;
    default :
      return flags & ~SEN_SYM_WITH_SIS;
    }
  }
}

sen_index *
sen_index_create(const char *path, int key_size,
                 int flags, int initial_n_segments, sen_encoding encoding)
{
  sen_index *i;
  char buffer[PATH_MAX];
  if (!path) { SEN_LOG(sen_log_warning, "sen_index_create: invalid argument"); return NULL; }
  if (initial_n_segments == 0) { initial_n_segments = SENNA_DEFAULT_INITIAL_N_SEGMENTS; }
  if (encoding == sen_enc_default) { encoding = sen_gctx.encoding; }
  if (strlen(path) > PATH_MAX - 4) {
    SEN_LOG(sen_log_warning, "sen_index_create: too long index path (%s)", path);
    return NULL;
  }
  if (!(i = SEN_GMALLOC(sizeof(sen_index)))) { return NULL; }
  SEN_LOG(sen_log_notice, "creating '%s' encoding=%s initial_n_segments=%d",
          path, sen_enctostr(encoding), initial_n_segments);
  strcpy(buffer, path);
  strcat(buffer, ".SEN");
  i->foreign_flags = 0;
  if ((i->keys = sen_sym_create(buffer, key_size, (flags & 0x70000), sen_enc_none))) {
    strcpy(buffer, path);
    strcat(buffer, ".SEN.l");
    if ((i->lexicon = sen_sym_create(buffer, 0, build_flags(flags), encoding))) {
      strcpy(buffer, path);
      strcat(buffer, ".SEN.i");
      if ((i->inv = sen_inv_create(buffer, i->lexicon, initial_n_segments))) {
        if ((flags & SEN_INDEX_WITH_VGRAM)) {
          strcpy(buffer, path);
          strcat(buffer, ".SEN.v");
          i->vgram= sen_vgram_create(buffer);
        } else {
          i->vgram = NULL;
        }
        if (!(flags & SEN_INDEX_WITH_VGRAM) || i->vgram) {
          SEN_LOG(sen_log_notice, "index created (%s) flags=%x", path, i->lexicon->flags);
          return i;
        }
        sen_inv_close(i->inv);
      }
      sen_sym_close(i->lexicon);
    }
    sen_sym_close(i->keys);
  }
  SEN_GFREE(i);
  return NULL;
}

sen_index *
sen_index_open(const char *path)
{
  sen_index *i;
  char buffer[PATH_MAX];
  if (!path) { SEN_LOG(sen_log_warning, "sen_index_open: invalid argument"); return NULL; }
  if (strlen(path) > PATH_MAX - 4) {
    SEN_LOG(sen_log_warning, "sen_index_open: too long index path (%s)", path);
    return NULL;
  }
  if (!(i = SEN_GMALLOC(sizeof(sen_index)))) { return NULL; }
  strcpy(buffer, path);
  strcat(buffer, ".SEN");
  i->foreign_flags = 0;
  if ((i->keys = sen_sym_open(buffer))) {
    strcpy(buffer, path);
    strcat(buffer, ".SEN.l");
    if ((i->lexicon = sen_sym_open(buffer))) {
      strcpy(buffer, path);
      strcat(buffer, ".SEN.i");
      if ((i->inv = sen_inv_open(buffer, i->lexicon))) {
        if ((i->lexicon->flags & SEN_INDEX_WITH_VGRAM)) {
          strcpy(buffer, path);
          strcat(buffer, ".SEN.v");
          i->vgram = sen_vgram_open(buffer);
        } else {
          i->vgram = NULL;
        }
        if (!(i->lexicon->flags & SEN_INDEX_WITH_VGRAM) || i->vgram) {
          SEN_LOG(sen_log_notice, "index opened (%p:%s) flags=%x", i, path, i->lexicon->flags);
          return i;
        }
        sen_inv_close(i->inv);
      }
      sen_sym_close(i->lexicon);
    }
    sen_sym_close(i->keys);
  }
  SEN_GFREE(i);
  return NULL;
}

#define FOREIGN_KEY     1
#define FOREIGN_LEXICON 2

sen_index *
sen_index_create_with_keys(const char *path, sen_sym *keys,
                           int flags, int initial_n_segments, sen_encoding encoding)
{
  sen_index *i;
  char buffer[PATH_MAX];
  if (!path || !keys) {
    SEN_LOG(sen_log_warning, "sen_index_create_with_keys: invalid argument");
    return NULL;
  }
  if (initial_n_segments == 0) { initial_n_segments = SENNA_DEFAULT_INITIAL_N_SEGMENTS; }
  if (encoding == sen_enc_default) { encoding = sen_gctx.encoding; }
  if (strlen(path) > PATH_MAX - 4) {
    SEN_LOG(sen_log_warning, "too long index path (%s)", path);
    return NULL;
  }
  if (!(i = SEN_GMALLOC(sizeof(sen_index)))) { return NULL; }
  SEN_LOG(sen_log_notice, "creating '%s' encoding=%s initial_n_segments=%d",
          path, sen_enctostr(encoding), initial_n_segments);
  i->keys = keys;
  i->foreign_flags = FOREIGN_KEY;
  strcpy(buffer, path);
  strcat(buffer, ".SEN.l");
  if ((i->lexicon = sen_sym_create(buffer, 0, build_flags(flags), encoding))) {
    strcpy(buffer, path);
    strcat(buffer, ".SEN.i");
    if ((i->inv = sen_inv_create(buffer, i->lexicon, initial_n_segments))) {
      if ((flags & SEN_INDEX_WITH_VGRAM)) {
        strcpy(buffer, path);
        strcat(buffer, ".SEN.v");
        i->vgram= sen_vgram_create(buffer);
      } else {
        i->vgram = NULL;
      }
      if (!(flags & SEN_INDEX_WITH_VGRAM) || i->vgram) {
        SEN_LOG(sen_log_notice, "index created (%s) flags=%x", path, i->lexicon->flags);
        return i;
      }
      sen_inv_close(i->inv);
    }
    sen_sym_close(i->lexicon);
  }
  SEN_GFREE(i);
  return NULL;
}

sen_index *
sen_index_open_with_keys(const char *path, sen_sym *keys)
{
  sen_index *i;
  char buffer[PATH_MAX];
  if (!path || !keys) {
    SEN_LOG(sen_log_warning, "sen_index_open_with_keys: invalid argument");
    return NULL;
  }
  if (strlen(path) > PATH_MAX - 4) {
    SEN_LOG(sen_log_warning, "too long index path (%s)", path);
    return NULL;
  }
  if (!(i = SEN_GMALLOC(sizeof(sen_index)))) { return NULL; }
  i->keys = keys;
  i->foreign_flags = FOREIGN_KEY;
  strcpy(buffer, path);
  strcat(buffer, ".SEN.l");
  if ((i->lexicon = sen_sym_open(buffer))) {
    strcpy(buffer, path);
    strcat(buffer, ".SEN.i");
    if ((i->inv = sen_inv_open(buffer, i->lexicon))) {
      if ((i->lexicon->flags & SEN_INDEX_WITH_VGRAM)) {
        strcpy(buffer, path);
        strcat(buffer, ".SEN.v");
        i->vgram = sen_vgram_open(buffer);
      } else {
        i->vgram = NULL;
      }
      if(!(i->lexicon->flags & SEN_INDEX_WITH_VGRAM) || i->vgram) {
        SEN_LOG(sen_log_notice, "index opened (%p:%s) flags=%x", i, path, i->lexicon->flags);
        return i;
      }
      sen_inv_close(i->inv);
    }
    sen_sym_close(i->lexicon);
  }
  SEN_GFREE(i);
  return NULL;
}

sen_index *
sen_index_create_with_keys_lexicon(const char *path, sen_sym *keys, sen_sym *lexicon,
                                   int initial_n_segments)
{
  sen_index *i;
  if (!keys || !path || !lexicon) {
    SEN_LOG(sen_log_warning, "sen_index_create_with_keys_lexicon: invalid argument");
    return NULL;
  }
  if (initial_n_segments == 0) { initial_n_segments = SENNA_DEFAULT_INITIAL_N_SEGMENTS; }
  if (!(i = SEN_GMALLOC(sizeof(sen_index)))) { return NULL; }
  SEN_LOG(sen_log_notice, "creating '%s' encoding=%s initial_n_segments=%d",
          path, sen_enctostr(lexicon->encoding), initial_n_segments);
  i->keys = keys;
  i->lexicon = lexicon;
  i->foreign_flags = FOREIGN_KEY|FOREIGN_LEXICON;
  i->vgram = NULL;
  if ((i->inv = sen_inv_create(path, i->lexicon, initial_n_segments))) {
    SEN_LOG(sen_log_notice, "index created (%s) flags=%x", path, i->lexicon->flags);
    return i;
  }
  SEN_GFREE(i);
  return NULL;
}

sen_index *
sen_index_open_with_keys_lexicon(const char *path, sen_sym *keys, sen_sym *lexicon)
{
  sen_index *i;
  if (!keys || !path || !lexicon) {
    SEN_LOG(sen_log_warning, "sen_index_open_with_keys_lexicon: invalid argument");
    return NULL;
  }
  if (!(i = SEN_GMALLOC(sizeof(sen_index)))) { return NULL; }
  i->keys = keys;
  i->lexicon = lexicon;
  i->foreign_flags = FOREIGN_KEY|FOREIGN_LEXICON;
  i->vgram = NULL;
  if ((i->inv = sen_inv_open(path, i->lexicon))) {
    SEN_LOG(sen_log_notice, "index opened (%p:%s) flags=%x", i, path, i->lexicon->flags);
    return i;
  }
  SEN_GFREE(i);
  return NULL;
}

sen_rc
sen_index_close(sen_index *i)
{
  if (!i) { return sen_invalid_argument; }
  if (!(i->foreign_flags & FOREIGN_KEY)) { sen_sym_close(i->keys); }
  if (!(i->foreign_flags & FOREIGN_LEXICON)) { sen_sym_close(i->lexicon); }
  sen_inv_close(i->inv);
  if (i->vgram) { sen_vgram_close(i->vgram); }
  SEN_GFREE(i);
  return sen_success;
}

sen_rc
sen_index_info(sen_index *i, int *key_size, int *flags,
               int *initial_n_segments, sen_encoding *encoding,
               unsigned *nrecords_keys, unsigned *file_size_keys,
               unsigned *nrecords_lexicon, unsigned *file_size_lexicon,
               unsigned long long *inv_seg_size, unsigned long long *inv_chunk_size)
{
  sen_rc rc = sen_success;

  if (!i) { return sen_invalid_argument; }
  if (key_size) { *key_size = i->keys->key_size; }
  if (flags) { *flags = i->lexicon->flags & ~SEN_SYM_WITH_SIS; }
  if (initial_n_segments) { *initial_n_segments = sen_inv_initial_n_segments(i->inv); }
  if (encoding) { *encoding = i->lexicon->encoding; }
  if (nrecords_keys || file_size_keys) {
    if ((rc = sen_sym_info(i->keys, NULL, NULL, NULL, nrecords_keys, file_size_keys))) { return rc; }
  }
  if (nrecords_lexicon || file_size_lexicon) {
    if ((rc = sen_sym_info(i->lexicon, NULL, NULL, NULL, nrecords_lexicon, file_size_lexicon))) { return rc; }
  }
  if (inv_seg_size || inv_chunk_size) {
    uint64_t seg_size, chunk_size;

    rc = sen_inv_info(i->inv, &seg_size, &chunk_size);

    if (inv_seg_size) {
      *inv_seg_size = seg_size;
    }

    if (inv_chunk_size) {
      *inv_chunk_size = chunk_size;
    }

    if (rc != sen_success) {
      return rc;
    }
  }
  return sen_success;
}

sen_rc
sen_index_lock(sen_index *i, int timeout)
{
  if (!i) { return sen_invalid_argument; }
  return sen_sym_lock(i->keys, timeout);
}

sen_rc
sen_index_unlock(sen_index *i)
{
  if (!i) { return sen_invalid_argument; }
  return sen_sym_unlock(i->keys);
}

sen_rc
sen_index_clear_lock(sen_index *i)
{
  if (!i) { return sen_invalid_argument; }
  return sen_sym_clear_lock(i->keys);
}

int
sen_index_path(sen_index *i, char *pathbuf, int bufsize)
{
  const char *invpath;
  int pathsize;
  if (!i) {
    SEN_LOG(sen_log_warning, "sen_index_path: invalid argument");
    return sen_invalid_argument;
  }
  invpath = sen_io_path(i->lexicon->io);
  pathsize = strlen(invpath) - 5;
  if (bufsize >= pathsize && pathbuf) {
    memcpy(pathbuf, invpath, pathsize - 1);
    pathbuf[pathsize - 1] = '\0';
  }
  return pathsize;
}

sen_rc
sen_index_upd(sen_index *i, const void *key,
              const char *oldvalue, unsigned int oldvalue_len,
              const char *newvalue, unsigned int newvalue_len)
{
  sen_id rid;
  sen_rc rc = sen_invalid_argument;
  if (!i || !key) {
    SEN_LOG(sen_log_warning, "sen_index_upd: invalid argument");
    return sen_invalid_argument;
  }
  if ((rc = sen_index_lock(i, -1))) {
    SEN_LOG(sen_log_crit, "sen_index_upd: index lock failed");
    return rc;
  }
  if (oldvalue && *oldvalue) {
    if (!(rid = sen_sym_at(i->keys, key))) {
      SEN_LOG(sen_log_error, "del : (%x) (invalid key)", key);
      goto exit;
    }
  } else if (newvalue && *newvalue) {
    if (!(rid = sen_sym_get(i->keys, key))) { goto exit; }
  } else {
    goto exit;
  }
  rc = sen_inv_upd(i->inv, rid, i->vgram, oldvalue, oldvalue_len, newvalue, newvalue_len);
exit :
  sen_index_unlock(i);
  return rc;
}

#define DELETE_FLAG 1

sen_rc
sen_index_del(sen_index *i, const void *key)
{
  sen_id rid;
  if (!i || !key) { SEN_LOG(sen_log_warning, "sen_index_del: invalid argument"); return sen_invalid_argument; }
  rid = sen_sym_at(i->keys, key);
  if (!rid) { return sen_invalid_argument; }
  return sen_sym_pocket_set(i->keys, rid, DELETE_FLAG);
}

sen_rc
sen_index_update(sen_index *i, const void *key, unsigned int section,
                 sen_values *oldvalues, sen_values *newvalues)
{
  sen_id rid;
  sen_rc rc = sen_invalid_argument;
  if (!i || !key) {
    SEN_LOG(sen_log_warning, "sen_index_update: invalid argument");
    return rc;
  }
  if ((rc = sen_index_lock(i, -1))) {
    SEN_LOG(sen_log_crit, "sen_index_update: index lock failed");
    return rc;
  }
  if (newvalues) {
    if (!(rid = sen_sym_get(i->keys, key))) { goto exit; }
  } else {
    if (!(rid = sen_sym_at(i->keys, key))) { goto exit; }
  }
  rc = sen_inv_update(i->inv, rid, i->vgram, section, oldvalues, newvalues);
exit :
  sen_index_unlock(i);
  return rc;
}

/* select */

sen_rc
sen_index_similar_search(sen_index *i, const char *string,
                         unsigned int string_len, sen_records *r,
                         sen_sel_operator op, sen_select_optarg *optarg)
{
  sen_rc rc;
  if (!i || !string || !r || !optarg) { return sen_invalid_argument; }
  r->keys = i->keys;
  optarg->max_size = sen_sym_size(i->keys) * sizeof(int);
  rc = sen_inv_similar_search(i->inv, string, string_len, r, op, optarg);
  sen_records_cursor_clear(r);
  return rc;
}

#define TERM_EXTRACT_EACH_POST 0
#define TERM_EXTRACT_EACH_TERM 1

sen_rc
sen_index_term_extract(sen_index *i, const char *string,
                       unsigned int string_len, sen_records *r,
                       sen_sel_operator op, sen_select_optarg *optarg)
{
  sen_rc rc;
  r->keys = i->keys;
  rc = sen_inv_term_extract(i->inv, string, string_len, r, op, optarg);
  sen_records_cursor_clear(r);
  if (!rc && optarg->max_interval == TERM_EXTRACT_EACH_POST) {
    sen_sort_optarg opt;
    opt.mode = sen_sort_ascending;
    opt.compar = NULL;
    opt.compar_arg = (void *)(intptr_t)r->key_size;
    sen_records_sort(r, 10000, &opt); /* todo : why 10000? */
  }
  return rc;
}

sen_rc
sen_index_select(sen_index *i, const char *string, unsigned int string_len,
                 sen_records *r, sen_sel_operator op, sen_select_optarg *optarg)
{
  sen_rc rc;
  if (!r || !i) { return sen_invalid_argument; }
  r->keys = i->keys;
  if (optarg) { optarg->max_size = sen_sym_size(i->keys) * sizeof(int); }
  rc = sen_inv_select(i->inv, string, string_len, r, op, optarg);
  sen_records_cursor_clear(r);
  return rc;
}

sen_records *
sen_index_sel(sen_index *i, const char *string, unsigned int string_len)
{
  sen_records *r;
  if (!i) { return NULL; }
  r = sen_inv_sel(i->inv, string, string_len);
  if (r) { r->keys = i->keys; }
  return r;
}

/* sen_records : sets with keys or/and cursor */

int
sen_rec_unit_size(sen_rec_unit unit, int rec_size)
{
  switch (unit) {
  case sen_rec_document :
    return sizeof(sen_id);
  case sen_rec_section :
    return sizeof(sen_id) + sizeof(int);
  case sen_rec_position :
    return sizeof(sen_id) + sizeof(int) + sizeof(int);
  default :
    return rec_size;
  }
}

sen_records *
sen_records_open(sen_rec_unit record_unit,
                 sen_rec_unit subrec_unit, unsigned int max_n_subrecs)
{
  sen_records *r;
  if (!(r = SEN_GMALLOC(sizeof(sen_records)))) { return NULL; }
  if (sen_rset_init(&sen_gctx, r, record_unit, 0, subrec_unit, 0, max_n_subrecs)) {
    SEN_GFREE(r);
    return NULL;
  }
  return r;
}

void
sen_records_cursor_clear(sen_records *r)
{
  sen_ctx *ctx = r->ctx;
  if (r->sorted) {
    SEN_FREE(r->sorted);
    r->sorted = NULL;
  }
  if (r->cursor) {
    sen_set_cursor_close(r->cursor);
    r->cursor = NULL;
  }
  r->curr_rec = NULL;
}

sen_rc
sen_records_close(sen_records *r)
{
  sen_ctx *ctx;
  if (!r) { return sen_invalid_argument; }
  ctx = r->ctx;
  if (r->curr_rec) {
    sen_id *rid;
    sen_rset_recinfo *ri;
    if (!sen_set_element_info(r, r->curr_rec, (void **)&rid, (void **)&ri)) {
      SEN_LOG(sen_log_debug, "curr_rec: %d:%d", *rid, ri->score);
      // sen_log("sen_record_next=%d,%d", *rid, *(_sen_sym_key(r->keys, *rid)));
    }
  }
  sen_records_cursor_clear(r);
  return sen_set_close(r);
}

sen_rc
sen_records_reopen(sen_records *r, sen_rec_unit record_unit,
                   sen_rec_unit subrec_unit, unsigned int max_n_subrecs)
{
  sen_ctx *ctx;
  int record_size = sen_rec_unit_size(record_unit, 0);
  int subrec_size = sen_rec_unit_size(subrec_unit, 0);
  if (!r || record_size < 0 || (max_n_subrecs && subrec_size <= record_size)) {
    return sen_invalid_argument;
  }
  ctx = r->ctx;
  sen_records_cursor_clear(r);
  sen_set_fin(r);
  return sen_rset_init(ctx, r, record_unit, 0, subrec_unit, 0, max_n_subrecs);
}

sen_rc
sen_records_sort(sen_records *r, int limit, sen_sort_optarg *optarg)
{
  if (!r) { return sen_invalid_argument; }
  sen_records_cursor_clear(r);
  if (optarg) {
    r->sorted = sen_set_sort(r, limit, optarg);
  } else {
    sen_set_sort_optarg arg;
    arg.mode = sen_sort_descending;
    arg.compar = NULL;
    arg.compar_arg = (void *)(intptr_t)r->key_size;
    r->sorted = sen_set_sort(r, limit, &arg);
  }
  return r->sorted ? sen_success : sen_internal_error;
}

sen_rc
sen_records_rewind(sen_records *r)
{
  if (!r) { return sen_invalid_argument; }
  if (r->sorted) {
    r->curr_rec = NULL;
    return sen_success;
  } else {
    sen_records_cursor_clear(r);
    r->cursor = sen_set_cursor_open(r);
    return r->cursor ? sen_success : sen_internal_error;
  }
}

int
sen_records_next(sen_records *r, void *keybuf, int bufsize, int *score)
{
  if (!r) { return 0; }
  if (r->sorted) {
    if (r->curr_rec) {
      if (r->curr_rec - r->sorted >= r->limit - 1) { return 0; }
      r->curr_rec++;
    } else {
      if (r->limit > 0) { r->curr_rec = r->sorted; }
    }
  } else {
    if (!r->cursor && sen_records_rewind(r)) { return 0; }
    r->curr_rec = sen_set_cursor_next(r->cursor, NULL, NULL);
  }
  if (r->curr_rec) {
    sen_id *rid;
    sen_rset_recinfo *ri;
    if (!sen_set_element_info(r, r->curr_rec,  (void **)&rid, (void **)&ri)) {
      if (score) { *score = ri->score; }
      if (r->record_unit != sen_rec_userdef) {
        return r->keys ? sen_sym_key(r->keys, *rid, keybuf, bufsize) : r->record_size;
      } else {
        if (r->record_size <= bufsize) { memcpy(keybuf, rid, r->record_size); }
        return r->record_size;
      }
    }
  }
  return 0;
}

int
sen_records_find(sen_records *r, const void *key)
{
  sen_id rid;
  sen_rset_recinfo *ri;
  if (!r || !r->keys || r->record_unit != sen_rec_document) { return 0; }
  if (!(rid = sen_sym_at(r->keys, key))) { return 0; }
  if (!sen_set_at(r, &rid, (void **) &ri)) { return 0; }
  return ri->score;
}

int
sen_records_nhits(sen_records *r)
{
  if (!r) { return 0; }
  return r->n_entries;
}

int
sen_records_curr_score(sen_records *r)
{
  sen_rset_recinfo *ri;
  if (!r || !r->curr_rec) { return 0; }
  if (sen_set_element_info(r, r->curr_rec, NULL, (void **)&ri)) { return 0; }
  return ri->score;
}

int
sen_records_curr_key(sen_records *r, void *keybuf, int bufsize)
{
  sen_id *rid;
  if (!r || !r->curr_rec) { return 0; }
  if (sen_set_element_info(r, r->curr_rec, (void **)&rid, NULL)) {
    return 0;
  }
  if (r->record_unit != sen_rec_userdef) {
    return sen_sym_key(r->keys, *rid, keybuf, bufsize);
  } else {
    if (r->record_size <= bufsize) { memcpy(keybuf, rid, r->record_size); }
    return r->record_size;
  }
}

const sen_recordh *
sen_records_curr_rec(sen_records *r)
{
  if (!r) { return NULL; }
  return r->curr_rec;
}

sen_records *
sen_records_union(sen_records *a, sen_records *b)
{
  if (!a || !b) { return NULL; }
  if (a->keys != b->keys) { return NULL; }
  if (!sen_set_union(a, b)) { return NULL; }
  sen_records_close(b);
  sen_records_cursor_clear(a);
  return a;
}

sen_records *
sen_records_subtract(sen_records *a, sen_records *b)
{
  if (!a || !b) { return NULL; }
  if (a->keys != b->keys) { return NULL; }
  if (!sen_set_subtract(a, b)) { return NULL; }
  sen_records_close(b);
  sen_records_cursor_clear(a);
  return a;
}

sen_records *
sen_records_intersect(sen_records *a, sen_records *b)
{
  if (!a || !b) { return NULL; }
  if (a->keys != b->keys) { return NULL; }
  if (a->n_entries > b->n_entries) {
    sen_set c;
    memcpy(&c, a, sizeof(sen_records));
    memcpy(a, b, sizeof(sen_records));
    memcpy(b, &c, sizeof(sen_records));
  }
  if (!sen_set_intersect(a, b)) { return NULL; }
  sen_records_close(b);
  sen_records_cursor_clear(a);
  return a;
}

int
sen_records_difference(sen_records *a, sen_records *b)
{
  int count;
  if (!a || !b) { return -1; }
  if (a->keys != b->keys) { return -1; }
  if ((count = sen_set_difference(a, b)) >= 0) {
    sen_records_cursor_clear(a);
    sen_records_cursor_clear(b);
  }
  return count;
}

#define SUBRECS_CMP(a,b,dir) (((a) - (b))*(dir) > 0)
#define SUBRECS_NTH(subrecs,size,n) ((int *)(subrecs + n * (size + SEN_RSET_SCORE_SIZE)))
#define SUBRECS_COPY(subrecs,size,n,src) \
  (memcpy(subrecs + n * (size + SEN_RSET_SCORE_SIZE), src, size + SEN_RSET_SCORE_SIZE))

sen_rc
sen_records_group(sen_records *r, int limit, sen_group_optarg *optarg)
{
  sen_ctx *ctx;
  sen_set *g;
  if (!(g = sen_rset_group(r, limit, optarg))) {
    return sen_internal_error;
  }
  ctx = r->ctx;
  sen_records_cursor_clear(r);
  sen_set_fin(r);
  memcpy(r, g, sizeof(sen_set));
  SEN_FREE(g);
  return sen_success;
}

const sen_recordh *
sen_records_at(sen_records *r, const void *key, unsigned section, unsigned pos,
               int *score, int *n_subrecs)
{
  sen_rset_recinfo *ri;
  const sen_recordh *res;
  if (!r || !r->keys) { return NULL; }
  if (r->record_unit == sen_rec_userdef) {
    res = r->curr_rec = sen_set_at(r, key, (void **) &ri);
  } else {
    sen_rset_posinfo pi;
    if (!(pi.rid = sen_sym_at(r->keys, key))) { return NULL; }
    pi.sid = section;
    pi.pos = pos;
    res = r->curr_rec = sen_set_at(r, &pi, (void **) &ri);
  }
  if (res) {
    if (score) { *score = ri->score; }
    if (n_subrecs) { *n_subrecs = ri->n_subrecs; }
  }
  return res;
}

sen_rc
sen_record_info(sen_records *r, const sen_recordh *rh,
                void *keybuf, int bufsize, int *keysize,
                int *section, int *pos, int *score, int *n_subrecs)
{
  sen_rc rc;
  sen_rset_posinfo *pi;
  sen_rset_recinfo *ri;
  if (!r || !rh) { return sen_invalid_argument; }
  rc = sen_set_element_info(r, rh, (void **)&pi, (void **)&ri);
  if (rc) { return rc; }
  switch (r->record_unit) {
  case sen_rec_document :
    if ((keybuf && bufsize) || keysize) {
      int len = sen_sym_key(r->keys, pi->rid, keybuf, bufsize);
      if (keysize) { *keysize = len; }
    }
    if (section) { *section = 0; }
    if (pos) { *pos = 0; }
    break;
  case sen_rec_section :
    if ((keybuf && bufsize) || keysize) {
      int len = sen_sym_key(r->keys, pi->rid, keybuf, bufsize);
      if (keysize) { *keysize = len; }
    }
    if (section) { *section = pi->sid; }
    if (pos) { *pos = 0; }
    break;
  case sen_rec_position :
    if ((keybuf && bufsize) || keysize) {
      int len = sen_sym_key(r->keys, pi->rid, keybuf, bufsize);
      if (keysize) { *keysize = len; }
    }
    if (section) { *section = pi->sid; }
    if (pos) { *pos = pi->pos; }
    break;
  case sen_rec_userdef :
    if ((keybuf && bufsize) || keysize) {
      unsigned int len = r->record_size;
      if (!len) { len = (unsigned int)strlen((const char *)pi) + 1; }
      if ((unsigned int)bufsize >= len) { memcpy(keybuf, pi, len); }
      if (keysize) { *keysize = len; }
    }
    if (section) { *section = 0; }
    if (pos) { *pos = 0; }
    break;
  default :
    return sen_invalid_format;
    break;
  }
  if (score) { *score = ri->score; }
  if (n_subrecs) { *n_subrecs = ri->n_subrecs; }
  return sen_success;
}

sen_rc
sen_record_subrec_info(sen_records *r, const sen_recordh *rh, int index,
                       void *keybuf, int bufsize, int *keysize,
                       int *section, int *pos, int *score)
{
  sen_rc rc;
  sen_id rid;
  void *subrec;
  if ((rc = sen_rset_subrec_info(r, rh, index, &rid, section, pos, score, &subrec))) {
    return rc;
  }
  if ((keybuf && bufsize) || keysize) {
    if (rid) {
      int len = sen_sym_key(r->keys, rid, keybuf, bufsize);
      if (keysize) { *keysize = len; }
    } else if (r->record_unit == sen_rec_userdef && r->subrec_unit == sen_rec_userdef) {
      int len = (int) r->subrec_size;
      if (!len) { len = (unsigned int)strlen((char *)subrec) + 1; }
      if ((unsigned int)bufsize >= len) { memcpy(keybuf, subrec, len); }
      if (keysize) { *keysize = len; }
    }
  }
  return sen_success;
}

#ifdef USE_QUERY_ABORT
void
sen_index_set_abort_callback(sen_index *i, int (*cb)(void*), void *arg)
{
  sen_inv_set_abort_callback(i->inv, cb, arg);
}
#endif /* USE_QUERY_ABORT */

/* sen_records_heap class */

struct _sen_records_heap {
  int n_entries;
  int n_bins;
  sen_records **bins;
  int limit;
  int curr;
  int dir;
  int (*compar)(sen_records *, sen_recordh *, sen_records *, sen_recordh *, void *);
  void *compar_arg;
};

inline static int
records_heap_cmp(sen_records_heap *h, sen_records *r1, sen_records *r2)
{
  sen_recordh *rh1 = (sen_recordh *)sen_records_curr_rec(r1);
  sen_recordh *rh2 = (sen_recordh *)sen_records_curr_rec(r2);
  if (!h->compar) {
    int off1, off2;
    if (h->compar_arg == (void *)-1) {
      off1 = (r1->key_size) / sizeof(int32_t);
      off2 = (r2->key_size) / sizeof(int32_t);
    } else {
      off1 = off2 = (int)(intptr_t)h->compar_arg;
    }
    return (((int32_t *)(rh2))[off2] - ((int32_t *)(rh1))[off1]) * h->dir > 0;
  }
  return h->compar(r1, rh1, r2, rh2, h->compar_arg) * h->dir > 0;
}

sen_records_heap *
sen_records_heap_open(int size, int limit, sen_sort_optarg *optarg)
{
  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
  sen_records_heap *h = SEN_MALLOC(sizeof(sen_records_heap));
  if (!h) { return NULL; }
  h->bins = SEN_MALLOC(sizeof(sen_records *) * size);
  if (!h->bins) {
    SEN_FREE(h);
    return NULL;
  }
  h->n_entries = 0;
  h->n_bins = size;
  h->limit = limit;
  h->curr = 0;
  if (optarg) {
    h->dir = (optarg->mode == sen_sort_ascending) ? 1 : -1;
    h->compar = optarg->compar;
    h->compar_arg = optarg->compar_arg;
  } else {
    h->dir = -1;
    h->compar = NULL;
    h->compar_arg = (void *) -1;
  }
  return h;
}

sen_rc
sen_records_heap_add(sen_records_heap *h, sen_records *r)
{
  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
  if (h->n_entries >= h->n_bins) {
    int size = h->n_bins * 2;
    sen_records **bins = SEN_REALLOC(h->bins, sizeof(sen_records *) * size);
    // sen_log("expanded sen_records_heap to %d,%p", size, bins);
    if (!bins) { return sen_memory_exhausted; }
    h->n_bins = size;
    h->bins = bins;
  }
  if (!sen_records_next(r, NULL, 0, NULL)) {
    sen_records_close(r);
    return sen_internal_error;
  }
  {
    int n, n2;
    sen_records *r2;
    n = h->n_entries++;
    while (n) {
      n2 = (n - 1) >> 1;
      r2 = h->bins[n2];
      if (records_heap_cmp(h, r, r2)) { break; }
      h->bins[n] = r2;
      n = n2;
    }
    h->bins[n] = r;
  }
  return sen_success;
}

int
sen_records_heap_next(sen_records_heap *h)
{
  if (!h || !h->n_entries) { return 0; }
  {
    sen_records *r = h->bins[0];
    if (!sen_records_next(r, NULL, 0, NULL)) {
      sen_records_close(r);
      r = h->bins[0] = h->bins[--h->n_entries];
    }
    {
      int n = 0, m = h->n_entries;
      if (m > 1) {
        for (;;) {
          int n1 = n * 2 + 1;
          int n2 = n1 + 1;
          sen_records *r1 = n1 < m ? h->bins[n1] : NULL;
          sen_records *r2 = n2 < m ? h->bins[n2] : NULL;
          if (r1 && records_heap_cmp(h, r, r1)) {
            if (r2 && records_heap_cmp(h, r, r2) && records_heap_cmp(h, r1, r2)) {
              h->bins[n] = r2;
              n = n2;
            } else {
              h->bins[n] = r1;
              n = n1;
            }
          } else {
            if (r2 && records_heap_cmp(h, r, r2)) {
              h->bins[n] = r2;
              n = n2;
            } else {
              h->bins[n] = r;
              break;
            }
          }
        }
      }
      h->curr++;
      return m;
    }
  }
}

sen_records *
sen_records_heap_head(sen_records_heap *h)
{
  return h->n_entries ? h->bins[0] : NULL;
}

sen_rc
sen_records_heap_close(sen_records_heap *h)
{
  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
  int i;
  if (!h) { return sen_invalid_argument; }
  for (i = h->n_entries; i--;) { sen_records_close(h->bins[i]); }
  SEN_FREE(h->bins);
  SEN_FREE(h);
  return sen_success;
}

sen_sym *
sen_index_keys(sen_index *i) {
  return (sen_sym *)i->keys;
}

sen_sym *
sen_index_lexicon(sen_index *i) {
  return (sen_sym *)i->lexicon;
}

sen_inv *
sen_index_inv(sen_index *i) {
  return (sen_inv *)i->inv;
}

#ifdef USE_SEN_INDEX2
/******** sen_index2 ********/

/**
 * sen_index2 and sen_records2 are API set
 * which resemble Senna basic API, but are based on DB API.
 *
 **/

#include "pat.h"
#include "ii.h"

struct _sen_index2 {
  sen_pat *keys;
  sen_pat *lexicon;
  sen_ii *inv;
};

struct _sen_records2 {
  sen_hash hash;
  sen_array *sorted;
  sen_obj *cursor;
  sen_pat *keys;
  sen_id curr_rec;
};

typedef struct _sen_index2 sen_index2;
typedef struct _sen_records2 sen_records2;

sen_records2 *sen_records2_open(sen_rec_unit record_unit,
                                sen_rec_unit subrec_unit, unsigned int max_n_subrecs);

sen_rc sen_records2_close(sen_records2 *r);

/*
static sen_obj *sen_index2_space;

sen_rc
sen_index2_init(void)
{
  sen_index2_space = NULL;
  return sen_success;
}

void
sen_index2_fin(void)
{
  if (sen_index2_space) { sen_space_close(&sen_gctx, (sen_space *)sen_index2_space); }
}

static sen_obj *
index_space(void)
{
  if (!sen_index2_space) {
    sen_index2_space = sen_space_create(&sen_gctx, NULL, sen_gctx.encoding);
  }
  return sen_index2_space;
}
*/

inline static uint32_t
build_flags2(int flags)
{
  uint32_t f = SEN_OBJ_PERSISTENT;
  switch (flags & SEN_INDEX_TOKENIZER_MASK) {
  case SEN_INDEX_MORPH_ANALYSE :
    f = SEN_OBJ_TOKEN_MECAB|SEN_SYM_WITH_SIS;
    break;
  case SEN_INDEX_NGRAM :
    f = SEN_OBJ_TOKEN_NGRAM;
    break;
  case SEN_INDEX_DELIMITED :
    f = SEN_OBJ_TOKEN_DELIMITED;
  default :
    f = SEN_OBJ_TOKEN_MECAB|SEN_SYM_WITH_SIS;
  }
  if (flags & SEN_INDEX_ENABLE_SUFFIX_SEARCH) {
    f |= SEN_OBJ_KEY_WITH_SIS;
  } else if (flags & SEN_INDEX_DISABLE_SUFFIX_SEARCH) {
    f &= ~SEN_OBJ_KEY_WITH_SIS;
  }
  if (f & SEN_INDEX_NGRAM) {
    if (flags & SEN_INDEX_SPLIT_ALPHA) {
      f |= SEN_OBJ_KEY_SPLIT_ALPHA;
    }
    if (flags & SEN_INDEX_SPLIT_DIGIT) {
      f |= SEN_OBJ_KEY_SPLIT_DIGIT;
    }
    if (flags & SEN_INDEX_SPLIT_SYMBOL) {
      f |= SEN_OBJ_KEY_SPLIT_SYMBOL;
    }
  }
  return f;
}

sen_index2 *
sen_index2_create(const char *path, int key_size,
                 int flags, int initial_n_segments, sen_encoding encoding)
{
  sen_index2 *i;
  char buffer[PATH_MAX];
  if (!path) {
    SEN_LOG(sen_log_warning, "sen_index2_create: invalid argument");
    return NULL;
  }
  if (initial_n_segments == 0) { initial_n_segments = SENNA_DEFAULT_INITIAL_N_SEGMENTS; }
  if (encoding == sen_enc_default) { encoding = sen_gctx.encoding; }
  if (strlen(path) > PATH_MAX - 4) {
    SEN_LOG(sen_log_warning, "sen_index2_create: too long index path (%s)", path);
    return NULL;
  }
  if (!(i = SEN_GMALLOC(sizeof(sen_index2)))) { return NULL; }
  SEN_LOG(sen_log_notice, "creating '%s' encoding=%s initial_n_segments=%d",
          path, sen_enctostr(encoding), initial_n_segments);
  strcpy(buffer, path);
  strcat(buffer, ".SEN");
  if ((i->keys = sen_pat_create(&sen_gctx,
                                buffer,
                                key_size ? key_size : SEN_PAT_MAX_KEY_SIZE,
                                0,
                                key_size
                                ? SEN_OBJ_PERSISTENT|SEN_OBJ_KEY_UINT
                                : SEN_OBJ_PERSISTENT|SEN_OBJ_KEY_VAR_SIZE,
                                sen_enc_none))) {
    strcpy(buffer, path);
    strcat(buffer, ".SEN.l");
    if ((i->lexicon = sen_pat_create(&sen_gctx,
                                     buffer,
                                     SEN_PAT_MAX_KEY_SIZE,
                                     0,
                                     build_flags2(flags), encoding))) {
      strcpy(buffer, path);
      strcat(buffer, ".SEN.i");
      if ((i->inv = sen_ii_create(&sen_gctx,
                                  buffer,
                                  (sen_obj *)i->lexicon,
                                  SEN_OBJ_COLUMN_INDEX|SEN_OBJ_PERSISTENT))) {
        SEN_LOG(sen_log_notice, "index created (%s) flags=%x", path, i->lexicon->flags);
        return i;
      }
      sen_pat_close(&sen_gctx, (sen_pat *)i->lexicon);
    }
    sen_pat_close(&sen_gctx, (sen_pat *)i->keys);
  }
  SEN_GFREE(i);
  return NULL;
}

sen_index2 *
sen_index2_open(const char *path)
{
  sen_index2 *i;
  char buffer[PATH_MAX];
  if (!path) { SEN_LOG(sen_log_warning, "sen_index2_open: invalid argument"); return NULL; }
  if (strlen(path) > PATH_MAX - 4) {
    SEN_LOG(sen_log_warning, "sen_index2_open: too long index path (%s)", path);
    return NULL;
  }
  if (!(i = SEN_GMALLOC(sizeof(sen_index2)))) { return NULL; }
  strcpy(buffer, path);
  strcat(buffer, ".SEN");
  if ((i->keys = sen_pat_open(&sen_gctx, buffer))) {
    strcpy(buffer, path);
    strcat(buffer, ".SEN.l");
    if ((i->lexicon = sen_pat_open(&sen_gctx, buffer))) {
      strcpy(buffer, path);
      strcat(buffer, ".SEN.i");
      if ((i->inv = sen_ii_open(&sen_gctx, buffer, (sen_obj *)i->lexicon))) {
        SEN_LOG(sen_log_notice, "index opened (%p:%s) flags=%x", i, path, i->lexicon->flags);
        return i;
      }
      sen_pat_close(&sen_gctx, i->lexicon);
    }
    sen_pat_close(&sen_gctx, i->keys);
  }
  SEN_GFREE(i);
  return NULL;
}

sen_rc
sen_index2_close(sen_index2 *i)
{
  if (!i) { return sen_invalid_argument; }
  sen_pat_close(&sen_gctx, i->keys);
  sen_pat_close(&sen_gctx, i->lexicon);
  sen_ii_close(&sen_gctx, i->inv);
  SEN_GFREE(i);
  return sen_success;
}

sen_rc
sen_index2_lock(sen_index2 *i, int timeout)
{
  if (!i) { return sen_invalid_argument; }
  return sen_io_lock(&sen_gctx, i->keys->io, timeout);
}

sen_rc
sen_index2_unlock(sen_index2 *i)
{
  if (!i) { return sen_invalid_argument; }
  sen_io_unlock(i->keys->io);
  return sen_success;
}

sen_rc
sen_index2_clear_lock(sen_index2 *i)
{
  if (!i) { return sen_invalid_argument; }
  sen_io_clear_lock(i->keys->io);
  return sen_success;
}

sen_rc
sen_index2_upd(sen_index2 *i, const void *key,
              const char *oldvalue, unsigned int oldvalue_len,
              const char *newvalue, unsigned int newvalue_len)
{
  sen_id rid;
  sen_rc rc = sen_invalid_argument;
  uint32_t key_size;
  if (!i || !key) {
    SEN_LOG(sen_log_warning, "sen_index2_upd: invalid argument");
    return sen_invalid_argument;
  }
  key_size = (i->keys->flags & SEN_OBJ_KEY_VAR_SIZE) ? strlen(key) : i->keys->key_size;
  if ((rc = sen_index2_lock(i, -1))) {
    SEN_LOG(sen_log_crit, "sen_index2_upd: index lock failed");
    return rc;
  }
  if (oldvalue && *oldvalue) {
    if (!(rid = sen_pat_at(&sen_gctx, i->keys, key, key_size, NULL))) {
      SEN_LOG(sen_log_error, "del : (%x) (invalid key)", key);
      goto exit;
    }
  } else if (newvalue && *newvalue) {
    sen_search_flags f = SEN_TABLE_ADD;
    if (!(rid = sen_pat_get(&sen_gctx, i->keys, key, key_size, NULL, &f))) { goto exit; }
  } else {
    goto exit;
  }
  rc = sen_ii_upd(&sen_gctx, i->inv, rid, NULL,
                  oldvalue, oldvalue_len, newvalue, newvalue_len);
exit :
  sen_index2_unlock(i);
  return rc;
}

sen_records2 *
sen_index2_sel(sen_index2 *i, const char *string, unsigned int string_len)
{
  sen_records2 *r;
  if (!i) { return NULL; }
  r = (sen_records2 *)sen_records2_open(sen_rec_document, sen_rec_none, 0);
  r->keys = i->keys;
  /* todo : r->obj.domain must be used */
  sen_ii_sel(&sen_gctx, i->inv, string, string_len, &r->hash);
  if (*r->hash.n_entries) {
    sen_records2_close((sen_records2 *)r);
    r = NULL;
  }
  return (sen_records2 *)r;
}

sen_index2 *
sen_index2_open_with_keys(const char *path, sen_sym *keys)
{
  // fixme
  return NULL;
}

sen_index2 *
sen_index2_open_with_keys_lexicon(const char *path, sen_sym *keys, sen_sym *lexicon)
{
  // fixme
  return NULL;
}

sen_rc
sen_index2_info(sen_index2 *i, int *key_size, int *flags,
               int *initial_n_segments, sen_encoding *encoding,
               unsigned *nrecords_keys, unsigned *file_size_keys,
               unsigned *nrecords_lexicon, unsigned *file_size_lexicon,
               unsigned long long *inv_seg_size, unsigned long long *inv_chunk_size)
{
  sen_rc rc = sen_success;
  // fixme
  return rc;
}

/* sen_records2 : wrapper of sen_hash */

sen_rc sen_rhash_init(sen_ctx *ctx, sen_hash *hash,
                      sen_rec_unit record_unit, int record_size,
                      sen_rec_unit subrec_unit, int subrec_size,
                      unsigned int max_n_subrecs);

sen_rc sen_rhash_fin(sen_ctx *ctx, sen_hash *hash);

sen_records2 *
sen_records2_open(sen_rec_unit record_unit,
                 sen_rec_unit subrec_unit, unsigned int max_n_subrecs)
{
  sen_records2 *r;
  if (!(r = SEN_GMALLOC(sizeof(sen_records2)))) { return NULL; }
  if (sen_rhash_init(&sen_gctx, &r->hash, record_unit, 0, subrec_unit, 0, max_n_subrecs)) {
    SEN_GFREE(r);
    return NULL;
  }
  r->sorted = NULL;
  r->cursor = NULL;
  r->keys = NULL;
  r->curr_rec = SEN_ID_NIL;
  return (sen_records2 *)r;
}

void
sen_records2_cursor_clear(sen_records2 *r2)
{
  sen_ctx *ctx;
  if (!r2) { return; }
  ctx = r2->hash.ctx;
  if (r2->sorted) {
    sen_array_close(ctx, r2->sorted);
    r2->sorted = NULL;
  }
  if (r2->cursor) {
    sen_obj_close(ctx, r2->cursor);
    r2->cursor = NULL;
  }
  r2->curr_rec = SEN_ID_NIL;
}

sen_rc
sen_records2_close(sen_records2 *r2)
{
  sen_rc rc;
  sen_ctx *ctx;
  if (!r2) { return sen_invalid_argument; }
  ctx = r2->hash.ctx;
  sen_records2_cursor_clear(r2);
  rc = sen_rhash_fin(ctx, &r2->hash);
  SEN_FREE(r2);
  return rc;
}

sen_rc
sen_records2_reopen(sen_records2 *r2, sen_rec_unit record_unit,
                   sen_rec_unit subrec_unit, unsigned int max_n_subrecs)
{
  sen_ctx *ctx;
  int record_size = sen_rec_unit_size(record_unit, 0);
  int subrec_size = sen_rec_unit_size(subrec_unit, 0);
  if (!r2 || record_size < 0 || (max_n_subrecs && subrec_size <= record_size)) {
    return sen_invalid_argument;
  }
  ctx = r2->hash.ctx;
  sen_records2_cursor_clear(r2);
  sen_rhash_fin(ctx, &r2->hash);
  return sen_rhash_init(ctx, &r2->hash, record_unit, 0, subrec_unit, 0, max_n_subrecs);
}

sen_rc
sen_records2_sort(sen_records2 *r2, int limit, sen_sort_optarg *optarg)
{
  sen_hash *h;
  sen_ctx *ctx;
  if (!r2) { return sen_invalid_argument; }
  sen_records2_cursor_clear(r2);
  h = &r2->hash;
  ctx = h->ctx;
  if (!(r2->sorted = sen_array_create(ctx, NULL, sizeof(sen_id), SEN_ARRAY_TINY))) {
    return ctx->rc;
  }
  {
    sen_table_sort_optarg arg;
    arg.flags = SEN_TABLE_SORT_BY_VALUE;
    arg.compar = NULL;
    arg.compar_arg = (void *)(intptr_t)h->key_size;
    arg.proc = NULL;
    arg.offset = 0;
    if (optarg) {
      if (optarg->mode == sen_sort_ascending) {
        arg.flags |= SEN_TABLE_SORT_ASC;
      }
    }
    sen_hash_sort(ctx, &r2->hash, limit, r2->sorted, &arg);
    return sen_success;
  }
}

sen_rc
sen_records2_rewind(sen_records2 *r2)
{
  sen_ctx *ctx;
  if (!r2) { return sen_invalid_argument; }
  ctx = r2->hash.ctx;
  r2->curr_rec = SEN_ID_NIL;
  if (r2->cursor) {
    sen_table_cursor_close(ctx, r2->cursor);
    r2->cursor = NULL;
  }
  if (r2->sorted) {
    r2->cursor = sen_table_cursor_open(ctx, (sen_obj *)r2->sorted, NULL, 0, NULL, 0, 0);
  } else {
    r2->cursor = sen_table_cursor_open(ctx, (sen_obj *)&r2->hash, NULL, 0, NULL, 0, 0);
  }
  return r2->cursor ? sen_success : sen_internal_error;
}

int
sen_records2_next(sen_records2 *r2, void *keybuf, int bufsize, int *score)
{
  sen_id id;
  sen_ctx *ctx;
  if (!r2) { return 0; }
  ctx = r2->hash.ctx;
  if (!r2->cursor && sen_records2_rewind(r2)) { return 0; }
  if (!(id = sen_table_cursor_next(ctx, r2->cursor))) { return 0; }
  if (r2->sorted) {
    void *array_value;
    if (!sen_table_cursor_get_value(ctx, r2->cursor, &array_value)) { return 0; }
    id = *((sen_id *)array_value);
  }
  if ((r2->curr_rec = id)) {
    sen_id *rid;
    sen_rset_recinfo *ri;
    sen_hash *h = &r2->hash;
    if (_sen_hash_get_key_value(ctx, h, id, (void **)&rid, (void **)&ri)) {
      if (score) { *score = ri->score; }
      if (h->record_unit != sen_rec_userdef) {
        return r2->keys ?
          sen_pat_get_key(ctx, r2->keys, *rid, keybuf, bufsize) : h->key_size;
      } else {
        if (h->key_size <= bufsize) { memcpy(keybuf, rid, h->key_size); }
        return h->key_size;
      }
    }
  }
  return 0;
}

int
sen_records2_find(sen_records2 *r2, const void *key)
{
  sen_id rid;
  sen_hash *h;
  sen_ctx *ctx;
  uint32_t key_size;
  sen_rset_recinfo *ri;
  if (!r2 || !r2->keys) { return 0; }
  h = &r2->hash;
  ctx = h->ctx;
  if (h->record_unit != sen_rec_document) { return 0; }
  key_size = (r2->keys->flags & SEN_OBJ_KEY_VAR_SIZE)
    ? strlen(key) : r2->keys->key_size;
  if (!(rid = sen_pat_at(ctx, r2->keys, key, key_size, NULL))) { return 0; }
  if (!sen_hash_at(ctx, h, &rid, sizeof(sen_id), (void **)&ri)) { return 0; }
  return ri->score;
}

int
sen_records2_nhits(sen_records2 *r2)
{
  sen_hash *h;
  if (!r2) { return 0; }
  h = &r2->hash;
  return *h->n_entries;
}

int
sen_records2_curr_score(sen_records2 *r2)
{
  sen_ctx *ctx;
  if (!r2) { return 0; }
  if (!r2->curr_rec) { return 0; }
  {
    sen_id *rid;
    sen_rset_recinfo *ri;
    sen_hash *h = &r2->hash;
    ctx = h->ctx;
    if (_sen_hash_get_key_value(ctx, h, r2->curr_rec, (void **)&rid, (void **)&ri)) {
      return ri->score;
    } else {
      return 0;
    }
  }
}

int
sen_records2_curr_key(sen_records2 *r2, void *keybuf, int bufsize)
{
  sen_ctx *ctx;
  if (!r2) { return 0; }
  if (!r2->curr_rec) { return 0; }
  {
    sen_id *rid;
    sen_rset_recinfo *ri;
    sen_hash *h = &r2->hash;
    ctx = h->ctx;
    if (!_sen_hash_get_key_value(ctx, h, r2->curr_rec, (void **)&rid, (void **)&ri)) {
      return 0;
    }
    if (h->record_unit != sen_rec_userdef) {
      return r2->keys ?
        sen_pat_get_key(ctx, r2->keys, *rid, keybuf, bufsize) : h->key_size;
    } else {
      if (h->key_size <= bufsize) { memcpy(keybuf, rid, h->key_size); }
      return h->key_size;
    }
  }
}

sen_id
sen_records2_curr_rec(sen_records2 *r2)
{
  if (!r2) { return SEN_ID_NIL; }
  return r2->curr_rec;
}

sen_hash * sen_rhash_group(sen_hash *s, int limit, sen_group_optarg *optarg);

sen_rc
sen_records2_group(sen_records2 *r2, int limit, sen_group_optarg *optarg)
{
  sen_hash *h;
  if (!r2) { return sen_invalid_argument; }
  sen_records2_cursor_clear(r2);
  h = &r2->hash;
  if (!sen_rhash_group(h, limit, optarg)) {
    return sen_internal_error;
  }
  return sen_success;
}

sen_id
sen_records2_at(sen_records2 *r2, const void *key, unsigned section, unsigned pos,
               int *score, int *n_subrecs)
{
  sen_id res;
  sen_hash *h;
  sen_ctx *ctx;
  sen_rset_recinfo *ri;
  if (!r2 || !r2->keys) { return SEN_ID_NIL; }
  h = &r2->hash;
  ctx = h->ctx;
  if (h->record_unit == sen_rec_userdef) {
    int key_size = (h->flags & SEN_OBJ_KEY_VAR_SIZE)
      ? strlen(key) : h->key_size;
    res = r2->curr_rec = sen_hash_at(ctx, h, key, key_size, (void **) &ri);
  } else {
    sen_rset_posinfo pi;
    int key_size = (r2->keys->flags & SEN_OBJ_KEY_VAR_SIZE)
      ? strlen(key) : r2->keys->key_size;
    if (!(res = sen_pat_at(ctx, r2->keys, key, key_size, NULL))) { return SEN_ID_NIL; }
    pi.sid = section;
    pi.pos = pos;
    res = r2->curr_rec = sen_hash_at(ctx, h, &pi, h->key_size, (void **) &ri);
  }
  if (res) {
    if (score) { *score = ri->score; }
    if (n_subrecs) { *n_subrecs = ri->n_subrecs; }
  }
  return res;
}

sen_rc
sen_record2_info(sen_records2 *r2, sen_id rh,
                void *keybuf, int bufsize, int *keysize,
                int *section, int *pos, int *score, int *n_subrecs)
{
  sen_rset_posinfo *pi;
  sen_rset_recinfo *ri;
  sen_hash *h;
  sen_ctx *ctx;
  if (!r2 || !rh) { return sen_invalid_argument; }
  h = &r2->hash;
  ctx = h->ctx;
  if (!_sen_hash_get_key_value(ctx, h, rh,
                               (void **)&pi, (void **)&ri)) {
    return sen_invalid_argument;
  }
  switch (h->record_unit) {
  case sen_rec_document :
    if ((keybuf && bufsize) || keysize) {
      int len;
      if (!r2->keys) { return sen_invalid_argument; }
      len = sen_pat_get_key(ctx, r2->keys, pi->rid, keybuf, bufsize);
      if (keysize) { *keysize = len; }
    }
    if (section) { *section = 0; }
    if (pos) { *pos = 0; }
    break;
  case sen_rec_section :
    if ((keybuf && bufsize) || keysize) {
      int len;
      if (!r2->keys) { return sen_invalid_argument; }
      len = sen_pat_get_key(ctx, r2->keys, pi->rid, keybuf, bufsize);
      if (keysize) { *keysize = len; }
    }
    if (section) { *section = pi->sid; }
    if (pos) { *pos = 0; }
    break;
  case sen_rec_position :
    if ((keybuf && bufsize) || keysize) {
      int len;
      if (!r2->keys) { return sen_invalid_argument; }
      len = sen_pat_get_key(ctx, r2->keys, pi->rid, keybuf, bufsize);
      if (keysize) { *keysize = len; }
    }
    if (section) { *section = pi->sid; }
    if (pos) { *pos = pi->pos; }
    break;
  case sen_rec_userdef :
    if ((keybuf && bufsize) || keysize) {
      unsigned int len = h->key_size;
      if (!len) { len = (unsigned int)strlen((const char *)pi) + 1; }
      if ((unsigned int)bufsize >= len) { memcpy(keybuf, pi, len); }
      if (keysize) { *keysize = len; }
    }
    if (section) { *section = 0; }
    if (pos) { *pos = 0; }
    break;
  default :
    return sen_invalid_format;
    break;
  }
  if (score) { *score = ri->score; }
  if (n_subrecs) { *n_subrecs = ri->n_subrecs; }
  return sen_success;
}

sen_rc sen_rhash_subrec_info(sen_hash *s, sen_id rh, int index, sen_id *rid,
                             int *section, int *pos, int *score, void **subrec);

sen_rc
sen_record2_subrec_info(sen_records2 *r2, sen_id rh, int index,
                       void *keybuf, int bufsize, int *keysize,
                       int *section, int *pos, int *score)
{
  sen_rc rc;
  sen_id rid;
  void *subrec;
  sen_hash *h;
  sen_ctx *ctx;
  if (!r2 || !rh) { return sen_invalid_argument; }
  h = &r2->hash;
  ctx = h->ctx;
  if ((rc = sen_rhash_subrec_info(h, rh, index, &rid, section, pos, score, &subrec))) {
    return rc;
  }
  if ((keybuf && bufsize) || keysize) {
    if (rid) {
      int len;
      if (!r2->keys) { return sen_invalid_argument; }
      len = sen_pat_get_key(ctx, r2->keys, rid, keybuf, bufsize);
      if (keysize) { *keysize = len; }
    } else if (h->record_unit == sen_rec_userdef && h->subrec_unit == sen_rec_userdef) {
      int len = (int) h->subrec_size;
      if (!len) { len = (unsigned int)strlen((char *)subrec) + 1; }
      if ((unsigned int)bufsize >= len) { memcpy(keybuf, subrec, len); }
      if (keysize) { *keysize = len; }
    }
  }
  return sen_success;
}
#endif /* USE_SEN_INDEX2 */
