/* Copyright(C) 2004 Brazil

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "senna_in.h"

#include <stdio.h>

#ifndef __USE_GNU
#define __USE_GNU /* O_DIRECT */
#endif /* __USE_GNU */
#include <fcntl.h>
#include <string.h>
#include <sys/stat.h>

#include "ctx.h"
#include "io.h"
#include "cache.h"
#include "hash.h"

#define SEN_IO_IDSTR "SENNA:IO:01.000"

/* VA hack */
/* max aio request (/proc/sys/fs/aio-max-nr) */
#define MAX_REQUEST     (64*1024)

#define MEM_ALIGN    sen_cache_block

/* end VA hack */

typedef struct _sen_io_fileinfo {
#ifdef WIN32
  HANDLE fh;
  HANDLE fmo;
  sen_mutex mutex;
#else /* WIN32 */
  int fd;
  dev_t dev;
  ino_t inode;
#endif /* WIN32 */
} fileinfo;

typedef struct _sen_io_header08 {
  char idstr[16];
  uint32_t header_size;
  uint32_t segment_size;
  uint32_t max_segment;
} io_header08;

typedef struct _sen_io_header {
  char idstr[16];
  uint32_t header_size;
  uint32_t segment_size;
  uint32_t max_segment;
  uint32_t n_arrays;
  uint64_t curr_size;
  uint32_t flags;
  uint32_t type;
  uint32_t lock;
} io_header;

#define IO_HEADER_SIZE 64

inline static sen_rc sen_open(fileinfo *fi, const char *path, int flags, size_t maxsize);
inline static void sen_fileinfo_init(fileinfo *fis, int nfis);
inline static int sen_opened(fileinfo *fi);
inline static sen_rc sen_close(fileinfo *fi);
#if defined(WIN32) && defined(WIN32_FMO_EACH)
inline static void * sen_mmap(sen_ctx *ctx, HANDLE *fmo, fileinfo *fi,
                              off_t offset, size_t length);
inline static int sen_munmap(sen_ctx *ctx, HANDLE *fmo, void *start, size_t length);
#define SEN_MMAP(ctx,fmo,fi,offset,length)\
  (sen_mmap((ctx), (fmo), (fi), (offset), (length)))
#define SEN_MUNMAP(ctx,fmo,start,length) (sen_munmap((ctx), (fmo), (start), (length)))
#else /* defined(WIN32) && defined(WIN32_FMO_EACH) */
inline static void * sen_mmap(sen_ctx *ctx, fileinfo *fi, off_t offset, size_t length);
inline static int sen_munmap(sen_ctx *ctx, void *start, size_t length);
#define SEN_MUNMAP(ctx,fmo,start,length) (sen_munmap((ctx), (start), (length)))
#ifdef USE_FAIL_MALLOC
inline static void * sen_fail_mmap(sen_ctx *ctx, fileinfo *fi,
                                   off_t offset, size_t length,
                                   const char* file, int line, const char *func);
#define SEN_MMAP(ctx,fmo,fi,offset,length)                              \
  (sen_fail_mmap((ctx), (fi), (offset), (length), __FILE__, __LINE__, __FUNCTION__))
#else /* USE_FAIL_MALLOC */
#define SEN_MMAP(ctx,fmo,fi,offset,length) (sen_mmap((ctx), (fi), (offset), (length)))
#endif /* USE_FAIL_MALLOC */
#endif  /* defined(WIN32) && defined(WIN32_FMO_EACH) */
inline static int sen_msync(void *start, size_t length);
inline static sen_rc sen_pread(fileinfo *fi, void *buf, size_t count, off_t offset);
inline static sen_rc sen_pwrite(fileinfo *fi, void *buf, size_t count, off_t offset);

#define V08      1
#define TEMPORAL 2
#define V08P(io) ((io)->flags & V08)

static sen_hash *sen_dls = NULL;

sen_rc
sen_io_init(void)
{
  sen_dls = sen_hash_create(&sen_gctx, NULL, PATH_MAX, sizeof(void *),
                            SEN_OBJ_KEY_VAR_SIZE, sen_enc_none);
  if (!sen_dls) { return sen_memory_exhausted; }
  return sen_success;
}

sen_rc
sen_io_fin(void)
{
  sen_ctx *ctx = &sen_gctx;
  if (!sen_dls) { return sen_invalid_argument; }
  SEN_HASH_EACH(sen_dls, id, NULL, NULL, NULL, { sen_dl_close(ctx, id); });
  return sen_hash_close(&sen_gctx, sen_dls);
}

sen_id
sen_dl_lookup(sen_ctx *ctx, const char *filename)
{
  sen_search_flags f = 0;
  return sen_hash_lookup(ctx, sen_dls, filename, strlen(filename), NULL, &f);
}

sen_io *
sen_io_create_tmp(uint32_t header_size, uint32_t segment_size,
                  uint32_t max_segment, sen_io_mode mode, uint32_t flags)
{
  sen_io *io;
  unsigned int b;
  uint32_t total_header_size;
  io_header *header;
  total_header_size = IO_HEADER_SIZE + header_size;
  if (!(flags & (SEN_IO_WO_SEGREF|SEN_IO_WO_NREF))) {
    total_header_size += max_segment * sizeof(uint32_t);
  }
  b = (total_header_size + sen_pagesize - 1) & ~(sen_pagesize - 1);
  if ((header = (io_header *)SEN_MMAP(&sen_gctx, NULL, NULL, 0, b))) {
    header->header_size = header_size;
    header->segment_size = segment_size;
    header->max_segment = max_segment;
    header->n_arrays = 0;
    header->flags = flags;
    header->lock = 0;
    memcpy(header->idstr, SEN_IO_IDSTR, 16);
    if ((io = SEN_GMALLOCN(sen_io, 1))) {
      sen_io_mapinfo *maps = NULL;
      if ((flags & SEN_IO_WO_SEGREF) ||
          ((maps = SEN_GMALLOCN(sen_io_mapinfo, max_segment)) &&
           memset(maps, 0, sizeof(sen_io_mapinfo) * max_segment))) {
        io->header = header;
        if (flags & (SEN_IO_WO_SEGREF|SEN_IO_WO_NREF)) {
          io->nrefs = NULL;
          io->user_header = (((byte *) header) + IO_HEADER_SIZE);
        } else {
          io->nrefs = (uint32_t *)(((byte *) header) + IO_HEADER_SIZE);
          io->user_header = ((byte *) io->nrefs) + max_segment * sizeof(uint32_t);
        }
        io->maps = maps;
        io->base = b;
        io->base_seg = 0;
        io->mode = mode;
        io->header->curr_size = b;
        io->fis = NULL;
        io->ainfo = NULL;
        io->max_map_seg = 0;
        io->nmaps = 0;
        io->count = 0;
        io->flags = TEMPORAL;
        io->lock = &header->lock;
        return io;
      }
      SEN_GFREE(io);
    }
    SEN_MUNMAP(&sen_gctx, NULL, header, b);
  }
  return NULL;
}

sen_io *
sen_io_create(const char *path, uint32_t header_size, uint32_t segment_size,
              uint32_t max_segment, sen_io_mode mode, uint32_t flags)
{
  sen_io *io;
  fileinfo *fis;
  unsigned int b, max_nfiles;
  uint32_t bs, total_header_size;
  io_header *header;
  if (!path) {
    return sen_io_create_tmp(header_size, segment_size, max_segment, mode, flags);
  }
  if (!*path || (strlen(path) > PATH_MAX - 4)) { return NULL; }
  total_header_size = IO_HEADER_SIZE + header_size;
  if (!(flags & (SEN_IO_WO_SEGREF|SEN_IO_WO_NREF))) {
    total_header_size += max_segment * sizeof(uint32_t);
  }
  b = (total_header_size + sen_pagesize - 1) & ~(sen_pagesize - 1);
  bs = (b + segment_size - 1) / segment_size;
  max_nfiles = (unsigned int)(
    ((uint64_t)segment_size * (max_segment + bs) + SEN_IO_FILE_SIZE - 1)
    / SEN_IO_FILE_SIZE);
  if ((fis = SEN_GMALLOCN(fileinfo, max_nfiles))) {
    sen_fileinfo_init(fis, max_nfiles);
    if (!sen_open(fis, path, O_RDWR|O_CREAT|O_TRUNC, SEN_IO_FILE_SIZE)) {
      if ((header = (io_header *)SEN_MMAP(&sen_gctx, &fis->fmo, fis, 0, b))) {
        header->header_size = header_size;
        header->segment_size = segment_size;
        header->max_segment = max_segment;
        header->n_arrays = 0;
        header->flags = flags;
        header->lock = 0;
        memcpy(header->idstr, SEN_IO_IDSTR, 16);
        sen_msync(header, b);
        if ((io = SEN_GMALLOCN(sen_io, 1))) {
          sen_io_mapinfo *maps = NULL;
          if ((flags & SEN_IO_WO_SEGREF) ||
              ((maps = SEN_GMALLOCN(sen_io_mapinfo, max_segment)) &&
               memset(maps, 0, sizeof(sen_io_mapinfo) * max_segment))) {
            strncpy(io->path, path, PATH_MAX);
            io->header = header;
            if (flags & (SEN_IO_WO_SEGREF|SEN_IO_WO_NREF)) {
              io->nrefs = NULL;
              io->user_header = (((byte *) header) + IO_HEADER_SIZE);
            } else {
              io->nrefs = (uint32_t *)(((byte *) header) + IO_HEADER_SIZE);
              io->user_header = ((byte *) io->nrefs) + max_segment * sizeof(uint32_t);
            }
            io->maps = maps;
            io->base = b;
            io->base_seg = bs;
            io->mode = mode;
            io->header->curr_size = b;
            io->fis = fis;
            io->ainfo = NULL;
            io->max_map_seg = 0;
            io->nmaps = 0;
            io->count = 0;
            io->flags = 0;
            io->lock = &header->lock;
            return io;
          }
          SEN_GFREE(io);
        }
        SEN_MUNMAP(&sen_gctx, &fis->fmo, header, b);
      }
      sen_close(fis);
    }
    SEN_GFREE(fis);
  }
  return NULL;
}

sen_rc
array_init_(sen_io *io, int n_arrays, size_t hsize, size_t msize)
{
  int i;
  uint32_t ws;
  byte *hp, *mp;
  sen_io_array_spec *array_specs = (sen_io_array_spec *)io->user_header;
  hp = io->user_header;
  if (!(mp = SEN_GCALLOC(msize))) {
    return sen_memory_exhausted;
  }
  io->ainfo = (sen_io_array_info *)mp;
  hp += sizeof(sen_io_array_spec) * n_arrays;
  mp += sizeof(sen_io_array_info) * n_arrays;
  for (ws = 0; (1 << ws) < io->header->segment_size; ws++);
  for (i = 0; i < n_arrays; i++) {
    uint32_t we = ws - array_specs[i].w_of_element;
    io->ainfo[i].w_of_elm_in_a_segment = we;
    io->ainfo[i].elm_mask_in_a_segment = (1 << we) - 1;
    io->ainfo[i].max_n_segments = array_specs[i].max_n_segments;
    io->ainfo[i].element_size = 1 << array_specs[i].w_of_element;
    io->ainfo[i].segments = (uint32_t *)hp;
    io->ainfo[i].addrs = (void **)mp;
    hp += sizeof(uint32_t) * array_specs[i].max_n_segments;
    mp += sizeof(void *) * array_specs[i].max_n_segments;
  }
  io->user_header += hsize;
  return sen_success;
}

sen_rc
array_init(sen_io *io, int n_arrays)
{
  if (n_arrays) {
    int i;
    sen_io_array_spec *array_specs = (sen_io_array_spec *)io->user_header;
    size_t hsize = sizeof(sen_io_array_spec) * n_arrays;
    size_t msize = sizeof(sen_io_array_info) * n_arrays;
    for (i = 0; i < n_arrays; i++) {
      hsize += sizeof(uint32_t) * array_specs[i].max_n_segments;
      msize += sizeof(void *) * array_specs[i].max_n_segments;
    }
    return array_init_(io, n_arrays, hsize, msize);
  }
  return sen_success;
}

sen_io *
sen_io_create_with_array(const char *path, uint32_t header_size, uint32_t segment_size,
                         sen_io_mode mode, int n_arrays, sen_io_array_spec *array_specs)
{
  if (n_arrays) {
    int i;
    sen_io *io;
    byte *hp;
    uint32_t nsegs = 0;
    size_t hsize = sizeof(sen_io_array_spec) * n_arrays;
    size_t msize = sizeof(sen_io_array_info) * n_arrays;
    for (i = 0; i < n_arrays; i++) {
      nsegs += array_specs[i].max_n_segments;
      hsize += sizeof(uint32_t) * array_specs[i].max_n_segments;
      msize += sizeof(void *) * array_specs[i].max_n_segments;
    }
    if ((io = sen_io_create(path, header_size + hsize,
                            segment_size, nsegs, mode, 0))) {
      hp = io->user_header;
      memcpy(hp, array_specs, sizeof(sen_io_array_spec) * n_arrays);
      io->header->n_arrays = n_arrays;
      if (!array_init_(io, n_arrays, hsize, msize)) {
        return io;
      }
      sen_io_close(io);
    }
  }
  return NULL;
}

inline static uint32_t
segment_alloc(sen_io *io)
{
  uint32_t n, s;
  sen_io_array_info *ai;
  char *used = SEN_GCALLOC(io->header->max_segment + 1);
  if (!used) { return 0; }
  for (n = io->header->n_arrays, ai = io->ainfo; n; n--, ai++) {
    for (s = 0; s < ai->max_n_segments; s++) {
      used[ai->segments[s]] = 1;
    }
  }
  for (s = 1; s <= io->header->max_segment; s++) {
    if (!used[s]) { goto exit; }
  }
  s = 0;
exit :
  SEN_GFREE(used);
  return s;
}

void
sen_io_segment_alloc(sen_io *io, sen_io_array_info *ai, uint32_t lseg, int *flags, void **p)
{
  uint32_t *sp = &ai->segments[lseg];
  if (!*sp) {
    if ((*flags & SEN_TABLE_ADD)) {
      if ((*sp = segment_alloc(io))) {
        *flags |= SEN_TABLE_ADDED;
      }
    }
  }
  if (*sp) {
    uint32_t pseg = *sp - 1;
    SEN_IO_SEG_REF(io, pseg, *p);
    if (*p) { SEN_IO_SEG_UNREF(io, pseg); };
  }
}

void *
sen_io_array_at(sen_io *io, uint32_t array, off_t offset, int *flags)
{
  void *res;
  SEN_IO_ARRAY_AT(io,array,offset,flags,res);
  return res;
}

uint32_t
sen_io_detect_type(sen_ctx *ctx, const char *path)
{
  io_header h;
  uint32_t res = 0;
  int fd = open(path, O_RDWR);
  if (fd != -1) {
    struct stat s;
    if (fstat(fd, &s) != -1 && s.st_size >= sizeof(io_header)) {
      if (read(fd, &h, sizeof(io_header)) == sizeof(io_header)) {
        if (!memcmp(h.idstr, SEN_IO_IDSTR, 16)) {
          res = h.type;
        } else {
          ERR(sen_invalid_format, "sen_io_detect_type failed");
        }
      } else {
        SERR(path);
      }
    } else {
      ERR(sen_invalid_format, "sen_io_detect_type failed");
    }
    close(fd);
  } else {
    SERR(path);
  }
  return res;
}

sen_io *
sen_io_open(const char *path, sen_io_mode mode)
{
  sen_io *io;
  struct stat s;
  fileinfo *fis;
  uint32_t flags = 0;
  unsigned int b, max_nfiles;
  uint32_t total_header_size;
  uint32_t header_size = 0, segment_size = 0, max_segment = 0, bs, v08p = 0;
  if (!path || !*path || (strlen(path) > PATH_MAX - 4)) { return NULL; }
  {
    io_header h;
    int fd = open(path, O_RDWR);
    if (fd == -1) { GSERR(path); return NULL; }
    if (fstat(fd, &s) != -1 && s.st_size >= sizeof(io_header)) {
      if (read(fd, &h, sizeof(io_header)) == sizeof(io_header)) {
        if (memcmp(h.idstr, SEN_IO_IDSTR, 16)) {
          /* regard as v0.8 format */
          v08p = 1;
        }
        header_size = h.header_size;
        segment_size = h.segment_size;
        max_segment = h.max_segment;
        flags = h.flags;
      }
    }
    close(fd);
    if (!segment_size) { return NULL; }
  }
  total_header_size = IO_HEADER_SIZE + header_size;
  if (!v08p && !(flags & (SEN_IO_WO_SEGREF|SEN_IO_WO_NREF))) {
    total_header_size += max_segment * sizeof(uint32_t);
  }
  b = (total_header_size + sen_pagesize - 1) & ~(sen_pagesize - 1);
  bs = (b + segment_size - 1) / segment_size;
  max_nfiles = (unsigned int)(
    ((uint64_t)segment_size * (max_segment + bs) + SEN_IO_FILE_SIZE - 1)
    / SEN_IO_FILE_SIZE);
  if (!(fis = SEN_GMALLOCN(fileinfo, max_nfiles))) { return NULL; }
  sen_fileinfo_init(fis, max_nfiles);
  if (!sen_open(fis, path, O_RDWR, SEN_IO_FILE_SIZE)) {
    io_header *header;
    if ((header = SEN_MMAP(&sen_gctx, &fis->fmo, fis, 0, b))) {
      if ((io = SEN_GMALLOC(sizeof(sen_io)))) {
        sen_io_mapinfo *maps = NULL;
        if ((header->flags & SEN_IO_WO_SEGREF) ||
            (maps = SEN_GCALLOC(sizeof(sen_io_mapinfo) * max_segment))) {
          strncpy(io->path, path, PATH_MAX);
          io->header = header;
          if (v08p) {
            io->nrefs = SEN_GCALLOC(sizeof(uint32_t) * max_segment);
            io->user_header = ((byte *) header) + sizeof(io_header08);
          } else {
            if (header->flags & (SEN_IO_WO_SEGREF|SEN_IO_WO_NREF)) {
              io->nrefs = NULL;
              io->user_header = (((byte *) header) + IO_HEADER_SIZE);
            } else {
              io->nrefs = (uint32_t *)(((byte *) header) + IO_HEADER_SIZE);
              io->user_header = ((byte *) io->nrefs) + max_segment * sizeof(uint32_t);
            }
          }
          if ((header->flags & (SEN_IO_WO_SEGREF|SEN_IO_WO_NREF)) || io->nrefs) {
            io->maps = maps;
            io->base = b;
            io->base_seg = bs;
            io->mode = mode;
            io->fis = fis;
            io->ainfo = NULL;
            io->max_map_seg = 0;
            io->nmaps = 0;
            io->count = 0;
            io->flags = V08;
            io->lock = &header->lock;
            if (!array_init(io, io->header->n_arrays)) {
              return io;
            }
          }
          if (io->maps) { SEN_GFREE(io->maps); }
        }
        SEN_GFREE(io);
      }
      SEN_MUNMAP(&sen_gctx, &fis->fmo, header, b);
    }
    sen_close(fis);
  }
  SEN_GFREE(fis);
  return NULL;
}

sen_rc
sen_io_close(sen_io *io)
{
  int i;
  sen_io_mapinfo *mi;
  fileinfo *fi;
  uint32_t bs = io->base_seg;
  uint32_t max_segment = io->header->max_segment;
  uint32_t segment_size = io->header->segment_size;
  unsigned int max_nfiles = (unsigned int)(
    ((uint64_t)segment_size * (max_segment + bs) + SEN_IO_FILE_SIZE - 1)
    / SEN_IO_FILE_SIZE);
  if (io->ainfo) { SEN_GFREE(io->ainfo); }
  if (io->maps) {
    for (mi = io->maps, i = max_segment; i; mi++, i--) {
      if (mi->map) {
        /* if (atomic_read(mi->nref)) { return STILL_IN_USE ; } */
#ifdef WIN32
        if ((io->flags & TEMPORAL)) {
          SEN_GFREE(mi->map);
        } else
#endif /* WIN32 */
        SEN_MUNMAP(&sen_gctx, &mi->fmo, mi->map, segment_size);
      }
    }
    SEN_GFREE(io->maps);
  }
#ifdef WIN32
  if ((io->flags & TEMPORAL)) {
    SEN_GFREE(io->header);
  } else
#endif /* WIN32 */
  SEN_MUNMAP(&sen_gctx, &io->fis->fmo, io->header, io->base);
  if (io->fis) {
    for (fi = io->fis, i = max_nfiles; i; fi++, i--) { sen_close(fi); }
    SEN_GFREE(io->fis);
  }
  SEN_GFREE(io);
  return sen_success;
}

uint32_t
sen_io_base_seg(sen_io *io)
{
  return io->base_seg;
}

const char *
sen_io_path(sen_io *io)
{
  return io->path;
}

void *
sen_io_header(sen_io *io)
{
  return io->user_header;
}

sen_rc
sen_io_set_type(sen_io *io, uint32_t type)
{
  if (!io || !io->header) {
    return sen_invalid_argument;
  }
  io->header->type = type;
  return sen_success;
}

inline static void
gen_pathname(const char *path, char *buffer, int fno)
{
  size_t len = strlen(path);
  memcpy(buffer, path, len);
  if (fno) {
    buffer[len] = '.';
    sen_itoh(fno, buffer + len + 1, 3);
  } else {
    buffer[len] = '\0';
  }
}

sen_rc
sen_io_size(sen_io *io, uint64_t *size)
{
  int fno;
  struct stat s;
  uint64_t tsize = 0;
  char buffer[PATH_MAX];
  uint32_t nfiles;
  if (io->header->curr_size) {
    nfiles = (uint32_t) ((io->header->curr_size + SEN_IO_FILE_SIZE - 1) / SEN_IO_FILE_SIZE);
  } else {
    uint32_t bs = io->base_seg;
    uint32_t max_segment = io->header->max_segment;
    uint32_t segment_size = io->header->segment_size;
    nfiles = (uint32_t) (((uint64_t)segment_size * (max_segment + bs) + SEN_IO_FILE_SIZE - 1)
                    / SEN_IO_FILE_SIZE);
  }
  for (fno = 0; fno < nfiles; fno++) {
    gen_pathname(io->path, buffer, fno);
    if (stat(buffer, &s)) {
      GSERR(buffer);
    } else {
      tsize += s.st_size;
    }
  }
  *size = tsize;
  return sen_success;
}

sen_rc
sen_io_remove(const char *path)
{
  struct stat s;
  if (stat(path, &s)) {
    SEN_LOG(sen_log_info, "stat failed '%s' (%s)", path, strerror(errno));
    return sen_file_operation_error;
  } else if (unlink(path)) {
    GSERR(path);
    return sen_file_operation_error;
  } else {
    int fno;
    char buffer[PATH_MAX];
    for (fno = 1; ; fno++) {
      gen_pathname(path, buffer, fno);
      if (!stat(buffer, &s)) {
        if (unlink(buffer)) { GSERR(buffer); }
      } else {
        break;
      }
    }
    return sen_success;
  }
}

sen_rc
sen_io_rename(const char *old_name, const char *new_name)
{
  struct stat s;
  if (stat(old_name, &s)) {
    SEN_LOG(sen_log_info, "stat failed '%s' (%s)", old_name, strerror(errno));
    return sen_file_operation_error;
  } else if (rename(old_name, new_name)) {
    GSERR(old_name);
    return sen_file_operation_error;
  } else {
    int fno;
    char old_buffer[PATH_MAX];
    char new_buffer[PATH_MAX];
    for (fno = 1; ; fno++) {
      gen_pathname(old_name, old_buffer, fno);
      if (!stat(old_buffer, &s)) {
        gen_pathname(new_name, new_buffer, fno);
        if (rename(old_buffer, new_buffer)) { GSERR(old_buffer); }
      } else {
        break;
      }
    }
    return sen_success;
  }
  return sen_file_operation_error;
}

typedef struct {
  sen_io_ja_ehead head;
  char body[256];
} ja_element;

sen_rc
sen_io_read_ja(sen_io *io, sen_ctx *ctx, sen_io_ja_einfo *einfo, uint32_t epos, uint32_t key,
               uint32_t segment, uint32_t offset, void **value, uint32_t *value_len)
{
  uint32_t rest = 0, size = *value_len + sizeof(sen_io_ja_ehead);
  uint32_t segment_size = io->header->segment_size;
  uint32_t segments_per_file = SEN_IO_FILE_SIZE / segment_size;
  uint32_t bseg = segment + io->base_seg;
  int fno = bseg / segments_per_file;
  fileinfo *fi = &io->fis[fno];
  off_t base = fno ? 0 : io->base - (uint64_t)segment_size * io->base_seg;
  off_t pos = (uint64_t)segment_size * (bseg % segments_per_file) + offset + base;
  ja_element *v = SEN_MALLOC(size);
  if (!v) {
    *value = NULL;
    *value_len = 0;
    return sen_memory_exhausted;
  }
  if (pos + size > SEN_IO_FILE_SIZE) {
    rest = pos + size - SEN_IO_FILE_SIZE;
    size = SEN_IO_FILE_SIZE - pos;
  }
  if (!sen_opened(fi)) {
    char path[PATH_MAX];
    gen_pathname(io->path, path, fno);
    if (sen_open(fi, path, O_RDWR|O_CREAT, SEN_IO_FILE_SIZE)) {
      *value = NULL;
      *value_len = 0;
      SEN_FREE(v);
      return sen_file_operation_error;
    }
  }
  if (sen_pread(fi, v, size, pos)) {
    *value = NULL;
    *value_len = 0;
    SEN_FREE(v);
    return sen_file_operation_error;
  }
  if (einfo->pos != epos) {
    SEN_LOG(sen_log_warning, "einfo pos changed %x => %x", einfo->pos, epos);
    *value = NULL;
    *value_len = 0;
    SEN_FREE(v);
    return sen_internal_error;
  }
  if (einfo->size != *value_len) {
    SEN_LOG(sen_log_warning, "einfo size changed %d => %d", einfo->size, *value_len);
    *value = NULL;
    *value_len = 0;
    SEN_FREE(v);
    return sen_internal_error;
  }
  if (v->head.key != key) {
    SEN_LOG(sen_log_error, "ehead key unmatch %x => %x", key, v->head.key);
    *value = NULL;
    *value_len = 0;
    SEN_FREE(v);
    return sen_invalid_format;
  }
  if (v->head.size != *value_len) {
    SEN_LOG(sen_log_error, "ehead size unmatch %d => %d", *value_len, v->head.size);
    *value = NULL;
    *value_len = 0;
    SEN_FREE(v);
    return sen_invalid_format;
  }
  if (rest) {
    byte *vr = (byte *)v + size;
    do {
      fi = &io->fis[++fno];
      if (!sen_opened(fi)) {
        char path[PATH_MAX];
        gen_pathname(io->path, path, fno);
        if (sen_open(fi, path, O_RDWR|O_CREAT, SEN_IO_FILE_SIZE)) {
          *value = NULL;
          *value_len = 0;
          SEN_FREE(v);
          return sen_file_operation_error;
        }
      }
      size = rest > SEN_IO_FILE_SIZE ? SEN_IO_FILE_SIZE : rest;
      if (sen_pread(fi, vr, size, 0)) {
        *value = NULL;
        *value_len = 0;
        SEN_FREE(v);
        return sen_file_operation_error;
      }
      vr += size;
      rest -= size;
    } while (rest);
  }
  *value = v->body;
  return sen_success;
}

sen_rc
sen_io_write_ja(sen_io *io, sen_ctx *ctx, uint32_t key,
                uint32_t segment, uint32_t offset, void *value, uint32_t value_len)
{
  sen_rc rc;
  uint32_t rest = 0, size = value_len + sizeof(sen_io_ja_ehead);
  uint32_t segment_size = io->header->segment_size;
  uint32_t segments_per_file = SEN_IO_FILE_SIZE / segment_size;
  uint32_t bseg = segment + io->base_seg;
  int fno = bseg / segments_per_file;
  fileinfo *fi = &io->fis[fno];
  off_t base = fno ? 0 : io->base - (uint64_t)segment_size * io->base_seg;
  off_t pos = (uint64_t)segment_size * (bseg % segments_per_file) + offset + base;
  if (pos + size > SEN_IO_FILE_SIZE) {
    rest = pos + size - SEN_IO_FILE_SIZE;
    size = SEN_IO_FILE_SIZE - pos;
  }
  if (!sen_opened(fi)) {
    char path[PATH_MAX];
    gen_pathname(io->path, path, fno);
    if ((rc = sen_open(fi, path, O_RDWR|O_CREAT, SEN_IO_FILE_SIZE))) { return rc; }
  }
  if (value_len <= 256) {
    ja_element je;
    je.head.size = value_len;
    je.head.key = key;
    memcpy(je.body, value, value_len);
    rc = sen_pwrite(fi, &je, size, pos);
  } else {
    sen_io_ja_ehead eh;
    eh.size = value_len;
    eh.key =  key;
    if ((rc = sen_pwrite(fi, &eh, sizeof(sen_io_ja_ehead), pos))) { return rc; }
    pos += sizeof(sen_io_ja_ehead);
    rc = sen_pwrite(fi, value, size - sizeof(sen_io_ja_ehead), pos);
  }
  if (rc) { return rc; }
  if (rest) {
    byte *vr = (byte *)value + size - sizeof(sen_io_ja_ehead);
    do {
      fi = &io->fis[++fno];
      if (!sen_opened(fi)) {
        char path[PATH_MAX];
        gen_pathname(io->path, path, fno);
        if ((rc = sen_open(fi, path, O_RDWR|O_CREAT, SEN_IO_FILE_SIZE))) { return rc; }
      }
      size = rest > SEN_IO_FILE_SIZE ? SEN_IO_FILE_SIZE : rest;
      if ((rc = sen_pwrite(fi, vr, size, 0))) { return rc; }
      vr += size;
      rest -= size;
    } while (rest);
  }
  return rc;
}

sen_rc
sen_io_write_ja_ehead(sen_io *io, sen_ctx *ctx, uint32_t key,
                      uint32_t segment, uint32_t offset, uint32_t value_len)
{
  sen_rc rc;
  uint32_t segment_size = io->header->segment_size;
  uint32_t segments_per_file = SEN_IO_FILE_SIZE / segment_size;
  uint32_t bseg = segment + io->base_seg;
  int fno = bseg / segments_per_file;
  fileinfo *fi = &io->fis[fno];
  off_t base = fno ? 0 : io->base - (uint64_t)segment_size + io->base_seg;
  off_t pos = (uint64_t)segment_size * (bseg % segments_per_file) + offset + base;
  if (!sen_opened(fi)) {
    char path[PATH_MAX];
    gen_pathname(io->path, path, fno);
    if ((rc = sen_open(fi, path, O_RDWR|O_CREAT, SEN_IO_FILE_SIZE))) { return rc; }
  }
  {
    sen_io_ja_ehead eh;
    eh.size = value_len;
    eh.key =  key;
    return sen_pwrite(fi, &eh, sizeof(sen_io_ja_ehead), pos);
  }
}

void *
sen_io_win_map(sen_io *io, sen_ctx *ctx, sen_io_win *iw, uint32_t segment,
               uint32_t offset, uint32_t size, sen_io_rw_mode mode)
{
  byte *p;
  off_t pos, base;
  int fno;
  uint32_t nseg, bseg;
  uint32_t segment_size = io->header->segment_size;
  uint32_t segments_per_file = SEN_IO_FILE_SIZE / segment_size;
  iw->ctx = ctx;
  iw->diff = 0;
  if (offset >= segment_size) {
    segment += offset / segment_size;
    offset = offset % segment_size;
  }
  nseg = (offset + size + segment_size - 1) / segment_size;
  bseg = segment + io->base_seg;
  fno = bseg / segments_per_file;
  base = fno ? 0 : io->base - (uint64_t)segment_size * io->base_seg;
  pos = (uint64_t)segment_size * (bseg % segments_per_file) + offset + base;
  if (!size || !io || segment + nseg > io->header->max_segment ||
      fno != (bseg + nseg - 1) / segments_per_file) {
    return NULL;
  }
  switch (mode) {
  case sen_io_rdonly:
    {
      fileinfo *fi = &io->fis[fno];
      if (!sen_opened(fi)) {
        char path[PATH_MAX];
        gen_pathname(io->path, path, fno);
        if (sen_open(fi, path, O_RDWR|O_CREAT, SEN_IO_FILE_SIZE)) {
          return NULL;
        }
      }
      if (!(p = SEN_MALLOC(size))) { return NULL; }
      if (sen_pread(fi, p, size, pos)) {
        SEN_FREE(p);
        return NULL;
      }
      iw->addr = p;
    }
    break;
  case sen_io_rdwr:
    // if (nseg > 1) { /* auto unmap is not implemented yet */
    if (nseg > 0) {
      fileinfo *fi = &io->fis[fno];
      if (!sen_opened(fi)) {
        char path[PATH_MAX];
        gen_pathname(io->path, path, fno);
        if (sen_open(fi, path, O_RDWR|O_CREAT, SEN_IO_FILE_SIZE)) {
          return NULL;
        }
      }
      if (!(p = SEN_MMAP(&sen_gctx, &iw->fmo, fi, pos, (uint64_t)segment_size * nseg))) {
        return NULL;
      }
      if (!V08P(io)) {
        uint64_t tail = io->base + (uint64_t)segment_size * segment + offset + size;
        if (tail > io->header->curr_size) { io->header->curr_size = tail; }
      }
    } else {
      SEN_LOG(sen_log_alert, "nseg == 0! in sen_io_win_map(%p, %u, %u, %u)", io, segment, offset, size);
      // SEN_IO_SEG_REF(io, segment, p); if (!p) { return NULL; }
      return NULL;
    }
    iw->addr = p + offset;
    break;
  case sen_io_wronly:
    if (!(p = SEN_MALLOC(size))) { return NULL; }
    memset(p, 0, size);
    iw->cached = 0;
    iw->addr = p;
    break;
  default :
    return NULL;
  }
  iw->io = io;
  iw->mode = mode;
  iw->segment = segment;
  iw->offset = offset;
  iw->nseg = nseg;
  iw->size = size;
  iw->pos = pos;
  return iw->addr;
}

#ifdef USE_AIO
sen_rc
sen_io_win_mapv(sen_io_win **list, sen_ctx *ctx, int nent)
{
  int i;
  sen_io_win *iw;
  struct aiocb *iocbs[MAX_REQUEST];
  struct aiocb iocb[MAX_REQUEST];
  CacheIOOper oper[MAX_REQUEST];
  int count = 0;

  sen_io_win **clist = list;
  int cl = 0;

retry:
  for (i = 0; i < nent; i++) {
    iw = list[i];
    if (sen_aio_enabled && iw->mode == sen_io_rdonly) {
        /* this block is same as sen_io_win_map() */
        sen_io *io = iw->io;
        uint32_t segment = iw->segment, offset = iw->offset, size = iw->size;
        byte *p;
        off_t pos, base;
        int fno;
        uint32_t nseg, bseg;
        uint32_t segment_size = io->header->segment_size;
        uint32_t segments_per_file = SEN_IO_FILE_SIZE / segment_size;
        fileinfo *fi;
        iw->diff = 0;
        if (offset >= segment_size) {
            segment += offset / segment_size;
            offset = offset % segment_size;
        }
        nseg = (offset + size + segment_size - 1) / segment_size;
        bseg = segment + io->base_seg;
        fno = bseg / segments_per_file;
        base = fno ? 0 : io->base - (uint64_t)segment_size * io->base_seg;
        pos = (uint64_t)segment_size * (bseg % segments_per_file) + offset + base;
        if (!size || !io || segment + nseg > io->header->max_segment ||
            fno != (bseg + nseg - 1) / segments_per_file) {
          return sen_abnormal_error;
        }
        fi = &io->fis[fno];
        if (!sen_opened(fi)) {
            char path[PATH_MAX];
            gen_pathname(io->path, path, fno);
            if (sen_open(fi, path, O_RDWR|O_CREAT|O_DIRECT, SEN_IO_FILE_SIZE)) {
                return sen_internal_error;
            }
        }
        {
            /* AIO + DIO + cache hack */
            /* calc alignment */
            // todo : calculate curr_size.
            off_t voffset = pos - (pos % MEM_ALIGN);
            uint32_t vsize = pos + size;

            vsize = ((vsize - 1) / MEM_ALIGN + 1) * MEM_ALIGN;
            vsize = vsize - voffset;

            /* diff of aligned offset */
            iw->diff = pos - voffset;

            dp ("pos: %lu, allocate: %d, really needed: %d\n", voffset, vsize, size);
            memset(&oper[count], 0, sizeof (CacheIOOper));
            memset(&iocb[count], 0, sizeof (struct aiocb));
            oper[count].iocb = &iocb[count];
            iocb[count].aio_fildes = fi->fd;
            iocb[count].aio_lio_opcode = LIO_READ;

            if (vsize <= MEM_ALIGN &&
                (p = sen_cache_read (&oper[count], fi->dev, fi->inode, voffset, vsize)) != NULL) {
                /* use cache process  */
                iw->cached = oper[count].cd->num;

                /* Cache require aio_read() or
                   already aio_read() by other process */
                if (oper[count].read == 1) {
                    iocbs[count] = &iocb[count];
                    count++; /* aio count */
                } else if (oper[count].cd->flag == CACHE_READ) {
                    /* this iw is ignored in this loop.
                       should re-check after AIO */
                    clist[cl++] = iw;
                }
            } else {
                /* This size cannot use Cache */
                dp ("Wont use cache offset=%lu size=%u\n", voffset, vsize);

                /* allocate aligned memory */
                if (posix_memalign(&p, MEM_ALIGN, vsize) != 0) {
                    return sen_external_error;
                }
                iocb[count].aio_buf = p;
                iocb[count].aio_nbytes = vsize;
                iocb[count].aio_offset = voffset;
                iocbs[count] = &iocb[count];

                /* This is not cached  */
                oper[count].cd = NULL;
                iw->cached = -1;

                /* aio count up */
                count++;
            }
            iw->addr = p;
            iw->segment = segment;
            iw->offset = offset;
            iw->nseg = nseg;
            iw->size = size;
            iw->pos = pos;
        } /* End  AIO + DIO + cache hack */
    } else {
        if (!sen_io_win_map(iw->io, ctx, iw, iw->segment, iw->offset, iw->size, iw->mode)) {
            return sen_internal_error;
        }
    }
  }

  if (sen_aio_enabled) {
      if (count > 0) {
          int c;

          /* aio_read () */
          if (lio_listio (LIO_WAIT, iocbs, count, NULL) < 0) {
              perror ("lio_listio");
              return sen_external_error;
          }
          for (c=0;c<count;c++) {
              /* cache data is now VALID */
              if (oper[c].cd) oper[c].cd->flag = CACHE_VALID;
          }
      }
      if (cl > 0) {
          /*
           *  if previous loop have CACHE_READ CacheData,
           *  it should retry.
           */
          dp ("-- Validate Reading state CacheData (%d) --\n", cl);
          /* update list and nent for loop */
          list = clist; /* list of iw which use CACHE_READ CacheData */
          nent = cl;    /* number of iw */
          cl = 0;
          count = 0;
          usleep(1);
          goto retry;
      } else
          dp("-- No Reading state CacheData. --\n");
  }
  return sen_success;
}
#endif /* USE_AIO */

sen_rc
sen_io_win_unmap(sen_io_win *iw)
{
  sen_rc rc = sen_success;
  sen_io *io = iw->io;
  sen_ctx *ctx = iw->ctx;
  uint32_t segment_size = io->header->segment_size;
  uint32_t segments_per_file = SEN_IO_FILE_SIZE / segment_size;
  int nseg = iw->nseg;
  switch (iw->mode) {
  case sen_io_rdonly:
#ifdef USE_AIO
    /* VA hack */
    if (!sen_aio_enabled || (iw->cached < 0 && iw->addr)) { SEN_FREE(iw->addr); }
    else if (iw->cached >= 0){
      /* this data is cached */
      sen_cache_data_unref (iw->cached);
      iw->cached = -1;
    }
    /* end VA hack */
#else /* USE_AIO */
    if (iw->addr) { SEN_FREE(iw->addr); }
#endif /* USE_AIO */
    iw->addr = NULL;
    break;
  case sen_io_rdwr:
    // if (nseg > 1) { /* auto unmap is not implemented yet */
    if (nseg > 0) {
      SEN_MUNMAP(&sen_gctx, &iw->fmo, ((byte *)iw->addr) - iw->offset, (uint64_t)segment_size * nseg);
    } else {
      if (iw->segment >= io->header->max_segment) {
        rc = sen_invalid_argument;
      } else {
        SEN_IO_SEG_UNREF(io, iw->segment);
      }
#ifdef USE_AIO
      if (sen_aio_enabled) {
        int fno = (iw->segment + io->base_seg) / segments_per_file;
        fileinfo *fi = &io->fis[fno];
        sen_cache_mark_invalid(fi->dev, fi->inode, iw->pos, iw->size);
      }
#endif /* USE_AIO */
    }
    iw->addr = NULL;
    break;
  case sen_io_wronly:
    {
      int fno = (iw->segment + io->base_seg) / segments_per_file;
      fileinfo *fi = &io->fis[fno];
      if (!sen_opened(fi)) {
        char path[PATH_MAX];
        gen_pathname(io->path, path, fno);
        rc = sen_open(fi, path, O_RDWR|O_CREAT, SEN_IO_FILE_SIZE);
      }
      if (!rc) {
        if (!(rc = sen_pwrite(fi, iw->addr, iw->size, iw->pos))) {
          if (!V08P(io)) {
            uint64_t tail = io->base + (uint64_t)segment_size * iw->segment + iw->offset + iw->size;
            if (tail > io->header->curr_size) { io->header->curr_size = tail; }
          }
          if (!iw->cached) { SEN_FREE(iw->addr); }
          iw->addr = NULL;
        }
#ifdef USE_AIO
        if (sen_aio_enabled) {
          sen_cache_mark_invalid(fi->dev, fi->inode, iw->pos, iw->size);
        }
#endif /* USE_AIO */
      }
    }
    break;
  default :
    rc = sen_invalid_argument;
  }
  return rc;
}

void *
sen_io_win_map2(sen_io *io, sen_ctx *ctx, sen_io_win *iw, uint32_t segment,
                uint32_t offset, uint32_t size, sen_io_rw_mode mode)
{
  uint32_t nseg, segment_size = io->header->segment_size;
  if (offset >= segment_size) {
    segment += offset / segment_size;
    offset = offset % segment_size;
  }
  nseg = (offset + size + segment_size - 1) / segment_size;
  if (!iw || !size || !ctx || segment + nseg > io->header->max_segment) { return NULL; }
  iw->ctx = ctx;
  iw->diff = 0;
  iw->io = io;
  iw->mode = mode;
  iw->segment = segment;
  iw->offset = offset;
  iw->nseg = nseg;
  iw->size = size;
  if (nseg == 1) {
    byte *addr;
    SEN_IO_SEG_MAP(io, segment, addr);
    if (!addr) { return NULL; }
    iw->cached = 1;
    iw->addr = addr + offset;
  } else {
    if (!(iw->addr = SEN_MALLOC(size))) { return NULL; }
    iw->cached = 0;
    switch (mode) {
    case sen_io_rdonly:
    case sen_io_rdwr:
      {
        byte *p, *q;
        uint32_t s, r;
        for (p = iw->addr, r = size; r; p += s, r -= s, segment++, offset = 0) {
          SEN_IO_SEG_MAP(io, segment, q);
          if (!q) {
            SEN_FREE(iw->addr);
            return NULL;
          }
          s = (offset + r > segment_size) ? segment_size - offset : r;
          memcpy(p, q + offset, s);
        }
      }
      break;
    case sen_io_wronly:
      break;
    default :
      return NULL;
    }
  }
  return iw->addr;
}

sen_rc
sen_io_win_unmap2(sen_io_win *iw)
{
  if (!iw || !iw->io ||!iw->ctx) { return sen_invalid_argument; }
  if (iw->cached) { return sen_success; }
  {
    sen_io *io = iw->io;
    sen_ctx *ctx = iw->ctx;
    switch (iw->mode) {
    case sen_io_rdonly:
      if (!iw->addr) { return sen_invalid_argument; }
      SEN_FREE(iw->addr);
      return sen_success;
    case sen_io_rdwr:
    case sen_io_wronly:
      {
        byte *p, *q;
        uint32_t segment_size = io->header->segment_size;
        uint32_t s, r, offset = iw->offset, segment = iw->segment;
        for (p = iw->addr, r = iw->size; r; p += s, r -= s, segment++, offset = 0) {
          SEN_IO_SEG_MAP(io, segment, q);
          if (!q) { return sen_memory_exhausted; }
          s = (offset + r > segment_size) ? segment_size - offset : r;
          memcpy(q + offset, p, s);
        }
      }
      SEN_FREE(iw->addr);
      return sen_success;
    default :
      return sen_invalid_argument;
    }
  }
}

#define DO_MAP(io,fmo,fi,pos,size,segno,res)\
{\
  if (((res) = SEN_MMAP(&sen_gctx, (fmo), (fi), (pos), (size)))) {\
    uint32_t nmaps;\
    if (io->max_map_seg < segno) { io->max_map_seg = segno; }\
    SEN_ATOMIC_ADD_EX(&io->nmaps, 1, nmaps);\
    if (!V08P(io)) {\
      uint64_t tail = io->base + (uint64_t)(size) * ((segno) + 1);\
      if (tail > io->header->curr_size) { io->header->curr_size = tail; }\
    }\
  }\
}

#define SEG_MAP(io,segno,info)\
{\
  uint32_t segment_size = io->header->segment_size;\
  if ((io->flags & TEMPORAL)) {\
    DO_MAP(io, &info->fmo, NULL, 0, segment_size, segno, info->map);\
  } else {\
    uint32_t segments_per_file = SEN_IO_FILE_SIZE / segment_size;\
    uint32_t bseg = segno + io->base_seg;\
    uint32_t fno = bseg / segments_per_file;\
    off_t base = fno ? 0 : io->base - (uint64_t)segment_size * io->base_seg;\
    off_t pos = (uint64_t)segment_size * (bseg % segments_per_file) + base;\
    fileinfo *fi = &io->fis[fno];\
    if (!sen_opened(fi)) {\
      char path[PATH_MAX];\
      gen_pathname(io->path, path, fno);\
      if (!sen_open(fi, path, O_RDWR|O_CREAT, SEN_IO_FILE_SIZE)) {\
        DO_MAP(io, &info->fmo, fi, pos, segment_size, segno, info->map);\
      }\
    } else {\
      DO_MAP(io, &info->fmo, fi, pos, segment_size, segno, info->map);\
    }\
  }\
}

void
sen_io_seg_map_(sen_io *io, uint32_t segno, sen_io_mapinfo *info)
{
  SEG_MAP(io, segno, info);
}

// deprecated
sen_rc
sen_io_seg_expire(sen_io *io, uint32_t segno, uint32_t nretry)
{
  uint32_t retry, *pnref;
  sen_io_mapinfo *info;
  if (!io->maps || segno >= io->header->max_segment) { return sen_invalid_argument; }
  info = &io->maps[segno];
  /* pnref = &io->nrefs[segno]; */
  pnref = &info->nref;
  for (retry = 0;; retry++) {
    uint32_t nref;
    SEN_ATOMIC_ADD_EX(pnref, 1, nref);
    if (nref) {
      SEN_ATOMIC_ADD_EX(pnref, -1, nref);
      if (retry >= SEN_IO_MAX_RETRY) {
        SEN_LOG(sen_log_crit, "deadlock detected! in sen_io_seg_expire(%p, %u, %u)", io, segno, nref);
        return sen_abnormal_error;
      }
    } else {
      SEN_ATOMIC_ADD_EX(pnref, SEN_IO_MAX_REF, nref);
      if (nref > 1) {
        SEN_ATOMIC_ADD_EX(pnref, -(SEN_IO_MAX_REF + 1), nref);
        if (retry >= SEN_IO_MAX_RETRY) {
          SEN_LOG(sen_log_crit, "deadlock detected!! in sen_io_seg_expire(%p, %u, %u)", io, segno, nref);
          return sen_abnormal_error;
        }
      } else {
        uint32_t nmaps;
        SEN_MUNMAP(&sen_gctx, &info->fmo, info->map, io->header->segment_size);
        info->map = NULL;
        SEN_ATOMIC_ADD_EX(pnref, -(SEN_IO_MAX_REF + 1), nref);
        SEN_ATOMIC_ADD_EX(&io->nmaps, -1, nmaps);
        return sen_success;
      }
    }
    if (retry >= nretry) { return sen_abnormal_error; }
    usleep(1000);
  }
}

uint32_t
sen_io_expire(sen_io *io, int count_thresh, uint32_t limit)
{
  uint32_t m, n = 0, ln = io->nmaps;
  if (io && io->maps) {
    sen_io_mapinfo *info = io->maps;
    for (m = io->max_map_seg; n < limit && m; info++, m--) {
      if (info->map && (sen_gtick - info->count) > count_thresh) {
        uint32_t nmaps, nref, *pnref = &info->nref;
        SEN_ATOMIC_ADD_EX(pnref, 1, nref);
        if (!nref && info->map && (sen_gtick - info->count) > count_thresh) {
          SEN_MUNMAP(&sen_gctx, &info->fmo, info->map, io->header->segment_size);
          SEN_ATOMIC_ADD_EX(&io->nmaps, -1, nmaps);
          info->map = NULL;
          info->count = sen_gtick;
          n++;
        }
        SEN_ATOMIC_ADD_EX(pnref, -1, nref);
      }
    }
  }
  SEN_LOG(sen_log_info, "expired:%08x max=%d (%d/%d)",
          sen_gtick, io->max_map_seg, n, ln);
  return n;
}

void *
sen_io_anon_map(sen_ctx *ctx, sen_io_mapinfo *mi, size_t length)
{
  return (mi->map = SEN_MMAP(ctx, &mi->fmo, NULL, 0, length));
}

void
sen_io_anon_unmap(sen_ctx *ctx, sen_io_mapinfo *mi, size_t length)
{
  /* support WIN32 */
  SEN_MUNMAP(ctx, &mi->fmo, mi->map, length);
}

sen_rc
sen_io_lock(sen_ctx *ctx, sen_io *io, int timeout)
{
  static int _ncalls = 0, _ncolls = 0;
  uint32_t count;
  _ncalls++;
  if (!io) { return sen_invalid_argument; }
  for (count = 0;; count++) {
    uint32_t lock;
    SEN_ATOMIC_ADD_EX(io->lock, 1, lock);
    if (lock) {
      SEN_ATOMIC_ADD_EX(io->lock, -1, lock);
      if (!timeout || (timeout > 0 && timeout == count)) { break; }
      if (!(++_ncolls % 1000000) && (_ncolls > _ncalls)) {
        if (_ncolls < 0 || _ncalls < 0) {
          _ncolls = 0; _ncalls = 0;
        } else {
          SEN_LOG(sen_log_notice, "io(%p) collisions(%d/%d)", io, _ncolls, _ncalls);
        }
      }
      usleep(1000);
      continue;
    }
    return sen_success;
  }
  return sen_other_error;
}

void
sen_io_unlock(sen_io *io)
{
  if (io) {
    uint32_t lock;
    SEN_ATOMIC_ADD_EX(io->lock, -1, lock);
  }
}

void
sen_io_clear_lock(sen_io *io)
{
  if (io) { *io->lock = 0; }
}

uint32_t
sen_io_is_locked(sen_io *io)
{
  return io ? *io->lock : 0;
}

/** mmap abstraction **/

static size_t mmap_size = 0;

#ifdef WIN32

#ifdef WIN32_FMO_EACH

inline static sen_rc
sen_open(fileinfo *fi, const char *path, int flags, size_t maxsize)
{
  fi->fh = CreateFile(path, GENERIC_READ | GENERIC_WRITE,
                      FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
                      OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0);
  if (fi->fh == INVALID_HANDLE_VALUE) {
    GERR(sen_file_operation_error, "CreateFile failed");
    return sen_file_operation_error;
  }
  if ((flags & O_TRUNC)) {
    CloseHandle(fi->fh);
    fi->fh = CreateFile(path, GENERIC_READ | GENERIC_WRITE,
                        FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
                        TRUNCATE_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
    if (fi->fh == INVALID_HANDLE_VALUE) {
      GERR(sen_file_operation_error, "CreateFile failed");
      return sen_file_operation_error;
    }
  }
  MUTEX_INIT(fi->mutex);
  return sen_success;
}

inline static void *
sen_mmap(sen_ctx *ctx, HANDLE *fmo, fileinfo *fi, off_t offset, size_t length)
{
  void *res;
  if (!fi) { return SEN_GCALLOC(length); }
  /* MUTEX_LOCK(fi->mutex); */
  /* try to create fmo */
  *fmo = CreateFileMapping(fi->fh, NULL, PAGE_READWRITE, 0, offset + length, NULL);
  if (!*fmo) { return NULL; }
  res = MapViewOfFile(*fmo, FILE_MAP_WRITE, 0, (DWORD)offset, (SIZE_T)length);
  if (!res) {
    sen_index_expire();
    res = MapViewOfFile(*fmo, FILE_MAP_WRITE, 0, (DWORD)offset, (SIZE_T)length);
    if (!res) {
      MERR("MapViewOfFile failed #%d <%d>", GetLastError(), mmap_size);
      return NULL;
    }
  }
  /* MUTEX_UNLOCK(fi->mutex); */
  mmap_size += length;
  return res;
}

inline static int
sen_munmap(sen_ctx *ctx, HANDLE *fmo, void *start, size_t length)
{
  int r = 0;
  if (UnmapViewOfFile(start)) {
    mmap_size -= length;
  } else {
    ERR(sen_other_error, "munmap(%p,%d) failed #%d <%d>", start, length, GetLastError(), mmap_size);
    r = -1;
  }
  if (*fmo) {
    if (!CloseHandle(*fmo)) {
      ERR(sen_other_error, "closehandle(%p,%d) failed #%d", start, length, GetLastError());
    }
    *fmo = NULL;
  } else {
    ERR(sen_other_error, "fmo not exists <%p,%d>", start, length);
  }
  return r;
}

inline static sen_rc
sen_close(fileinfo *fi)
{
  if (fi->fmo != NULL) {
    SEN_LOG(sen_log_alert, "file mapping object exists");
  }
  if (fi->fh != INVALID_HANDLE_VALUE) {
    CloseHandle(fi->fh);
    MUTEX_DESTROY(fi->mutex);
    fi->fh = INVALID_HANDLE_VALUE;
  }
  return sen_success;
}

#else /* WIN32_FMO_EACH */
inline static sen_rc
sen_open(fileinfo *fi, const char *path, int flags, size_t maxsize)
{
  /* may be wrong if flags is just only O_RDWR */
  fi->fh = CreateFile(path, GENERIC_READ | GENERIC_WRITE,
                      FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
                      OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0);
  if (fi->fh == INVALID_HANDLE_VALUE) {
    GERR(sen_file_operation_error, "CreateFile failed");
    return sen_file_operation_error;
  }
  if ((flags & O_TRUNC)) {
    CloseHandle(fi->fh);
    /* unable to assign OPEN_ALWAYS and TRUNCATE_EXISTING at once */
    fi->fh = CreateFile(path, GENERIC_READ | GENERIC_WRITE,
                        FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
                        TRUNCATE_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
    if (fi->fh == INVALID_HANDLE_VALUE) {
      GERR(sen_file_operation_error, "CreateFile failed");
      return sen_file_operation_error;
    }
  }
  /* signature may be wrong.. */
  fi->fmo = OpenFileMapping(FILE_MAP_ALL_ACCESS, FALSE, NULL);
  /* open failed */
  if (fi->fmo == NULL) {
    // flock
    /* retry to open */
    fi->fmo = OpenFileMapping(FILE_MAP_ALL_ACCESS, FALSE, NULL);
    /* failed again */
    if (fi->fmo == NULL) {
      /* try to create fmo */
      fi->fmo = CreateFileMapping(fi->fh, NULL, PAGE_READWRITE, 0, SEN_IO_FILE_SIZE, NULL);
    }
    // funlock
  }
  if (fi->fmo != NULL) {
    if (GetLastError() != ERROR_ALREADY_EXISTS ) {
      MUTEX_INIT(fi->mutex);
      return sen_success;
    } else {
      SEN_LOG(sen_log_error, "fmo object already exists! handle=%d", fi->fh);
      CloseHandle(fi->fmo);
    }
  } else {
    SEN_LOG(sen_log_alert, "failed to get FileMappingObject #%d", GetLastError());
  }
  CloseHandle(fi->fh);
  GERR(sen_file_operation_error, "OpenFileMapping failed");
  return sen_file_operation_error;
}

inline static void *
sen_mmap(sen_ctx *ctx, fileinfo *fi, off_t offset, size_t length)
{
  void *res;
  if (!fi) { return SEN_GCALLOC(length); }
  /* file must be exceeded to SEN_IO_FILE_SIZE when FileMappingObject created.
     and, after fmo created, it's not allowed to expand the size of file.
  DWORD tail = (DWORD)(offset + length);
  DWORD filesize = GetFileSize(fi->fh, NULL);
  if (filesize < tail) {
    if (SetFilePointer(fi->fh, tail, NULL, FILE_BEGIN) != tail) {
      sen_log("SetFilePointer failed");
      return NULL;
    }
    if (!SetEndOfFile(fi->fh)) {
      sen_log("SetEndOfFile failed");
      return NULL;
    }
    filesize = tail;
  }
  */
  res = MapViewOfFile(fi->fmo, FILE_MAP_WRITE, 0, (DWORD)offset, (SIZE_T)length);
  if (!res) {
    sen_index_expire();
    res = MapViewOfFile(fi->fmo, FILE_MAP_WRITE, 0, (DWORD)offset, (SIZE_T)length);
    if (!res) {
      MERR("MapViewOfFile failed #%d  <%d>", GetLastError(), mmap_size);
      return NULL;
    }
  }
  mmap_size += length;
  return res;
}

inline static int
sen_munmap(sen_ctx *ctx, void *start, size_t length)
{
  if (UnmapViewOfFile(start)) {
    mmap_size -= length;
    return 0;
  } else {
    ERR(sen_other_error, "munmap(%p,%d) failed <%d>", start, length, mmap_size);
    return -1;
  }
}

inline static sen_rc
sen_close(fileinfo *fi)
{
  if (fi->fmo != NULL) {
    CloseHandle(fi->fmo);
    fi->fmo = NULL;
  }
  if (fi->fh != INVALID_HANDLE_VALUE) {
    CloseHandle(fi->fh);
    MUTEX_DESTROY(fi->mutex);
    fi->fh = INVALID_HANDLE_VALUE;
  }
  return sen_success;
}
#endif /* WIN32_FMO_EACH */

inline static void
sen_fileinfo_init(fileinfo *fis, int nfis)
{
  for (; nfis--; fis++) {
    fis->fh = INVALID_HANDLE_VALUE;
    fis->fmo = NULL;
  }
}

inline static int
sen_opened(fileinfo *fi)
{
  return fi->fh != INVALID_HANDLE_VALUE;
}

inline static int
sen_msync(void *start, size_t length)
{
  /* return value may be wrong... */
  return FlushViewOfFile(start, length);
}

inline static sen_rc
sen_pread(fileinfo *fi, void *buf, size_t count, off_t offset)
{
  DWORD r, len;
  sen_rc rc = sen_success;
  MUTEX_LOCK(fi->mutex);
  r = SetFilePointer(fi->fh, offset, NULL, FILE_BEGIN);
  if (r == INVALID_SET_FILE_POINTER) {
    rc = sen_file_operation_error;
    SEN_LOG(sen_log_alert, "SetFilePointer error(%d)", GetLastError());
  } else {
    if (!ReadFile(fi->fh, buf, (DWORD)count, &len, NULL)) {
      rc = sen_file_operation_error;
      SEN_LOG(sen_log_alert, "ReadFile error(%d)", GetLastError());
    } else if (len != count) {
      SEN_LOG(sen_log_alert, "ReadFile %d != %d", count, len);
      rc = sen_file_operation_error;
    }
  }
  MUTEX_UNLOCK(fi->mutex);
  return rc;
}

inline static sen_rc
sen_pwrite(fileinfo *fi, void *buf, size_t count, off_t offset)
{
  DWORD r, len;
  sen_rc rc = sen_success;
  MUTEX_LOCK(fi->mutex);
  r = SetFilePointer(fi->fh, offset, NULL, FILE_BEGIN);
  if (r == INVALID_SET_FILE_POINTER) {
    rc = sen_file_operation_error;
    SEN_LOG(sen_log_alert, "SetFilePointer error(%d)", GetLastError());
  } else {
    if (!WriteFile(fi->fh, buf, (DWORD)count, &len, NULL)) {
      SEN_LOG(sen_log_alert, "WriteFile error(%d)", GetLastError());
      rc = sen_file_operation_error;
    } else if (len != count) {
      SEN_LOG(sen_log_alert, "WriteFile %d != %d", count, len);
      rc = sen_file_operation_error;
    }
  }
  MUTEX_UNLOCK(fi->mutex);
  return rc;
}

sen_id
sen_dl_open(sen_ctx *ctx, const char *filename)
{
  sen_id id;
  HMODULE dl, *dlp;
  sen_search_flags f = 0;
  if ((id = sen_hash_lookup(ctx, sen_dls, filename, strlen(filename),
                            (void **)&dlp, &f))) {
    return id;
  }
  if ((dl = LoadLibrary(filename))) {
    f = SEN_TABLE_ADD;
    if ((id = sen_hash_lookup(ctx, sen_dls, filename, strlen(filename),
                              (void *)&dlp, &f))) {
      *dlp = dl;
    } else {
      ERR(sen_other_error, "sen_hash_lookup failed");
      if (!FreeLibrary(dl)) {
        ERR(sen_other_error, "FreeLibrary return %d", GetLastError());
      }
    }
  } else {
    ERR(sen_other_error, "LoadLibrary failed %d", GetLastError());
  }
  return id;
}

sen_rc
sen_dl_close(sen_ctx *ctx, sen_id id)
{
  HMODULE *dlp;
  if (!sen_hash_get_value(ctx, sen_dls, id, &dlp)) {
    return sen_invalid_argument;
  }
  if (!FreeLibrary(*dlp)) {
    ERR(sen_other_error, "FreeLibrary return %d", GetLastError());
  }
  return sen_hash_delete_by_id(ctx, sen_dls, id, NULL);
}

void *
sen_dl_sym(sen_ctx *ctx, sen_id id, const char *symbol)
{
  HMODULE *dlp;
  FARPROC func;
  if (!sen_hash_get_value(ctx, sen_dls, id, &dlp)) {
    return sen_invalid_argument;
  }
  if (!(func = GetProcAddress(dlp, symbol))) {
    ERR(sen_other_error, "GetProcAddress failed %d", GetLastError());
  }
  return func;
}

#else /* WIN32 */

inline static sen_rc
sen_open(fileinfo *fi, const char *path, int flags, size_t maxsize)
{
  struct stat st;
  if ((fi->fd = open(path, flags, 0666)) == -1) {
    GSERR(path);
    return sen_file_operation_error;
  }
  if (fstat(fi->fd, &st) == -1) {
    GSERR(path);
    return sen_file_operation_error;
  }
  fi->dev = st.st_dev;
  fi->inode = st.st_ino;
  return sen_success;
}

inline static void
sen_fileinfo_init(fileinfo *fis, int nfis)
{
  for (; nfis--; fis++) { fis->fd = -1; }
}

inline static int
sen_opened(fileinfo *fi)
{
  return fi->fd != -1;
}

inline static sen_rc
sen_close(fileinfo *fi)
{
  if (fi->fd != -1) {
    if (close(fi->fd) == -1) {
      GSERR("close");
      return sen_file_operation_error;
    }
    fi->fd = -1;
  }
  return sen_success;
}

#if defined(MAP_ANON) && !defined(MAP_ANONYMOUS)
#define MAP_ANONYMOUS MAP_ANON
#endif

#include <sys/mman.h>

inline static void *
sen_mmap(sen_ctx *ctx, fileinfo *fi, off_t offset, size_t length)
{
  void *res;
  int fd, flags;
  if (fi) {
    struct stat s;
    off_t tail = offset + length;
    fd = fi->fd;
    if ((fstat(fd, &s) == -1) || (s.st_size < tail && ftruncate(fd, tail) == -1)) {
      ERR(sen_other_error, "fstat or ftruncate failed %d", fd);
      return NULL;
    }
    flags = MAP_SHARED;
  } else {
    fd = -1;
    flags = MAP_PRIVATE|MAP_ANONYMOUS;
  }
  res = mmap(NULL, length, PROT_READ|PROT_WRITE, flags, fd, offset);
  if (MAP_FAILED == res) {
    sen_index_expire();
    res = mmap(NULL, length, PROT_READ|PROT_WRITE, flags, fd, offset);
    if (MAP_FAILED == res) {
      MERR("mmap(%zu,%d,%d)=%s <%zu>", length, fd, offset, strerror(errno), mmap_size);
    }
    return NULL;
  }
  mmap_size += length;
  return res;
}

#ifdef USE_FAIL_MALLOC
int
fail_malloc_check(size_t size, const char *file, int line, const char *func);
inline static void *
sen_fail_mmap(sen_ctx *ctx, fileinfo *fi, off_t offset, size_t length,
              const char* file, int line, const char *func)
{
  if (fail_malloc_check(length, file, line, func)) {
    return sen_mmap(ctx, fi, offset, length);
  } else {
    sen_index_expire();
    MERR("fail_mmap(%zu,%d,%llu) (%s:%d@%s) <%zu>",
          length, fi ? fi->fd : 0, offset, file, line, func, mmap_size);
    return NULL;
  }
}
#endif /* USE_FAIL_MALLOC */

inline static int
sen_msync(void *start, size_t length)
{
  int r = msync(start, length, MS_SYNC);
  if (r == -1) { GSERR("msync"); }
  return r;
}

inline static int
sen_munmap(sen_ctx *ctx, void *start, size_t length)
{
  int res;
  res = munmap(start, length);
  if (res) {
    ERR(sen_other_error, "munmap(%p,%zu) failed <%zu>", start, length, mmap_size);
  } else {
    mmap_size -= length;
  }
  return res;
}

inline static sen_rc
sen_pread(fileinfo *fi, void *buf, size_t count, off_t offset)
{
  ssize_t r = pread(fi->fd, buf, count, offset);
  if (r != count) {
    if (r == -1) {
      GSERR("pread");
    } else {
      GERR(sen_file_operation_error, "pread returned %d != %d", r, count);
    }
    return sen_file_operation_error;
  }
  return sen_success;
}

inline static sen_rc
sen_pwrite(fileinfo *fi, void *buf, size_t count, off_t offset)
{
  ssize_t r = pwrite(fi->fd, buf, count, offset);
  if (r != count) {
    if (r == -1) {
      GSERR("pwrite");
    } else {
      GERR(sen_file_operation_error, "pwrite returned %d != %d", r, count);
    }
    return sen_file_operation_error;
  }
  return sen_success;
}

#include <dlfcn.h>

sen_id
sen_dl_open(sen_ctx *ctx, const char *filename)
{
  sen_id id;
  void *dl, **dlp;
  sen_search_flags f = 0;
  if ((id = sen_hash_lookup(ctx, sen_dls, filename, strlen(filename),
                            (void **)&dlp, &f))) {
    return id;
  }
  if ((dl = dlopen(filename, 0))) {
    f = SEN_TABLE_ADD;
    if ((id = sen_hash_lookup(ctx, sen_dls, filename, strlen(filename),
                              (void **)&dlp, &f))) {
      *dlp = dl;
    } else {
      int r;
      ERR(sen_other_error, "sen_hash_lookup failed");
      if ((r = dlclose(dl))) {
        ERR(sen_other_error, "dlclose return %d (%s)", r, dlerror());
      }
    }
  } else {
    ERR(sen_other_error, "dlopen failed (%s)", dlerror());
  }
  return id;
}

sen_rc
sen_dl_close(sen_ctx *ctx, sen_id id)
{
  int r;
  void **dlp;
  if (!sen_hash_get_value(ctx, sen_dls, id, &dlp)) {
    return sen_invalid_argument;
  }
  if ((r = dlclose(*dlp))) {
    ERR(sen_other_error, "dlclose return %d (%s)", r, dlerror());
  }
  return sen_hash_delete_by_id(ctx, sen_dls, id, NULL);
}

void *
sen_dl_sym(sen_ctx *ctx, sen_id id, const char *symbol)
{
  char *msg;
  void *func, **dlp;
  if (!sen_hash_get_value(ctx, sen_dls, id, &dlp)) {
    return NULL;
  }
  dlerror(); /* clear */
  func = dlsym(*dlp, symbol);
  msg = dlerror();
  if (msg) {
    ERR(sen_other_error, "dlsym failed (%s)", msg);
  }
  return func;
}

#endif /* WIN32 */
