/* Copyright(C) 2008 Brazil

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "senna_in.h"
#include <string.h>
#include <limits.h>
#include "ctx.h"
#include "pat.h"

#define SEN_IDSTR "SENNA:PAT:01.00"
#define SEN_PAT_DELETED (SEN_ID_MAX + 1)

#define SEN_PAT_SEGMENT_SIZE 0x400000
#define W_OF_KEY_IN_A_SEGMENT 22
#define W_OF_PAT_IN_A_SEGMENT 18
#define W_OF_SIS_IN_A_SEGMENT 19
#define KEY_MASK_IN_A_SEGMENT 0x3fffff
#define PAT_MASK_IN_A_SEGMENT 0x3ffff
#define SIS_MASK_IN_A_SEGMENT 0x7ffff
#define SEG_NOT_ASSIGNED 0xffff
#define SEN_PAT_MAX_SEGMENT 0x1000
#define SEN_PAT_NDELINFOS 0x100
#define SEN_PAT_MDELINFOS (SEN_PAT_NDELINFOS - 1)

#define SEN_PAT_BIN_KEY 0x70000

typedef struct {
  sen_id d;
  sen_id ld;
  uint32_t stat;
  uint32_t shared;
} sen_pat_delinfo;

struct sen_pat_header {
  char idstr[16];
  uint32_t flags;
  sen_encoding encoding;
  uint32_t key_size;
  uint32_t value_size;
  uint32_t nrecords;
  uint32_t curr_rec;
  int32_t curr_key;
  int32_t curr_del;
  int32_t curr_del2;
  int32_t curr_del3;
  uint8_t ngram_unit;
  uint32_t reserved[1007];
  sen_pat_delinfo delinfos[SEN_PAT_NDELINFOS];
  sen_id garbages[SEN_PAT_MAX_KEY_SIZE + 1];
};

typedef struct {
  sen_id lr[2];
  uint32_t key;
  uint16_t check;
  int16_t bits;
  /* length : 13, deleting : 1, immediate : 1 */
} pat_node;

#define PAT_DELETING  (1<<1)
#define PAT_IMMEDIATE (1<<2)

#define PAT_DEL(x) ((x)->bits & PAT_DELETING)
#define PAT_IMD(x) ((x)->bits & PAT_IMMEDIATE)
#define PAT_LEN(x) (((x)->bits >> 3) + 1)
#define PAT_CHK(x) ((x)->check)
#define PAT_DEL_ON(x) ((x)->bits |= PAT_DELETING)
#define PAT_IMD_ON(x) ((x)->bits |= PAT_IMMEDIATE)
#define PAT_DEL_OFF(x) ((x)->bits &= ~PAT_DELETING)
#define PAT_IMD_OFF(x) ((x)->bits &= ~PAT_IMMEDIATE)
#define PAT_LEN_SET(x,v) ((x)->bits = ((x)->bits & ((1<<3) - 1))|(((v) - 1) << 3))
#define PAT_CHK_SET(x,v) ((x)->check = (v))

typedef struct {
  sen_id children;
  sen_id sibling;
} sis_node;

enum {
  segment_key = 0,
  segment_pat = 1,
  segment_sis = 2
};

/* bit operation */

#define nth_bit(key,n,l) ((((key)[(n)>>4]) >> (7 - (((n)>>1) & 7))) & 1)

/* segment operation */

/* patricia array operation */

#define PAT_AT(pat,id,n) {\
  int flags = 0;\
  SEN_IO_ARRAY_AT(pat->io, segment_pat, id, &flags, n);\
}

inline static pat_node *
pat_get(sen_pat *pat, sen_id id)
{
  pat_node *res;
  int flags = SEN_TABLE_ADD;
  if (id > SEN_ID_MAX) { return NULL; }
  SEN_IO_ARRAY_AT(pat->io, segment_pat, id, &flags, res);
  return res;
}

inline static pat_node *
pat_node_new(sen_pat *pat, sen_id *id)
{
  uint32_t n = pat->header->curr_rec + 1;
  pat_node *res;
  if (n > SEN_ID_MAX) { return NULL; }
  if ((res = pat_get(pat, n))) {
    pat->header->curr_rec = n;
    pat->header->nrecords++;
  }
  if (id) { *id = n; }
  return res;
}

/* sis operation */

inline static sis_node *
sis_at(sen_pat *pat, sen_id id)
{
  sis_node *res;
  int flags = 0;
  if (id > SEN_ID_MAX) { return NULL; }
  SEN_IO_ARRAY_AT(pat->io, segment_sis, id, &flags, res);
  return res;
}

inline static sis_node *
sis_get(sen_pat *pat, sen_id id)
{
  sis_node *res;
  int flags = SEN_TABLE_ADD;
  if (id > SEN_ID_MAX) { return NULL; }
  SEN_IO_ARRAY_AT(pat->io, segment_sis, id, &flags, res);
  return res;
}

#define MAX_LEVEL 16

static void
sis_collect(sen_ctx *ctx, sen_pat *pat, sen_hash *h, sen_id id, uint32_t level)
{
  uint32_t *offset;
  sis_node *sl = sis_at(pat, id);
  if (sl) {
    sen_search_flags flags;
    sen_id sid = sl->children;
    while (sid && sid != id) {
      flags = SEN_TABLE_ADD;
      if (sen_hash_lookup(ctx, h, &sid, sizeof(sen_id), (void **) &offset, &flags)) {
        *offset = level;
        if (level < MAX_LEVEL) { sis_collect(ctx, pat, h, sid, level + 1); }
        if (!(sl = sis_at(pat, sid))) { break; }
        sid = sl->sibling;
      } else {
        /* todo : must be handled */
      }
    }
  }
}

/* key operation */

#define KEY_AT(pat,pos,ptr,addp)\
{\
  int flags = addp;\
  SEN_IO_ARRAY_AT(pat->io, segment_key, pos, &flags, ptr);\
}

inline static uint32_t
key_put(sen_pat *pat, const uint8_t *key, int len)
{
  uint32_t res, ts;
//  if (len >= SEN_PAT_SEGMENT_SIZE) { return 0; /* error */ }
  res = pat->header->curr_key;
  ts = (res + len) >> W_OF_KEY_IN_A_SEGMENT;
  if (res >> W_OF_KEY_IN_A_SEGMENT != ts) {
    res = pat->header->curr_key = ts << W_OF_KEY_IN_A_SEGMENT;
  }
  {
    uint8_t *dest;
    KEY_AT(pat, res, dest, SEN_TABLE_ADD);
    if (!dest) { return 0; }
    memcpy(dest, key, len);
  }
  pat->header->curr_key += len;
  return res;
}

inline static uint8_t *
pat_node_get_key(sen_pat *pat, pat_node *n)
{
  if (PAT_IMD(n)) {
    return (uint8_t *) &n->key;
  } else {
    uint8_t *res;
    KEY_AT(pat, n->key, res, 0);
    return res;
  }
}

inline static sen_rc
pat_node_set_key(sen_pat *pat, pat_node *n, const uint8_t *key, unsigned int len)
{
  if (!key || !len) { return sen_invalid_argument; }
  PAT_LEN_SET(n, len);
  if (len <= sizeof(uint32_t)) {
    PAT_IMD_ON(n);
    memcpy(&n->key, key, len);
  } else {
    PAT_IMD_OFF(n);
    n->key = key_put(pat, key, len);
  }
  return sen_success;
}

/* delinfo operation */

enum {
  DL_EMPTY = 0,
  DL_PHASE1,
  DL_PHASE2
};

inline static sen_pat_delinfo *
delinfo_search(sen_pat *pat, sen_id id)
{
  int i;
  sen_pat_delinfo *di;
  for (i = (pat->header->curr_del2) & SEN_PAT_MDELINFOS;
       i != pat->header->curr_del;
       i = (i + 1) & SEN_PAT_MDELINFOS) {
    di = &pat->header->delinfos[i];
    if (di->stat != DL_PHASE1) { continue; }
    if (di->ld == id) { return di; }
    if (di->d == id) { return di; }
  }
  return NULL;
}

inline static sen_rc
delinfo_turn_2(sen_pat *pat, sen_pat_delinfo *di)
{
  sen_id d, *p = NULL;
  pat_node *ln, *dn;
  // sen_log("delinfo_turn_2> di->d=%d di->ld=%d stat=%d", di->d, di->ld, di->stat);
  if (di->stat != DL_PHASE1) { return sen_success; }
  PAT_AT(pat, di->ld, ln);
  if (!ln) { return sen_invalid_argument; }
  if (!(d = di->d)) { return sen_invalid_argument; }
  PAT_AT(pat, d, dn);
  if (!dn) { return sen_invalid_argument; }
  PAT_DEL_OFF(ln);
  PAT_DEL_OFF(dn);
  {
    sen_id r, *p0;
    pat_node *rn;
    int c0 = -1, c;
    uint32_t len = PAT_LEN(dn) * 16;
    const uint8_t *key = pat_node_get_key(pat, dn);
    if (!key) { return sen_invalid_argument; }
    PAT_AT(pat, 0, rn);
    p0 = &rn->lr[1];
    while ((r = *p0)) {
      if (r == d) {
        p = p0;
        break;
      }
      PAT_AT(pat, r, rn);
      if (!rn) { return sen_invalid_format; }
      c = PAT_CHK(rn);
      if (c <= c0 || len <= c) { break; }
      if (c & 1) {
        p0 = (c + 1 < len) ? &rn->lr[1] : &rn->lr[0];
      } else {
        p0 = &rn->lr[nth_bit((uint8_t *)key, c, len)];
      }
      c0 = c;
    }
  }
  if (p) {
    PAT_CHK_SET(ln, PAT_CHK(dn));
    ln->lr[1] = dn->lr[1];
    ln->lr[0] = dn->lr[0];
    *p = di->ld;
  } else {
    /* debug */
    int j;
    sen_id dd;
    sen_pat_delinfo *ddi;
    SEN_LOG(sen_log_debug, "failed to find d=%d", d);
    for (j = (pat->header->curr_del2 + 1) & SEN_PAT_MDELINFOS;
         j != pat->header->curr_del;
         j = (j + 1) & SEN_PAT_MDELINFOS) {
      ddi = &pat->header->delinfos[j];
      if (ddi->stat != DL_PHASE1) { continue; }
      PAT_AT(pat, ddi->ld, ln);
      if (!ln) { continue; }
      if (!(dd = ddi->d)) { continue; }
      if (d == ddi->ld) {
        SEN_LOG(sen_log_debug, "found!!!, d(%d) become ld of (%d)", d, dd);
      }
    }
    /* debug */
  }
  di->stat = DL_PHASE2;
  di->d = d;
  // sen_log("delinfo_turn_2< di->d=%d di->ld=%d", di->d, di->ld);
  return sen_success;
}

inline static sen_rc
delinfo_turn_3(sen_pat *pat, sen_pat_delinfo *di)
{
  pat_node *dn;
  uint32_t size;
  if (di->stat != DL_PHASE2) { return sen_success; }
  PAT_AT(pat, di->d, dn);
  if (!dn) { return sen_invalid_argument; }
  if (di->shared) {
    PAT_IMD_ON(dn);
    size = 0;
  } else {
    if (PAT_IMD(dn)) {
      size = 0;
    } else {
      size = PAT_LEN(dn);
    }
  }
  di->stat = DL_EMPTY;
  //  dn->lr[1] = SEN_PAT_DELETED;
  dn->lr[0] = pat->header->garbages[size];
  pat->header->garbages[size] = di->d;
  return sen_success;
}

inline static sen_pat_delinfo *
delinfo_new(sen_pat *pat)
{
  sen_pat_delinfo *res = &pat->header->delinfos[pat->header->curr_del];
  uint32_t n = (pat->header->curr_del + 1) & SEN_PAT_MDELINFOS;
  int gap = ((n + SEN_PAT_NDELINFOS - pat->header->curr_del2) & SEN_PAT_MDELINFOS)
            - (SEN_PAT_NDELINFOS / 2);
  while (gap-- > 0) {
    if (delinfo_turn_2(pat, &pat->header->delinfos[pat->header->curr_del2])) {
      SEN_LOG(sen_log_crit, "d2 failed: %d", pat->header->delinfos[pat->header->curr_del2].ld);
    }
    pat->header->curr_del2 = (pat->header->curr_del2 + 1) & SEN_PAT_MDELINFOS;
  }
  if (n == pat->header->curr_del3) {
    if (delinfo_turn_3(pat, &pat->header->delinfos[pat->header->curr_del3])) {
      SEN_LOG(sen_log_crit, "d3 failed: %d", pat->header->delinfos[pat->header->curr_del3].ld);
    }
    pat->header->curr_del3 = (pat->header->curr_del3 + 1) & SEN_PAT_MDELINFOS;
  }
  pat->header->curr_del = n;
  return res;
}

/* pat operation */

sen_pat *
sen_pat_create(sen_ctx *ctx, const char *path, uint32_t key_size,
               uint32_t value_size, uint32_t flags, sen_encoding encoding)
{
  sen_io *io;
  sen_pat *pat;
  pat_node *node0;
  struct sen_pat_header *header;
  uint32_t entry_size, w_of_element;
  if (flags & SEN_OBJ_KEY_WITH_SIS) {
    entry_size = sizeof(sis_node) + value_size;
  } else {
    entry_size = value_size;
  }
  for (w_of_element = 0; (1 << w_of_element) < entry_size; w_of_element++);
  {
    sen_io_array_spec array_spec[3];
    array_spec[segment_key].w_of_element = 0;
    array_spec[segment_key].max_n_segments = 0x400;
    array_spec[segment_pat].w_of_element = 4;
    array_spec[segment_pat].max_n_segments = 1 << (30 - (22 - 4));
    array_spec[segment_sis].w_of_element = w_of_element;
    array_spec[segment_sis].max_n_segments = 1 << (30 - (22 - w_of_element));
    io = sen_io_create_with_array(path, sizeof(struct sen_pat_header),
                                  SEN_PAT_SEGMENT_SIZE, sen_io_auto, 3, array_spec);
  }
  if (!io) { return NULL; }
  if (encoding == sen_enc_default) { encoding = sen_gctx.encoding; }
  header = sen_io_header(io);
  memcpy(header->idstr, SEN_IDSTR, 16);
  sen_io_set_type(io, SEN_TABLE_PAT_KEY);
  header->flags = flags;
  header->encoding = encoding;
  header->key_size = key_size;
  header->value_size = value_size;
  header->nrecords = 0;
  header->curr_rec = 0;
  header->curr_key = -1;
  header->curr_del = 0;
  header->curr_del2 = 0;
  header->curr_del3 = 0;
  header->ngram_unit = SEN_TABLE_DEFAULT_NGRAM_UNIT_SIZE;
  if (!(pat = SEN_MALLOC(sizeof(sen_pat)))) {
    sen_io_close(io);
    return NULL;
  }
  SEN_OBJ_DB_SET_TYPE(pat, SEN_TABLE_PAT_KEY);
  pat->io = io;
  pat->header = header;
  pat->key_size = key_size;
  pat->value_size = value_size;
  pat->ngram_unit = header->ngram_unit;
  pat->encoding = encoding;
  pat->obj.flags = flags;
  if (!(node0 = pat_get(pat, 0))) {
    sen_io_close(io);
    SEN_FREE(pat);
    return NULL;
  }
  node0->lr[1] = 0;
  node0->lr[0] = 0;
  node0->key = 0;
  return pat;
}

sen_pat *
sen_pat_open(sen_ctx *ctx, const char *path)
{
  sen_io *io;
  sen_pat *pat;
  pat_node *node0;
  struct sen_pat_header *header;
  io = sen_io_open(path, sen_io_auto);
  if (!io) { return NULL; }
  header = sen_io_header(io);
  if (memcmp(header->idstr, SEN_IDSTR, 16)) {
    SEN_LOG(sen_log_notice, "invalid pat file. pat_idstr (%s)", header->idstr);
    sen_io_close(io);
    return NULL;
  }
  if (!(pat = SEN_MALLOC(sizeof(sen_pat)))) {
    sen_io_close(io);
    return NULL;
  }
  SEN_OBJ_DB_SET_TYPE(pat, SEN_TABLE_PAT_KEY);
  pat->io = io;
  pat->header = header;
  pat->key_size = header->key_size;
  pat->value_size = header->value_size;
  pat->encoding = header->encoding;
  pat->obj.flags = header->flags;
  pat->ngram_unit = header->ngram_unit;
  PAT_AT(pat, 0, node0);
  if (!node0) {
    sen_io_close(io);
    SEN_GFREE(pat);
    return NULL;
  }
  return pat;
}

sen_rc
sen_pat_close(sen_ctx *ctx, sen_pat *pat)
{
  sen_rc rc;
  if ((rc = sen_io_close(pat->io))) {
    ERR(rc, "sen_io_close failed");
  } else {
    SEN_FREE(pat);
  }
  return rc;
}

sen_rc
sen_pat_remove(sen_ctx *ctx, const char *path)
{
  if (!path) {
    ERR(sen_invalid_argument, "path is null");
    return sen_invalid_argument;
  }
  return sen_io_remove(path);
}

inline static sen_id
_sen_pat_get(sen_pat *pat, const uint8_t *key, uint32_t size, uint32_t *new, uint32_t *lkey)
{
  sen_id r, r0, *p0, *p1 = NULL;
  pat_node *rn, *rn0;
  int c, c0 = -1, c1 = -1, len;
  *new = 0;
  len = (int)size * 16;
  PAT_AT(pat, 0, rn0);
  p0 = &rn0->lr[1];
  if (*p0) {
    uint32_t size2;
    int xor, mask;
    const uint8_t *s, *d;
    for (;;) {
      if (!(r0 = *p0)) {
        if (!(s = pat_node_get_key(pat, rn0))) { return 0; }
        size2 = PAT_LEN(rn0);
        break;
      }
      PAT_AT(pat, r0, rn0);
      if (!rn0) { return SEN_ID_NIL; }
      if (c0 < rn0->check && rn0->check < len) {
        c1 = c0; c0 = rn0->check;
        p1 = p0;
        if (c0 & 1) {
          p0 = (c0 + 1 < len) ? &rn0->lr[1] : &rn0->lr[0];
        } else {
          p0 = &rn0->lr[nth_bit(key, c0, len)];
        }
      } else {
        if (!(s = pat_node_get_key(pat, rn0))) { return 0; }
        size2 = PAT_LEN(rn0);
        if (size == size2 && !memcmp(s, key, size)) { return r0; }
        break;
      }
    }
    {
      uint32_t min = size > size2 ? size2 : size;
      for (c = 0, d = key; min && *s == *d; c += 16, s++, d++, min--);
      if (min) {
        for (xor = *s ^ *d, mask = 0x80; !(xor & mask); mask >>= 1, c += 2);
      } else {
        c--;
      }
    }
    if (c == c0 && !*p0) {
      if (c < len - 2) { c += 2; }
    } else {
      if (c < c0) {
        if (c > c1) {
          p0 = p1;
        } else {
          PAT_AT(pat, 0, rn0);
          p0 = &rn0->lr[1];
          while ((r0 = *p0)) {
            PAT_AT(pat, r0, rn0);
            if (!rn0) { return 0; }
            c0 = PAT_CHK(rn0);
            if (c < c0) { break; }
            if (c0 & 1) {
              p0 = (c0 + 1 < len) ? &rn0->lr[1] : &rn0->lr[0];
            } else {
              p0 = &rn0->lr[nth_bit(key, c0, len)];
            }
          }
        }
      }
    }
    if (c >= len) { return 0; }
  } else {
    c = len - 2;
  }
  {
    uint32_t size2 = size > sizeof(uint32_t) ? size : 0;
    if (*lkey && size2) {
      if (pat->header->garbages[0]) {
        r = pat->header->garbages[0];
        pat->header->nrecords++;
        PAT_AT(pat, r, rn);
        if (!rn) { return 0; }
        pat->header->garbages[0] = rn->lr[0];
      } else {
        if (!(rn = pat_node_new(pat, &r))) { return 0; }
      }
      PAT_IMD_OFF(rn);
      PAT_LEN_SET(rn, size);
      rn->key = *lkey;
    } else {
      if (pat->header->garbages[size2]) {
        uint8_t *keybuf;
        r = pat->header->garbages[size2];
        pat->header->nrecords++;
        PAT_AT(pat, r, rn);
        if (!rn) { return 0; }
        pat->header->garbages[size2] = rn->lr[0];
        if (!(keybuf = pat_node_get_key(pat, rn))) { return 0; }
        PAT_LEN_SET(rn, size);
        memcpy(keybuf, key, size);
      } else {
        if (!(rn = pat_node_new(pat, &r))) { return 0; }
        pat_node_set_key(pat, rn, key, size);
      }
      *lkey = rn->key;
    }
  }
  PAT_CHK_SET(rn, c);
  PAT_DEL_OFF(rn);
  if ((c & 1) ? (c + 1 < len) : nth_bit(key, c, len)) {
    rn->lr[1] = r;
    rn->lr[0] = *p0;
  } else {
    rn->lr[1] = *p0;
    rn->lr[0] = r;
  }
  *p0 = r;
  *new = 1;
  return r;
}

inline static int
chop(sen_pat *pat, const char **key, const char *end, uint32_t *lkey)
{
  size_t len = sen_str_charlen_nonnull(*key, end, pat->encoding);
  if (len) {
    *lkey += len;
    *key += len;
    return **key;
  } else {
    return 0;
  }
}

#define MAX_FIXED_KEY_SIZE (sizeof(int64_t))

#define KEY_NEEDS_CONVERT(pat,size) \
  (!((pat)->obj.flags & SEN_OBJ_KEY_VAR_SIZE) && (size) <= MAX_FIXED_KEY_SIZE)

#define KEY_ENC(pat,keybuf,key,size) {\
  switch ((pat)->obj.flags & SEN_OBJ_KEY_MASK) {\
  case SEN_OBJ_KEY_UINT :\
    sen_hton((keybuf), (key), (size));\
    break;\
  case SEN_OBJ_KEY_INT :\
    sen_hton((keybuf), (key), (size));\
    *((uint8_t *)(keybuf)) ^= 0x80;\
    break;\
  case SEN_OBJ_KEY_FLOAT :\
    if ((size) == sizeof(int64_t)) {\
      int64_t v = *(int64_t *)(key);\
      v ^= ((v >> 63)|(1LL << 63));\
      sen_hton((keybuf), &v, (size));\
    }\
    break;\
  }\
}

#define KEY_DEC(pat,keybuf,key,size) {\
  switch ((pat)->obj.flags & SEN_OBJ_KEY_MASK) {\
  case SEN_OBJ_KEY_UINT :\
    sen_ntoh((keybuf), (key), (size));\
    break;\
  case SEN_OBJ_KEY_INT :\
    sen_ntohi((keybuf), (key), (size));\
    break;\
  case SEN_OBJ_KEY_FLOAT :\
    if ((size) == sizeof(int64_t)) {\
      int64_t v;\
      sen_hton(&v, (key), (size));\
      *((int64_t *)(keybuf)) = v ^ (((v^(1LL<<63))>> 63)|(1LL<<63));  \
    }\
    break;\
  }\
}

#define KEY_ENCODE(pat,keybuf,key,size) \
if (KEY_NEEDS_CONVERT(pat,size)) {\
  KEY_ENC((pat), (keybuf), (key), (size));\
  (key) = (keybuf);\
}

sen_id
sen_pat_get(sen_ctx *ctx, sen_pat *pat, const void *key, uint32_t key_size,
            void **value, sen_search_flags *flags)
{
  uint32_t new, lkey = 0;
  sen_id r0;
  uint8_t keybuf[MAX_FIXED_KEY_SIZE];
  KEY_ENCODE(pat, keybuf, key, key_size);
  r0 = _sen_pat_get(pat, (uint8_t *)key, key_size, &new, &lkey);
  if (new && flags) { *flags |= SEN_TABLE_ADDED; }
  if (r0 && (pat->obj.flags & SEN_OBJ_KEY_WITH_SIS) &&
      (*((uint8_t *)key) & 0x80)) { // todo: refine!!
    sis_node *sl, *sr;
    sen_id l = r0, r;
    if (new && (sl = sis_get(pat, l))) {
      const char *sis = key, *end = sis + key_size;
      sl->children = l;
      sl->sibling = 0;
      while (chop(pat, &sis, end, &lkey)) {
        if (!(*sis & 0x80)) { break; }
        if (!(r = _sen_pat_get(pat, (uint8_t *)sis, end - sis, &new, &lkey))) {
          break;
        }
        if (!(sr = sis_get(pat, r))) { break; }
        if (new) {
          sl->sibling = r;
          sr->children = l;
          sr->sibling = 0;
        } else {
          sl->sibling = sr->children;
          sr->children = l;
          break;
        }
        l = r;
        sl = sr;
      }
    }
  }
  if (r0 && value) {
    byte *v = (byte *)sis_get(pat, r0);
    if (pat->obj.flags & SEN_OBJ_KEY_WITH_SIS) {
      *value = v + sizeof(sis_node);
    } else {
      *value = v;
    }
  }
  return r0;
}

sen_id
sen_pat_at(sen_ctx *ctx, sen_pat *pat, const void *key, uint32_t key_size, void **value)
{
  sen_id r;
  pat_node *rn;
  int c0 = -1, c;
  uint32_t len = key_size * 16;
  uint8_t keybuf[MAX_FIXED_KEY_SIZE];
  KEY_ENCODE(pat, keybuf, key, key_size);
  PAT_AT(pat, 0, rn);
  for (r = rn->lr[1]; r;) {
    PAT_AT(pat, r, rn);
    if (!rn) { break; /* corrupt? */ }
    c = PAT_CHK(rn);
    if (len <= c) { break; }
    if (c <= c0) {
      const uint8_t *k = pat_node_get_key(pat, rn);
      if (k && key_size == PAT_LEN(rn) && !memcmp(k, key, key_size)) {
        if (value) {
          byte *v = (byte *)sis_get(pat, r);
          if (pat->obj.flags & SEN_OBJ_KEY_WITH_SIS) {
            *value = v + sizeof(sis_node);
          } else {
            *value = v;
          }
        }
        return r;
      }
      break;
    }
    if (c & 1) {
      r = (c + 1 < len) ? rn->lr[1] : rn->lr[0];
    } else {
      r = rn->lr[nth_bit((uint8_t *)key, c, len)];
    }
    c0 = c;
  }
  return SEN_ID_NIL;
}

sen_id
sen_pat_lookup(sen_ctx *ctx, sen_pat *pat, const void *key, int key_size,
               void **value, sen_search_flags *flags)
{
  if (!pat || !key || !key_size || pat->key_size < key_size) { return SEN_ID_NIL; }
  if (*flags & SEN_TABLE_ADD) {
    return sen_pat_get(ctx, pat, key, key_size, value, flags);
  } else {
    return sen_pat_at(ctx, pat, key, key_size, value);
  }
}

sen_id
sen_pat_nextid(sen_ctx *ctx, sen_pat *pat, const void *key, uint32_t key_size)
{
  sen_id r = SEN_ID_NIL;
  if (pat && key) {
    if (!(r = pat->header->garbages[key_size > sizeof(uint32_t) ? key_size : 0])) {
      r = pat->header->curr_rec + 1;
    }
  }
  return r;
}

static void
get_tc(sen_ctx *ctx, sen_pat *pat, sen_hash *h, pat_node *rn)
{
  sen_id id;
  pat_node *node;
  sen_search_flags flags = SEN_TABLE_ADD;
  id = rn->lr[1];
  if (id) {
    PAT_AT(pat, id, node);
    if (node) {
      if (PAT_CHK(node) > PAT_CHK(rn)) {
        get_tc(ctx, pat, h, node);
      } else {
        sen_hash_lookup(ctx, h, &id, sizeof(sen_id), NULL, &flags);
      }
    }
  }
  id = rn->lr[0];
  if (id) {
    PAT_AT(pat, id, node);
    if (node) {
      if (PAT_CHK(node) > PAT_CHK(rn)) {
        get_tc(ctx, pat, h, node);
      } else {
        sen_hash_lookup(ctx, h, &id, sizeof(sen_id), NULL, &flags);
      }
    }
  }
}

sen_rc
sen_pat_prefix_search(sen_ctx *ctx, sen_pat *pat,
                      const void *key, uint32_t key_size, sen_hash *h)
{
  int c0 = -1, c;
  const uint8_t *k;
  uint32_t len = key_size * 16;
  sen_id r;
  pat_node *rn;
  uint8_t keybuf[MAX_FIXED_KEY_SIZE];
  KEY_ENCODE(pat, keybuf, key, key_size);
  PAT_AT(pat, 0, rn);
  r = rn->lr[1];
  while (r) {
    PAT_AT(pat, r, rn);
    if (!rn) { return sen_invalid_format; }
    c = PAT_CHK(rn);
    if (c0 < c && c < len - 1) {
      if (c & 1) {
        r = (c + 1 < len) ? rn->lr[1] : rn->lr[0];
      } else {
        r = rn->lr[nth_bit((uint8_t *)key, c, len)];
      }
      c0 = c;
      continue;
    }
    if (!(k = pat_node_get_key(pat, rn))) { break; }
    if (PAT_LEN(rn) < key_size) { break; }
    if (!memcmp(k, key, key_size)) {
      if (c >= len - 1) {
        get_tc(ctx, pat, h, rn);
      } else {
        sen_search_flags flags = SEN_TABLE_ADD;
        sen_hash_lookup(ctx, h, &r, sizeof(sen_id), NULL, &flags);
      }
      return sen_success;
    }
    break;
  }
  return sen_end_of_data;
}

sen_hash *
sen_pat_prefix_search2(sen_ctx *ctx, sen_pat *pat, const void *key, uint32_t key_size)
{
  sen_hash *h;
  if (!pat || !key) { return NULL; }
  if ((h = sen_hash_create(ctx, NULL, sizeof(sen_id), 0, 0, 0))) {
    if (sen_pat_prefix_search(ctx, pat, key, key_size, h)) {
      sen_hash_close(ctx, h);
      h = NULL;
    }
  }
  return h;
}

sen_rc
sen_pat_suffix_search(sen_ctx *ctx, sen_pat *pat,
                      const void *key, uint32_t key_size, sen_hash *h)
{
  sen_id r;
  if ((r = sen_pat_at(ctx, pat, key, key_size, NULL))) {
    uint32_t *offset;
    sen_search_flags flags = SEN_TABLE_ADD;
    if (sen_hash_lookup(ctx, h, &r, sizeof(sen_id), (void **) &offset, &flags)) {
      *offset = 0;
      if (pat->obj.flags & SEN_OBJ_KEY_WITH_SIS) { sis_collect(ctx, pat, h, r, 1); }
      return sen_success;
    }
  }
  return sen_end_of_data;
}

sen_hash *
sen_pat_suffix_search2(sen_ctx *ctx, sen_pat *pat, const void *key, uint32_t key_size)
{
  sen_hash *h;
  if (!pat || !key) { return NULL; }
  if ((h = sen_hash_create(ctx, NULL, sizeof(sen_id), sizeof(uint32_t), 0, 0))) {
    if (sen_pat_suffix_search(ctx, pat, key, key_size, h)) {
      sen_hash_close(ctx, h);
      h = NULL;
    }
  }
  return h;
}

sen_id
sen_pat_lcp_search(sen_ctx *ctx, sen_pat *pat, const void *key, uint32_t key_size)
{
  pat_node *rn;
  sen_id r, r2 = SEN_ID_NIL;
  uint32_t len = key_size * 16;
  int c0 = -1, c;
  if (!pat || !key || !(pat->obj.flags & SEN_OBJ_KEY_VAR_SIZE)) { return SEN_ID_NIL; }
  PAT_AT(pat, 0, rn);
  for (r = rn->lr[1]; r;) {
    PAT_AT(pat, r, rn);
    if (!rn) { break; /* corrupt? */ }
    c = PAT_CHK(rn);
    if (c <= c0) {
      if (PAT_LEN(rn) <= key_size) {
        uint8_t *p = pat_node_get_key(pat, rn);
        if (!p) { break; }
        if (!memcmp(p, key, PAT_LEN(rn))) { return r; }
      }
      break;
    }
    if (len <= c) { break; }
    if (c & 1) {
      uint8_t *p = pat_node_get_key(pat, rn);
      if (!p) { break; }
      if (PAT_LEN(rn) <= key_size && !memcmp(p, key, PAT_LEN(rn))) { r2 = r; }
      r = (c + 1 < len) ? rn->lr[1] : rn->lr[0];
    } else {
      r = rn->lr[nth_bit((uint8_t *)key, c, len)];
    }
    c0 = c;
  }
  return r2;
}

inline static sen_rc
__sen_pat_del(sen_ctx *ctx, sen_pat *pat, const char *key, uint32_t key_size, int shared,
             sen_table_delete_optarg *optarg)
{
  sen_pat_delinfo *di;
  uint8_t direction;
  pat_node *rn, *rn0 = NULL, *rno;
  int c, c0 = -1, ch;
  uint32_t len = key_size * 16;
  sen_id r, otherside, *p, *p0 = NULL;
  di = delinfo_new(pat); /* must be called before find rn */
  di->shared = shared;
  PAT_AT(pat, 0, rn);
  c = -1;
  p = &rn->lr[1];
  for (;;) {
    if (!(r = *p)) { return sen_invalid_argument; }
    PAT_AT(pat, r, rn);
    if (!rn) { return sen_invalid_format; }
    ch = PAT_CHK(rn);
    if (len <= ch) { return sen_invalid_argument; }
    if (c >= ch) {
      const uint8_t *k = pat_node_get_key(pat, rn);
      if (!k) { return sen_invalid_argument; }
      if (key_size == PAT_LEN(rn) && !memcmp(k, key, key_size)) {
        break; /* found */
      } else {  return sen_invalid_argument; }
    }
    c0 = c;
    p0 = p;
    if ((c = ch) & 1) {
      p = (c + 1 < len) ? &rn->lr[1] : &rn->lr[0];
    } else {
      p = &rn->lr[nth_bit((uint8_t *)key, c, len)];
    }
    rn0 = rn;
  }
  if (optarg && optarg->func &&
      !optarg->func(ctx, (sen_obj *)pat, r, optarg->func_arg)) {
    return sen_success;
  }
  direction = (rn0->lr[1] == r);
  otherside = direction ? rn0->lr[0] : rn0->lr[1];
  if (rn == rn0) {
    di->stat = DL_PHASE2;
    di->d = r;
    if (otherside) {
      PAT_AT(pat, otherside, rno);
      if (rno && c0 < PAT_CHK(rno) && PAT_CHK(rno) <= c) {
        if (!delinfo_search(pat, otherside)) {
          SEN_LOG(sen_log_error, "no delinfo found %d", otherside);
        }
        PAT_CHK_SET(rno, 0);
      }
    }
    *p0 = otherside;
  } else {
    sen_pat_delinfo *ldi = NULL, *ddi = NULL;
    if (PAT_DEL(rn)) { ldi = delinfo_search(pat, r); }
    if (PAT_DEL(rn0)) { ddi = delinfo_search(pat, *p0); }
    if (ldi) {
      PAT_DEL_OFF(rn);
      di->stat = DL_PHASE2;
      if (ddi) {
        PAT_DEL_OFF(rn0);
        ddi->stat = DL_PHASE2;
        if (ddi == ldi) {
          if (r != ddi->ld) {
            SEN_LOG(sen_log_error, "r(%d) != ddi->ld(%d)", r, ddi->ld);
          }
          di->d = r;
        } else {
          ldi->ld = ddi->ld;
          di->d = r;
        }
      } else {
        PAT_DEL_ON(rn0);
        ldi->ld = *p0;
        di->d = r;
      }
    } else {
      PAT_DEL_ON(rn);
      if (ddi) {
        if (ddi->d != *p0) {
          SEN_LOG(sen_log_error, "ddi->d(%d) != *p0(%d)", ddi->d, *p0);
        }
        PAT_DEL_OFF(rn0);
        ddi->stat = DL_PHASE2;
        di->stat = DL_PHASE1;
        di->ld = ddi->ld;
        di->d = r;
        /*
        PAT_DEL_OFF(rn0);
        ddi->d = r;
        di->stat = DL_PHASE2;
        di->d = *p0;
        */
      } else {
        PAT_DEL_ON(rn0);
        di->stat = DL_PHASE1;
        di->ld = *p0;
        di->d = r;
        // sen_log("pat_del d=%d ld=%d stat=%d", r, *p0, DL_PHASE1);
      }
    }
    if (*p0 == otherside) {
      PAT_CHK_SET(rn0, 0);
    } else {
      if (otherside) {
        PAT_AT(pat, otherside, rno);
        if (rno && c0 < PAT_CHK(rno) && PAT_CHK(rno) <= c) {
          if (!delinfo_search(pat, otherside)) {
            SEN_LOG(sen_log_error, "no delinfo found %d", otherside);
          }
          PAT_CHK_SET(rno, 0);
        }
      }
      *p0 = otherside;
    }
  }
  pat->header->nrecords--;
  return sen_success;
}

static sen_rc
_sen_pat_delete(sen_ctx *ctx, sen_pat *pat, const void *key, uint32_t key_size,
                sen_table_delete_optarg *optarg)
{
  if (pat->obj.flags & SEN_OBJ_KEY_WITH_SIS) {
    sen_id id = sen_pat_at(ctx, pat, key, key_size, NULL);
    if (id && sen_pat_delete_with_sis(ctx, pat, id, optarg)) {
      return sen_success;
    }
    return sen_invalid_argument;
  }
  return __sen_pat_del(ctx, pat, key, key_size, 0, optarg);
}

sen_rc
sen_pat_delete(sen_ctx *ctx, sen_pat *pat, const void *key, uint32_t key_size,
               sen_table_delete_optarg *optarg)
{
  uint8_t keybuf[MAX_FIXED_KEY_SIZE];
  if (!pat || !key || !key_size) { return sen_invalid_argument; }
  KEY_ENCODE(pat, keybuf, key, key_size);
  return _sen_pat_delete(ctx, pat, key, key_size, optarg);
}

uint32_t
sen_pat_size(sen_ctx *ctx, sen_pat *pat)
{
  if (!pat) { return sen_invalid_argument; }
  return pat->header->nrecords;
}

const char *
_sen_pat_key(sen_pat *pat, sen_id id, uint32_t *key_size)
{
  pat_node *node;
  PAT_AT(pat, id, node);
  if (!node) { return NULL; }
  *key_size = PAT_LEN(node);
  return (const char *) pat_node_get_key(pat, node);
}

sen_rc
sen_pat_delete_by_id(sen_ctx *ctx, sen_pat *pat, sen_id id,
                     sen_table_delete_optarg *optarg)
{
  if (!pat || !id) { return sen_invalid_argument; }
  {
    uint32_t key_size;
    const char *key = _sen_pat_key(pat, id, &key_size);
    return _sen_pat_delete(ctx, pat, key, key_size, optarg);
  }
}

int
sen_pat_get_key(sen_ctx *ctx, sen_pat *pat, sen_id id, void *keybuf, int bufsize)
{
  int len;
  uint8_t *key;
  pat_node *node;
  if (!pat) { return sen_invalid_argument; }
  PAT_AT(pat, id, node);
  if (!node) { return 0; }
  if (!(key = pat_node_get_key(pat, node))) { return 0; }
  len = PAT_LEN(node);
  if (keybuf && bufsize >= len) {
    if (KEY_NEEDS_CONVERT(pat, len)) {
      KEY_DEC(pat, keybuf, key, len);
    } else {
      memcpy(keybuf, key, len);
    }
  }
  return len;
}

int
sen_pat_get_key2(sen_ctx *ctx, sen_pat *pat, sen_id id, sen_obj *bulk)
{
  uint32_t len;
  uint8_t *key;
  pat_node *node;
  if (!pat) { return sen_invalid_argument; }
  PAT_AT(pat, id, node);
  if (!node) { return 0; }
  if (!(key = pat_node_get_key(pat, node))) { return 0; }
  len = PAT_LEN(node);
  if (KEY_NEEDS_CONVERT(pat, len)) {
    if (bulk->header.flags & SEN_OBJ_DO_SHALLOW_COPY) {
      SEN_OBJ_INIT(bulk, SEN_BULK, 0);
    }
    if (!sen_bulk_reserve(ctx, bulk, len)) {
      KEY_DEC(pat, bulk->u.b.curr, key, len);
      bulk->u.b.curr += len;
    }
  } else {
    if (bulk->header.flags & SEN_OBJ_DO_SHALLOW_COPY) {
      bulk->u.b.head = (char *)key;
      bulk->u.b.curr = (char *)key + len;
    } else {
      sen_bulk_write(ctx, bulk, (char *)key, len);
    }
  }
  return len;
}

int
sen_pat_get_value(sen_ctx *ctx, sen_pat *pat, sen_id id, void *valuebuf)
{
  int value_size = (int)pat->value_size;
  if (value_size) {
    byte *v = (byte *)sis_at(pat, id);
    if (v) {
      if (valuebuf) {
        if (pat->obj.flags & SEN_OBJ_KEY_WITH_SIS) {
          memcpy(valuebuf, v + sizeof(sis_node), value_size);
        } else {
          memcpy(valuebuf, v, value_size);
        }
      }
      return value_size;
    }
  }
  return 0;
}

const char *
sen_pat_get_value_(sen_ctx *ctx, sen_pat *pat, sen_id id, uint32_t *size)
{
  const char *value = NULL;
  if ((*size = pat->value_size)) {
    if ((value = (const char *)sis_at(pat, id))
        && (pat->obj.flags & SEN_OBJ_KEY_WITH_SIS)) {
      value += sizeof(sis_node);
    }
  }
  return value;
}

sen_rc
sen_pat_set_value(sen_ctx *ctx, sen_pat *pat, sen_id id, void *value, int flags)
{
  if (value) {
    uint32_t value_size = pat->value_size;
    if (value_size) {
      byte *v = (byte *)sis_get(pat, id);
      if (v) {
        if (pat->obj.flags & SEN_OBJ_KEY_WITH_SIS) { v += sizeof(sis_node); }
        switch ((flags & SEN_OBJ_SET_MASK)) {
        case SEN_OBJ_SET :
          memcpy(v, value, value_size);
          return sen_success;
        default :
          // todo : support other types.
          return sen_invalid_argument;
        }
      } else {
        return sen_memory_exhausted;
      }
    }
  }
  return sen_invalid_argument;
}

sen_rc
sen_pat_info(sen_ctx *ctx, sen_pat *pat, int *key_size, unsigned *flags,
             sen_encoding *encoding, unsigned *nrecords, unsigned *file_size)
{
  ERRCLR(NULL);
  if (!pat) { return sen_invalid_argument; }
  if (key_size) { *key_size = pat->key_size; }
  if (flags) { *flags = pat->obj.flags; }
  if (encoding) { *encoding = pat->encoding; }
  if (nrecords) { *nrecords = pat->header->nrecords; }
  if (file_size) {
    sen_rc rc;
    uint64_t tmp = 0;
    if ((rc = sen_io_size(pat->io, &tmp))) {
      return rc;
    }
    *file_size = (unsigned) tmp; /* FIXME: inappropriate cast */
  }
  return sen_success;
}

int
sen_pat_delete_with_sis(sen_ctx *ctx, sen_pat *pat, sen_id id,
                        sen_table_delete_optarg *optarg)
{
  int level = 0, shared;
  const char *key = NULL, *_key;
  sis_node *sp, *ss = NULL, *si = sis_at(pat, id);
  while (id) {
    pat_node *rn;
    uint32_t key_size;
    if ((si && si->children && si->children != id) ||
        (optarg && optarg->func &&
         !optarg->func(ctx, (sen_obj *)pat, id, optarg->func_arg))) {
      break;
    }
    PAT_AT(pat, id, rn);
    if (!(_key = (char *)pat_node_get_key(pat, rn))) { return 0; }
    if (_key == key) {
      shared = 1;
    } else {
      key = _key;
      shared = 0;
    }
    key_size = PAT_LEN(rn);
    if (key && key_size) { __sen_pat_del(ctx, pat, key, key_size, shared, NULL); }
    if (si) {
      sen_id *p, sid;
      uint32_t lkey = 0;
      if ((*key & 0x80) && chop(pat, &key, key + key_size, &lkey)) {
        if ((sid = sen_pat_at(ctx, pat, key, key_size - lkey, NULL)) &&
            (ss = sis_at(pat, sid))) {
          for (p = &ss->children; *p && *p != sid; p = &sp->sibling) {
            if (*p == id) {
              *p = si->sibling;
              break;
            }
            if (!(sp = sis_at(pat, *p))) { break; }
          }
        }
      } else {
        sid = SEN_ID_NIL;
      }
      si->sibling = 0;
      si->children = 0;
      id = sid;
      si = ss;
    } else {
      id = SEN_ID_NIL;
    }
    level++;
  }
  if (level) {
    uint32_t lkey = 0;
    while (id && key) {
      uint32_t key_size;
      if (_sen_pat_key(pat, id, &key_size) != key) { break; }
      {
        pat_node *rn;
        PAT_AT(pat, id, rn);
        if (!rn) { break; }
        if (lkey) {
          rn->key = lkey;
        } else {
          pat_node_set_key(pat, rn, (uint8_t *)key, key_size);
          lkey = rn->key;
        }
      }
      {
        const char *end = key + key_size;
        if (!((*key & 0x80) && chop(pat, &key, end, &lkey))) { break; }
        id = sen_pat_at(ctx, pat, key, end - key, NULL);
      }
    }
  }
  return level;
}

sen_id
sen_pat_next(sen_ctx *ctx, sen_pat *pat, sen_id id)
{
  while (++id <= pat->header->curr_rec) {
    uint32_t key_size;
    if (id == sen_pat_at(ctx, pat, _sen_pat_key(pat, id, &key_size), key_size, NULL)) {
      return id;
    }
  }
  return SEN_ID_NIL;
}

sen_id
sen_pat_curr_id(sen_ctx *ctx, sen_pat *pat)
{
  return pat->header->curr_rec;
}

int
sen_pat_scan(sen_ctx *ctx, sen_pat *pat, const char *str, unsigned int str_len,
             sen_pat_scan_hit *sh, unsigned int sh_size, const char **rest)
{
  int n = 0;
  sen_id tid;
  if (pat->obj.flags & SEN_OBJ_KEY_NORMALIZE) {
    sen_nstr *nstr = sen_nstr_open(str, str_len, pat->encoding, SEN_STR_WITH_CHECKS);
    if (nstr) {
      int16_t *cp = nstr->checks;
      unsigned int offset = 0, offset0 = 0;
      const char *sp = nstr->norm, *se = nstr->norm + nstr->norm_blen;
      while (n < sh_size) {
        if ((tid = sen_pat_lcp_search(ctx, pat, sp, se - sp))) {
          uint32_t len;
          _sen_pat_key(pat, tid, &len);
          sh[n].id = tid;
          sh[n].offset = (*cp > 0) ? offset : offset0;
          while (len--) {
            if (*cp > 0) { offset0 = offset; offset += *cp; }
            sp++; cp++;
          }
          sh[n].length = offset - sh[n].offset;
          n++;
        } else {
          if (*cp > 0) { offset0 = offset; offset += *cp; }
          do {
            sp++; cp++;
          } while (sp < se && !*cp);
        }
        if (se <= sp) { offset = str_len; break; }
      }
      if (rest) { *rest = nstr->orig + offset; }
      sen_nstr_close(nstr);
    } else {
      n = -1;
      if (rest) { *rest = str; }
    }
  } else {
    uint32_t len;
    const char *sp, *se = str + str_len;
    for (sp = str; sp < se && n < sh_size; sp += len) {
      if ((tid = sen_pat_lcp_search(ctx, pat, sp, se - sp))) {
        _sen_pat_key(pat, tid, &len);
        sh[n].id = tid;
        sh[n].offset = sp - str;
        sh[n].length = len;
        n++;
      } else {
        len = sen_str_charlen_nonnull(sp, se, pat->encoding);
      }
      if (!len) { break; }
    }
    if (rest) { *rest = sp; }
  }
  return n;
}

#define INITIAL_SIZE 512

inline static void
push(sen_pat_cursor *c, sen_id id, uint16_t check)
{
  sen_ctx *ctx = c->ctx;
  sen_pat_cursor_entry *se;
  if (c->size <= c->sp) {
    if (c->ss) {
      uint32_t size = c->size * 4;
      sen_pat_cursor_entry *ss = SEN_REALLOC(c->ss, size);
      if (!ss) { return; /* give up */ }
      c->ss = ss;
      c->size = size;
    } else {
      if (!(c->ss = SEN_MALLOC(sizeof(sen_pat_cursor_entry) * INITIAL_SIZE))) {
        return; /* give up */
      }
      c->size = INITIAL_SIZE;
    }
  }
  se = &c->ss[c->sp++];
  se->id = id;
  se->check = check;
}

inline static sen_pat_cursor_entry *
pop(sen_pat_cursor *c)
{
  return c->sp ? &c->ss[--c->sp] : NULL;
}

sen_id
sen_pat_cursor_next(sen_ctx *ctx, sen_pat_cursor *c)
{
  pat_node *node;
  sen_pat_cursor_entry *se;
  while ((se = pop(c))) {
    sen_id id = se->id;
    int check = se->check, ch;
    for (;;) {
      if (id) {
        PAT_AT(c->pat, id, node);
        if (node) {
          ch = PAT_CHK(node);
          if (ch > check) {
            if (c->obj.flags & SEN_CURSOR_ASCENDING) {
              push(c, node->lr[1], ch);
              id = node->lr[0];
            } else {
              push(c, node->lr[0], ch);
              id = node->lr[1];
            }
            check = ch;
            continue;
          } else {
            if (id == c->limit) {
              c->sp = 0;
            } else {
              if (!c->curr_rec && c->limit) {
                uint32_t lmin, lmax;
                pat_node *nmin, *nmax;
                const uint8_t *kmin, *kmax;
                if (c->obj.flags & SEN_CURSOR_ASCENDING) {
                  PAT_AT(c->pat, id, nmin);
                  PAT_AT(c->pat, c->limit, nmax);
                } else {
                  PAT_AT(c->pat, c->limit, nmin);
                  PAT_AT(c->pat, id, nmax);
                }
                lmin = PAT_LEN(nmin);
                lmax = PAT_LEN(nmax);
                kmin = pat_node_get_key(c->pat, nmin);
                kmax = pat_node_get_key(c->pat, nmax);
                if ((lmin < lmax) ?
                    (memcmp(kmin, kmax, lmin) > 0) :
                    (memcmp(kmin, kmax, lmax) >= 0)) {
                  c->sp = 0;
                  break;
                }
              }
            }
            c->curr_rec = id;
            return id;
          }
        }
      } else {
        break;
      }
    }
  }
  return SEN_ID_NIL;
}

void
sen_pat_cursor_close(sen_ctx *ctx, sen_pat_cursor *c)
{
  SEN_ASSERT(c->ctx == ctx);
  if (c->ss) { SEN_FREE(c->ss); }
  SEN_FREE(c);
}

inline static int
bitcmp(const void *s1, const void *s2, int offset, int length)
{
  int r, rest = length + (offset & 7) - 8, bl = offset >> 3, mask = 0xff >> (offset & 7);
  unsigned char *a = (unsigned char *)s1 + bl, *b = (unsigned char *)s2 + bl;
  if (rest <= 0) {
    mask &= 0xff << -rest;
    return (*a & mask) - (*b & mask);
  }
  if ((r = (*a & mask) - (*b & mask))) { return r; }
  a++; b++;
  if ((bl = rest >> 3)) {
    if ((r = memcmp(a, b, bl))) { return r; }
    a += bl; b += bl;
  }
  mask = 0xff << (8 - (rest & 7));
  return (*a & mask) - (*b & mask);
}

inline static sen_rc
set_cursor_ascend(sen_pat *pat, sen_pat_cursor *c,
                  const void *key, uint32_t key_size, int flags)
{
  sen_id id;
  pat_node *node;
  const uint8_t *k;
  int r, check = -1, ch, c2;
  uint32_t len = key_size * 16;
  uint8_t keybuf[MAX_FIXED_KEY_SIZE];
  KEY_ENCODE(pat, keybuf, key, key_size);
  PAT_AT(pat, 0, node);
  for (id = node->lr[1]; id;) {
    PAT_AT(pat, id, node);
    if (!node) { return sen_invalid_format; }
    ch = PAT_CHK(node);
    if (ch <= check) {
      if (!(k = pat_node_get_key(pat, node))) { return sen_invalid_format; }
      {
        uint32_t l = PAT_LEN(node);
        if (l == key_size) {
          if (flags & SEN_CURSOR_GT) {
            if (memcmp(key, k, l) < 0) { push(c, id, check); }
          } else {
            if (memcmp(key, k, l) <= 0) { push(c, id, check); }
          }
        } else if (l < key_size) {
          if (memcmp(key, k, l) < 0) { push(c, id, check); }
        } else {
          if (memcmp(key, k, key_size) <= 0) { push(c, id, check); }
        }
      }
      break;
    }
    c2 = len < ch ? len : ch;
    if ((check += 2) < c2) {
      if (!(k = pat_node_get_key(pat, node))) { return sen_invalid_format; }
      if ((r = bitcmp(key, k, check >> 1, (c2 - check) >> 1))) {
        if (r < 0) {
          push(c, node->lr[1], ch);
          push(c, node->lr[0], ch);
        }
        break;
      }
    }
    check = ch;
    if (len <= check) { break; }
    if (check & 1) {
      if (check + 1 < len) {
        id = node->lr[1];
      } else {
        push(c, node->lr[1], check);
        id = node->lr[0];
      }
    } else {
      if (nth_bit((uint8_t *)key, check, len)) {
        id = node->lr[1];
      } else {
        push(c, node->lr[1], check);
        id = node->lr[0];
      }
    }
  }
  return sen_success;
}

inline static sen_rc
set_cursor_descend(sen_pat *pat, sen_pat_cursor *c,
                   const void *key, uint32_t key_size, int flags)
{
  sen_id id;
  pat_node *node;
  const uint8_t *k;
  int r, check = -1, ch, c2;
  uint32_t len = key_size * 16;
  uint8_t keybuf[MAX_FIXED_KEY_SIZE];
  KEY_ENCODE(pat, keybuf, key, key_size);
  PAT_AT(pat, 0, node);
  for (id = node->lr[1]; id;) {
    PAT_AT(pat, id, node);
    if (!node) { return sen_invalid_format; }
    ch = PAT_CHK(node);
    if (ch <= check) {
      if (!(k = pat_node_get_key(pat, node))) { return sen_invalid_format; }
      {
        uint32_t l = PAT_LEN(node);
        if (l <= key_size) {
          if ((flags & SEN_CURSOR_LT) && l == key_size) {
            if (memcmp(key, k, l) > 0) { push(c, id, check); }
          } else {
            if (memcmp(key, k, l) >= 0) { push(c, id, check); }
          }
        } else {
          if (memcmp(key, k, key_size) > 0) { push(c, id, check); }
        }
      }
      break;
    }
    c2 = len < ch ? len : ch;
    if ((check += 2) < c2) {
      if (!(k = pat_node_get_key(pat, node))) { return sen_invalid_format; }
      if ((r = bitcmp(key, k, check >> 1, (c2 - check) >> 1))) {
        if (r >= 0) {
          push(c, node->lr[0], ch);
          push(c, node->lr[1], ch);
        }
        break;
      }
    }
    check = ch;
    if (len <= check) {
      push(c, node->lr[0], check);
      push(c, node->lr[1], check);
      break;
    }
    if (check & 1) {
      if (check + 1 < len) {
        push(c, node->lr[0], check);
        id = node->lr[1];
      } else {
        id = node->lr[0];
      }
    } else {
      if (nth_bit((uint8_t *)key, check, len)) {
        push(c, node->lr[0], check);
        id = node->lr[1];
      } else {
        id = node->lr[0];
      }
    }
  }
  return sen_success;
}

sen_pat_cursor *
sen_pat_cursor_open(sen_ctx *ctx, sen_pat *pat,
                    const void *min, uint32_t min_size,
                    const void *max, uint32_t max_size, int flags)
{
  sen_id id;
  pat_node *node;
  sen_pat_cursor *c;
  if (!pat || !ctx) { return NULL; }
  if (!(c = SEN_MALLOCN(sen_pat_cursor, 1))) { return NULL; }
  SEN_OBJ_DB_SET_TYPE(c, SEN_CURSOR_TABLE_PAT_KEY);
  c->pat = pat;
  c->ctx = ctx;
  c->size = 0;
  c->sp = 0;
  c->ss = NULL;
  c->limit = 0;
  c->curr_rec = SEN_ID_NIL;
  if (flags & SEN_CURSOR_ASCENDING) {
    if (max) {
      set_cursor_descend(pat, c, max, max_size, flags);
      c->obj.flags = SEN_CURSOR_DESCENDING;
      c->limit = sen_pat_cursor_next(ctx, c);
      c->sp = 0;
      if (!c->limit) { goto exit; }
    }
    if (min) {
      set_cursor_ascend(pat, c, min, min_size, flags);
    } else {
      PAT_AT(pat, 0, node);
      if (!node) {
        sen_pat_cursor_close(ctx, c);
        return NULL;
      }
      if ((id = node->lr[1])) {
        PAT_AT(pat, id, node);
        if (node) {
          int ch = PAT_CHK(node);
          push(c, node->lr[1], ch);
          push(c, node->lr[0], ch);
        }
      }
    }
  } else {
    if (min) {
      set_cursor_ascend(pat, c, min, min_size, flags);
      c->obj.flags = SEN_CURSOR_ASCENDING;
      c->limit = sen_pat_cursor_next(ctx, c);
      c->sp = 0;
      if (!c->limit) { goto exit; }
    }
    if (max) {
      set_cursor_descend(pat, c, max, max_size, flags);
    } else {
      PAT_AT(pat, 0, node);
      if (!node) {
        sen_pat_cursor_close(ctx, c);
        return NULL;
      }
      if ((id = node->lr[1])) {
        PAT_AT(pat, id, node);
        if (node) {
          int ch = PAT_CHK(node);
          push(c, node->lr[0], ch);
          push(c, node->lr[1], ch);
        }
      }
    }
  }
exit :
  c->obj.flags = flags;
  c->curr_rec = SEN_ID_NIL;
  return c;
}

int
sen_pat_cursor_get_key(sen_ctx *ctx, sen_pat_cursor *c, void **key)
{
  *key = c->curr_key;
  return sen_pat_get_key(ctx, c->pat, c->curr_rec, *key, SEN_TABLE_MAX_KEY_SIZE);
}

int
sen_pat_cursor_get_value(sen_ctx *ctx, sen_pat_cursor *c, void **value)
{
  int value_size = (int)c->pat->value_size;
  if (value_size) {
    byte *v = (byte *)sis_at(c->pat, c->curr_rec);
    if (v) {
      if (c->pat->obj.flags & SEN_OBJ_KEY_WITH_SIS) {
        *value = v + sizeof(sis_node);
      } else {
        *value = v;
      }
    } else {
      *value = NULL;
    }
  }
  return value_size;
}

int
sen_pat_cursor_get_key_value(sen_ctx *ctx, sen_pat_cursor *c,
                             void **key, uint32_t *key_size, void **value)
{
  int value_size = (int)c->pat->value_size;
  if (key_size) {
    *key_size = (uint32_t) sen_pat_get_key(ctx, c->pat, c->curr_rec, c->curr_key,
                                           SEN_TABLE_MAX_KEY_SIZE);
    if (key) { *key = c->curr_key; }
  }
  if (value_size) {
    byte *v = (byte *)sis_at(c->pat, c->curr_rec);
    if (v) {
      if (c->pat->obj.flags & SEN_OBJ_KEY_WITH_SIS) {
        *value = v + sizeof(sis_node);
      } else {
        *value = v;
      }
    } else {
      *value = NULL;
    }
  }
  return value_size;
}

sen_rc
sen_pat_cursor_set_value(sen_ctx *ctx, sen_pat_cursor *c, void *value, int flags)
{
  return sen_pat_set_value(ctx, c->pat, c->curr_rec, value, flags);
}

sen_rc
sen_pat_cursor_delete(sen_ctx *ctx, sen_pat_cursor *c,
                      sen_table_delete_optarg *optarg)
{
  return sen_pat_delete_by_id(ctx, c->pat, c->curr_rec, optarg);
}
