/* Copyright(C) 2008 Brazil

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "senna_in.h"
#include <string.h>
#include "ql.h"
#include "db.h"
#include "pat.h"
#include "snip.h"

#define OBJ_DB(obj) ((sen_obj_db *)obj)
#define RVALUE(obj) (sen_ctx_get(ctx, (obj)->u.o.id))
#define INTERN2(s,l) (sen_ql_mk_symbol(ctx, s, l))

static sen_cell *ha_object(sen_ctx *ctx, sen_cell *args, sen_ql_co *co);
static sen_cell *ha_snip(sen_ctx *ctx, sen_cell *args, sen_ql_co *co);
static sen_cell *ha_verses(sen_ctx *ctx, sen_cell *args, sen_ql_co *co);

#define STRBUF_SIZE (SEN_TABLE_MAX_KEY_SIZE + 1)

static sen_rc
obj2str(sen_cell *o, char *buf, uint16_t *size)
{
  if (SYMBOLP(o)) {
    const char *r = _sen_hash_strkey_by_val(o, size);
    if (r && *r == ':') {
      r++;
      (*size)--;
    }
    memcpy(buf, r, *size);
    buf[*size] = '\0';
  } else if (o->header.type == SEN_CELL_STR) {
    if (STRSIZE(o) > SEN_TABLE_MAX_KEY_SIZE) {
      memcpy(buf, STRVALUE(o), SEN_TABLE_MAX_KEY_SIZE);
      buf[SEN_TABLE_MAX_KEY_SIZE] = '\0';
      *size = SEN_TABLE_MAX_KEY_SIZE + 1;
    } else {
      *size = STRSIZE(o);
      memcpy(buf, STRVALUE(o), *size);
      buf[*size] = '\0';
    }
  } else if (o->header.type == SEN_CELL_INT) {
    char *rest;
    sen_lltoa(IVALUE(o), buf, buf + SEN_TABLE_MAX_KEY_SIZE, &rest);
    *size = rest - buf;
    *rest = '\0';
  } else {
    return sen_invalid_argument;
  }
  return sen_success;
}

enum {
  SEN_QL_INT = 1,
  SEN_QL_UINT,
  SEN_QL_INT64,
  SEN_QL_FLOAT,
  SEN_QL_TIME,
  SEN_QL_SHORTTEXT,
  SEN_QL_TEXT,
  SEN_QL_LONGTEXT
};

static void
obj_obj_bind(sen_cell *obj, sen_id domain, sen_id self)
{
  obj->header.type = SEN_CELL_OBJECT;
  obj->header.impl_flags = SEN_CELL_NATIVE;
  obj->u.o.id = self;
  obj->header.domain = domain;
  obj->u.o.func = ha_object;
}

static sen_cell *
obj2cell(sen_ctx *ctx, sen_obj *obj, sen_cell *cell)
{
  if (SEN_OBJ_DBP(obj)) {
    sen_id id = OBJ_DB(obj)->id;
    if (id & SEN_OBJ_TMP_OBJECT) {
      sen_tmp_obj_db *tmp_obj;
      if ((tmp_obj = _sen_array_get_value(ctx, ctx->impl->values,
                                          id & ~SEN_OBJ_TMP_OBJECT))) {
        return &tmp_obj->cell;
      } else {
        return F;
      }
    } else {
      char buf[STRBUF_SIZE];
      int len = sen_obj_name(ctx, obj, buf, STRBUF_SIZE);
      return INTERN2(buf, len);
    }
  }
  if (!cell) { if (!(cell = sen_cell_new(ctx))) { return F; } }
  switch (obj->header.type) {
  case SEN_BULK :
    {
      void *v = SEN_BULK_HEAD(obj);
      sen_id rid = obj->header.domain;
      sen_obj *range = sen_ctx_get(ctx, rid);
      if (range && range->header.type == SEN_TYPE) {
        switch (rid) {
        case SEN_QL_INT :
          SETINT(cell, *((int32_t *)v));
          break;
        case SEN_QL_UINT :
          SETINT(cell, *((uint32_t *)v));
          break;
        case SEN_QL_INT64 :
          SETINT(cell, *((int64_t *)v));
          break;
        case SEN_QL_FLOAT :
          SETFLOAT(cell, *((double *)v));
          break;
        case SEN_QL_TIME :
          SETTIME(cell, v);
          break;
        default :
          {
            uint32_t size = SEN_BULK_VSIZE(obj);
            void *value = SEN_MALLOC(size);
            if (!value) { return F; }
            cell->header.impl_flags |= SEN_OBJ_ALLOCATED;
            memcpy(value, v, size);
            SETBULK(cell, value, size);
          }
          break;
        }
      } else {
        obj_obj_bind(cell, rid, *((sen_id *)v));
      }
    }
    break;
  case SEN_VERSES :
    {
      sen_obj *verses = sen_obj_graft(ctx, obj);
      if (verses) {
        cell->header.type = SEN_VERSES;
        cell->header.impl_flags |= SEN_OBJ_ALLOCATED|SEN_CELL_NATIVE;
        cell->u.p.value = verses;
        cell->u.p.func = ha_verses;
      }
    }
    break;
  }
  return cell;
}

static int
keywordp(sen_cell *x)
{
  uint16_t name_size;
  const char *name;
  return SYMBOLP(x)
    && (name = _sen_hash_strkey_by_val(x, &name_size))
    && name_size
    && *name == ':';
}

static int
symbolp(sen_cell *x)
{
  uint16_t name_size;
  const char *name;
  return SYMBOLP(x)
    && (name = _sen_hash_strkey_by_val(x, &name_size))
    && (!name_size || *name != ':');
}

static int
ascp(sen_cell *x)
{
  uint16_t name_size;
  const char *name;
  return SYMBOLP(x)
    && (name = _sen_hash_strkey_by_val(x, &name_size))
    && (name_size > 1)
    && (*name == ':')
    && (name[1] == 'a');
}

/* column_exp */

typedef struct {
  sen_cell *expr;
  sen_table_sort_key *keys;
  sen_cell **cells;
  int n_keys;
  int n_applys;
} column_exp;

static sen_obj *
get_accessor(sen_ctx *ctx, sen_obj *table, sen_cell *e)
{
  uint16_t msg_size;
  char msg[STRBUF_SIZE];
  obj2str(e, msg, &msg_size);
  return sen_obj_get_accessor(ctx, table, msg, msg_size);
}

#define COLUMN_EXPP(x,ce) ((x) == (ce)->cells[(ce)->n_keys])

static sen_cell *
column_exp_build_(sen_ctx *ctx, sen_obj *table, sen_cell *e, column_exp *ce)
{
  if (PAIRP(e)) {
    sen_cell *x, *r, *l = NIL, **d;
    POP(x, e);
    if (x == NIL && keywordp(CAR(e))) {
      if (ce->keys[ce->n_keys].key) { ce->n_keys++; }
      ce->keys[ce->n_keys].key = get_accessor(ctx, table, CAR(e));
      d = &ce->cells[ce->n_keys];
      if (!*d) { *d = sen_cell_new(ctx); }
      r = *d;
    } else {
      r = column_exp_build_(ctx, table, x, ce);
      if (COLUMN_EXPP(r, ce) && keywordp(CAR(e))) {
        get_accessor(ctx, ce->keys[ce->n_keys].key, CAR(e));
      } else {
        r = CONS(r, NIL);
        d = &CDR(r);
        while (PAIRP(e)) {
          POP(x, e);
          if (COLUMN_EXPP(l, ce) && ascp(x)) {
            ce->keys[ce->n_keys].flags |= SEN_TABLE_SORT_ASC;
          }
          l = column_exp_build_(ctx, table, x, ce);
          *d = CONS(l, NIL);
          d = &CDR(*d);
        }
      }
    }
    return r;
  } else {
    return e;
  }
}

static void
column_exp_build(sen_ctx *ctx, sen_obj *table, sen_cell *e, column_exp *ce)
{
  sen_cell *x, *r = e, *l = NIL, **d = &r;
  ce->n_keys = 0;
  while (PAIRP(e)) {
    POP(x, e);
    if (COLUMN_EXPP(l, ce) && ascp(x)) {
      ce->keys[ce->n_keys].flags |= SEN_TABLE_SORT_ASC;
    }
    l = column_exp_build_(ctx, table, x, ce);
    *d = CONS(l, NIL);
    d = &CDR(*d);
  }
  if (ce->keys[ce->n_keys].key) { ce->n_keys++; }
  ce->expr = r;
  sen_ql_obj_mark(ctx, ce->expr);
}

#define EVAL_BY_FUNCALLP(ce) \
 ((ce)->n_applys == 1 && PAIRP((ce)->expr) && NATIVE_FUNCP(CAR((ce)->expr)))

static void
column_exp_check(sen_cell *e, int *ns, int *ne)
{
  if (PAIRP(e)) {
    sen_cell *x;
    POP(x, e);
    if (x == NIL) {
      (*ns)++;
    } else if (NATIVE_FUNCP(x)) {
      (*ne)++;
    } else {
      column_exp_check(x, ns, ne);
    }
    while (PAIRP(e)) {
      POP(x, e);
      column_exp_check(x, ns, ne);
    }
  } else {
    if (symbolp(e)) { (*ne)++; }
  }
}

static column_exp *
column_exp_open(sen_ctx *ctx, sen_obj *table, sen_cell *expr)
{
  column_exp *ce = SEN_CALLOC(sizeof(column_exp));
  if (ce) {
    column_exp_check(expr, &ce->n_keys, &ce->n_applys);
    if (ce->n_keys) {
      if (!(ce->keys = SEN_CALLOC(sizeof(sen_table_sort_key) * ce->n_keys))) {
        SEN_FREE(ce);
        return NULL;
      }
      if (!(ce->cells = SEN_CALLOC(sizeof(sen_cell *) * ce->n_keys))) {
        SEN_FREE(ce->keys);
        SEN_FREE(ce);
        return NULL;
      }
      column_exp_build(ctx, table, expr, ce);
    } else {
      ce->expr = expr;
    }
  }
  return ce;
}

static sen_rc
column_exp_exec(sen_ctx *ctx, column_exp *ce, sen_id id)
{
  int i;
  sen_obj v, *vp;
  SEN_OBJ_INIT(&v, SEN_BULK, 0);
  for (i = 0; i < ce->n_keys; i++) {
    sen_cell *c = ce->cells[i];
    SEN_BULK_REWIND(&v);
    vp = sen_obj_get_value(ctx, ce->keys[i].key, id, &v);
    sen_cell_clear(ctx, c);
    obj2cell(ctx, vp, c);
  }
  sen_obj_close(ctx, &v);
  return sen_success;
}

static sen_rc
column_exp_close(sen_ctx *ctx, column_exp *ce)
{
  int i;
  sen_ql_obj_unmark(ctx, ce->expr);
  for (i = 0; i < ce->n_keys; i++) {
    sen_obj_close(ctx, ce->keys[i].key);
  }
  if (ce->keys) { SEN_FREE(ce->keys); }
  if (ce->cells) { SEN_FREE(ce->cells); }
  SEN_FREE(ce);
  return sen_success;
}

static sen_cell *
ha_void(sen_ctx *ctx, sen_cell *args, sen_ql_co *co)
{
  if (!ctx->impl->code) { return F; }
  return ctx->impl->code;
}

sen_cell *
sen_ql_obj_new(sen_ctx *ctx, sen_id domain, sen_id self)
{
  sen_cell *o;
  SEN_CELL_NEW(ctx, o);
  obj_obj_bind(o, domain, self);
  return o;
}

static sen_cell *
get_cell(sen_ctx *ctx, sen_obj *obj_db)
{
  if (obj_db) {
    sen_id id = OBJ_DB(obj_db)->id;
    if (id & SEN_OBJ_TMP_OBJECT) {
      sen_tmp_obj_db *tmp_obj;
      if ((tmp_obj = _sen_array_get_value(ctx, ctx->impl->values,
                                          id & ~SEN_OBJ_TMP_OBJECT))) {
        return &tmp_obj->cell;
      } else {
        return F;
      }
    } else {
      char buf[STRBUF_SIZE];
      int len = sen_obj_name(ctx, obj_db, buf, STRBUF_SIZE);
      return INTERN2(buf, len);
    }
  } else {
    return F;
  }
}

static sen_obj *
get_obj(sen_ctx *ctx, sen_cell *cell)
{
  return SEN_OBJ_DBP(cell) ? sen_ctx_get(ctx, cell->u.o.id) : NULL;
}

static sen_obj *
get_domain(sen_ctx *ctx, sen_cell *cell)
{
  return sen_ctx_get(ctx, cell->header.domain);
}

static void
unesc(sen_ctx *ctx, sen_cell *obj)
{
  char *src, *dest, *end = STRVALUE(obj) + STRSIZE(obj);
  for (src = dest = STRVALUE(obj);;) {
    unsigned int len;
    if (!(len = sen_str_charlen_nonnull(src, end, ctx->encoding))) { break; }
    if (src[0] == '\\' && src + 1 < end && len == 1) {
      src++;
      switch (*src) {
      case 'n' :
        *dest++ = '\n';
        break;
      case 'r' :
        *dest++ = '\r';
        break;
      case 't' :
        *dest++ = '\t';
        break;
      default :
        *dest++ = *src;
        break;
      }
      src++;
    } else {
      while (len--) { *dest++ = *src++; }
    }
  }
  STRSIZE(obj) = dest - STRVALUE(obj);
}

static sen_cell *
sen_ql_table_at(sen_ctx *ctx, sen_obj *table, const void *key, unsigned key_size,
                sen_search_flags flags, sen_cell *res)
{
  sen_id id = sen_table_lookup(ctx, table, key, key_size, &flags);
  if (id) {
    if (!res) { SEN_CELL_NEW(ctx, res); }
    obj_obj_bind(res, OBJ_DB(table)->id, id);
    return res;
  } else {
    return F;
  }
}

static sen_obj *
get_column(sen_ctx *ctx, sen_id tid, char *msg, unsigned msg_size, sen_id *id)
{
  sen_obj *table = sen_ctx_get(ctx, tid);
  for (;;) {
    sen_obj *domain = sen_ctx_get(ctx, OBJ_DB(table)->header.domain);
    if (!domain || domain->header.type == SEN_TYPE) { break; }
    if (id) {
      if (!sen_table_get_key(ctx, table, *id, id, sizeof(sen_id))) { return NULL; }
    }
    table = domain;
  }
  return sen_table_column(ctx, table, msg, msg_size);
}

static sen_cell *
table_column(sen_ctx *ctx, sen_id base, char *msg, unsigned msg_size)
{
  sen_obj *column = get_column(ctx, base, msg, msg_size, NULL);
  return column ? get_cell(ctx, column) : F;
}

static sen_cell *
ha_verses(sen_ctx *ctx, sen_cell *args, sen_ql_co *co)
{
  uint16_t msg_size;
  sen_cell *car, *res;
  char msg[STRBUF_SIZE];
  if (!(res = ctx->impl->code)) { QLERR("invalid receiver"); }
  POP(car, args);
  if (obj2str(car, msg, &msg_size)) { QLERR("invalid argument"); }
  switch (*msg) {
  case ':' :
    switch (msg[1]) {
    case 's' : /* :sexp */
      {
        sen_obj *verses = (sen_obj *)res->u.p.value;
        int n = verses->u.v.n_verses;
        sen_verse *vp = &verses->u.v.verses[n];
        sen_cell *a = INTERN("@");
        sen_cell *d = INTERN(":dic");
        sen_cell *w = INTERN(":weight");
        sen_cell *v = INTERN(":value");
        res = NIL;
        while (vp--, n--) {
          sen_cell *vs = NIL;
          if (vp->weight || vp->domain) {
            vs = CONS(v, CONS(sen_ql_mk_string(ctx, vp->str, vp->str_len), vs));
            if (vp->weight) {
              sen_cell *nv;
              SEN_CELL_NEW(ctx, nv);
              SETINT(nv, vp->weight);
              vs = CONS(w, CONS(nv, vs));
            }
            if (vp->domain) {
              sen_cell *nv;
              SEN_CELL_NEW(ctx, nv);
              SETINT(nv, vp->domain);
              vs = CONS(d, CONS(nv, vs));
            }
            vs = CONS(a, vs);
          } else {
            vs = sen_ql_mk_string(ctx, vp->str, vp->str_len);
          }
          res = CONS(vs, res);
        }
      }
      break;
    default :
      QLERR("invalid argument");
    }
    break;
  default :
    QLERR("invalid argument");
  }
  return res;
}

#define QLWARN(...) ERRSET(ctx, SEN_WARN, sen_invalid_argument, __VA_ARGS__)

#define STR2DBL(str,len,val) do {\
  char *end, buf0[128], *buf = (len) < 128 ? buf0 : SEN_MALLOC((len) + 1);\
  if (buf) {\
    double d;\
    memcpy(buf, (str), (len));\
    buf[len] = '\0';\
    errno = 0;\
    d = strtod(buf, &end);\
    if (!((len) < 128)) { SEN_FREE(buf); }\
    if (!errno && buf + (len) == end) {\
      (val) = d;\
    } else { QLWARN("cast failed"); }\
  } else { QLWARN("buf alloc failed"); }\
} while (0)

static sen_obj *
cell2obj(sen_ctx *ctx, sen_cell *cell, sen_obj *column, sen_obj *obj)
{
  if ((column->header.flags & SEN_OBJ_COLUMN_TYPE_MASK) == SEN_OBJ_COLUMN_SCALAR) {
    sen_id rid = sen_obj_get_range(ctx, column);
    sen_obj *range = sen_ctx_get(ctx, rid);
    if (range && range->header.type == SEN_TYPE) {
      switch (rid) {
      case SEN_QL_INT :
        {
          int32_t v;
          switch (cell->header.type) {
          case SEN_CELL_STR :
            {
              int64_t iv = sen_atoll(STRVALUE(cell), STRVALUE(cell) + STRSIZE(cell), NULL);
              v = (int32_t) iv;
            }
            break;
          case SEN_CELL_INT :
            v = (int32_t) IVALUE(cell);
            break;
          case SEN_CELL_FLOAT :
            v = (int32_t) FVALUE(cell);
            break;
          case SEN_CELL_TIME :
            v = (int32_t) cell->u.tv.tv_usec;
            break;
          }
          if (!obj) { if (!(obj = sen_obj_open(ctx, SEN_BULK, 0))) { return NULL; }}
          sen_bulk_write(ctx, obj, (const char *)(&v), sizeof(int32_t));
        }
        break;
      case SEN_QL_UINT :
        {
          uint32_t v;
          switch (cell->header.type) {
          case SEN_CELL_STR :
            {
              int64_t iv = sen_atoll(STRVALUE(cell), STRVALUE(cell) + STRSIZE(cell), NULL);
              v = (uint32_t) iv;
            }
            break;
          case SEN_CELL_INT :
            v = (uint32_t) IVALUE(cell);
            break;
          case SEN_CELL_FLOAT :
            v = (uint32_t) FVALUE(cell);
            break;
          case SEN_CELL_TIME :
            v = (uint32_t) cell->u.tv.tv_usec;
            break;
          }
          if (!obj) { if (!(obj = sen_obj_open(ctx, SEN_BULK, 0))) { return NULL; }}
          sen_bulk_write(ctx, obj, (const char *)(&v), sizeof(uint32_t));
        }
        break;
      case SEN_QL_INT64 :
        {
          int64_t v;
          switch (cell->header.type) {
          case SEN_CELL_STR :
            v = sen_atoll(STRVALUE(cell), STRVALUE(cell) + STRSIZE(cell), NULL);
            break;
          case SEN_CELL_INT :
            v = IVALUE(cell);
            break;
          case SEN_CELL_FLOAT :
            v = (int64_t) FVALUE(cell);
            break;
          case SEN_CELL_TIME :
            v = (int32_t) cell->u.tv.tv_usec;
            break;
          }
          if (!obj) { if (!(obj = sen_obj_open(ctx, SEN_BULK, 0))) { return NULL; }}
          sen_bulk_write(ctx, obj, (const char *)(&v), sizeof(int64_t));
        }
        break;
      case SEN_QL_FLOAT :
        {
          double v;
          switch (cell->header.type) {
          case SEN_CELL_STR :
            { /* todo : support #i notation */
              char *str = STRVALUE(cell);
              int len = STRSIZE(cell);
              STR2DBL(str, len, v);
            }
            break;
          case SEN_CELL_INT :
            v = (double) IVALUE(cell);
            break;
          case SEN_CELL_FLOAT :
            v = FVALUE(cell);
            break;
          case SEN_CELL_TIME :
            v = ((double) cell->u.tv.tv_usec) / 1000000 + cell->u.tv.tv_sec;
            break;
          }
          if (!obj) { if (!(obj = sen_obj_open(ctx, SEN_BULK, 0))) { return NULL; }}
          sen_bulk_write(ctx, obj, (const char *)(&v), sizeof(double));
        }
        break;
      case SEN_QL_TIME :
        {
          sen_timeval v;
          switch (cell->header.type) {
          case SEN_CELL_STR :
            {
              int len = STRSIZE(cell);
              char *str = STRVALUE(cell);
              if (sen_str2timeval(str, len, &v)) {
                if (len > 3 && *str == '#' && str[1] == ':' && str[2] == '<') {
                  const char *cur;
                  v.tv_sec = sen_atoi(str + 3, str + len, &cur);
                  if (cur >= str + len || *cur != '.') {
                    QLWARN("illegal time format '%s'", str);
                  }
                  v.tv_usec = sen_atoi(cur + 1, str + len, &cur);
                  if (cur >= str + len || *cur != '>') {
                    QLWARN("illegal time format '%s'", str);
                  }
                } else {
                  double dval;
                  char *str = STRVALUE(cell);
                  int len = cell->u.b.size;
                  STR2DBL(str, len, dval);
                  v.tv_sec = (int32_t) dval;
                  v.tv_usec = (int32_t) ((dval - v.tv_sec) * 1000000);
                }
              }
            }
            break;
          case SEN_CELL_INT :
            v.tv_sec = (int32_t) IVALUE(cell);
            v.tv_usec = 0;
            break;
          case SEN_CELL_FLOAT :
            v.tv_sec = (int32_t) FVALUE(cell);
            v.tv_usec = (int32_t) ((FVALUE(cell) - v.tv_sec) * 1000000);
            break;
          case SEN_CELL_TIME :
            v.tv_sec = cell->u.tv.tv_sec;
            v.tv_usec = cell->u.tv.tv_usec;
            break;
          }
          if (!obj) { if (!(obj = sen_obj_open(ctx, SEN_BULK, 0))) { return NULL; }}
          sen_bulk_write(ctx, obj, (const char *)(&v), sizeof(sen_timeval));
        }
        break;
      default :
        if (BULKP(cell)) {
          if (!obj) { if (!(obj = sen_obj_open(ctx, SEN_BULK, 0))) { return NULL; }}
          sen_bulk_write(ctx, obj, STRVALUE(cell), STRSIZE(cell));
        } else {
          if (obj) { SEN_BULK_REWIND(obj); }
        }
        break;
      }
    } else {
      /* todo : sen_type_any support
      sen_bulk_write(ctx, bulk, (const char *)(&cell->header.domain), sizeof(sen_id));
      */
      /* todo : nested table support.
       */
      sen_id id;
      switch (cell->header.type) {
      case SEN_CELL_OBJECT :
        if (cell->header.domain != rid) { return NULL; }
        id = cell->u.o.id;
        break;
      case SEN_CELL_STR :
        {
          sen_search_flags flags = SEN_TABLE_ADD;
          id = sen_table_get(ctx, range, STRVALUE(cell), STRSIZE(cell), NULL, &flags);
        }
        break;
      default :
        if (VOIDP(cell)) { id = SEN_ID_NIL; }
        break;
      }
      if (!obj) { if (!(obj = sen_obj_open(ctx, SEN_BULK, 0))) { return NULL; }}
      sen_bulk_write(ctx, obj, (const char *)&id, sizeof(sen_id));
    }
  } else {
    switch (cell->header.type) {
    case SEN_CELL_STR :
      /* todo */
      break;
    case SEN_CELL_INT :
      /* todo */
      break;
    case SEN_CELL_FLOAT :
      /* todo */
      break;
    case SEN_CELL_TIME :
      /* todo */
      break;
    case SEN_VERSES :
      obj = (sen_obj *)cell->u.p.value;
      break;
    }
  }
  return obj;
}

inline static sen_cell *
column_value(sen_ctx *ctx, sen_obj *column, sen_id obj,
             sen_cell *args, sen_cell *res)
{
  sen_obj *value;
  if (VOIDP(args) || (PAIRP(args) && VOIDP(CAR(args)))) {
    if ((value = sen_obj_get_value(ctx, column, obj, NULL))) {
      res = obj2cell(ctx, value, res);
      sen_obj_close(ctx, value);
    } else {
      res = F;
    }
  } else {
    sen_cell *car;
    int flags = SEN_OBJ_SET;
    POP(car, args);
    if ((value = cell2obj(ctx, car, column, NULL))) {
      res = sen_obj_set_value(ctx, column, obj, value, flags) ? F : car;
      sen_obj_close(ctx, value);
    } else {
      res = F;
    }
  }
  return res;
}

static sen_cell *
register_cell(sen_ctx *ctx, sen_obj *obj_db, const char *name, unsigned name_size)
{
  if (name && name_size) {
    return INTERN2(name, name_size);
    // sen_ql_obj_bind(obj_db, res);
  } else {
    sen_cell *r = get_cell(ctx, obj_db);
    if (r != F) {
      r->header.impl_flags = 0;
      r->header.flags = 0;
      sen_ql_obj_bind(obj_db, r);
    }
    return r;
  }
}

static sen_cell *
table_create(sen_ctx *ctx, const char *name, unsigned name_size,
             sen_obj_flags flags, sen_obj *domain,
             unsigned value_size, sen_encoding encoding)
{
  sen_obj *table = sen_table_create(ctx, name, name_size,
                                    NULL, flags, domain, value_size, encoding);
  if (!table) { QLERR("table create failed"); }
  return register_cell(ctx, table, name, name_size);
}

static sen_cell *
rec_obj_new(sen_ctx *ctx, sen_obj *domain, unsigned value_size)
{
  return table_create(ctx, NULL, 0, SEN_OBJ_TABLE_HASH_KEY|SEN_OBJ_WITH_SUBREC,
                      domain, value_size, sen_enc_none);
}

typedef struct {
  column_exp *ce;
  sen_cell *func;
  sen_cell *exprs;
  sen_cell *args;
  sen_sel_operator op;
  sen_cell *objs;
  int32_t offset;
  int32_t limit;
  int mode;
  char *from;
  unsigned fromsize;
  char *to;
  unsigned tosize;
} match_spec;

inline static sen_cell *
match_prepare(sen_ctx *ctx, match_spec *spec, sen_id base, sen_cell *args)
{
  sen_cell *car, *expr;
  sen_obj *r = sen_ctx_get(ctx, base);
  POP(expr, args);
  spec->ce = column_exp_open(ctx, r, expr);
  if (EVAL_BY_FUNCALLP(spec->ce)) {
    spec->func = CAR(expr);
  } else {
    spec->func = NULL;
  }
  spec->offset = 0;
  spec->limit = 0;
  spec->mode = 0;
  POP(expr, args);
  if (RECORDSP(expr)) {
    char ops[STRBUF_SIZE];
    uint16_t ops_size;
    if (expr->header.domain != base) { QLERR("table unmatch"); }
    POP(car, args);
    spec->op = sen_sel_and;
    if (!obj2str(car, ops, &ops_size)) {
      switch (*ops) {
      case '+': spec->op = sen_sel_or; break;
      case '-': spec->op = sen_sel_but; break;
      case '*': spec->op = sen_sel_and; break;
      case '>': spec->op = sen_sel_adjust; break;
      }
    }
  } else {
    char str[STRBUF_SIZE];
    uint16_t str_size;
     sen_obj *table = sen_ctx_get(ctx, base);
    if (INTP(expr)) { spec->offset = IVALUE(expr); }
    POP(expr, args);
    if (INTP(expr)) { spec->limit = IVALUE(expr); }
    if (spec->limit <= 0) { spec->limit += sen_table_size(ctx, table); }
    POP(expr, args);
    if (!obj2str(expr, str, &str_size)) {
      uint16_t i;
      for (i = 0; i < str_size; i++) {
        switch (str[i]) {
        case 'a' : spec->mode |= SEN_CURSOR_ASCENDING; break;
        case 'g' : spec->mode |= SEN_CURSOR_GT; break;
        case 'l' : spec->mode |= SEN_CURSOR_LT; break;
        }
      }
    }
    POP(expr, args);
    if (BULKP(expr)) {
      spec->from = STRVALUE(expr);
      spec->fromsize = STRSIZE(expr);
    } else {
      spec->from = NULL;
      spec->fromsize = 0;
    }
    POP(expr, args);
    if (BULKP(expr)) {
      spec->to = STRVALUE(expr);
      spec->tosize = STRSIZE(expr);
    } else {
      spec->to = NULL;
      spec->tosize = 0;
    }
    expr = rec_obj_new(ctx, table, 0);
    if (ERRP(ctx, SEN_WARN)) { return F; }
    spec->op = sen_sel_or;
  }
  spec->objs = CONS(expr, spec->exprs);
  return expr;
}

sen_obj *
sen_ql_obj_key(sen_ctx *ctx, sen_cell *obj, sen_obj *value)
{
  sen_obj *table = sen_ctx_get(ctx, obj->header.domain);
  sen_obj *accessor = sen_obj_get_accessor(ctx, table, ":key", 4);
  if (accessor) {
    value = sen_obj_get_value(ctx, accessor, obj->u.o.id, value);
    sen_obj_close(ctx, accessor);
  }
  return value;
}

inline static sen_cell *
obj2oid(sen_ctx *ctx, sen_cell *obj, sen_cell *res)
{
  char buf[32];
  sen_obj bogus_buf;
  bogus_buf.u.b.head = buf;
  bogus_buf.u.b.curr = buf;
  bogus_buf.u.b.tail = buf + 32;
  if (obj->header.type != SEN_CELL_OBJECT) { return F; }
  sen_obj_inspect(ctx, obj, &bogus_buf, SEN_OBJ_INSPECT_ESC);
  if (res) {
    uint32_t size = SEN_BULK_VSIZE(&bogus_buf);
    char *value = SEN_MALLOC(size + 1);
    if (!value) { return F; }
    sen_cell_clear(ctx, res);
    res->header.impl_flags = SEN_OBJ_ALLOCATED;
    res->header.type = SEN_CELL_STR;
    res->u.b.size = size;
    res->u.b.value = value;
    memcpy(res->u.b.value, buf, res->u.b.size + 1);
  } else {
    if (!(res = sen_ql_mk_string(ctx, buf, SEN_BULK_VSIZE(&bogus_buf)))) { return F; }
  }
  return res;
}

inline static int
match_exec(sen_ctx *ctx, match_spec *spec, sen_id base, sen_id id)
{
  sen_cell *res;
  column_exp_exec(ctx, spec->ce, id);
  if (spec->func) {
    sen_cell *code = ctx->impl->code;
    ctx->impl->code = spec->func;
    res = spec->func->u.o.func(ctx, CDR(spec->ce->expr), &ctx->impl->co);
    ctx->impl->code = code;
  } else {
    res = sen_ql_eval(ctx, spec->ce->expr, NIL);
  }
  return res != F;
}

sen_rc
match_close(sen_ctx *ctx, match_spec *spec)
{
  return column_exp_close(ctx, spec->ce);
}

// todo : refine
#define MAXCOLUMNS 0x100

struct _ins_stat {
  sen_cell *columns;
  int ncolumns;
  int nrecs;
};

static sen_cell *
ha_table(sen_ctx *ctx, sen_cell *args, sen_ql_co *co)
{
  sen_id base;
  int load = 0;
  sen_obj *table;
  sen_cell *car, *res;
  char msg[STRBUF_SIZE];
  uint16_t msg_size;
  if (!(res = ctx->impl->code)) { QLERR("invalid receiver"); }
  base = ctx->impl->code->u.o.id;
  if (!(table = sen_ctx_get(ctx, base))) { QLERR("invalid table"); }
  SEN_QL_CO_BEGIN(co);
  POP(car, args);
  if (obj2str(car, msg, &msg_size)) { QLERR("invalid argument"); }
  switch (*msg) {
  case '\0' : /* get instance by key */
    {
      POP(car, args);
      if (obj2str(car, msg, &msg_size)) { QLERR("invalid argument"); }
      res = sen_ql_table_at(ctx, table, msg, msg_size, 0, NULL);
    }
    break;
  case ':' :
    switch (msg[1]) {
    case 'c' :
    case 'C' :
      switch (msg[2]) {
      case 'l' : /* :clearlock */
      case 'L' :
        {
          res = sen_obj_is_locked(ctx, table) ? T : F;
          sen_obj_clear_lock(ctx, table);
        }
        break;
      case 'o' : /* :common-prefix-search */
      case 'O' :
        switch (msg[3]) {
        case 'm' :
        case 'M' :
          {
            sen_id id;
            sen_search_flags flags = SEN_SEARCH_LCP;
            POP(car, args);
            if (!BULKP(car)) { QLERR("invalid argument"); }
            id = sen_table_lookup(ctx, table, STRVALUE(car), STRSIZE(car), &flags);
            res = id ? sen_ql_obj_new(ctx, base, id) : F;
          }
          break;
        case 'l' : /* :columns */
        case 'L' :
          POP(car, args);
          if (RECORDSP(car)) {
            msg_size = 0;
            res = car;
          } else {
            if (obj2str(car, msg, &msg_size)) { QLERR("invalid argument"); }
            POP(res, args);
            if (res == NIL) {
              if (!(res = rec_obj_new(ctx, NULL, 0))) { QLERR("rec_obj_new failed"); }
            } else if (!RECORDSP(res)) {
              QLERR("records object expected");
            }
          }
          sen_table_columns(ctx, table, msg, msg_size, get_obj(ctx, res));
          break;
        default :
          res = F;
          break;
        }
        break;
      }
      break;
    case 'd' :
    case 'D' :
      switch (msg[2]) {
      case 'e' :
      case 'E' :
        switch (msg[3]) {
        case 'f' : /* :def */
        case 'F' :
          {
            uint16_t name_size;
            int nsources = 0;
            sen_id sources[MAXCOLUMNS];
            sen_obj *column, *type;
            char name[STRBUF_SIZE];
            sen_obj_flags flags = SEN_OBJ_PERSISTENT; /* default */
            POP(car, args);
            if (obj2str(car, name, &name_size)) { QLERR("invalid argument"); }
            if (sen_table_column(ctx, table, name, name_size)) { return T; }
            POP(car, args);
            type = get_obj(ctx, car);
            while (PAIRP(args)) {
              POP(car, args);
              if (PAIRP(car)) {
                car = CAR(car);
                if (!obj2str(car, msg, &msg_size)) {
                  sen_obj *source = sen_table_column(ctx, type, msg, msg_size);
                  if (source) { sources[nsources++] = OBJ_DB(source)->id; }
                }
              } else {
                if (obj2str(car, msg, &msg_size)) { QLERR("invalid argument"); }
                switch (*msg) {
                case 'a' :
                case 'A' :
                  flags |= SEN_OBJ_COLUMN_ARRAY;
                  break;
                case 'i' :
                case 'I' :
                  flags |= SEN_OBJ_COLUMN_INDEX;
                  break;
                case 'p' :
                case 'P' :
                  switch (msg[1]) {
                  case 'e' :
                  case 'E' :
                    flags |= SEN_OBJ_PERSISTENT;
                    break;
                  case 'o' :
                  case 'O' :
                    flags |= SEN_OBJ_COLUMN_POSTINGS;
                    break;
                  }
                  break;
                case 's' :
                case 'S' :
                  flags &= ~SEN_OBJ_COLUMN_TYPE_MASK;
                  break;
                case 't' :
                case 'T' :
                  flags &= ~SEN_OBJ_PERSISTENT;
                  break;
                case 'v' :
                case 'V' :
                  flags |= SEN_OBJ_COLUMN_VERSES;
                  break;
                }
              }
            }
            column = sen_column_create(ctx, table, name, name_size, NULL, flags, type);
            if (column) {
              if (nsources) {
                sen_obj source;
                SEN_OBJ_INIT(&source, SEN_BULK, SEN_OBJ_DO_SHALLOW_COPY);
                SEN_BULK_SET(ctx, &source, sources, nsources * sizeof(sen_id));
                sen_obj_set_info(ctx, column, SEN_INFO_SOURCE, &source);
              }
              msg_size = sen_obj_name(ctx, column, msg, SEN_PAT_MAX_KEY_SIZE);
              res = INTERN2(msg, msg_size);
            } else {
              res = F;
            }
          }
          break;
        case 'l' : /* :delete */
        case 'L' :
          POP(car, args);
          if (obj2str(car, msg, &msg_size)) { QLERR("invalid argument"); }
          res = sen_table_delete(ctx, table, msg, msg_size) ? T : F;
          break;
        default :
          res = F;
          break;
        }
        break;
      case 'i' : /* :difference */
      case 'I' :
        {
          if (PAIRP(args)) {
            POP(car, args);
            if (TABLEP(car)) {
              sen_obj *table2 = RVALUE(car);
              res = sen_table_difference(ctx, table, table2, table, table2) ? T : F;
            }
          }
        }
        break;
      default :
        res = F;
        break;
      }
      break;
    case 'e' : /* :extract */
    case 'E' :
      POP(car, args);
      if (!BULKP(car)) { QLERR("string expected"); }
      POP(res, args);
      if (res == NIL) {
        if (!(res = rec_obj_new(ctx, table, 0))) { QLERR("rec_obj_new failed"); }
      } else if (!RECORDSP(res)) {
        QLERR("records object expected");
      }
      if (sen_table_search(ctx, table, STRVALUE(car), STRSIZE(car),
                           SEN_SEARCH_TERM_EXTRACT, get_obj(ctx, res), sen_sel_or)) {
        QLERR("term_extract failed");
      }
      break;
    case 'g' : /* :group */
    case 'G' :
      {
        column_exp *ce;
        POP(car, args);
        ce = column_exp_open(ctx, table, car);
        if (ce) {
          int n_results = 0;
          sen_table_group_result *rp, results[256];
          POP(car, args);
          for (rp = results; TABLEP(car); rp++) {
            memset(rp, 0, sizeof(sen_table_group_result));
            rp->table = get_obj(ctx, car);
            POP(car, args);
            if (INTP(car)) {
              rp->key_begin = IVALUE(car);
              POP(car, args);
              if (INTP(car)) {
                rp->key_end = IVALUE(car);
              } else {
                rp->key_end = rp->key_begin + 1;
              }
            } else {
              rp->key_begin = n_results;
              rp->key_end = rp->key_begin + 1;
            }
            if (++n_results >= 256) { break; }
          }
          if (!n_results) {
            for (rp = results; n_results < ce->n_keys; rp++, n_results++) {
              sen_id range = sen_obj_get_range(ctx, ce->keys[n_results].key);
              rp->table = get_obj(ctx, rec_obj_new(ctx, sen_ctx_get(ctx, range), 0));
              rp->key_begin = n_results;
              rp->key_end = rp->key_begin + 1;
            }
          }
          sen_table_group(ctx, table, ce->keys, ce->n_keys, results, n_results);
          column_exp_close(ctx, ce);
          res = NIL;
          while (n_results--) {
            res = CONS(get_cell(ctx, results[n_results].table), res);
          }
        } else {
          QLERR("group key parse error");
        }
      }
      break;
    case 'i' : /* :intersect */
    case 'I' :
      {
        while (PAIRP(args)) {
          POP(car, args);
          if (!TABLEP(car)) { continue; }
          sen_table_setoperation(ctx, table, RVALUE(car), table, sen_sel_and);
        }
      }
      break;
    case 'l' : /* :load */
    case 'L' :
      load = 1;
      break;
    case 'n' :
    case 'N' :
      {
        switch (msg[2]) {
        case 'e' : /* :new */
        case 'E' :
          {
            // todo : support array
            POP(car, args);
            if (obj2str(car, msg, &msg_size)) { return F; }
            res = sen_ql_table_at(ctx, table, msg, msg_size, SEN_TABLE_ADD, NULL);
            if (res != F) {
              sen_cell cons, dummy;
              sen_obj *column;
              cons.header.type = SEN_CELL_LIST;
              cons.header.impl_flags = 0;
              cons.u.l.cdr = NIL;
              while (PAIRP(args)) {
                POP(car, args);
                if (obj2str(car, msg, &msg_size)) { break; }
                POP(car, args);
                if (!(column = sen_table_column(ctx, table, msg, msg_size))) { break; }
                cons.u.l.car = car;
                column_value(ctx, column, res->u.o.id, &cons, &dummy);
              }
            }
          }
          break;
        case 'r' : /* :nrecs */
        case 'R' :
          SEN_CELL_NEW(ctx, res);
          SETINT(res, sen_table_size(ctx, table));
          break;
        default :
          /* ambiguous message. todo : return error */
          res = F;
        }
      }
      break;
    case 'p' : /* :prefix-search */
    case 'P' :
      POP(car, args);
      if (!BULKP(car)) { QLERR("string expected"); }
      POP(res, args);
      if (res == NIL) {
        if (!(res = rec_obj_new(ctx, table, 0))) { QLERR("rec_obj_new failed"); }
      } else if (!RECORDSP(res)) {
        QLERR("records object expected");
      }
      if (sen_table_search(ctx, table, STRVALUE(car), STRSIZE(car),
                           SEN_SEARCH_PREFIX, get_obj(ctx, res), sen_sel_or)) {
        QLERR("prefix search failed");
      }
      break;
    case 's' :
    case 'S' :
      switch (msg[2]) {
      case 'c' :
      case 'C' :
        switch (msg[3]) {
        case 'a' : /* :scan-select */
        case 'A' :
          {
            sen_rset_recinfo *ri;
            sen_id *rid;
            match_spec spec;
            sen_obj *rec;
            res = match_prepare(ctx, &spec, base, args);
            if (ERRP(ctx, SEN_WARN)) { return F; }
            rec = RVALUE(res);
            switch (spec.op) {
            case sen_sel_or :
              {
                sen_id id;
                sen_table_cursor *c;
                sen_rset_posinfo *pi = (sen_rset_posinfo *) &id;
                int n = 0, o = spec.offset, l = spec.limit;
                if (l) {
                  if ((c = sen_table_cursor_open(ctx, table,
                                                 spec.from, spec.fromsize,
                                                 spec.to, spec.tosize,
                                                 spec.mode))) {
                    while ((id = sen_table_cursor_next(ctx, c))) {
                      if (match_exec(ctx, &spec, base, id)) {
                        if (n++ >= o) {
                          /* todo : use SEN_SET_INT_ADD if !n_entries */
                          sen_search_flags fl = SEN_TABLE_ADD;
                          sen_table_get(ctx, rec, pi, sizeof(sen_id), (void **)&ri, &fl);
                          // sen_table_add_subrec(rp->table, value, ri ? ri->score : 0, NULL, 0)
                          if (!--l) { break; }
                        }
                      }
                    }
                    sen_table_cursor_close(ctx, c);
                  }
                }
              }
              break;
            case sen_sel_and :
              SEN_HASH_EACH((sen_hash *)rec, id, &rid, NULL, &ri, {
                if (!match_exec(ctx, &spec, base, *rid)) {
                  sen_hash_delete_by_id(ctx, (sen_hash *)rec, id, NULL);
                }
              });
              break;
            case sen_sel_but :
              SEN_HASH_EACH((sen_hash *)rec, id, &rid, NULL, &ri, {
                if (match_exec(ctx, &spec, base, *rid)) {
                  sen_hash_delete_by_id(ctx, (sen_hash *)rec, id, NULL);
                }
              });
              break;
            case sen_sel_adjust :
              /* todo : support it */
              break;
            }
            match_close(ctx, &spec);
          }
          break;
        case 'h' : /* :schema */
        case 'H' :
          {
            sen_obj_flags flags;
            sen_encoding encoding;
            uint8_t ngram_unit;
            res = NIL;
            if (!(msg_size = sen_obj_name(ctx, table, msg, STRBUF_SIZE))) {
              return F;
            }
            if (sen_table_get_info(ctx, table, &flags, &encoding, &ngram_unit)) {
              return F;
            }
            if (flags & SEN_OBJ_TOKEN_DELIMITED) { res = CONS(INTERN(":delimited"), res); }
            if (flags & SEN_OBJ_TOKEN_NGRAM) { res = CONS(INTERN(":ngram"), res); }
            if (flags & SEN_OBJ_KEY_NORMALIZE) { res = CONS(INTERN(":normalize"), res); }
            switch (flags & SEN_OBJ_TABLE_TYPE_MASK) {
            case SEN_OBJ_TABLE_HASH_KEY :
              res = CONS(INTERN(":hash"), res);
              break;
            case SEN_OBJ_TABLE_PAT_KEY :
              res = CONS(INTERN(":pat"), res);
              break;
            case SEN_OBJ_TABLE_NO_KEY :
              res = CONS(INTERN(":surrogate"), res);
              break;
            }
            {
              char encstr[32] = ":";
              strcpy(encstr + 1, sen_enctostr(encoding));
              res = CONS(INTERN(encstr), res);
            }
            res = CONS(INTERN("ptable"),
                       CONS(CONS(INTERN("quote"),
                                 CONS(INTERN2(msg, msg_size), NIL)), res));
          }
          break;
        default :
          res = F;
          break;
        }
        break;
      case 'o' : /* :sort */
      case 'O' :
        {
          int limit = 0;
          column_exp *ce;
          POP(car, args);
          ce = column_exp_open(ctx, table, car);
          POP(car, args);
          if (!sen_obj2int(ctx, car)) { limit = car->u.i.i; }
          if (limit <= 0) { limit += sen_table_size(ctx, table); }
          POP(res, args);
          if (!RECORDSP(res)) {
            res = table_create(ctx, NULL, 0, SEN_OBJ_TABLE_NO_KEY,
                               table, sizeof(sen_id), sen_enc_none);
          }
          if (ce) {
            sen_table_sort(ctx, table, limit, get_obj(ctx, res), ce->keys, ce->n_keys);
            column_exp_close(ctx, ce);
          } else {
            sen_table_sort(ctx, table, limit, get_obj(ctx, res), NULL, 0);
          }
        }
        break;
      case 'u' :
      case 'U' :
        switch (msg[3]) {
        case 'f' : /* :suffix-search */
        case 'F' :
          POP(car, args);
          if (!BULKP(car)) { QLERR("string expected"); }
          POP(res, args);
          if (res == NIL) {
            if (!(res = rec_obj_new(ctx, table, 0))) { QLERR("rec_obj_new failed"); }
          } else if (!RECORDSP(res)) {
            QLERR("records object expected");
          }
          if (sen_table_search(ctx, table, STRVALUE(car), STRSIZE(car),
                               SEN_SEARCH_SUFFIX, get_obj(ctx, res), sen_sel_or)) {
            QLERR("suffix search failed");
          }
          break;
        case 'b' : /* :subtract */
        case 'B' :
          {
            while (PAIRP(args)) {
              POP(car, args);
              if (!TABLEP(car)) { continue; }
              sen_table_setoperation(ctx, table, RVALUE(car), table, sen_sel_but);
            }
          }
          break;
        default :
          res = F;
          break;
        }
        break;
      default :
        res = F;
        break;
      }
      break;
    case 't' :
      {
        int n;
        sen_id id = SEN_ID_NIL;
        for (n = 0; (id = sen_table_next(ctx, table, id)); n++) ;
        res = (n == sen_table_size(ctx, table)) ? T : F;
      }
      break;
    case 'T' :
      {
        int n = 0;
        sen_id id;
        sen_table_cursor *c;
        if ((c = sen_table_cursor_open(ctx, table, NULL, 0, NULL, 0,
                                       SEN_CURSOR_DESCENDING))) {
          for (; (id = sen_table_cursor_next(ctx, c)); n++) ;
          sen_table_cursor_close(ctx, c);
        }
        res = (n == sen_table_size(ctx, table)) ? T : F;
      }
      break;
    case 'u' : /* :undef */
    case 'U' :
      switch (msg[2]) {
      case 'n' :
      case 'N' :
        switch (msg[3]) {
        case 'd' :
        case 'D' :
          {
            sen_obj *column;
            POP(car, args);
            if (obj2str(car, msg, &msg_size)) { return F; }
            if (!(column = sen_table_column(ctx, table, msg, msg_size))) { return F; }
            if (!(msg_size = sen_obj_name(ctx, column, msg, STRBUF_SIZE))) { return F; }
            res = sen_table_delete(ctx, ctx->impl->db, msg, msg_size) ? F : T;
          }
          break;
        case 'i' : /* :union */
        case 'I' :
          {
            while (PAIRP(args)) {
              POP(car, args);
              if (!TABLEP(car)) { continue; }
              sen_table_setoperation(ctx, table, RVALUE(car), table, sen_sel_or);
            }
          }
          break;
        default :
          res = F;
          break;
        }
        break;
      default :
        res = F;
        break;
      }
      break;
    case '+' : /* :+ (iterator next) */
      {
        sen_id id;
        POP(res, args);
        if (res->header.type == SEN_CELL_OBJECT &&
            res->header.domain == OBJ_DB(table)->id &&
            (id = sen_table_next(ctx, table, res->u.o.id))) {
          res->u.o.id = id;
        } else {
          res = F;
        }
      }
      break;
    case '\0' : /* : (iterator begin) */
      {
        sen_id id;
        id = sen_table_next(ctx, table, SEN_ID_NIL);
        if (id == SEN_ID_NIL) {
          res = F;
        } else {
          SEN_CELL_NEW(ctx, res);
          obj_obj_bind(res, OBJ_DB(table)->id, id);
        }
      }
      break;
    }
    break;
  default : /* :columnname */
    res = table_column(ctx, base, msg, msg_size);
    break;
  }
  if (load) {
    int i;
    sen_cell *s;
    struct _ins_stat *stat;
    for (s = args, i = 0; PAIRP(s); s = CDR(s), i++) {
      car = CAR(s);
      if (obj2str(car, msg, &msg_size)) { QLERR("invalid argument"); }
      CAR(s) = table_column(ctx, base, msg, msg_size);
      if (CAR(s) == F) { QLERR("invalid argument"); }
    }
    if (!(s = sen_cell_alloc(ctx, sizeof(struct _ins_stat)))) { return F; }
    stat = (struct _ins_stat *)s->u.b.value; // todo : not GC safe
    stat->columns = args;
    stat->ncolumns = i + 1;
    stat->nrecs = 0;
    do {
      SEN_QL_CO_WAIT(co, stat);
      if (BULKP(args) && STRSIZE(args)) {
        char *tokbuf[MAXCOLUMNS];
        sen_obj *column;
        sen_cell val, obj, cons, dummy;
        cons.header.type = SEN_CELL_LIST;
        cons.header.impl_flags = 0;
        cons.u.l.car = &val;
        cons.u.l.cdr = NIL;
        val.header.type = SEN_CELL_STR;
        if (sen_str_tok(args->u.b.value, STRSIZE(args), '\t', tokbuf, MAXCOLUMNS, NULL) == stat->ncolumns) {
          sen_cell *o;
          if (sen_obj_lock(ctx, ctx->impl->db, SEN_ID_NIL, -1)) {
            SEN_LOG(sen_log_crit, "ha_table::load lock failed");
          } else {
            o = sen_ql_table_at(ctx, table, args->u.b.value,
                                tokbuf[0] - args->u.b.value,
                                SEN_TABLE_ADD, &obj);
            if (o != F) {
              for (s = stat->columns, i = 1; i < stat->ncolumns; s = CDR(s), i++) {
                val.u.b.value = tokbuf[i - 1] + 1;
                val.u.b.size = tokbuf[i] - val.u.b.value;
                unesc(ctx, &val);
                if (!(column = sen_ctx_get(ctx, CAR(s)->u.o.id))) { /* todo */ }
                column_value(ctx, column, obj.u.o.id, &cons, &dummy); // todo : refine cons
              }
              stat->nrecs++;
            }
            sen_obj_unlock(ctx, ctx->impl->db, SEN_ID_NIL);
          }
        }
      } else {
        co->mode |= SEN_CTX_TAIL;
      }
    } while (!(co->mode & (SEN_CTX_HEAD|SEN_CTX_TAIL)));
    if ((res = sen_cell_new(ctx))) {
      res->header.type = SEN_CELL_INT;
      res->u.i.i = stat->nrecs;
    } else {
      res = F;
    }
  }
  SEN_QL_CO_END(co);
  return res;
}

static sen_cell *
ha_object(sen_ctx *ctx, sen_cell *args, sen_ql_co *co)
{
  char msg[STRBUF_SIZE];
  uint16_t msg_size;
  sen_cell *obj, *car, *res;
  if (!(obj = res = ctx->impl->code)) { QLERR("invalid receiver"); }
  while (PAIRP(args)) {
    POP(car, args);
    if (obj2str(car, msg, &msg_size)) { QLERR("invalid message"); }
    if (*msg == ':' && msg[1] == 'i') {
      res = obj2oid(ctx, obj, NULL);
    } else {
      sen_obj *domain = sen_ctx_get(ctx, obj->header.domain);
      sen_obj *column = sen_obj_get_accessor(ctx, domain, msg, msg_size);
      if (!column) {
        QLERR("invalid column %s", msg);
      }
      res = column_value(ctx, column, obj->u.o.id, args, NULL);
      sen_obj_close(ctx, column);
      POP(car, args);
    }
  }
  return res;
}

static sen_cell *
ha_query(sen_ctx *ctx, sen_cell *args, sen_ql_co *co)
{
  /* args: (str1@bulk) (str2@bulk) .. */
  sen_rc rc;
  sen_cell *x;
  const char **strs;
  sen_query *q;
  unsigned int *str_lens;
  int nstrs, found = 0, score = 0;
  if (!PAIRP(args) || !BULKP(CAR(args))) { QLERR("invalid argument"); }
  for (x = args, nstrs = 0; PAIRP(x) && BULKP(CAR(x)); x = CDR(x)) { nstrs++; }
  if (!(strs = SEN_MALLOC(sizeof(intptr_t) * nstrs * 2))) {
    QLERR("malloc failed");
  }
  str_lens = (unsigned int *)&strs[nstrs];
  for (x = args, nstrs = 0; PAIRP(x) && BULKP(CAR(x)); x = CDR(x)) {
    strs[nstrs] = STRVALUE(CAR(x));
    str_lens[nstrs] = STRSIZE(CAR(x));
    nstrs++;
  }
  q = (sen_query *)ctx->impl->code->u.p.value;
  rc = sen_query_scan(q, strs, str_lens, nstrs, SEN_QUERY_SCAN_NORMALIZE, &found, &score);
  SEN_FREE(strs);
  if (rc) { QLERR("sen_query_scan failed"); }
  if (!found) { return F; }
  SEN_CELL_NEW(ctx, x);
  SETINT(x, score);
  return x;
}

static sen_cell *
sen_obj_query(sen_ctx *ctx, const char *str, unsigned int str_len,
              sen_sel_operator default_op, int max_exprs, sen_encoding encoding)
{
  sen_query *q;
  sen_cell *res = sen_cell_new(ctx);
  if (!res || !(q = sen_query_open(str, str_len, default_op, max_exprs, encoding))) {
    return NULL;
  }
  res->header.type = SEN_QUERY;
  res->header.impl_flags = SEN_CELL_NATIVE|SEN_OBJ_ALLOCATED;
  res->u.p.value = (sen_obj *)q;
  res->u.p.func = ha_query;
 return res;
}

static sen_cell *
nf_toquery(sen_ctx *ctx, sen_cell *args, sen_ql_co *co)
{
  sen_cell *o = F, *s;
  POP(s, args);
  if (BULKP(s)) {
    /* TODO: operator, exprs, encoding */
    if (!(o = sen_obj_query(ctx, s->u.b.value, s->u.b.size, sen_sel_and, 32, ctx->encoding))) {
      QLERR("query_obj_new failed");
    }
  }
  return o;
}

static sen_cell *
nf_toverses(sen_ctx *ctx, sen_cell *args, sen_ql_co *co)
{
  sen_cell *o = F, *s, *car, *key, *value;
  POP(s, args);
  if (PAIRP(s)) {
    sen_obj verses;
    SEN_OBJ_INIT(&verses, SEN_VERSES, 0);
    while (PAIRP(s)) {
      POP(car, s);
      if (BULKP(car)) {
        sen_verses_add(ctx, &verses, STRVALUE(car), STRSIZE(car), 0, SEN_ID_NIL);
      } else if (PAIRP(car)) {
        const char *str = NULL;
        sen_id domain = SEN_ID_NIL;
        unsigned int weight = 0, str_len = 0;
        POP(key, car);
        if (key == INTERN("@")) {
          while (PAIRP(car)) {
            POP(key, car);
            POP(value, car);
            if (key == INTERN(":dic")) {
              /* todo */
            } else if (key == INTERN(":weight")) {
              sen_obj2int(ctx, value);
              weight = (uint32_t) IVALUE(value);
            } else if (key == INTERN(":value") && BULKP(value)) {
              str = STRVALUE(value);
              str_len = STRSIZE(value);
            }
          }
          if (str) {
            sen_verses_add(ctx, &verses, str, str_len, weight, domain);
          }
        }
      }
    }
    SEN_CELL_NEW(ctx, o);
    obj2cell(ctx, &verses, o);
  }
  return o;
}

static sen_cell *
ha_column(sen_ctx *ctx, sen_cell *args, sen_ql_co *co)
{
  char msg[STRBUF_SIZE];
  uint16_t msg_size;
  sen_id base;
  sen_cell *car, *res;
  sen_obj *column;
  if (!(res = ctx->impl->code)) { QLERR("invalid receiver"); }
  base = ctx->impl->code->u.o.id;
  if (!(column = sen_ctx_get(ctx, base))) { QLERR("sen_ctx_get failed"); }
  POP(car, args);
  if (obj2str(car, msg, &msg_size)) { QLERR("invalid message"); }
  switch (*msg) {
  case '\0' :
    {
      if (IDX_COLUMNP(ctx->impl->code)) {
        sen_cell *q;
        sen_sel_operator op;
        POP(q, args);
        if (!QUERYP(q)) {
          if (!BULKP(q)) { return F; }
          if (!(q = sen_obj_query(ctx, q->u.b.value, q->u.b.size, sen_sel_and, 32, ctx->encoding))) {
            QLERR("query_obj_new failed");
          }
        }
        /* TODO: specify record unit */
        /* (idxcolumn query ((column1 weight1) (column2 weight2) ...) records operator+ */
        POP(car, args);
        /* TODO: handle weights */
        POP(res, args);
        if (RECORDSP(res)) {
          char ops[STRBUF_SIZE];
          uint16_t ops_size;
          op = sen_sel_and;
          POP(car, args);
          if (!obj2str(car, ops, &ops_size)) {
            switch (*ops) {
            case '+': op = sen_sel_or; break;
            case '-': op = sen_sel_but; break;
            case '*': op = sen_sel_and; break;
            case '>': op = sen_sel_adjust; break;
            }
          }
        } else {
          sen_obj *table;
          if (!(table = sen_ctx_get(ctx, OBJ_DB(column)->range))) { return F; }
          res = rec_obj_new(ctx, table, 0);
          if (ERRP(ctx, SEN_WARN)) { return F; }
          op = sen_sel_or;
        }
        sen_obj_search(ctx, column, (sen_obj *)q->u.p.value, RVALUE(res), op, NULL);
      } else {
        char name[STRBUF_SIZE];
        uint16_t name_size;
        sen_obj *table;
        POP(car, args);
        if (obj2str(car, name, &name_size)) { return F; }
        if (!(table = sen_column_table(ctx, column))) { return F; }
        res = sen_ql_table_at(ctx, table, name, name_size, 0, NULL);
        if (res != F) {
          // sen_obj_lock(ctx, ctx->impl->db, SEN_ID_NIL, -1))
          column_value(ctx, column, res->u.o.id, args, res);
          // sen_obj_unlock(ctx, ctx->impl->db, SEN_ID_NIL);
        }
      }
    }
    break;
  case ':' :
    switch (msg[1]) {
    case 's' : /* :schema */
    case 'S' :
      {
        switch (column->header.type) {
        case SEN_COLUMN_FIX_SIZE  :
        case SEN_COLUMN_VAR_SIZE  :
          res = CONS(get_cell(ctx, sen_ctx_get(ctx, OBJ_DB(column)->range)), NIL);
          break;
        case SEN_COLUMN_INDEX :
          {
            /* todo :
            sen_db_trigger *t;
            res = CONS(INTERN("::match"), CONS(NIL, NIL));
            for (t = column->triggers; t; t = t->next) {
              if (t->type == sen_db_index_target) {
                res = CONS(get_cell(ctx, sen_ctx_get(ctx, t->target)), res);
              }
            }
            */
            res = CONS(INTERN(":index"), CONS(CONS(INTERN("quote"), CONS(res, NIL)), NIL));
          }
          break;
        case SEN_CELL_PSEUDO_COLUMN    :
          QLERR("not supported yet");
          break;
        default :
          QLERR("invalid column type");
          break;
        }
        /*
        {
          char *p, buf[SEN_PAT_MAX_KEY_SIZE];
          strcpy(buf, _sen_pat_key(ctx->impl->db->keys, base));
          if (!(p = strchr(buf, '.'))) { QLERR("invalid columnname %s", buf); }
          *p = ':';
          res = CONS(INTERN("::def"), CONS(INTERN(p), res));
          *p = '\0';
          res = CONS(INTERN(buf), res);
        }
        */
      }
      break;
    }
    break;
  }
  return res;
}

void
sen_ql_obj_bind(sen_obj *obj, sen_cell *symbol)
{
  symbol->header.type = obj->header.type;
  symbol->header.impl_flags |= SEN_CELL_NATIVE;
  symbol->u.o.id = OBJ_DB(obj)->id;
  symbol->header.domain = OBJ_DB(obj)->header.domain;
  switch (symbol->header.type) {
  case SEN_TABLE_HASH_KEY :
  case SEN_TABLE_PAT_KEY :
  case SEN_TABLE_NO_KEY :
    symbol->u.o.func = ha_table;
    break;
  case SEN_COLUMN_FIX_SIZE :
    symbol->u.o.func = ha_column;
    break;
  case SEN_COLUMN_VAR_SIZE :
    symbol->u.o.func = ha_column;
    break;
  case SEN_COLUMN_INDEX :
    symbol->u.o.func = ha_column;
    break;
  default :
    symbol->u.o.func = ha_void;
    break;
  }
}

static sen_cell *
nf_db(sen_ctx *ctx, sen_cell *args, sen_ql_co *co)
{
  sen_obj *table;
  char msg[STRBUF_SIZE];
  uint16_t msg_size;
  sen_cell *car, *res = ctx->impl->code;
  POP(car, args);
  if (obj2str(car, msg, &msg_size)) { QLERR("invalid argument"); }
  if (*msg == ':') {
    switch (msg[1]) {
    case 'c' : /* :clearlock */
    case 'C' :
      {
        sen_obj *obj;
        void *key, *value;
        uint32_t key_size;
        sen_pat *keys = (sen_pat *)sen_space_keys(ctx->impl->db);
        SEN_PAT_EACH(keys, id, &key, &key_size, &value, {
          obj = sen_ctx_get(ctx, id);
          sen_obj_clear_lock(ctx, obj);
        });
        res = sen_obj_clear_lock(ctx, ctx->impl->db) ? F : T;
      }
      break;
    case 'd' : /* :drop */
    case 'D' :
      {
        sen_hash *columns;
        const char *path;
        char pathbuf[PATH_MAX];
        POP(car, args);
        if (obj2str(car, msg, &msg_size)) { QLERR("invalid argument"); }
        if (!(table = sen_ctx_lookup(ctx, msg, msg_size))) {
          QLERR("Invalid table");
        }
        if (!(columns = sen_hash_create(ctx, NULL, sizeof(sen_id), 0,
                                        SEN_OBJ_TABLE_HASH_KEY|SEN_HASH_TINY,
                                        sen_enc_none))) {
          QLERR("sen_hash_create failed");
        }
        if (sen_table_columns(ctx, table, "", 0, (sen_obj *)columns)) {
          sen_id *key;
          SEN_HASH_EACH(columns, id, &key, NULL, NULL, {
            sen_obj *col = sen_ctx_get(ctx, *key);
            if (col) {
              if ((path = sen_obj_path(ctx, col))) {
                strncpy(pathbuf, path, PATH_MAX);
                sen_obj_close(ctx, col);
                sen_obj_remove(ctx, pathbuf);
              } else {
                sen_obj_close(ctx, col);
              }
            }
          });
        }
        sen_hash_close(ctx, columns);
        if ((path = sen_obj_path(ctx, table))) {
          strncpy(pathbuf, path, PATH_MAX);
          sen_obj_close(ctx, table);
          sen_obj_remove(ctx, pathbuf);
        } else {
          sen_obj_close(ctx, table);
        }
      }
      break;
    case 'p' : /* :prefix-search */
    case 'P' :
      {
        sen_hash *r;
        sen_pat *keys = (sen_pat *)sen_space_keys(ctx->impl->db);
        POP(car, args);
        if (obj2str(car, msg, &msg_size)) { QLERR("invalid argument"); }
        if ((res = rec_obj_new(ctx, ctx->impl->db, 0)) == F) {
          QLERR("rec_obj_new failed.");
        }
        r = (sen_hash *)sen_ctx_get(ctx, res->u.o.id);
        sen_pat_prefix_search(ctx, keys, msg, msg_size, r);
        {
          sen_id *rid;
          SEN_HASH_EACH(r, id, &rid, NULL, NULL, {
            uint32_t key_size;
            const char *key = _sen_pat_key(keys, *rid, &key_size);
            while (key_size--) {
              if (*key++ == '.') {
                sen_hash_delete(ctx, r, rid, sizeof(sen_id), NULL);
              }
            }
          });
        }
      }
      break;
    case 't' : /* :typedef */
    case 'T' :
      {
        uint32_t size;
        sen_obj *type;
        sen_cell *flag;
        sen_obj_flags flags = 0;
        POP(car, args);
        POP(flag, args);
        if (obj2str(flag, msg, &msg_size)) { QLERR("invalid argument"); }
        flags = (*msg == 'v' || *msg == 'V') ? SEN_OBJ_KEY_VAR_SIZE : 0;
        if (obj2str(car, msg, &msg_size)) { QLERR("invalid argument"); }
        POP(car, args);
        size = (uint32_t) IVALUE(car);
        type = sen_type_create(ctx, msg, msg_size, flags, size);
        if (!type) { QLERR("type_create failed"); }
        if ((res = INTERN2(msg, msg_size)) != F) {
          sen_ql_obj_bind(type, res);
        }
      }
      break;
    case '+' : /* :+ (iterator next) */
      {
        sen_pat *keys = (sen_pat *)sen_space_keys(ctx->impl->db);
        POP(res, args);
        if (res->header.type == SEN_TYPE ||
            res->header.type == SEN_TABLE_PAT_KEY ||
            res->header.type == SEN_COLUMN_FIX_SIZE ||
            res->header.type == SEN_COLUMN_VAR_SIZE ||
            res->header.type == SEN_COLUMN_INDEX ||
            res->header.type == SEN_CELL_PSEUDO_COLUMN) {
          sen_id id = res->u.o.id;
          while ((id = sen_pat_next(ctx, keys, id))) {
            if ((msg_size = sen_pat_get_key(ctx, keys, id, msg, STRBUF_SIZE))) { break; }
          }
          if (id == SEN_ID_NIL) {
            res = F;
          } else {
            res = INTERN2(msg, msg_size);
          }
        } else {
          res = F;
        }
      }
      break;
    case '\0' : /* : (iterator begin) */
      {
        sen_id id = SEN_ID_NIL;
        sen_pat *keys = (sen_pat *)sen_space_keys(ctx->impl->db);
        while ((id = sen_pat_next(ctx, keys, id))) {
          if ((msg_size = sen_pat_get_key(ctx, keys, id, msg, STRBUF_SIZE))) { break; }
        }
        if (id == SEN_ID_NIL) {
          res = F;
        } else {
          res = INTERN2(msg, msg_size);
        }
      }
      break;
    }
  }
  return res;
}

static sen_cell *
nf_table_(sen_ctx *ctx, sen_cell *args, const char *name, uint16_t name_size)
{
  uint32_t value_size = 0;
  sen_obj_flags flags = (name && name_size) ? SEN_OBJ_PERSISTENT : SEN_OBJ_TEMPORARY;
  sen_encoding encoding = sen_enc_default;
  sen_obj *domain = sen_ctx_get(ctx, SEN_QL_SHORTTEXT);
  sen_cell *car;
  char msg[STRBUF_SIZE];
  uint16_t msg_size;
  while (PAIRP(args)) {
    POP(car, args);
    switch (car->header.type) {
    case SEN_TABLE_HASH_KEY :
    case SEN_TABLE_PAT_KEY :
    case SEN_TABLE_NO_KEY :
    case SEN_TYPE :
      domain = sen_ctx_get(ctx, car->u.o.id);
      break;
    case SEN_CELL_INT :
      value_size = (uint32_t) IVALUE(car);
      break;
    default :
      if (obj2str(car, msg, &msg_size)) { QLERR("invalid argument"); }
      switch (*msg) {
      case 'd' :
      case 'D' :
        switch (msg[2]) {
        case 'l' :  /* delimited */
        case 'L' :
          flags |= SEN_OBJ_TOKEN_DELIMITED;
          break;
        case 'f' :  /* default */
        case 'F' :
          encoding = sen_enc_default;
          break;
        }
        break;
      case 'e' : /* euc-jp */
      case 'E' :
        encoding = sen_enc_euc_jp;
        break;
      case 'k' : /* koi8r */
      case 'K' :
        encoding = sen_enc_koi8r;
        break;
      case 'l' : /* latin1 */
      case 'L' :
        encoding = sen_enc_latin1;
        break;
      case 'n' :
      case 'N' :
        switch (msg[1]) {
        case 'g' : /* ngram */
        case 'G' :
          flags |= SEN_OBJ_TOKEN_NGRAM;
          break;
        case 'o' : /* normalize */
        case 'O' :
          flags |= SEN_OBJ_KEY_NORMALIZE;
          break;
        default :
          QLERR("ambiguous option %s", msg);
        }
        break;
      case 'p' :
      case 'P' :
        switch (msg[1]) {
        case 'a' :
        case 'A' :
          flags |= SEN_OBJ_TABLE_PAT_KEY;
          break;
        case 'e' :
        case 'E' :
          flags |= SEN_OBJ_PERSISTENT;
          break;
        default :
          QLERR("ambiguous option %s", msg);
        }
        break;
      case 's' :
      case 'S' :
        switch (msg[1]) {
        case 'j' : /* shift-jis */
        case 'J' :
          encoding = sen_enc_sjis;
          break;
        case 'i' : /* with-sis */
        case 'I' :
          flags |= SEN_OBJ_KEY_WITH_SIS;
          break;
        case 'u' : /* surrogate-key */
        case 'U' :
          flags |= SEN_OBJ_TABLE_NO_KEY;
          break;
        default :
          QLERR("ambiguous option %s", msg);
        }
        break;
      case 'u' : /* utf8 */
      case 'U' :
        encoding = sen_enc_utf8;
        break;
      case 'v' : /* view */
      case 'V' :
        /* todo */
        break;
      default :
        QLERR("illegal option");
        break;
      }
    }
  }
  return table_create(ctx, name, name_size, flags, domain, value_size, encoding);
}

static sen_cell *
nf_ptable(sen_ctx *ctx, sen_cell *args, sen_ql_co *co)
{
  char name[STRBUF_SIZE];
  uint16_t name_size;
  sen_cell *car;
  POP(car, args);
  if (obj2str(car, name, &name_size)) { QLERR("invalid argument"); }
  if (sen_ctx_lookup(ctx, name, name_size)) { return T; }
  return nf_table_(ctx, args, name, name_size);
}

static sen_cell *
nf_table(sen_ctx *ctx, sen_cell *args, sen_ql_co *co)
{
  return nf_table_(ctx, args, NULL, 0);
}

#define PVALUE(obj,type) ((type *)((obj)->u.p.value))

inline static void
snip_obj_bind(sen_cell *obj, sen_snip *snip)
{
  obj->header.type = SEN_SNIP;
  obj->header.impl_flags = SEN_CELL_NATIVE|SEN_OBJ_ALLOCATED;
  obj->u.p.value = (sen_obj *)snip;
  obj->u.p.func = ha_snip;
}

struct _patsnip_spec {
  sen_obj *table;
  int width;
  int max_results;
  column_exp *ce;
};

void
sen_obj_patsnip_spec_close(sen_ctx *ctx, patsnip_spec *spec)
{
  if (spec) {
    column_exp_close(ctx, spec->ce);
    SEN_FREE(spec);
  }
}

static sen_cell *
nf_snippet(sen_ctx *ctx, sen_cell *args, sen_ql_co *co)
{
  /* args: (cond width@int max_results@int opentag1@bulk closetag1@bulk..)
     cond: (keyword@bulk..) or cond@query or cond@table */

  sen_cell *res, *cur, *cond;
  sen_snip *s;
  unsigned int width = 100, max_results = 3;
  if (!PAIRP(args)) { QLERR("cond expected"); }
  POP(cond, args);
  if (PAIRP(args)) {
    POP(cur, args);
    if (sen_obj2int(ctx, cur)) { QLERR("snippet failed (width expected)"); }
    width = IVALUE(cur);
  }
  if (PAIRP(args)) {
    POP(cur, args);
    if (sen_obj2int(ctx, cur)) { QLERR("snipped failed (max_result expected)"); }
    max_results = IVALUE(cur);
  }
  switch (cond->header.type) {
  case SEN_CELL_LIST :
    {
      sen_cell *tags = args;
      /* FIXME: mapping */
      if (!(s = sen_snip_open(ctx->encoding, SEN_SNIP_NORMALIZE, width, max_results,
                              NULL, 0, NULL, 0, (sen_snip_mapping *)-1))) {
        QLERR("sen_snip_open failed");
      }
      SEN_CELL_NEW(ctx, res);
      snip_obj_bind(res, s);
      s->flags |= SEN_SNIP_COPY_TAG;
      while (PAIRP(cond)) {
        char *ot = NULL, *ct = NULL;
        uint32_t ot_l = 0, ct_l = 0;
        sen_cell *kw;
        POP(kw, cond);
        if (!BULKP(kw)) { QLERR("snippet failed (invalid kw)"); }
        if (!PAIRP(args)) { args = tags; }
        POP(cur, args);
        if (BULKP(cur)) {
          ot = STRVALUE(cur);
          ot_l = STRSIZE(cur);
        }
        POP(cur, args);
        if (BULKP(cur)) {
          ct = STRVALUE(cur);
          ct_l = STRSIZE(cur);
        }
        if ((sen_snip_add_cond(s, kw->u.b.value, kw->u.b.size, ot, ot_l, ct, ct_l))) {
          QLERR("sen_snip_add_cond failed");
        }
      }
    }
    break;
  case SEN_QUERY :
    {
      sen_cell *x;
      sen_query *q;
      unsigned int n_tags = 0;
      const char **opentags, **closetags;
      unsigned int *opentag_lens, *closetag_lens;
      q = (sen_query *)cond->u.p.value;
      for (x = args; PAIRP(x); x = CDR(x)) { n_tags++; }
      if (!n_tags) { n_tags++; }
      if (!(opentags = SEN_MALLOC((sizeof(char *) + sizeof(unsigned int)) * 2 * n_tags))) {
        QLERR("malloc failed");
      }
      closetags = &opentags[n_tags];
      opentag_lens = (unsigned int *)&closetags[n_tags];
      closetag_lens = &opentag_lens[n_tags];
      n_tags = 0;
      for (x = args; PAIRP(x); x = CDR(x)) {
        if (BULKP(CAR(x))) {
          opentags[n_tags] = STRVALUE(CAR(x));
          opentag_lens[n_tags] = STRSIZE(CAR(x));
          if (PAIRP(CDR(x))) {
            x = CDR(x);
            if (BULKP(CAR(x))) {
              closetags[n_tags] = STRVALUE(CAR(x));
              closetag_lens[n_tags] = STRSIZE(CAR(x));
              n_tags++;
            }
          }
        }
      }
      if (!n_tags) {
        n_tags++;
        opentags[0] = NULL;
        closetags[0] = NULL;
        opentag_lens[0] = 0;
        closetag_lens[0] = 0;
      }
      s = sen_query_snip(q, SEN_SNIP_NORMALIZE|SEN_SNIP_COPY_TAG, width, max_results, n_tags,
                         opentags, opentag_lens, closetags, closetag_lens,
                         (sen_snip_mapping *)-1);
      SEN_FREE(opentags);
      if (!s) { QLERR("sen_query_snip failed"); }
      SEN_CELL_NEW(ctx, res);
      snip_obj_bind(res, s);
    }
    break;
  case SEN_TABLE_PAT_KEY :
    {
      patsnip_spec *spec;
      sen_obj *table = sen_ctx_get(ctx, cond->u.o.id);
      if (!table) { QLERR("table get failed."); }
      SEN_CELL_NEW(ctx, res);
      if (!(spec = SEN_MALLOC(sizeof(patsnip_spec)))) {
        QLERR("patsnip_spec malloc failed");
      }
      res->header.type = SEN_PATSNIP;
      res->header.impl_flags = SEN_CELL_NATIVE|SEN_OBJ_ALLOCATED;
      res->u.p.value = (sen_obj *)spec;
      res->u.p.func = ha_snip;
      spec->table = table;
      spec->width = width;
      spec->max_results = max_results;
      spec->ce = column_exp_open(ctx, table, args);
    }
    break;
  default :
    QLERR("snippet failed. cond or query expected");
    break;
  }
  return res;
}

static sen_cell *
ha_snip(sen_ctx *ctx, sen_cell *args, sen_ql_co *co)
{
  /* args: (str@bulk) */
  if (!PAIRP(args) || !BULKP(CAR(args))) { QLERR("invalid argument"); }
  switch (ctx->impl->code->header.type) {
  case SEN_SNIP :
    {
      sen_obj buf;
      unsigned int i, len, max_len, nresults;
      sen_snip *s = PVALUE(ctx->impl->code, sen_snip);
      sen_cell *v, *str = CAR(args), *spc = PAIRP(CDR(args)) ? CADR(args) : NIL;
      if ((sen_snip_exec(s, str->u.b.value, str->u.b.size, &nresults, &max_len))) {
        QLERR("sen_snip_exec failed");
      }
      if (sen_bulk_init(ctx, &buf, max_len)) { QLERR("sen_bulk_init failed"); }
      if (nresults) {
        for (i = 0; i < nresults; i++) {
          if (i && spc != NIL) { sen_obj_inspect(ctx, spc, &buf, 0); }
          if (sen_bulk_reserve(ctx, &buf, max_len)) {
            sen_bulk_fin(ctx, &buf);
            QLERR("sen_bulk_space failed");
          }
          if ((sen_snip_get_result(s, i, buf.u.b.curr, &len))) {
            sen_bulk_fin(ctx, &buf);
            QLERR("sen_snip_get_result failed");
          }
          buf.u.b.curr += len;
        }
      } else {
        char *ss = str->u.b.value, *se = str->u.b.value + str->u.b.size;
        if (sen_substring(&ss, &se, 0, s->width, ctx->encoding)) {
          QLERR("sen_substring failed");
        }
        sen_bulk_write(ctx, &buf, ss, se - ss);
      }
      SEN_STR2OBJ(ctx, &buf, v);
      return v;
    }
    break;
  case SEN_PATSNIP :
    {
      sen_obj buf;
      patsnip_spec *spec = PVALUE(ctx->impl->code, patsnip_spec);
      off_t off = 0;
      const char *rest;
      sen_pat_scan_hit sh[1024];
      sen_cell *v, *expr, *str = CAR(args);
      char *string = STRVALUE(str);
      size_t len = STRSIZE(str);
      if (sen_bulk_init(ctx, &buf, len)) { QLERR("sen_bulk_init failed."); }
      while (off < len) {
        sen_obj *table = spec->table;
        int i, nhits = sen_pat_scan(ctx, (sen_pat *)table, string + off, len - off,
                                    sh, 1024, &rest);
        for (i = 0, off = 0; i < nhits; i++) {
          if (sh[i].offset < off) { continue; } /* skip overlapping region. */
          sen_bulk_write(ctx, &buf, string + off, sh[i].offset - off);
          column_exp_exec(ctx, spec->ce, sh[i].id);
          expr = spec->ce->expr;
          //          expr = sen_ql_eval(ctx, spec->ce->expr, NIL);
          POP(v, expr);
          sen_obj_inspect(ctx, sen_ql_eval(ctx, v, NIL), &buf, 0);
          sen_bulk_write(ctx, &buf, string + sh[i].offset, sh[i].length);
          POP(v, expr);
          sen_obj_inspect(ctx, sen_ql_eval(ctx, v, NIL), &buf, 0);
          off = sh[i].offset + sh[i].length;
        }
        if (string + off < rest) {
          sen_bulk_write(ctx, &buf, string + off, rest - (string + off));
        }
        off = rest - string;
      }
      SEN_STR2OBJ(ctx, &buf, v);
      return v;
    }
    break;
  default :
    QLERR("snip failed. invalid expr");
  }
}

static void disp_j(sen_ctx *ctx, sen_cell *obj, sen_obj *buf);

static void
disp_j_with_format(sen_ctx *ctx, sen_cell *args, sen_obj *buf)
{
  sen_cell *car;
  POP(car, args);
  switch (car->header.type) {
  case SEN_TABLE_PAT_KEY :
  case SEN_TABLE_HASH_KEY :
  case SEN_TABLE_NO_KEY :
    {
      sen_id id, base;
      sen_cell *columns;
      int i, offset = 0, limit = 10, flags = SEN_CURSOR_ASCENDING;
      sen_obj *r = RVALUE(car);
      column_exp *ce;
      base = car->u.o.id;
      POP(columns, args);
      if (!PAIRP(columns)) {
        disp_j(ctx, car, buf);
        return;
      }
      if ((ce = column_exp_open(ctx, r, columns))) {
        POP(car, args);
        if (!sen_obj2int(ctx, car)) { offset = car->u.i.i; }
        POP(car, args);
        if (!sen_obj2int(ctx, car)) { limit = car->u.i.i; }
        if (limit <= 0) { limit += sen_table_size(ctx, r); }
        POP(car, args);
        {
          char msg[STRBUF_SIZE];
          uint16_t msg_size;
          if (!obj2str(car, msg, &msg_size) && (*msg == 'd')) {
            flags &= ~SEN_CURSOR_ASCENDING;
          }
        }
        {
          sen_table_cursor *tc = sen_table_cursor_open_by_id(ctx, r, offset,
                                                             offset + limit, flags);
          SEN_BULK_PUTC(ctx, buf, '[');
          for (i = 0; (id = sen_table_cursor_next(ctx, tc)); i++) {
            if (i) { SEN_BULK_PUTS(ctx, buf, ", "); }
            column_exp_exec(ctx, ce, id);
            disp_j(ctx, ce->expr, buf);
          }
          SEN_BULK_PUTC(ctx, buf, ']');
          sen_table_cursor_close(ctx, tc);
        }
        column_exp_close(ctx, ce);
      }
    }
    break;
  case SEN_CELL_OBJECT :
    {
      sen_cell *columns;
      column_exp *ce;
      sen_obj *r = get_domain(ctx, car);
      POP(columns, args);
      if (!PAIRP(columns)) {
        disp_j(ctx, car, buf);
        return;
      }
      if ((ce = column_exp_open(ctx, r, columns))) {
        column_exp_exec(ctx, ce, car->u.o.id);
        disp_j(ctx, ce->expr, buf);
        column_exp_close(ctx, ce);
      }
    }
    break;
  default :
    disp_j(ctx, car, buf);
    if (ERRP(ctx, SEN_WARN)) { return; }
    break;
  }
}

static void
disp_j(sen_ctx *ctx, sen_cell *obj, sen_obj *buf)
{
  if (!obj || obj == NIL) {
    SEN_BULK_PUTS(ctx, buf, "[]");
  } else if (obj == T) {
    SEN_BULK_PUTS(ctx, buf, "true");
  } else if (obj == F) {
    SEN_BULK_PUTS(ctx, buf, "false");
  } else {
    switch (obj->header.type) {
    case SEN_VOID :
      if (SYMBOLP(obj) && obj != INTERN("null")) {
        uint16_t size = 0;
        const char *r = _sen_hash_strkey_by_val(obj, &size);
        if (size && (*r == ':')) {
          r++;
          size--;
        }
        sen_bulk_esc(ctx, buf, r, size, ctx->encoding);
      } else {
        SEN_BULK_PUTS(ctx, buf, "null");
      }
      break;
    case SEN_TABLE_HASH_KEY :
    case SEN_TABLE_PAT_KEY :
    case SEN_TABLE_NO_KEY :
      {
        int i;
        sen_id id;
        sen_cell o;
        sen_obj *r = RVALUE(obj);
        {
          sen_table_cursor *tc = sen_table_cursor_open_by_id(ctx, r, 0, 0,
                                                             SEN_CURSOR_ASCENDING);
          SEN_BULK_PUTC(ctx, buf, '[');
          for (i = 0; (id = sen_table_cursor_next(ctx, tc)); i++) {
            if (i) { SEN_BULK_PUTS(ctx, buf, ", "); }
            obj_obj_bind(&o, obj->u.o.id, id);
            disp_j(ctx, &o, buf);
            if (ERRP(ctx, SEN_WARN)) { return; }
          }
          SEN_BULK_PUTC(ctx, buf, ']');
          sen_table_cursor_close(ctx, tc);
        }
      }
      break;
    case SEN_CELL_LIST :
      if (CAR(obj) == INTERN(":")) {
        disp_j_with_format(ctx, CDR(obj), buf);
        if (ERRP(ctx, SEN_WARN)) { return; }
      } else if (CAR(obj) == INTERN("@")) {
        int o;
        SEN_BULK_PUTC(ctx, buf, '{');
        for (obj = CDR(obj), o = 0;; o = 1 - o) {
          if (PAIRP(obj)) {
            disp_j(ctx, CAR(obj), buf);
            if (ERRP(ctx, SEN_WARN)) { return; }
          }
          if ((obj = CDR(obj)) && (obj != NIL)) {
            if (PAIRP(obj)) {
              SEN_BULK_PUTS(ctx, buf, o ? ", " : ": ");
            } else {
              SEN_BULK_PUTS(ctx, buf, " . ");
              disp_j(ctx, obj, buf);
              if (ERRP(ctx, SEN_WARN)) { return; }
              SEN_BULK_PUTC(ctx, buf, '}');
              break;
            }
          } else {
            SEN_BULK_PUTC(ctx, buf, '}');
            break;
          }
        }
      } else {
        SEN_BULK_PUTC(ctx, buf, '[');
        for (;;) {
          disp_j(ctx, CAR(obj), buf);
          if (ERRP(ctx, SEN_WARN)) { return; }
          if ((obj = CDR(obj)) && (obj != NIL)) {
            if (PAIRP(obj)) {
              SEN_BULK_PUTS(ctx, buf, ", ");
            } else {
              SEN_BULK_PUTS(ctx, buf, " . ");
              disp_j(ctx, obj, buf);
              if (ERRP(ctx, SEN_WARN)) { return; }
              SEN_BULK_PUTC(ctx, buf, ']');
              break;
            }
          } else {
            SEN_BULK_PUTC(ctx, buf, ']');
            break;
          }
        }
      }
      break;
    case SEN_CELL_OBJECT :
      {
        sen_obj key;
        SEN_OBJ_INIT(&key, SEN_BULK, 0);
        sen_ql_obj_key(ctx, obj, &key);
        sen_bulk_esc(ctx, buf, SEN_BULK_HEAD(&key), SEN_BULK_VSIZE(&key), ctx->encoding);
        sen_obj_close(ctx, &key);
      }
      break;
    case SEN_CELL_TIME :
      {
        double dv= obj->u.tv.tv_sec;
        dv += obj->u.tv.tv_usec / 1000000.0;
        sen_bulk_ftoa(ctx, buf, dv);
      }
      break;
    default :
      sen_obj_inspect(ctx, obj, buf, SEN_OBJ_INSPECT_ESC|SEN_OBJ_INSPECT_SYMBOL_AS_STR);
      break;
    }
  }
}

static void disp_t(sen_ctx *ctx, sen_cell *obj, sen_obj *buf, int *f);

static void
disp_t_with_format(sen_ctx *ctx, sen_cell *args, sen_obj *buf, int *f)
{
  sen_cell *car;
  POP(car, args);
  switch (car->header.type) {
  case SEN_TABLE_PAT_KEY :
  case SEN_TABLE_HASH_KEY :
  case SEN_TABLE_NO_KEY :
    {
      sen_id id, base;
      sen_cell *columns;
      int i, offset = 0, limit = 10, flags = SEN_CURSOR_ASCENDING;
      sen_obj *r = RVALUE(car);
      column_exp *ce;
      base = car->u.o.id;
      POP(columns, args);
      if (!PAIRP(columns)) {
        disp_t(ctx, car, buf, f);
        return;
      }
      if ((ce = column_exp_open(ctx, r, columns))) {
        POP(car, args);
        if (!sen_obj2int(ctx, car)) { offset = car->u.i.i; }
        POP(car, args);
        if (!sen_obj2int(ctx, car)) { limit = car->u.i.i; }
        if (limit <= 0) { limit += sen_table_size(ctx, r); }
        POP(car, args);
        {
          char msg[STRBUF_SIZE];
          uint16_t msg_size;
          if (!obj2str(car, msg, &msg_size) && (*msg == 'd')) {
            flags &= ~SEN_CURSOR_ASCENDING;
          }
        }
        {
          sen_table_cursor *tc = sen_table_cursor_open_by_id(ctx, r, offset,
                                                             offset + limit, flags);
          for (i = 0; (id = sen_table_cursor_next(ctx, tc)); i++) {
            if (*f) { ctx->impl->output(ctx, SEN_CTX_MORE, ctx->impl->data.ptr); *f = 0; }
            column_exp_exec(ctx, ce, id);
            disp_t(ctx, ce->expr, buf, f);
          }
          sen_table_cursor_close(ctx, tc);
        }
        column_exp_close(ctx, ce);
      }
    }
    break;
  case SEN_CELL_OBJECT :
    {
      sen_cell *columns;
      column_exp *ce;
      sen_obj *r = get_domain(ctx, car);
      POP(columns, args);
      if (!PAIRP(columns)) {
        disp_t(ctx, car, buf, f);
        return;
      }
      if ((ce = column_exp_open(ctx, r, columns))) {
        column_exp_exec(ctx, ce, car->u.o.id);
        disp_t(ctx, ce->expr, buf, f);
        column_exp_close(ctx, ce);
      }
    }
    break;
  default :
    disp_t(ctx, car, buf, f);
    break;
  }
}

inline static void
bulk_tsv_esc(sen_ctx *ctx, sen_obj *buf, const char *s, int len, sen_encoding encoding)
{
  const char *e;
  unsigned int l;
  for (e = s + len; s < e; s += l) {
    if (!(l = sen_str_charlen_nonnull(s, e, encoding))) { break; }
    if (l == 1) {
      switch (*s) {
      case '\t' :
        sen_bulk_write(ctx, buf, "\\t", 2);
        break;
#ifdef SEN_QL_ESCAPE_NEWLINE
      case '\n' :
        sen_bulk_write(ctx, buf, "\\n", 2);
        break;
      case '\r' :
        sen_bulk_write(ctx, buf, "\\r", 2);
        break;
#endif /* SEN_QL_ESCAPE_NEWLINE */
      case '\\' :
        sen_bulk_write(ctx, buf, "\\\\", 2);
        break;
      default :
        SEN_BULK_PUTC(ctx, buf, *s);
      }
    } else {
      sen_bulk_write(ctx, buf, s, l);
    }
  }
}

static void
disp_t(sen_ctx *ctx, sen_cell *obj, sen_obj *buf, int *f)
{
  if (!obj || obj == NIL) {
    SEN_BULK_PUTS(ctx, buf, "()"); *f = 1;
  } else if (obj == T) {
    SEN_BULK_PUTS(ctx, buf, "#t"); *f = 1;
  } else if (obj == F) {
    SEN_BULK_PUTS(ctx, buf, "#f"); *f = 1;
  } else {
    switch (obj->header.type) {
    case SEN_TABLE_HASH_KEY :
    case SEN_TABLE_PAT_KEY :
    case SEN_TABLE_NO_KEY :
      {
        int i;
        sen_id id;
        sen_cell o;
        sen_obj *r = RVALUE(obj);
        sen_table_cursor *tc = sen_table_cursor_open_by_id(ctx, r, 0, 0,
                                                           SEN_CURSOR_ASCENDING);
        for (i = 0; (id = sen_table_cursor_next(ctx, tc)); i++) {
          obj_obj_bind(&o, obj->u.o.id, id);
          if (*f) { ctx->impl->output(ctx, SEN_CTX_MORE, ctx->impl->data.ptr); *f = 0; }
          disp_t(ctx, &o, buf, f);
        }
        sen_table_cursor_close(ctx, tc);
      }
      break;
    case SEN_CELL_LIST :
      if (CAR(obj) == INTERN(":")) {
        disp_t_with_format(ctx, CDR(obj), buf, f);
      } else if (CAR(obj) == INTERN("@")) {
        int o0, o;
        sen_cell *val = CDR(obj);
        for (o0 = 0; o0 <= 1; o0++) {
          if (*f) { ctx->impl->output(ctx, SEN_CTX_MORE, ctx->impl->data.ptr); *f = 0; }
          for (obj = val, o = o0;; o = 1 - o) {
            if (!o) { disp_t(ctx, CAR(obj), buf, f); }
            if ((obj = CDR(obj)) && (obj != NIL)) {
              if (PAIRP(obj)) {
                if (!o && PAIRP(CDR(obj))) { SEN_BULK_PUTC(ctx, buf, '\t'); *f = 1; }
              } else {
                if (!o) {
                  SEN_BULK_PUTC(ctx, buf, '\t'); *f = 1; /* dot pair */
                  disp_t(ctx, obj, buf, f);
                }
                break;
              }
            } else {
              break;
            }
          }
        }
      } else {
        sen_cell *car;
        for (;;) {
          POP(car, obj);
          if (PAIRP(car)) {
            car = sen_ql_eval(ctx, car, NIL);
          }
          disp_t(ctx, car, buf, f);
          if ((obj != NIL)) {
            if (PAIRP(obj)) {
              SEN_BULK_PUTC(ctx, buf, '\t'); *f = 1;
            } else {
              SEN_BULK_PUTC(ctx, buf, '\t'); *f = 1; /* dot pair */
              disp_t(ctx, obj, buf, f);
              break;
            }
          } else {
            break;
          }
        }
      }
      break;
    case SEN_CELL_STR :
      bulk_tsv_esc(ctx, buf, obj->u.b.value, obj->u.b.size, ctx->encoding);
      *f = 1;
      break;
    case SEN_CELL_TIME :
      {
        double dv= obj->u.tv.tv_sec;
        dv += obj->u.tv.tv_usec / 1000000.0;
        sen_bulk_ftoa(ctx, buf, dv);
        *f = 1;
      }
      break;
    default :
      sen_obj_inspect(ctx, obj, buf, 0); *f = 1;
      break;
    }
  }
}

static sen_cell *
nf_disp(sen_ctx *ctx, sen_cell *args, sen_ql_co *co)
{
  char str[STRBUF_SIZE];
  uint16_t str_size;
  int f = 0;
  sen_cell *val, *fmt;
  POP(val, args);
  POP(fmt, args);
  if (!obj2str(fmt, str, &str_size)) {
    switch (str[0]) {
    case 'j' : /* json */
    case 'J' :
      disp_j(ctx, val, &ctx->impl->outbuf);
      f = 1;
      if (ERRP(ctx, SEN_WARN)) { return F; }
      break;
    case 's' : /* sexp */
    case 'S' :
      break;
    case 't' : /* tsv */
    case 'T' :
      disp_t(ctx, val, &ctx->impl->outbuf, &f);
      if (ERRP(ctx, SEN_WARN)) { return F; }
      break;
    case 'x' : /* xml */
    case 'X' :
      break;
    }
  } else {
    QLERR("Few arguments");
  }
  if (f) {
    ctx->impl->output(ctx, SEN_CTX_MORE, ctx->impl->data.ptr);
    if (ERRP(ctx, SEN_WARN)) { return F; }
  }
  return T;
}

typedef struct {
  sen_encoding encoding;
  char *cur;
  char *str_end;
} jctx;

inline static sen_cell *
mk_atom(sen_ctx *ctx, char *str, unsigned int len)
{
  const char *cur, *str_end = str + len;
  int64_t ivalue = sen_atoll(str, str_end, &cur);
  if (cur == str_end) {
    sen_cell *x;
    SEN_CELL_NEW(ctx, x);
    SETINT(x, ivalue);
    return x;
  }
  switch (*str) {
  case 't' :
    if (len == 4 && !memcmp(str, "true", 4)) { return T; }
    break;
  case 'f' :
    if (len == 5 && !memcmp(str, "false", 5)) { return F; }
    break;
    /*
  case 'n' :
    if (len == 4 && !memcmp(str, "null", 4)) { return NIL; }
    break;
    */
  }
  if (0 < len && len < SEN_PAT_MAX_KEY_SIZE - 1) {
    char buf[SEN_PAT_MAX_KEY_SIZE];
    memcpy(buf, str, len);
    buf[len] = '\0';
    return INTERN(buf);
  } else {
    return F;
  }
}

inline static sen_cell *
json_readstr(sen_ctx *ctx, jctx *jc)
{
  char *start, *end;
  for (start = end = jc->cur;;) {
    unsigned int len;
    /* null check and length check */
    if (!(len = sen_str_charlen_nonnull(end, jc->str_end, jc->encoding))) {
      jc->cur = jc->str_end;
      break;
    }
    if (sen_isspace(end, jc->encoding)
        || *end == ':' || *end == ','
        || *end == '[' || *end == '{'
        || *end == ']' || *end == '}') {
      jc->cur = end;
      break;
    }
    end += len;
  }
  if (start < end || jc->cur < jc->str_end) {
    return mk_atom(ctx, start, end - start);
  } else {
    return F;
  }
}

inline static sen_cell *
json_readstrexp(sen_ctx *ctx, jctx *jc, int keyp)
{
  sen_cell *res;
  char *start, *src, *dest;
  for (start = src = dest = jc->cur;;) {
    unsigned int len;
    /* null check and length check */
    if (!(len = sen_str_charlen_nonnull(src, jc->str_end, jc->encoding))) {
      jc->cur = jc->str_end;
      if (start < dest) {
        res = keyp
          ? sen_ql_mk_symbol2(ctx, start, dest - start, 1)
          : sen_ql_mk_string(ctx, start, dest - start);
        return res ? res : F;
      }
      return F;
    }
    if (src[0] == '"' && len == 1) {
      jc->cur = src + 1;
      res = keyp
        ? sen_ql_mk_symbol2(ctx, start, dest - start, 1)
        : sen_ql_mk_string(ctx, start, dest - start);
      return res ? res : F;
    } else if (src[0] == '\\' && src + 1 < jc->str_end && len == 1) {
      src++;
      *dest++ = *src++;
    } else {
      while (len--) { *dest++ = *src++; }
    }
  }
}

static sen_cell *
json_read(sen_ctx *ctx, jctx *jc, int keyp)
{
  for (;;) {
    SKIPSPACE(jc);
    if (jc->cur >= jc->str_end) { return NULL; }
    switch (*jc->cur) {
    case '[':
      jc->cur++;
      {
        sen_cell *o, *r = NIL, **p = &r;
        while ((o = json_read(ctx, jc, 0)) && o != F) {
          *p = CONS(o, NIL);
          if (ERRP(ctx, SEN_WARN)) { return F; }
          p = &CDR(*p);
        }
        return r;
      }
    case '{':
      jc->cur++;
      {
        sen_cell *o, *r = CONS(INTERN("@"), NIL), **p = &(CDR(r));
        int i = 0;
        while ((o = json_read(ctx, jc, ((++i)&1))) && o != F) {
          *p = CONS(o, NIL);
          if (ERRP(ctx, SEN_WARN)) { return F; }
          p = &CDR(*p);
        }
        return r;
      }
    case '}':
    case ']':
      jc->cur++;
      return NULL;
    case ',':
      jc->cur++;
      break;
    case ':':
      jc->cur++;
      break;
    case '"':
      jc->cur++;
      return json_readstrexp(ctx, jc, keyp);
    default:
      return json_readstr(ctx, jc);
    }
  }
}

static sen_cell *
nf_json_read(sen_ctx *ctx, sen_cell *args, sen_ql_co *co)
{
  sen_cell *car;
  POP(car, args); // todo : delete when called with (())
  if (BULKP(car)) {
    sen_cell *r;
    jctx jc;
    jc.encoding = ctx->encoding;
    jc.cur = car->u.b.value;
    jc.str_end = car->u.b.value + car->u.b.size;
    if ((r = json_read(ctx, &jc, 0))) { return r; }
  }
  return F;
}

static sen_obj *
deftype(sen_ctx *ctx, const char *name,
        sen_obj_flags flags,  unsigned int size)
{
  sen_obj *o = sen_ctx_lookup(ctx, name, strlen(name));
  if (!o) { o = sen_type_create(ctx, name, strlen(name), flags, size); }
  return o;
}

sen_rc
sen_ql_def_builtin_types(sen_ctx *ctx)
{
  sen_obj *obj;
  obj = deftype(ctx, "<int>",
                SEN_OBJ_KEY_INT, sizeof(int32_t));
  if (!obj || OBJ_DB(obj)->id != SEN_QL_INT) { return sen_invalid_format; }
  obj = deftype(ctx, "<uint>",
                SEN_OBJ_KEY_UINT, sizeof(uint32_t));
  if (!obj || OBJ_DB(obj)->id != SEN_QL_UINT) { return sen_invalid_format; }
  obj = deftype(ctx, "<int64>",
                SEN_OBJ_KEY_INT, sizeof(int64_t));
  if (!obj || OBJ_DB(obj)->id != SEN_QL_INT64) { return sen_invalid_format; }
  obj = deftype(ctx, "<float>",
                SEN_OBJ_KEY_FLOAT, sizeof(double));
  if (!obj || OBJ_DB(obj)->id != SEN_QL_FLOAT) { return sen_invalid_format; }
  obj = deftype(ctx, "<time>",
                SEN_OBJ_KEY_UINT, sizeof(sen_timeval));
  if (!obj || OBJ_DB(obj)->id != SEN_QL_TIME) { return sen_invalid_format; }
  obj = deftype(ctx, "<shorttext>",
                SEN_OBJ_KEY_VAR_SIZE, SEN_TABLE_MAX_KEY_SIZE);
  if (!obj || OBJ_DB(obj)->id != SEN_QL_SHORTTEXT) { return sen_invalid_format; }
  obj = deftype(ctx, "<text>",
                SEN_OBJ_KEY_VAR_SIZE, 1 << 16);
  if (!obj || OBJ_DB(obj)->id != SEN_QL_TEXT) { return sen_invalid_format; }
  obj = deftype(ctx, "<longtext>",
                SEN_OBJ_KEY_VAR_SIZE, 1 << 31);
  if (!obj || OBJ_DB(obj)->id != SEN_QL_LONGTEXT) { return sen_invalid_format; }
  return sen_success;
}

sen_rc
sen_ql_def_db_funcs(sen_ctx *ctx)
{
  sen_ql_def_native_func(ctx, "<db>", nf_db);
  sen_ql_def_native_func(ctx, "table", nf_table);
  sen_ql_def_native_func(ctx, "ptable", nf_ptable);
  sen_ql_def_native_func(ctx, "snippet", nf_snippet);
  sen_ql_def_native_func(ctx, "disp", nf_disp);
  sen_ql_def_native_func(ctx, "json-read", nf_json_read);
  sen_ql_def_native_func(ctx, "x->query", nf_toquery);
  sen_ql_def_native_func(ctx, "x->verses", nf_toverses);
  return (ctx->impl->db) ? sen_ql_def_builtin_types(ctx) : sen_success;
}
