/* Copyright(C) 2004-2007 Brazil

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "senna_in.h"
#include "str.h"
#include "inv.h"
#include "sym.h"
#include "store.h"
#include "ctx.h"
#include "ql.h"
#include <string.h>

/* rectangular arrays */

#define SEN_RA_IDSTR "SENNA:RA:01.000"
#define SEN_RA_SEGMENT_SIZE (1 << 22)

sen_ra *
sen_ra_create(sen_ctx *ctx, const char *path, unsigned int element_size)
{
  sen_io *io;
  int max_segments, n_elm, w_elm;
  sen_ra *ra = NULL;
  struct sen_ra_header *header;
  unsigned actual_size;
  if (element_size > SEN_RA_SEGMENT_SIZE) {
    SEN_LOG(sen_log_error, "element_size too large (%d)", element_size);
    return NULL;
  }
  for (actual_size = 1; actual_size < element_size; actual_size *= 2) ;
  max_segments = ((SEN_ID_MAX + 1) / SEN_RA_SEGMENT_SIZE) * actual_size;
  io = sen_io_create(path, sizeof(struct sen_ra_header),
                     SEN_RA_SEGMENT_SIZE, max_segments, sen_io_auto, 0);
  if (!io) { return NULL; }
  header = sen_io_header(io);
  memcpy(header->idstr, SEN_RA_IDSTR, 16);
  sen_io_set_type(io, SEN_COLUMN_FIX_SIZE);
  header->element_size = actual_size;
  header->curr_max = 0;
  if (!(ra = SEN_GMALLOC(sizeof(sen_ra)))) {
    sen_io_close(io);
    return NULL;
  }
  n_elm = SEN_RA_SEGMENT_SIZE / header->element_size;
  for (w_elm = 22; (1 << w_elm) > n_elm; w_elm--);
  SEN_OBJ_DB_SET_TYPE(ra, SEN_COLUMN_FIX_SIZE);
  ra->io = io;
  ra->header = header;
  ra->element_mask =  n_elm - 1;
  ra->element_width = w_elm;
  return ra;
}

sen_ra *
sen_ra_open(sen_ctx *ctx, const char *path)
{
  sen_io *io;
  int n_elm, w_elm;
  sen_ra *ra = NULL;
  struct sen_ra_header *header;
  io = sen_io_open(path, sen_io_auto);
  if (!io) { return NULL; }
  header = sen_io_header(io);
  if (memcmp(header->idstr, SEN_RA_IDSTR, 16)) {
    SEN_LOG(sen_log_error, "ra_idstr (%s)", header->idstr);
    sen_io_close(io);
    return NULL;
  }
  if (!(ra = SEN_GMALLOC(sizeof(sen_ra)))) {
    sen_io_close(io);
    return NULL;
  }
  n_elm = SEN_RA_SEGMENT_SIZE / header->element_size;
  for (w_elm = 22; (1 << w_elm) > n_elm; w_elm--);
  SEN_OBJ_DB_SET_TYPE(ra, SEN_COLUMN_FIX_SIZE);
  ra->io = io;
  ra->header = header;
  ra->element_mask =  n_elm - 1;
  ra->element_width = w_elm;
  return ra;
}

sen_rc
sen_ra_info(sen_ctx *ctx, sen_ra *ra, unsigned int *element_size, sen_id *curr_max)
{
  if (!ra) { return sen_invalid_argument; }
  if (element_size) { *element_size = ra->header->element_size; }
  if (curr_max) { *curr_max = ra->header->curr_max; }
  return sen_success;
}

sen_rc
sen_ra_close(sen_ctx *ctx, sen_ra *ra)
{
  sen_rc rc;
  if (!ra) { return sen_invalid_argument; }
  rc = sen_io_close(ra->io);
  SEN_GFREE(ra);
  return rc;
}

sen_rc
sen_ra_remove(sen_ctx *ctx, const char *path)
{
  if (!path) { return sen_invalid_argument; }
  return sen_io_remove(path);
}

void *
sen_ra_get(sen_ctx *ctx, sen_ra *ra, sen_id id)
{
  void *p;
  uint16_t seg;
  if (id > SEN_ID_MAX) { return NULL; }
  seg = id >> ra->element_width;
  SEN_IO_SEG_MAP(ra->io, seg, p);
  if (!p) { return NULL; }
  if (id > ra->header->curr_max) { ra->header->curr_max = id; }
  return (void *)(((byte *)p) + ((id & ra->element_mask) * ra->header->element_size));
}

void *
sen_ra_at(sen_ctx *ctx, sen_ra *ra, sen_id id)
{
  void *p;
  uint16_t seg;
  static char buf[8] = {0, 0, 0, 0, 0, 0, 0, 0};
  if (id > ra->header->curr_max) { return (void *) buf; }
  seg = id >> ra->element_width;
  SEN_IO_SEG_MAP(ra->io, seg, p);
  if (!p) { return NULL; }
  return (void *)(((byte *)p) + ((id & ra->element_mask) * ra->header->element_size));
}

/**** jagged arrays ****/

#define USE_JA01

#define SEG_NOT_ASSIGNED 0xffffffff

#ifdef USE_JA01

#define SEN_JA_IDSTR "SENNA:JA:01.000"

#define W_OF_JA_MAX 38
#define W_OF_JA_SEGMENT 22
#define W_OF_JA_MAX_SEGMENTS (W_OF_JA_MAX - W_OF_JA_SEGMENT)

#define W_OF_JA_EINFO 3
#define W_OF_JA_EINFO_IN_A_SEGMENT (W_OF_JA_SEGMENT - W_OF_JA_EINFO)
#define N_OF_JA_EINFO_IN_A_SEGMENT (1U << W_OF_JA_EINFO_IN_A_SEGMENT)
#define JA_EINFO_MASK (N_OF_JA_EINFO_IN_A_SEGMENT - 1)

#define N_ELEMENT_VARIATION (W_OF_JA_SEGMENT - W_OF_JA_EINFO + 1)

#define JA_SEGMENT_SIZE (1U << W_OF_JA_SEGMENT)
#define JA_MAX_SEGMENTS (1U << W_OF_JA_MAX_SEGMENTS)

#define JA_N_BSEGMENTS (1U << (W_OF_JA_MAX_SEGMENTS - 7))
#define JA_N_ESEGMENTS (1U << 9)

typedef struct _sen_ja_einfo sen_ja_einfo;

struct _sen_ja_einfo {
  uint16_t seg;
  uint16_t pos;
  uint16_t size;
  uint8_t tail[2];
};

#define EINFO_SET(e,_seg,_pos,_size) {\
  (e)->seg = _seg;\
  (e)->pos = (_pos) >> 4;\
  (e)->size = _size;\
  (e)->tail[0] = (((_pos) >> 14) & 0xc0) + ((_size) >> 16);\
  (e)->tail[1] = 0;\
}

#define EINFO_GET(e,_seg,_pos,_size) {\
  _seg = (e)->seg;\
  _pos = ((e)->pos + (((e)->tail[0] & 0xc0) << 10)) << 4;\
  _size = (e)->size + (((e)->tail[0] & 0x3f) << 16);\
}

typedef struct {
  uint32_t seg;
  uint32_t pos;
} ja_pos;

#define N_GARBAGES 524280

typedef struct {
  uint32_t head;
  uint32_t tail;
  uint32_t nrecs;
  uint32_t next;
  ja_pos recs[N_GARBAGES];
} sen_ja_ginfo;

struct sen_ja_header {
  char idstr[16];
  unsigned max_element_size;
  unsigned max_segments;
  ja_pos free_elements[N_ELEMENT_VARIATION];
  uint32_t garbages[N_ELEMENT_VARIATION];
  uint32_t ngarbages[N_ELEMENT_VARIATION];
  uint8_t segments[JA_MAX_SEGMENTS >> 3];
  uint32_t esegs[JA_N_ESEGMENTS];
  //  uint32_t bsegs[JA_N_BSEGMENTS];
};

#define JA_SEG_ESEG 1 /* entry info */
#define JA_SEG_GSEG 2 /* garbage info */
#define JA_SEG_HSEG 3 /* huge records */

#define JA_IMMEDIATE 1
#define JA_HUGE      2

#define HEADER_SEGMENTS_AT(ja,offset) \
  ((((ja)->header->segments[((offset) >> 3)]) >> ((offset) & 7)) & 1)

#define HEADER_SEGMENTS_ON(ja,offset) \
  (((ja)->header->segments[((offset) >> 3)]) |= (1 << ((offset) & 7)))

#define HEADER_SEGMENTS_OFF(ja,offset) \
  (((ja)->header->segments[((offset) >> 3)]) &= ~(1 << ((offset) & 7)))

sen_ja *
sen_ja_create(sen_ctx *ctx, const char *path, unsigned int max_element_size, uint32_t flags)
{
  int i;
  sen_io *io;
  int max_segments;
  sen_ja *ja = NULL;
  struct sen_ja_header *header;
  max_segments = max_element_size * 128;
  if (max_segments > JA_MAX_SEGMENTS) { max_segments = JA_MAX_SEGMENTS; }
  io = sen_io_create(path, sizeof(struct sen_ja_header),
                     JA_SEGMENT_SIZE, max_segments, sen_io_auto, 0);
  if (!io) { return NULL; }
  header = sen_io_header(io);
  memcpy(header->idstr, SEN_JA_IDSTR, 16);
  sen_io_set_type(io, SEN_COLUMN_VAR_SIZE);
  for (i = 0; i < JA_N_ESEGMENTS; i++) { header->esegs[i] = SEG_NOT_ASSIGNED; }
  //  for (i = 0; i < JA_N_BSEGMENTS; i++) { header->bsegs[i] = SEG_NOT_ASSIGNED; }
  if (!(ja = SEN_GMALLOC(sizeof(sen_ja)))) {
    sen_io_close(io);
    return NULL;
  }
  SEN_OBJ_DB_SET_TYPE(ja, SEN_COLUMN_VAR_SIZE);
  ja->io = io;
  ja->header = header;
  header->max_element_size = max_element_size;
  header->max_segments = max_segments;
  HEADER_SEGMENTS_ON(ja, 0);
  header->esegs[0] = 0;
  return ja;
}

sen_ja *
sen_ja_open(sen_ctx *ctx, const char *path)
{
  sen_io *io;
  sen_ja *ja = NULL;
  struct sen_ja_header *header;
  io = sen_io_open(path, sen_io_auto);
  if (!io) { return NULL; }
  header = sen_io_header(io);
  if (memcmp(header->idstr, SEN_JA_IDSTR, 16)) {
    SEN_LOG(sen_log_error, "ja_idstr (%s)", header->idstr);
    sen_io_close(io);
    return NULL;
  }
  if (!(ja = SEN_GMALLOC(sizeof(sen_ja)))) {
    sen_io_close(io);
    return NULL;
  }
  SEN_OBJ_DB_SET_TYPE(ja, SEN_COLUMN_VAR_SIZE);
  ja->io = io;
  ja->header = header;
  return ja;
}

sen_rc
sen_ja_info(sen_ctx *ctx, sen_ja *ja, unsigned int *max_element_size)
{
  if (!ja) { return sen_invalid_argument; }
  if (max_element_size) { *max_element_size = ja->header->max_element_size; }
  return sen_success;
}

sen_rc
sen_ja_close(sen_ctx *ctx, sen_ja *ja)
{
  sen_rc rc;
  if (!ja) { return sen_invalid_argument; }
  rc = sen_io_close(ja->io);
  SEN_GFREE(ja);
  return rc;
}

sen_rc
sen_ja_remove(sen_ctx *ctx, const char *path)
{
  if (!path) { return sen_invalid_argument; }
  return sen_io_remove(path);
}

void *
sen_ja_ref(sen_ctx *ctx, sen_ja *ja, sen_id id, uint32_t *value_len)
{
  sen_ja_einfo *einfo;
  uint32_t lseg, *pseg, pos;
  lseg = id >> W_OF_JA_EINFO_IN_A_SEGMENT;
  pos = id & JA_EINFO_MASK;
  pseg = &ja->header->esegs[lseg];
  if (*pseg == SEG_NOT_ASSIGNED) { *value_len = 0; return NULL; }
  SEN_IO_SEG_MAP(ja->io, *pseg, einfo);
  if (!einfo) { *value_len = 0; return NULL; }
  if (einfo[pos].tail[1] & JA_IMMEDIATE) {
    *value_len = einfo[pos].tail[1] >> 1;
    return (void *) &einfo[pos];
  }
  {
    void *value;
    sen_io_win iw;
    uint32_t jag, vpos, vsize;
    EINFO_GET(&einfo[pos], jag, vpos, vsize);
    value = sen_io_win_map2(ja->io, ctx, &iw, jag, vpos, vsize, sen_io_rdonly);
    // printf("at id=%d value=%p jag=%d vpos=%d ei=%p(%d:%d)\n", id, value, jag, vpos, &einfo[pos], einfo[pos].pos, einfo[pos].tail[0]);
    if (!value) { *value_len = 0; return NULL; }
    *value_len = vsize;
    return value;
  }
}

sen_rc
sen_ja_unref(sen_ctx *ctx, sen_ja *ja, sen_id id, void *value, uint32_t value_len)
{
  if (!value) { return sen_invalid_argument; }
  // todo : depend on sen_io_win_map2 implementation. unmap2 should be used
  if (value_len > JA_SEGMENT_SIZE) { SEN_FREE(value); }
  return sen_success;
}

sen_rc
sen_ja_free(sen_ctx *ctx, sen_ja *ja, sen_ja_einfo *einfo)
{
  sen_ja_ginfo *ginfo;
  uint32_t seg, pos, element_size, size, m, *gseg;
  if (!einfo->size) { return sen_success; }
  if (einfo->tail[1] & JA_IMMEDIATE) { return sen_success; }
  EINFO_GET(einfo, seg, pos, element_size);
  if (element_size > JA_SEGMENT_SIZE) {
    int n = (pos + element_size + JA_SEGMENT_SIZE - 1) >> W_OF_JA_SEGMENT;
    for (; n--; seg++) { HEADER_SEGMENTS_OFF(ja, seg); }
    return sen_success;
  }
  {
    int es = element_size - 1;
    SEN_BIT_SCAN_REV(es, m);
    m++;
    size = 1 << m;
  }
  gseg = &ja->header->garbages[m - W_OF_JA_EINFO];
  while (*gseg) {
    SEN_IO_SEG_MAP(ja->io, *gseg, ginfo);
    if (!ginfo) { return sen_memory_exhausted; }
    if (ginfo->nrecs < N_GARBAGES) { break; }
    gseg = &ginfo->next;
  }
  if (!*gseg) {
    uint32_t i = 0;
    while (HEADER_SEGMENTS_AT(ja, i)) {
      if (++i >= ja->header->max_segments) { return sen_memory_exhausted; }
    }
    HEADER_SEGMENTS_ON(ja, i);
    *gseg = i;
    SEN_IO_SEG_MAP(ja->io, *gseg, ginfo);
    if (!ginfo) { return sen_memory_exhausted; }
    ginfo->head = 0;
    ginfo->tail = 0;
    ginfo->nrecs = 0;
    ginfo->next = 0;
  }
  ginfo->recs[ginfo->head].seg = seg;
  ginfo->recs[ginfo->head].pos = pos;
  if (++ginfo->head == N_GARBAGES) { ginfo->head = 0; }
  ginfo->nrecs++;
  ja->header->ngarbages[m - W_OF_JA_EINFO]++;
  return sen_success;
}

sen_rc
sen_ja_replace(sen_ctx *ctx, sen_ja *ja, sen_id id, sen_ja_einfo *ei)
{
  uint32_t lseg, *pseg, pos;
  sen_ja_einfo *einfo, eback;
  lseg = id >> W_OF_JA_EINFO_IN_A_SEGMENT;
  pos = id & JA_EINFO_MASK;
  pseg = &ja->header->esegs[lseg];
  if (*pseg == SEG_NOT_ASSIGNED) {
    int i = 0;
    while (HEADER_SEGMENTS_AT(ja, i)) {
      if (++i >= ja->header->max_segments) { return sen_memory_exhausted; }
    }
    HEADER_SEGMENTS_ON(ja, i);
    *pseg = i;
  }
  SEN_IO_SEG_MAP(ja->io, *pseg, einfo);
  if (!einfo) { return sen_memory_exhausted; }
  eback = einfo[pos];
  einfo[pos] = *ei;
  SEN_SET_64BIT(&einfo[pos], *ei);
  sen_ja_free(ctx, ja, &eback);
  return sen_success;
}

#define N_GARBAGES_TH 10

// todo : sen_io_win_map2 cause verbose copy when nseg > 1, it should be copied directly.
sen_rc
sen_ja_alloc(sen_ctx *ctx, sen_ja *ja, int element_size, sen_ja_einfo *einfo, sen_io_win *iw)
{
  void *addr;
  iw->io = ja->io;
  iw->ctx = ctx;
  iw->cached = 1;
  if (element_size < 8) {
    einfo->tail[1] = element_size * 2 + JA_IMMEDIATE;
    iw->addr = (void *)einfo;
    return sen_success;
  }
  if (element_size > JA_SEGMENT_SIZE) {
    int i, j, n = (element_size + JA_SEGMENT_SIZE - 1) >> W_OF_JA_SEGMENT;
    for (i = 0, j = -1; i < ja->header->max_segments; i++) {
      if (HEADER_SEGMENTS_AT(ja, i)) {
        j = i;
      } else {
        if (i == j + n) {
          j++;
          addr = sen_io_win_map2(ja->io, ctx, iw, j, 0, element_size, sen_io_wronly);
          if (!addr) { return sen_memory_exhausted; }
          EINFO_SET(einfo, j, 0, element_size);
          // einfo->tail[1] = JA_HUGE;
          for (; j <= i; j++) { HEADER_SEGMENTS_ON(ja, j); }
          return sen_success;
        }
      }
    }
    SEN_LOG(sen_log_crit, "ja full. requested element_size=%d.", element_size);
    return sen_memory_exhausted;
  } else {
    ja_pos *vp;
    int m, aligned_size, es = element_size - 1;
    SEN_BIT_SCAN_REV(es, m);
    m++;
    aligned_size = 1 << m;
    if (ja->header->ngarbages[m - W_OF_JA_EINFO] > N_GARBAGES_TH) {
      sen_ja_ginfo *ginfo;
      uint32_t seg, pos, *gseg;
      gseg = &ja->header->garbages[m - W_OF_JA_EINFO];
      while (*gseg) {
        SEN_IO_SEG_MAP(ja->io, *gseg, ginfo);
        if (!ginfo) { return sen_memory_exhausted; }
        if (ginfo->next || ginfo->nrecs > N_GARBAGES_TH) {
          seg = ginfo->recs[ginfo->tail].seg;
          pos = ginfo->recs[ginfo->tail].pos;
          SEN_IO_SEG_MAP(ja->io, seg, addr);
          if (!addr) { return sen_memory_exhausted; }
          EINFO_SET(einfo, seg, pos, element_size);
          iw->addr = (byte *)addr + pos;
          if (++ginfo->tail == N_GARBAGES) { ginfo->tail = 0; }
          ginfo->nrecs--;
          ja->header->ngarbages[m - W_OF_JA_EINFO]--;
          if (!ginfo->nrecs) {
            HEADER_SEGMENTS_OFF(ja, *gseg);
            *gseg = ginfo->next;
          }
          return sen_success;
        }
        gseg = &ginfo->next;
      }
    }
    vp = &ja->header->free_elements[m - W_OF_JA_EINFO];
    if (!vp->seg) {
      int i = 0;
      while (HEADER_SEGMENTS_AT(ja, i)) {
        if (++i >= ja->header->max_segments) { return sen_memory_exhausted; }
      }
      HEADER_SEGMENTS_ON(ja, i);
      vp->seg = i;
      vp->pos = 0;
    }
    EINFO_SET(einfo, vp->seg, vp->pos, element_size);
    SEN_IO_SEG_MAP(ja->io, vp->seg, addr);
    // printf("addr=%p seg=%d pos=%d\n", addr, vp->seg, vp->pos);
    if (!addr) { return sen_memory_exhausted; }
    iw->addr = (byte *)addr + vp->pos;
    if ((vp->pos += aligned_size) == JA_SEGMENT_SIZE) {
      vp->seg = 0;
      vp->pos = 0;
    }
    return sen_success;
  }
}

sen_rc
sen_ja_put(sen_ctx *ctx, sen_ja *ja, sen_id id, void *value, int value_len, int flags)
{
  int rc;
  sen_io_win iw;
  sen_ja_einfo einfo;
  if ((flags & SEN_ST_APPEND)) {
    if (value_len) {
      uint32_t old_len;
      void *oldvalue = sen_ja_ref(ctx, ja, id, &old_len);
      if (oldvalue) {
        if ((rc = sen_ja_alloc(ctx, ja, value_len + old_len, &einfo, &iw))) { return rc; }
        memcpy(iw.addr, oldvalue, old_len);
        memcpy((byte *)iw.addr + old_len, value, value_len);
        sen_ja_unref(ctx, ja, id, oldvalue, old_len);
        sen_io_win_unmap2(&iw);
      } else {
        if ((rc = sen_ja_alloc(ctx, ja, value_len, &einfo, &iw))) { return rc; }
        memcpy(iw.addr, value, value_len);
        sen_io_win_unmap2(&iw);
      }
    }
  } else {
    if (value_len) {
      if ((rc = sen_ja_alloc(ctx, ja, value_len, &einfo, &iw))) { return rc; }
      // printf("put id=%d, value_len=%d value=%p ei=%p(%d:%d)\n", id, value_len, buf, &einfo, einfo.pos, einfo.tail[0]);
      memcpy(iw.addr, value, value_len);
      sen_io_win_unmap2(&iw);
    } else {
      memset(&einfo, 0, sizeof(sen_ja_einfo));
    }
  }
  return sen_ja_replace(ctx, ja, id, &einfo);
}

sen_rc
sen_ja_putv(sen_ctx *ctx, sen_ja *ja, sen_id id, sen_obj *vector, int flags)
{
  sen_obj header;
  uint32_t last = 0;
  sen_rc rc = sen_success;
  sen_vector *v = (sen_vector *)vector;
  int i, n = sen_vector_size(ctx, vector);
  SEN_OBJ_INIT(&header, SEN_BULK, 0);
  sen_bulk_benc(ctx, &header, n);
  for (i = 0; i < n; i++) {
    sen_bulk_benc(ctx, &header, v->offsets[i] - last);
    last = v->offsets[i];
  }
  {
    sen_io_win iw;
    sen_ja_einfo einfo;
    size_t sizeh = SEN_BULK_VSIZE(&header);
    size_t sizev = SEN_BULK_VSIZE(vector);
    if ((rc = sen_ja_alloc(ctx, ja, sizeh + sizev, &einfo, &iw))) { goto exit; }
    memcpy(iw.addr, header.u.b.head, sizeh);
    memcpy((char *)iw.addr + sizeh, vector->u.b.head, sizev);
    sen_io_win_unmap2(&iw);
    rc = sen_ja_replace(ctx, ja, id, &einfo);
  }
exit :
  SEN_OBJ_FIN(ctx, &header);
  return rc;
}

int
sen_ja_size(sen_ctx *ctx, sen_ja *ja, sen_id id)
{
  sen_ja_einfo *einfo;
  uint32_t lseg, *pseg, pos;
  lseg = id >> W_OF_JA_EINFO_IN_A_SEGMENT;
  pos = id & JA_EINFO_MASK;
  pseg = &ja->header->esegs[lseg];
  if (*pseg == SEG_NOT_ASSIGNED) { return -1; }
  SEN_IO_SEG_MAP(ja->io, *pseg, einfo);
  if (!einfo) { return -1; }
  if (einfo[pos].tail[1] & JA_IMMEDIATE) {
    return einfo[pos].tail[1] >> 1;
  } else {
    return einfo[pos].size + ((einfo[pos].tail[0] & 0x3f) << 16);
  }
}

int
sen_ja_defrag(sen_ctx *ctx, sen_ja *ja, int threshold)
{
  return 0;
}

#else /* USE_JA01 */

#define SEN_JA_IDSTR "SENNA:JA:02.000"

struct sen_ja_header {
  char idstr[16];
  uint32_t flags;
  uint32_t align_width;
  uint32_t segment_width;
  uint32_t element_size;
  uint32_t curr_pos;
};

#define SEN_JA_DEFAULT_ALIGN_WIDTH    4
#define SEN_JA_DEFAULT_SEGMENT_WIDTH 20

#define JA_SEGMENT_SIZE (1 << SEN_JA_DEFAULT_SEGMENT_WIDTH)

#define JA_ALIGN_WIDTH (ja->header->align_width)
#define JA_ALIGN_MASK ((1U << JA_ALIGN_WIDTH) - 1)

#define JA_DSEG_WIDTH (ja->header->segment_width)
#define JA_DSEG_MASK ((1U << JA_DSEG_WIDTH) - 1)

#define JA_ESEG_WIDTH (JA_DSEG_WIDTH - 3)
#define JA_ESEG_MASK ((1U << JA_ESEG_WIDTH) - 1)

#define JA_EPOS_WIDTH (JA_DSEG_WIDTH - JA_ALIGN_WIDTH)
#define JA_EPOS_MASK ((1U << JA_EPOS_WIDTH) - 1)

// segment_width, align_width, flags
sen_ja *
sen_ja_create(sen_ctx *ctx, const char *path,
              unsigned int max_element_size, uint32_t flags)
{
  int i;
  sen_io *io;
  sen_ja *ja;
  struct sen_ja_header *header;
  uint32_t align_width = SEN_JA_DEFAULT_ALIGN_WIDTH;
  uint32_t segment_width = SEN_JA_DEFAULT_SEGMENT_WIDTH;
  uint32_t max_dsegs = 1 << (32 - (segment_width - align_width));
  uint32_t max_esegs = 1 << (31 - segment_width);
  if (align_width > segment_width || align_width + 32 < segment_width) {
    SEN_LOG(sen_log_error, "invalid align_width, segment_width value");
    return NULL;
  }
  io = sen_io_create(path,
                     sizeof(struct sen_ja_header) +
                     max_dsegs * sizeof(int32_t) +
                     max_esegs * sizeof(int32_t),
                     1 << segment_width, max_dsegs, sen_io_auto, 0);
  if (!io) { return NULL; }
  header = sen_io_header(io);
  memcpy(header->idstr, SEN_JA_IDSTR, 16);
  sen_io_set_type(io, SEN_COLUMN_VAR_SIZE);
  header->flags = flags;
  header->align_width = align_width;
  header->segment_width = segment_width;
  header->element_size = max_element_size;
  header->curr_pos = 0;
  if (!(ja = SEN_GMALLOC(sizeof(sen_ja)))) {
    sen_io_close(io);
    return NULL;
  }
  ja->io = io;
  ja->obj.header.type = SEN_COLUMN_VAR_SIZE;
  ja->header = header;
  ja->dsegs = (uint32_t *)(((uintptr_t) header) + sizeof(struct sen_ja_header));
  ja->esegs = &ja->dsegs[max_dsegs];
  for (i = 0; i < max_esegs; i++) { ja->esegs[i] = SEG_NOT_ASSIGNED; }
  return ja;
}

sen_ja *
sen_ja_open(sen_ctx *ctx, const char *path)
{
  sen_io *io;
  sen_ja *ja = NULL;
  struct sen_ja_header *header;
  io = sen_io_open(path, sen_io_auto);
  if (!io) { return NULL; }
  header = sen_io_header(io);
  if (memcmp(header->idstr, SEN_JA_IDSTR, 16)) {
    SEN_LOG(sen_log_error, "ja_idstr (%s)", header->idstr);
    sen_io_close(io);
    return NULL;
  }
  if (!(ja = SEN_GMALLOC(sizeof(sen_ja)))) {
    sen_io_close(io);
    return NULL;
  }
  ja->io = io;
  ja->obj.header.type = SEN_COLUMN_VAR_SIZE;
  ja->header = header;
  ja->dsegs = (uint32_t *)(((uintptr_t) header) + sizeof(struct sen_ja_header));
  ja->esegs = &ja->dsegs[1 << (32 - JA_EPOS_WIDTH)];
  return ja;
}

sen_rc
sen_ja_info(sen_ctx *ctx, sen_ja *ja, unsigned int *max_element_size)
{
  if (!ja) { return sen_invalid_argument; }
  if (max_element_size) { *max_element_size = 4294967295U; }
  return sen_success;
}

sen_rc
sen_ja_close(sen_ctx *ctx, sen_ja *ja)
{
  sen_rc rc;
  if (!ja) { return sen_invalid_argument; }
  rc = sen_io_close(ja->io);
  SEN_GFREE(ja);
  return rc;
}

sen_rc
sen_ja_remove(sen_ctx *ctx, const char *path)
{
  if (!path) { return sen_invalid_argument; }
  return sen_io_remove(path);
}

inline static void *
sen_ja_ref_raw(sen_ctx *ctx, sen_ja *ja, sen_id id, uint32_t *value_len)
{
  sen_io_ja_einfo *ep;
  uint32_t lseg = id >> JA_ESEG_WIDTH;
  uint32_t lpos = id & JA_ESEG_MASK;
  uint32_t *pseg = &ja->esegs[lseg];
  if (*pseg == SEG_NOT_ASSIGNED) { *value_len = 0; return NULL; }
  SEN_IO_SEG_MAP(ja->io, *pseg, ep);
  if (!ep) { *value_len = 0; return NULL; }
  ep += lpos;
  for (;;) {
    void *value;
    uint32_t epos = ep->pos;
    uint32_t dseg = epos >> JA_EPOS_WIDTH;
    uint32_t doff = (epos & JA_EPOS_MASK) << JA_ALIGN_WIDTH;
    if (!(*value_len = ep->size)) { return NULL; }
    if (sen_io_read_ja(ja->io, ctx, ep, epos, id, dseg, doff, &value, value_len)
        != sen_internal_error) {
      return value;
    }
  }
}

#ifndef NO_ZLIB
#include <zlib.h>

inline static void *
sen_ja_ref_zlib(sen_ctx *ctx, sen_ja *ja, sen_id id, uint32_t *value_len)
{
  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
  z_stream zstream;
  void *value, *zvalue;
  uint32_t zvalue_len = *value_len + sizeof (uint64_t);
  if (!(zvalue = sen_ja_ref_raw(ctx, ja, id, &zvalue_len))) {
    *value_len = 0; return NULL;
  }
  zstream.next_in = (Bytef *)((uint64_t *)zvalue + 1);
  zstream.avail_in = zvalue_len - sizeof (uint64_t);
  zstream.zalloc = Z_NULL;
  zstream.zfree = Z_NULL;
  if (inflateInit2(&zstream, 15 /* windowBits */) != Z_OK) {
    SEN_FREE((sen_io_ja_ehead *)zvalue - 1);
    *value_len = 0;
    return NULL;
  }
  if (!(value = SEN_MALLOC(*(uint64_t *)zvalue + sizeof (sen_io_ja_ehead)))) {
    inflateEnd(&zstream);
    SEN_FREE((sen_io_ja_ehead *)zvalue - 1);
    *value_len = 0;
    return NULL;
  }
  zstream.next_out = (Bytef *)((sen_io_ja_ehead *)value + 1);
  zstream.avail_out = *(uint64_t *)zvalue;
  if (inflate(&zstream, Z_FINISH) != Z_STREAM_END) {
    inflateEnd(&zstream);
    SEN_FREE(value);
    SEN_FREE((sen_io_ja_ehead *)zvalue - 1);
    *value_len = 0;
    return NULL;
  }
  *value_len = zstream.total_out;
  if (inflateEnd(&zstream) != Z_OK) {
    SEN_FREE(value);
    SEN_FREE((sen_io_ja_ehead *)zvalue - 1);
    *value_len = 0;
    return NULL;
  }
  *(sen_io_ja_ehead *)value = ((sen_io_ja_ehead *)zvalue)[-1];
  SEN_FREE((sen_io_ja_ehead *)zvalue - 1);
  return (sen_io_ja_ehead *)value + 1;
}
#endif /* NO_ZLIB */

#ifndef NO_LZO
#include <lzo/lzo1x.h>

inline static void *
sen_ja_ref_lzo(sen_ctx *ctx, sen_ja *ja, sen_id id, uint32_t *value_len)
{
  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
  void *value, *lvalue;
  uint32_t lvalue_len = *value_len + sizeof (uint64_t);
  lzo_uint loutlen;
  if (!(lvalue = sen_ja_ref_raw(ctx, ja, id, &lvalue_len))) {
    *value_len = 0; return NULL;
  }
  if (!(value = SEN_MALLOC(*(uint64_t *)lvalue + sizeof (sen_io_ja_ehead)))) {
    SEN_FREE((sen_io_ja_ehead *)lvalue - 1);
    *value_len = 0;
    return NULL;
  }
  loutlen = *(uint64_t *)lvalue;
  switch (lzo1x_decompress((lzo_bytep)((uint64_t *)lvalue + 1), lvalue_len - sizeof (uint64_t),
                           (lzo_bytep)((sen_io_ja_ehead *)value + 1), &loutlen, NULL)) {
  case LZO_E_OK :
  case LZO_E_INPUT_NOT_CONSUMED :
    break;
  default :
    SEN_FREE(value);
    SEN_FREE((sen_io_ja_ehead *)lvalue - 1);
    *value_len = 0;
    return NULL;
  }
  *value_len = loutlen;
  *(sen_io_ja_ehead *)value = ((sen_io_ja_ehead *)lvalue)[-1];
  SEN_FREE((sen_io_ja_ehead *)lvalue - 1);
  return (sen_io_ja_ehead *)value + 1;
}
#endif /* NO_LZO */

void *
sen_ja_ref(sen_ctx *ctx, sen_ja *ja, sen_id id, uint32_t *value_len)
{
#ifndef NO_ZLIB
  if (ja->header->flags & SEN_OBJ_COMPRESS_ZLIB) {
    return sen_ja_ref_zlib(ctx, ja, id, value_len);
  }
#endif /* NO_ZLIB */
#ifndef NO_LZO
  if (ja->header->flags & SEN_OBJ_COMPRESS_LZO) {
    return sen_ja_ref_lzo(ctx, ja, id, value_len);
  }
#endif /* NO_LZO */
  return sen_ja_ref_raw(ctx, ja, id, value_len);
}

sen_rc
sen_ja_unref(sen_ctx *ctx, sen_ja *ja, sen_id id, void *value, uint32_t value_len)
{
  SEN_FREE((void *)(((uintptr_t) value) - sizeof(sen_io_ja_ehead)));
  return sen_success;
}

#define SEG_ESEG 0xffffffff

sen_rc
assign_eseg(sen_ja *ja, uint32_t lseg)
{
  int i;
  uint32_t max_dsegs = 1 << (32 - JA_EPOS_WIDTH);
  for (i = 0; i < max_dsegs; i++) {
    if (ja->dsegs[i] == 0) {
      if (ja->header->curr_pos && i == (ja->header->curr_pos >> JA_EPOS_WIDTH)) {
        // todo : curr_pos must be moved.
      }
      ja->dsegs[i] = SEG_ESEG;
      ja->esegs[lseg] = i;
      return sen_success;
    }
  }
  return sen_internal_error;
}

static sen_rc
assign_rec(sen_ja *ja, int value_len)
{
  if (ja->header->curr_pos) {
    uint32_t doff = (ja->header->curr_pos & JA_EPOS_MASK) << JA_ALIGN_WIDTH;
    if (doff + value_len + sizeof(sen_io_ja_ehead) <= (1 << JA_DSEG_WIDTH)) {
      return sen_success;
    }
  }
  {
    uint32_t max_dsegs = 1 << (32 - JA_EPOS_WIDTH);
    /* if (value_len <= JA_DSEG_MASK) { todo : } else */
    {
      uint32_t i, j;
      uint32_t nseg = (value_len + sizeof(sen_io_ja_ehead) + JA_DSEG_MASK) >> JA_DSEG_WIDTH;
      for (i = 0, j = 0; i < max_dsegs; i++) {
        if (ja->dsegs[i] == 0) {
          if (++j == nseg) {
            ja->header->curr_pos = (i + 1 - j) << JA_EPOS_WIDTH;
            return sen_success;
          }
        } else {
          j = 0;
        }
      }
    }
    return sen_other_error;
  }
}

inline static sen_rc
sen_ja_put_raw(sen_ctx *ctx, sen_ja *ja, sen_id id, void *value, int value_len, int flags)
{
  sen_rc rc = sen_success;
  uint32_t newpos = 0;
  sen_io_ja_einfo *ep;
  {
    uint32_t lseg = id >> JA_ESEG_WIDTH;
    uint32_t lpos = id & JA_ESEG_MASK;
    uint32_t *pseg = &ja->esegs[lseg];
    if (*pseg == SEG_NOT_ASSIGNED && (rc = assign_eseg(ja, lseg))) { return rc; }
    SEN_IO_SEG_MAP(ja->io, *pseg, ep);
    if (!ep) { return sen_other_error; }
    ep += lpos;
  }
  if (value_len) {
    uint32_t dseg, doff, inc;
    if ((rc = assign_rec(ja, value_len))) { return rc; }
    newpos = ja->header->curr_pos;
    dseg = newpos >> JA_EPOS_WIDTH;
    doff = (newpos & JA_EPOS_MASK) << JA_ALIGN_WIDTH;
    rc = sen_io_write_ja(ja->io, ctx, id, dseg, doff, value, value_len);
    if (rc) { return rc; }
    inc = ((value_len + sizeof(sen_io_ja_ehead) + JA_ALIGN_MASK) >> JA_ALIGN_WIDTH);
    ja->header->curr_pos = ((newpos + inc) & JA_EPOS_MASK) ? (newpos + inc) : 0;
    {
      uint32_t max = 1U << JA_EPOS_WIDTH;
      while (ja->dsegs[dseg] + inc > max) {
        inc -= max - ja->dsegs[dseg];
        ja->dsegs[dseg++] = max;
      }
      ja->dsegs[dseg] += inc;
    }
  }
  {
    sen_io_ja_einfo ne, oe;
    oe = *ep;
    ne.pos = newpos;
    ne.size = value_len;
    *ep = ne; // atomic_set64 maybe more suitable.
    if (oe.size) {
      uint32_t dseg, off, dec, max = 1U << JA_EPOS_WIDTH;
      dseg = oe.pos >> JA_EPOS_WIDTH;
      off = (oe.pos & JA_EPOS_MASK);
      dec = ((oe.size + sizeof(sen_io_ja_ehead) + JA_ALIGN_MASK) >> JA_ALIGN_WIDTH);
      if (off + dec > max) {
        dec -= max - off;
        ja->dsegs[dseg++] -= max - off;
        while (dec > max) {
          dec -= max;
          ja->dsegs[dseg++] = 0;
        }
        {
          uint32_t vsize = (dec << JA_ALIGN_WIDTH) - sizeof(sen_io_ja_ehead);
          rc = sen_io_write_ja_ehead(ja->io, ctx, 0, dseg, 0, vsize);
          // todo : handle rc..
        }
      }
      ja->dsegs[dseg] -= dec;
    }
  }
  return rc;
}

#ifndef NO_ZLIB
inline static sen_rc
sen_ja_put_zlib(sen_ctx *ctx, sen_ja *ja, sen_id id, void *value, int value_len, int flags)
{
  sen_rc rc;
  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
  z_stream zstream;
  void *zvalue;
  int zvalue_len;
  zstream.next_in = value;
  zstream.avail_in = value_len;
  zstream.zalloc = Z_NULL;
  zstream.zfree = Z_NULL;
  if (deflateInit2(&zstream, Z_DEFAULT_COMPRESSION, Z_DEFLATED, 15 /* windowBits */,
                   8 /* memLevel */, Z_DEFAULT_STRATEGY) != Z_OK) { return sen_other_error; }
  zvalue_len = deflateBound(&zstream, value_len);
  if (!(zvalue = SEN_MALLOC(zvalue_len + sizeof (uint64_t)))) { deflateEnd(&zstream); return sen_memory_exhausted; }
  zstream.next_out = (Bytef *)((uint64_t *)zvalue + 1);
  zstream.avail_out = zvalue_len;
  if (deflate(&zstream, Z_FINISH) != Z_STREAM_END) {
    deflateEnd(&zstream);
    SEN_FREE(zvalue);
    return sen_other_error;
  }
  zvalue_len = zstream.total_out;
  if (deflateEnd(&zstream) != Z_OK) { SEN_FREE(zvalue); return sen_other_error; }
  *(uint64_t *)zvalue = value_len;
  rc = sen_ja_put_raw(ctx, ja, id, zvalue, zvalue_len + sizeof (uint64_t), flags);
  SEN_FREE(zvalue);
  return rc;
}
#endif /* NO_ZLIB */

#ifndef NO_LZO
inline static sen_rc
sen_ja_put_lzo(sen_ctx *ctx, sen_ja *ja, sen_id id, void *value, int value_len, int flags)
{
  sen_rc rc;
  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
  void *lvalue, *lwork;
  lzo_uint lvalue_len = value_len + value_len / 16 + 64 + 3;
  if (!(lvalue = SEN_MALLOC(lvalue_len + sizeof (uint64_t)))) { return sen_memory_exhausted; }
  if (!(lwork = SEN_MALLOC(LZO1X_1_MEM_COMPRESS))) { SEN_FREE(lvalue); return sen_memory_exhausted; }
  if (lzo1x_1_compress(value, value_len, (lzo_bytep)((uint64_t *)lvalue + 1), &lvalue_len, lwork) != LZO_E_OK) {
    SEN_FREE(lwork);
    SEN_FREE(lvalue);
    return sen_other_error;
  }
  SEN_FREE(lwork);
  *(uint64_t *)lvalue = value_len;
  rc = sen_ja_put_raw(ctx, ja, id, lvalue, lvalue_len + sizeof (uint64_t), flags);
  SEN_FREE(lvalue);
  return rc;
}
#endif /* NO_LZO */

sen_rc
sen_ja_put(sen_ctx *ctx, sen_ja *ja, sen_id id, void *value, int value_len, int flags)
{
#ifndef NO_ZLIB
  if (ja->header->flags & SEN_OBJ_COMPRESS_ZLIB) {
    return sen_ja_put_zlib(ctx, ja, id, value, value_len, flags);
  }
#endif /* NO_ZLIB */
#ifndef NO_LZO
  if (ja->header->flags & SEN_OBJ_COMPRESS_LZO) {
    return sen_ja_put_lzo(ctx, ja, id, value, value_len, flags);
  }
#endif /* NO_LZO */
  return sen_ja_put_raw(ctx, ja, id, value, value_len, flags);
}

static sen_rc
sen_ja_defrag_seg(sen_ctx *ctx, sen_ja *ja, uint32_t dseg)
{
  sen_rc rc = sen_success;
  sen_io_win iw;
  uint32_t epos = dseg << JA_EPOS_WIDTH;
  uint32_t segsize = (1 << JA_DSEG_WIDTH);
  byte *v = sen_io_win_map(ja->io, ctx, &iw, dseg, 0, segsize, sen_io_rdonly);
  byte *ve = v + segsize;
  uint32_t *segusage = &ja->dsegs[dseg];
  if (!v) { return sen_internal_error; }
  while (v < ve && *segusage) {
    sen_io_ja_ehead *eh = (sen_io_ja_ehead *) v;
    uint32_t inc = ((eh->size + sizeof(sen_io_ja_ehead) + JA_ALIGN_MASK) >> JA_ALIGN_WIDTH);
    sen_io_ja_einfo *ep;
    uint32_t lseg = eh->key >> JA_ESEG_WIDTH;
    uint32_t lpos = eh->key & JA_ESEG_MASK;
    uint32_t *pseg = &ja->esegs[lseg];
    if (*pseg == SEG_NOT_ASSIGNED) {
      SEN_LOG(sen_log_error, "ep is not assigned (%x)", lseg);
      rc = sen_internal_error;
      goto exit;
    }
    SEN_IO_SEG_MAP(ja->io, *pseg, ep);
    if (!ep) {
      SEN_LOG(sen_log_error, "ep map failed (%x)", lseg);
      rc = sen_internal_error;
      goto exit;
    }
    ep += lpos;
    if (ep->pos == epos) {
      if (ep->size != eh->size) {
        SEN_LOG(sen_log_error, "ep->size(%d) != eh->size(%d)", ep->size, eh->size);
        rc = sen_internal_error;
        goto exit;
      }
      if ((rc = sen_ja_put(ctx, ja, eh->key, v + sizeof(sen_io_ja_ehead), eh->size, 0))) {
        goto exit;
      }
    }
    epos += inc;
    v += (inc << JA_ALIGN_WIDTH);
  }
  if (*segusage) {
    SEN_LOG(sen_log_error, "ja->dsegs[dseg] = %d after defrag", ja->dsegs[dseg]);
    rc = sen_internal_error;
  }
exit :
  sen_io_win_unmap(&iw);
  return rc;
}

int
sen_ja_defrag(sen_ctx *ctx, sen_ja *ja, int threshold)
{
  int nsegs = 0;
  uint32_t ts = 1U << (JA_EPOS_WIDTH - threshold);
  uint32_t dseg, max_dsegs = 1 << (32 - JA_EPOS_WIDTH);
  for (dseg = 0; dseg < max_dsegs; dseg++) {
    if (dseg == ja->header->curr_pos >> JA_EPOS_WIDTH) { continue; }
    if (ja->dsegs[dseg] && ja->dsegs[dseg] < ts) {
      if (!sen_ja_defrag_seg(ctx, ja, dseg)) { nsegs++; }
    }
  }
  return nsegs;
}

#endif /* USE_JA01 */

/**** vgram ****/

static int len_sum = 0;
static int img_sum = 0;
static int simple_sum = 0;
static int skip_sum = 0;

sen_vgram *
sen_vgram_create(const char *path)
{
  sen_vgram *s;
  if (!(s = SEN_GMALLOC(sizeof(sen_vgram)))) { return NULL; }
  s->vgram = sen_sym_create(path, sizeof(sen_id) * 2, 0, sen_enc_none);
  if (!s->vgram) {
    SEN_GFREE(s);
    return NULL;
  }
  return s;
}

sen_vgram *
sen_vgram_open(const char *path)
{
  sen_vgram *s;
  if (!(s = SEN_GMALLOC(sizeof(sen_vgram)))) { return NULL; }
  s->vgram = sen_sym_open(path);
  if (!s->vgram) {
    SEN_GFREE(s);
    return NULL;
  }
  return s;
}

sen_vgram_buf *
sen_vgram_buf_open(size_t len)
{
  sen_vgram_buf *b;
  if (!(b = SEN_GMALLOC(sizeof(sen_vgram_buf)))) { return NULL; }
  b->len = len;
  b->tvs = b->tvp = SEN_GMALLOC(sizeof(sen_id) * len);
  if (!b->tvp) { SEN_GFREE(b); return NULL; }
  b->tve = b->tvs + len;
  b->vps = b->vpp = SEN_GMALLOC(sizeof(sen_vgram_vnode) * len * 2);
  if (!b->vpp) { SEN_GFREE(b->tvp); SEN_GFREE(b); return NULL; }
  b->vpe = b->vps + len;
  return b;
}

sen_rc
sen_vgram_buf_add(sen_vgram_buf *b, sen_id tid)
{
  uint8_t dummybuf[8], *dummyp;
  if (b->tvp < b->tve) { *b->tvp++ = tid; }
  dummyp = dummybuf;
  SEN_B_ENC(tid, dummyp);
  simple_sum += dummyp - dummybuf;
  return sen_success;
}

typedef struct {
  sen_id vid;
  sen_id tid;
} vgram_key;

sen_rc
sen_vgram_update(sen_vgram *vgram, sen_id rid, sen_vgram_buf *b, sen_hash *terms)
{
  /*
  sen_inv_updspec **u;
  if (b && b->tvs < b->tvp) {
    sen_id *t0, *tn;
    for (t0 = b->tvs; t0 < b->tvp - 1; t0++) {
      sen_vgram_vnode *v, **vp;
      if (sen_set_at(terms, t0, (void **) &u)) {
        vp = &(*u)->vnodes;
        for (tn = t0 + 1; tn < b->tvp; tn++) {
          for (v = *vp; v && v->tid != *tn; v = v->cdr) ;
          if (!v) {
            if (b->vpp < b->vpe) {
              v = b->vpp++;
            } else {
              // todo;
              break;
            }
            v->car = NULL;
            v->cdr = *vp;
            *vp = v;
            v->tid = *tn;
            v->vid = 0;
            v->freq = 0;
            v->len = tn - t0;
          }
          v->freq++;
          if (v->vid) {
            vp = &v->car;
          } else {
            break;
          }
        }
      }
    }
    {
      sen_set *th = sen_set_open(sizeof(sen_id), sizeof(int), 0);
      if (!th) { return sen_memory_exhausted; }
      if (t0 == b->tvp) { SEN_LOG(sen_log_debug, "t0 == tvp"); }
      for (t0 = b->tvs; t0 < b->tvp; t0++) {
        sen_id vid, vid0 = *t0, vid1 = 0;
        sen_vgram_vnode *v, *v2 = NULL, **vp;
        if (sen_set_at(terms, t0, (void **) &u)) {
          vp = &(*u)->vnodes;
          for (tn = t0 + 1; tn < b->tvp; tn++) {
            for (v = *vp; v; v = v->cdr) {
              if (!v->vid && (v->freq < 2 || v->freq * v->len < 4)) {
                *vp = v->cdr;
                v->freq = 0;
              }
              if (v->tid == *tn) { break; }
              vp = &v->cdr;
            }
            if (v) {
              if (v->freq) {
                v2 = v;
                vid1 = vid0;
                vid0 = v->vid;
              }
              if (v->vid) {
                vp = &v->car;
                continue;
              }
            }
            break;
          }
        }
        if (v2) {
          if (!v2->vid) {
            vgram_key key;
            key.vid = vid1;
            key.tid = v2->tid;
            if (!(v2->vid = sen_sym_get(vgram->vgram, (char *)&key))) {
              sen_set_close(th);
              return sen_memory_exhausted;
            }
          }
          vid = *t0 = v2->vid * 2 + 1;
          memset(t0 + 1, 0, sizeof(sen_id) * v2->len);
          t0 += v2->len;
        } else {
          vid = *t0 *= 2;
        }
        {
          int *tf;
          if (!sen_set_get(th, &vid, (void **) &tf)) {
            sen_set_close(th);
            return sen_memory_exhausted;
          }
          (*tf)++;
        }
      }
      if (!th->n_entries) { SEN_LOG(sen_log_debug, "th->n_entries == 0"); }
      {
        int j = 0;
        int skip = 0;
        sen_set_eh *ehs, *ehp, *ehe;
        sen_set_sort_optarg arg;
        uint8_t *ps = SEN_GMALLOC(b->len * 2), *pp, *pe;
        if (!ps) {
          sen_set_close(th);
          return sen_memory_exhausted;
        }
        pp = ps;
        pe = ps + b->len * 2;
        arg.mode = sen_sort_descending;
        arg.compar = NULL;
        arg.compar_arg = (void *)(intptr_t)sizeof(sen_id);
        ehs = sen_set_sort(th, 0, &arg);
        if (!ehs) {
          SEN_GFREE(ps);
          sen_set_close(th);
          return sen_memory_exhausted;
        }
        SEN_B_ENC(th->n_entries, pp);
        for (ehp = ehs, ehe = ehs + th->n_entries; ehp < ehe; ehp++, j++) {
          int *id = (int *)SEN_SET_INTVAL(*ehp);
          SEN_B_ENC(*SEN_SET_INTKEY(*ehp), pp);
          *id = j;
        }
        for (t0 = b->tvs; t0 < b->tvp; t0++) {
          if (*t0) {
            int *id;
            if (!sen_set_at(th, t0, (void **) &id)) {
              SEN_LOG(sen_log_error, "lookup error (%d)", *t0);
            }
            SEN_B_ENC(*id, pp);
          } else {
            skip++;
          }
        }
        len_sum += b->len;
        img_sum += pp - ps;
        skip_sum += skip;
        SEN_GFREE(ehs);
        SEN_GFREE(ps);
      }
      sen_set_close(th);
    }
  }
  */
  return sen_success;
}

sen_rc
sen_vgram_buf_close(sen_vgram_buf *b)
{
  if (!b) { return sen_invalid_argument; }
  if (b->tvs) { SEN_GFREE(b->tvs); }
  if (b->vps) { SEN_GFREE(b->vps); }
  SEN_GFREE(b);
  return sen_success;
}

sen_rc
sen_vgram_close(sen_vgram *vgram)
{
  if (!vgram) { return sen_invalid_argument; }
  SEN_LOG(sen_log_debug, "len=%d img=%d skip=%d simple=%d", len_sum, img_sum, skip_sum, simple_sum);
  sen_sym_close(vgram->vgram);
  SEN_GFREE(vgram);
  return sen_success;
}
